#! /usr/bin/python

# Usage: create_accesslog <accesslog> <logstats> 
#                         <startdate> <enddate> <format>
#  accesslog -- write the accesslog to the given file
#  logstats  -- write the corresponding logstats to the given file
#  startdate -- specify the starting date for the accesslog entries
#  enddate   -- specify the end date for the accesslog entries
#  format    -- specify the format of the accesslog (apache or squid)
#
#  The content length for each request is randomly generated. About
#  1/2 the requests will be gzipped, and about 1/4 of the requests
#  will be cached.


import sys, time, os, os.path, types, random

monthnames = [ "", "Jan", "Feb", "Mar", "Apr", "May", "Jun", 
               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" ]

NON_GZIP = 0
GZIP = 1
ORIG = 2
MISS = 3
HIT = 4


def days_in_month(year, month):
    days = [ 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ] 
    assert(month >= 1 and month <= 12)
    if (year % 4 == 0 and year != 2000):
        days[2] = 29
    return days[month] 


def Monthstats_new(year, month):
    data = {}
    data["year"] = year
    data["month"] = month
    data["names"] = [ "month_size", "month_count"]
    for i in range(1, days_in_month(year, month) + 1):
        data["names"].append("day_size" + str(i))
    for i in range(1, days_in_month(year, month) + 1):
        data["names"].append("day_count" + str(i))
    for i in range(1, 8):
        data["names"].append("dayweek_size" + str(i))
    for i in range(1, 8):
        data["names"].append("dayweek_count" + str(i))
    for i in range(0, 24):
        data["names"].append("hour_size" + str(i))
    for i in range(0, 24):
        data["names"].append("hour_count" + str(i))
 
    for name in data["names"]:
        data[name] = [ 0, 0, 0, 0, 0 ]
    return data


def write_apache_line(file, thetime, origsize, gzipsize, iscachehit):
    ident = ""
    if (iscachehit):
        ident = "hit" + str(origsize)
    else:
        ident = "miss" + str(origsize)

    date = time.localtime(thetime)
    year = date[0]; mon = monthnames[date[1]]; day = date[2];
    hour = date[3]; min = date[4]; sec = date[5] 

    s = "192.168.0.10 %s - [%.2d/%s/%d:%.2d:%.2d:%.2d -0700] \"GET /sample.html \
HTTP/1.0\" 200 %d \"http://www.referer.com/\" \"Mozilla 4.0 (IE 6.0 compatible)\"\n" % (ident, day, mon, year, hour, min, sec, gzipsize)
    file.write(s)
    return len(s)

def write_squid_line(file, thetime, origsize, gzipsize, iscachehit):
    hit = ""
    if (iscachehit):
        hit = "TCP_HIT"
        server = "NONE/-"
    else:
        hit = "TCP_MISS"
        server = "DIRECT/192.168.0.12"

    s = "%d.000 453 192.168.0.10 %s/200 %d GET http://www.host.com/sample.html \
%d %s text/html\n" % (thetime, hit, gzipsize, origsize, server)
    file.write(s)
    return len(s)


def write_log_file(stats, filename, starttime, endtime, format):
    curr = starttime
    iscachehit = 0
    count = 0
    numlines = 0
    filesize = 0

    file = open(filename, "w")

    while (curr < endtime):
        # Generate a random content length
        origsize = random.randint(10000, 80000)
        # Make half of replies gzipped. 
        if (count % 2 == 0):
            gzipsize = origsize / 2
        else:
            gzipsize = origsize
        # Make 1/4 of replies cache hits
        if (count % 4 == 0):
            iscachehit = 1
        else:
            iscachehit = 0
        if (format == "apache"):
            n = write_apache_line(file, curr, origsize, gzipsize, iscachehit)
        elif (format == "squid"):
            n = write_squid_line(file, curr, origsize, gzipsize, iscachehit)
        filesize += n
        numlines += 1
        compute_stats_line(stats, curr, origsize, gzipsize, iscachehit)
        # The request timestamps will be 10 minutes apart
        curr += 600
        count += 1

    sd = time.localtime(starttime)
    ed = time.localtime(curr - 600)
    if (stats.has_key(filename) == 0 or stats[filename] != filename):
        stats[filename] = filename
        stats["filesize"] = filesize
        stats["numlines"] = numlines
        stats["startdate"] = [sd[0], sd[1], sd[2], sd[3], sd[4], sd[5], sd[8] ]
        stats["enddate"] = [ed[0], ed[1], ed[2], ed[3], ed[4], ed[5], ed[8] ]
    else:
        stats["filesize"] += filesize
        stats["numlines"] += numlines
        stats["enddate"] = [ed[0], ed[1], ed[2], ed[3], ed[4], ed[5], ed[8] ]
    file.close()


def compute_stats_line(stats, curr, origsize, gzipsize, iscachehit):
    date = time.localtime(curr)
    year = date[0] 
    month = date[1] 
    day = date[2] 
    hour = date[3]
    # in python, monday is 0, tues = 1, ...
    dayweek = ((date[6] + 1) % 7) + 1;

    key = year * 100 + month
    if (not stats.has_key(key)):
        stats[key] = Monthstats_new(year, month)
    m = stats[key]
    sizes = [0, 0, 0, 0, 0]
    counts = [0, 0, 0, 0, 0]

    if (origsize == gzipsize):
        sizes[NON_GZIP] = origsize
        counts[NON_GZIP] = 1 
    else:
        sizes[GZIP] = gzipsize
        sizes[ORIG] = origsize
        counts[GZIP] = 1
        counts[ORIG] = 1
    if (iscachehit):
        sizes[HIT] = origsize
        counts[HIT] = 1
    else:
        sizes[MISS] = origsize
        counts[MISS] = 1

    monthsize = m["month_size"] 
    monthcount = m["month_count"]
    daysize = m["day_size" + str(day)]
    daycount = m["day_count" + str(day)]
    dayweeksize = m["dayweek_size" + str(dayweek)]
    dayweekcount = m["dayweek_count" + str(dayweek)]
    hoursize = m["hour_size" + str(hour)]
    hourcount = m["hour_count" + str(hour)]
 
    for i in range(0, 5):
        s = sizes[i]
        c = counts[i]
        monthsize[i] += s
        monthcount[i] += c
        daysize[i] += s
        daycount[i] += c
        dayweeksize[i] += s
        dayweekcount[i] += c
        hoursize[i] += s
        hourcount[i] += c
 

def write_int_array(file, name, array):
    file.write(name)
    for n in array:
        file.write(" " + str(n))
    file.write("\n")


def write_logstats_file(stats, filename):
    file = open(filename, "w")
    file.write("""# logstats.txt
#
# This single text file contains caching and compression statistics.
# The format is as follows:
#
# lastupdate
#     information about the accesslog when the last update was done
# filesize
#     The size of the accesslog (in bytes)
# numlines
#     The number of lines in the accesslog
# startdate <year> <month> <day> <hour> <min> <sec> <is daylight savings
#     The date of the first entry in the accesslog
# enddate <year> <month> <day> <hour> <min> <sec> <is daylight savings
#     The date of the last entry in the accesslog
# endlastupdate
#
# yearmonth <year> <month>
#     Statistics for a specific month
# month_size    <non-gzip> <gzip> <orig> <miss> <hit>
# day_sizeN     <non-gzip> <gzip> <orig> <miss> <hit>
# dayweek_sizeN <non-gzip> <gzip> <orig> <miss> <hit>
# hour_sizeN    <non-gzip> <gzip> <orig> <miss> <hit>
#
#     Bandwidth statistics for a particular time period.
#     month   -- The entire month
#     day     -- A specific day in the month. N = 1 to 31
#     dayweek -- A specific day of the week. N = 1 (Sunday) to 7
#     hour    -- A specific hour of the day. N = 0 to 23
#
#     The 5 statistics measured are given below:
#     non-gzip -- Amount of non-gzipped data written to clients
#     gzip     -- Amount of gzipped data written to clients
#     orig     -- Original size of gzipped data written to clients
#     miss     -- Amount of data read from origin websites
#     hit      -- Amount of data read from the cache
#     All measurements are in bytes
#
# month_count    <non-gzip> <gzip> <orig> <miss> <hit>
# day_countN     <non-gzip> <gzip> <orig> <miss> <hit>
# dayweek_countN <non-gzip> <gzip> <orig> <miss> <hit>
# hour_countN    <non-gzip> <gzip> <orig> <miss> <hit>
#
#     These statistics are similar to the ones above, except we are
#     measuring the number of requests, rather than the bandwidth
#
""")
    file.write("lastupdate\n")
    file.write("filesize " + str(stats["filesize"]) + "\n")
    file.write("numlines " + str(stats["numlines"]) + "\n")
    write_int_array(file, "startdate", stats["startdate"])
    write_int_array(file, "enddate", stats["enddate"])
    file.write("endlastupdate\n")

    dates = stats.keys()
    dates = filter(lambda x: type(x) == types.IntType, dates)
    dates.sort()

    for date in dates:
        year = date / 100
        month = date % 100
        m = stats[date]
        file.write("yearmonth " + str(year) + " " + str(month) + "\n")
        for name in m["names"]:
            write_int_array(file, name, m[name])
        file.write("endyearmonth\n")
    file.close()


def main(argv):
    if (len(argv) < 6):
        s = """
Usage: create_accesslog <accesslog> <logstats> 
                        <startdate> <enddate> <format>
  accesslog -- write the accesslog to the given file
  logstats  -- write the corresponding logstats to the given file
  startdate -- specify the starting date for the accesslog entries
  enddate   -- specify the end date for the accesslog entries
  format    -- specify the format of the accesslog (apache or squid)

  The content length for each request is randomly generated. About
  1/2 the requests will be gzipped, and about 1/4 of the requests
  will be cached.
"""
        print(s)
        sys.exit(0) 

    accesslogfile = argv[1]
    logstatsfile = argv[2]

    start = argv[3].split(":")
    for i in range(0, len(start)):
        start[i] = int(start[i])

    end = argv[4].split(":")
    for i in range(0, len(end)):
        end[i] = int(end[i])

    format = argv[5]

    stats = {}
    start.extend([0, 0, 0])
    end.extend([0, 0, 0])
    starttime = time.mktime(start)
    endtime = time.mktime(end)
    write_log_file(stats, accesslogfile, starttime, endtime, format)
    write_logstats_file(stats, logstatsfile)


main(sys.argv)

