#! /bin/sh

# Copyright (c) 2003-2007 Visolve
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License version 2.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.


# This is the functional test for the update_log_stats program
# To run, do
#
# ./update_logstats_test.sh <path to update_log_stats>
# ./update_logstats_test.sh /usr/local/vicompress/bin/update_log_stats


# void echo_noline(string message)
# print the message without a newline (\n)
function echo_noline {
    uname -a | grep -i -E "linux|cygwin" > /dev/null
    if [ $? = 0 ]
    then
        echo -n $1
    else
        echo "$1\c"
    fi
}

# void check_results(origlogstats.txt, logstats.txt)
# Verify that:
# - The original logstats.txt matches the generated logstats.txt
# - The stats.html files don't contain '$' variables.
#   The update_log_stats should've replaced all $ variables with
#   actual numerical values. 
function check_results {
    diff -b $1 $2 > /dev/null
    if [ $? != 0 ]
    then
        echo "Fail"
        echo "$1 != $2"; 
        diff -b $1 $2
        exit 1
    fi
    grep '\$' *stats.html > /dev/null 2>&1
    if [ $? = 0 ]
    then
        echo "Fail" 
        echo "The html files contain \$ variables";
        grep '\$' *stats.html 
        exit 1
    fi
}


if [ "$1" = "" ]
then
    echo "Usage: update_logstats_test.sh <path to update_log_stats>"
    echo "Usage: update_logstats_test.sh /usr/local/vicompress/bin/update_log_stats"
    exit 1 
fi

update_log_stats=$1

if [ ! -e $update_log_stats ]
then
    echo "Error: $update_log_stats doesnt exist"
    exit 1
fi

PATH=$PATH:/usr/sbin   # for useradd, userdel

cp ../logstats/template.html .

# Create the following files:
# apache.accesslog    - an apache formatted accesslog
# apache.logstats.txt - the corresponding logstats.txt for the log file
# squid.accesslog     - a squid formatted accesslog
# squid.logstats.txt  - the corresponding logstats.txt for the log file
#
# The test cases below will use these accesslog files and logstats.txt.
# The main idea is to run update_log_stats against the accesslog, and
# make sure the generated logstats.txt matches the corresponding
# apache/squid logstats.txt
#
echo_noline "creating apache logfile......................."
./create_accesslog.py apache.accesslog apache.logstats.txt \
   2003:10:1:0:3:0 2004:3:15:0:3:0 apache
echo "Done"

echo_noline "creating squid logfile........................"
./create_accesslog.py squid.accesslog squid.logstats.txt \
   2003:10:1:0:3:0 2004:3:15:0:3:0 squid
echo "Done"

############################################################
echo_noline "test usage...................................."
# Check that usage is printed when update_log_stats is not 
# passed arguments
############################################################
$update_log_stats | grep -i usage > /dev/null
if [ $? = 0 ]
then
    echo "Pass"
else
    echo "Fail"
fi


############################################################
echo_noline "test single accesslog (apache)................"
# Check that update_log_stats works correctly for a single 
# apache format accesslog file. 
############################################################

rm -f accesslog* logstats* *stats.html
cp apache.accesslog accesslog
$update_log_stats accesslog .
check_results apache.logstats.txt logstats.txt
rm -f accesslog* logstats* *stats.html
echo "Pass" 

############################################################
echo_noline "test single accesslog (squid)................."
# Check that update_log_stats works correctly for a single 
# squid format accesslog file.  
############################################################

rm -f accesslog* logstats* *stats.html
cp squid.accesslog accesslog
$update_log_stats accesslog . 
check_results squid.logstats.txt logstats.txt
rm -f accesslog* logstats* *stats.html
echo "Pass" 

############################################################
echo_noline "test IPv6 accesslog (apache).................."
# Check that update_log_stats works correctly for a single 
# apache format accesslog file, with IPv6 addresses.
############################################################

rm -f accesslog* logstats* *stats.html
cat apache.accesslog | sed 's|192.168.0.10|fe80::213:21ff:fe7c:7471|g' > accesslog
$update_log_stats accesslog .

# remove the filesize info from both logstats.txt files
# They will not be equal
cat apache.logstats.txt | grep -v -E "filesize" > tmp.txt
cat logstats.txt | grep -v -E "filesize" > tmp2.txt
mv tmp2.txt logstats.txt
check_results tmp.txt logstats.txt
rm -f accesslog* logstats* *stats.html tmp.txt
echo "Pass" 

############################################################
echo_noline "test IPv6 accesslog (squid)..................."
# Check that update_log_stats works correctly for a single 
# squid format accesslog file, with IPv6 addresses.
############################################################

rm -f accesslog* logstats* *stats.html
cat squid.accesslog | sed 's|192.168.0.10|fe80::213:21ff:fe7c:7471|g' | \
  sed 's|192.168.0.12|fe80::213:21ff:fe7c:8512|g' > accesslog
$update_log_stats accesslog . 

# remove the filesize info from both logstats.txt files
# They will not be equal
cat squid.logstats.txt | grep -v -E "filesize" > tmp.txt
cat logstats.txt | grep -v -E "filesize" > tmp2.txt
mv tmp2.txt logstats.txt
check_results tmp.txt logstats.txt
rm -f accesslog* logstats* *stats.html tmp.txt
echo "Pass" 



############################################################
echo_noline "test no accesslog ............................"
# Check that update_log_stats simply exits if no accesslog
# file exists.
############################################################

rm -f accesslog* logstats* *stats.html
output=`$update_log_stats accesslog .`
if [ "$output" != "" ]
then
    echo "Fail"; echo "The output was: $output"
    exit 1
fi
rm -f accesslog* logstats* *stats.html
echo "Pass"

############################################################
echo_noline "test blank accesslog ........................."
# Check that update_log_stats simply exits if an empty
# accesslog exists.
############################################################

rm -f accesslog* logstats* *stats.html
touch accesslog
output=`$update_log_stats accesslog .`
if [ "$output" != "" ]
then
    echo "Fail"; echo "The output was: $output"
    exit 1
fi
rm -f accesslog* logstats* *stats.html
echo "Pass"


############################################################
echo_noline "test multiple accesslog (apache).............."
# Check that update_log_stats works correctly with multiple 
# apache format accesslog files. 
############################################################

rm -f accesslog* logstats* *stats.html
cp apache.accesslog accesslog.tmp
# split accesslog into 5 logs
total=`cat accesslog.tmp | wc -l`
nlines=`expr $total / 3`
nlines=`expr $nlines + 1`
split -l $nlines accesslog.tmp accesslog.
mv accesslog.aa accesslog.2
mv accesslog.ab accesslog.1
mv accesslog.ac accesslog

$update_log_stats accesslog . 

# remove the lastupdate info from both logstats.txt files
# They will not be equal
cat apache.logstats.txt | \
  grep -v -E "filesize|numlines|startdate|enddate" > tmp.txt
cat logstats.txt | \
  grep -v -E "filesize|numlines|startdate|enddate" > tmp2.txt
mv tmp2.txt logstats.txt

check_results tmp.txt logstats.txt
rm -f accesslog* logstats* *stats.html tmp.txt
echo "Pass"

############################################################
echo_noline "test multiple accesslog (squid)..............."
# Check that update_log_stats works correctly with multiple 
# squid format accesslog files. 
############################################################

rm -f accesslog* logstats* *stats.html
cp apache.accesslog accesslog.tmp
# split accesslog into 5 logs
total=`cat accesslog.tmp | wc -l`
nlines=`expr $total / 3`
nlines=`expr $nlines + 1`
split -l $nlines accesslog.tmp accesslog.
mv accesslog.aa accesslog.2
mv accesslog.ab accesslog.1
mv accesslog.ac accesslog

$update_log_stats accesslog . 

# remove the lastupdate info from both logstats.txt files
# They will not be equal
cat apache.logstats.txt | \
  grep -v -E "filesize|numlines|startdate|enddate" > tmp.txt
cat logstats.txt | \
  grep -v -E "filesize|numlines|startdate|enddate" > tmp2.txt
mv tmp2.txt logstats.txt

check_results tmp.txt logstats.txt
rm -f accesslog* logstats* *stats.html tmp.txt
echo "Pass"


############################################################
echo_noline "test quotes in url ..........................."
# Check that update_log_stats correctly parses apache
# log lines which contain escaped quotes \" in the
# url.  It should not give an error
############################################################

echo "217.223.201.164 miss25 - [29/Mar/2006:02:19:28 -0800] \"GET /path/\\\" hasquote and space.html HTTP/1.0\" 400 25 \"-\" \"-\"" > accesslog

output=`$update_log_stats accesslog .`
rm -f accesslog* logstats* *stats.html tmp.txt

if [ "$output" != "" ]
then
    echo "Fail"; echo "output is: $output"
    exit 1
fi
echo "Pass"

############################################################
echo_noline "test multiple updates (apache) ..............."
# Check that update_log_stats works correclty when run multiple 
# times, as the accesslog increases.  On each run, the 
# update_log_stats will use the previous logstats.txt to skip
# to the last accesslog line where it left off.
############################################################

rm -f accesslog* logstats* *stats.html tmp.txt

cp apache.accesslog accesslog.tmp
# Use just the first half of the accesslog
total=`cat accesslog.tmp | wc -l`
nlines=`expr $total / 2`
nlines=`expr $nlines + 1`
split -l $nlines accesslog.tmp accesslog.

mv accesslog.aa accesslog

$update_log_stats accesslog . 
rm -f accesslog*
cp apache.accesslog accesslog
$update_log_stats accesslog .
check_results apache.logstats.txt logstats.txt
rm -f accesslog* logstats* *stats.html tmp.txt
echo "Pass"


############################################################
echo_noline "test multiple updates (squid) ................"
# Check that update_log_stats works correclty when run multiple 
# times, as the accesslog increases.  On each run, the 
# update_log_stats will use the previous logstats.txt to skip
# to the last accesslog line where it left off.
############################################################

rm -f accesslog* logstats* *stats.html tmp.txt

cp squid.accesslog accesslog.tmp
# Use just the first half of the accesslog
total=`cat accesslog.tmp | wc -l`
nlines=`expr $total / 2`
nlines=`expr $nlines + 1`
split -l $nlines accesslog.tmp accesslog.

mv accesslog.aa accesslog

$update_log_stats accesslog . 
rm -f accesslog*
cp squid.accesslog accesslog
$update_log_stats accesslog .
check_results squid.logstats.txt logstats.txt
rm -f accesslog* logstats* *stats.html tmp.txt
echo "Pass"


############################################################
echo_noline "test bad first entry.........................."
# Check that update_log_stats exits with an error if
# the first entry in the accesslog is not valid.
############################################################

rm -f accesslog* logstats* *stats.html tmp.txt
echo "hello there" > accesslog
output=`$update_log_stats accesslog . 2>&1`
rm -f accesslog* logstats* *stats.html tmp.txt

if [ "$output" != "Error: unknown log format file accesslog" ]
then
    echo "Fail";
    echo "Bad output: $output"
    exit 1
fi
echo "Pass"


############################################################
echo_noline "test bad entries (apache)....................."
# Check that update_log_stats still works even if there are 
# a couple invalid entries in the accesslog. 
############################################################

rm -f accesslog* logstats* *stats.html tmp.txt

cp apache.accesslog accesslog.tmp
# split accesslog into 2 logs
total=`cat accesslog.tmp | wc -l`
nlines=`expr $total / 2`
nlines=`expr $nlines + 1`
split -l $nlines accesslog.tmp accesslog.

echo "bad line" >> tmp.txt
nlines=`expr $nlines + 1`
echo "Warning file accesslog line $nlines: Expected at least 10 fields, got 2" > exp_output.txt

echo "stuff stuff stuff stuff stuff stuff stuff stuff stuff stuff" >> tmp.txt
nlines=`expr $nlines + 1`
echo "Warning file accesslog line $nlines: Invalid date: stuff" >> exp_output.txt

echo "15.13.130.10 miss100 - [21/Aug/2003:17:26:45 -0700] \"GET /index.html HTTP/1.0\" 200 badcl \"-\" \"-\"" >> tmp.txt
nlines=`expr $nlines + 1`
echo "Warning file accesslog line $nlines: Invalid content length: badcl" >> exp_output.txt

cat accesslog.aa tmp.txt accesslog.ab > accesslog
rm -f accesslog.* tmp.txt

$update_log_stats accesslog .  > output.txt 2>&1


# remove filesize and numlines in both logstats.txt. 
# They will not match
cat apache.logstats.txt | grep -v -E "filesize|numlines" > tmp.txt
cat logstats.txt | grep -v -E "filesize|numlines" > tmp2.txt
mv tmp2.txt logstats.txt

check_results tmp.txt logstats.txt
check_results output.txt exp_output.txt

rm -f *output.txt
rm -f accesslog* logstats* *stats.html tmp.txt
echo "Pass"


############################################################
echo_noline "test bad entries (squid)......................"
# Check that update_log_stats still works even if there are 
# a couple invalid entries in the accesslog. 
############################################################

rm -f accesslog* logstats* *stats.html tmp.txt

cp squid.accesslog accesslog.tmp
# split accesslog into 2 logs
total=`cat accesslog.tmp | wc -l`
nlines=`expr $total / 2`
nlines=`expr $nlines + 1`
split -l $nlines accesslog.tmp accesslog.

echo "bad line" >> tmp.txt
nlines=`expr $nlines + 1`
echo "Warning file accesslog line $nlines: Expected at least 10 fields, got 2" > exp_output.txt

echo "stuff stuff stuff stuff stuff stuff stuff stuff stuff stuff" >> tmp.txt
nlines=`expr $nlines + 1`
echo "Warning file accesslog line $nlines: Invalid date: stuff" >> exp_output.txt

echo "1112387949.000 529 15.13.130.10 TCP_MISS/200 badcl GET http://host/index.html - DIRECT/15.0.110.12 text/html" >> tmp.txt
nlines=`expr $nlines + 1`
echo "Warning file accesslog line $nlines: Invalid content length: badcl" >> exp_output.txt

cat accesslog.aa tmp.txt accesslog.ab > accesslog
rm -f accesslog.* tmp.txt

$update_log_stats accesslog .  > output.txt 2>&1

# remove filesize and numlines in both logstats.txt. 
# They will not match
cat squid.logstats.txt | grep -v -E "filesize|numlines" > tmp.txt
cat logstats.txt | grep -v -E "filesize|numlines" > tmp2.txt
mv tmp2.txt logstats.txt

check_results tmp.txt logstats.txt
check_results output.txt exp_output.txt

rm -f *output.txt
rm -f accesslog* logstats* *stats.html tmp.txt
echo "Pass"


############################################################
echo_noline "test bad logstats.txt........................."
# Check that update_log_stats prints an error if it encounters
# bad entries in the logstats.txt file
############################################################

rm -f accesslog* logstats* *stats.html 
touch accesslog

echo "yearmonth 2005" > logstats.txt
echo "Error file ./logstats.txt line 1: yearmonth requires 2 arguments" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt

echo "yearmonth hello 9" > logstats.txt
echo "Error file ./logstats.txt line 1: Invalid year hello" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "yearmonth 2005 hello" > logstats.txt
echo "Error file ./logstats.txt line 1: Invalid month hello" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "lastupdate" > logstats.txt
echo "startdate 2005" >> logstats.txt
echo "Error file ./logstats.txt line 2: startdate requires 7 arguments" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt



echo "lastupdate" > logstats.txt
echo "startdate 0 0 0 0 0 0 0" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid startdate" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "lastupdate" > logstats.txt
echo "enddate 2005" >> logstats.txt
echo "Error file ./logstats.txt line 2: enddate requires 7 arguments" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "lastupdate" > logstats.txt
echo "enddate 0 0 0 0 0 0 0" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid enddate" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "lastupdate" > logstats.txt
echo "Error file ./logstats.txt line 1: No endlastupdate entry found" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "yearmonth 2005 11" > logstats.txt
echo "month_size 2" >> logstats.txt
echo "endyearmonth" >> logstats.txt
echo "Error file ./logstats.txt line 2: month_size requires 5 arguments" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt



echo "yearmonth 2005 11" > logstats.txt
echo "day_size99 0 0 0 0 0" >> logstats.txt
echo "endyearmonth" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid name day_size99" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "yearmonth 2005 11" > logstats.txt
echo "day_count99 0 0 0 0 0" >> logstats.txt
echo "endyearmonth" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid name day_count99" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "yearmonth 2005 11" > logstats.txt
echo "dayweek_size99 0 0 0 0 0" >> logstats.txt
echo "endyearmonth" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid name dayweek_size99" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "yearmonth 2005 11" > logstats.txt
echo "dayweek_count99 0 0 0 0 0" >> logstats.txt
echo "endyearmonth" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid name dayweek_count99" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "yearmonth 2005 11" > logstats.txt
echo "hour_size99 0 0 0 0 0" >> logstats.txt
echo "endyearmonth" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid name hour_size99" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "yearmonth 2005 11" > logstats.txt
echo "hour_count99 0 0 0 0 0" >> logstats.txt
echo "endyearmonth" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid name hour_count99" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt



echo "yearmonth 2005 11" > logstats.txt
echo "hello 0 0 0 0 0" >> logstats.txt
echo "endyearmonth" >> logstats.txt
echo "Error file ./logstats.txt line 2: Invalid name hello" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt


echo "yearmonth 2005 11" > logstats.txt
echo "month_size 0 0 0 0 0" >> logstats.txt
echo "Error file ./logstats.txt line 2: No endyearmonth entry found" > exp_output.txt
$update_log_stats accesslog . > output.txt 2>&1
check_results exp_output.txt output.txt

rm -f exp_output.txt output.txt
rm -f accesslog* logstats* *stats.html 
echo "Pass"


############################################################
echo_noline "test cant read accesslog ....................."
# Check that update_log_stats prints an error if it
# cant read the accesslog
############################################################

me=`whoami`
if [ "$me" != "root" ]
then
    echo "Skip"
    echo "This test must be run as root user"
else
    echo "hello" > accesslog
    chmod 600 accesslog

    useradd vicomp
    su vicomp -c "$update_log_stats accesslog ." > output.txt 2>&1
    userdel vicomp

    echo "Error: cannot open file accesslog: Permission denied" > exp_output.txt
    echo "Error: unknown log format file accesslog" >> exp_output.txt
    check_results exp_output.txt output.txt
    rm -f exp_output.txt output.txt
    rm -f accesslog* logstats* *stats.html 
    echo "Pass"
fi


############################################################
echo_noline "test cant write logstats.txt.................."
# Check that update_log_stats prints an error if it
# cant write the logstats.txt file
############################################################

me=`whoami`
if [ "$me" != "root" ]
then
    echo "Skip"
    echo "This test must be run as root user"
else
    cp apache.accesslog accesslog
    mkdir tempdir
    chmod 755 tempdir
    cp template.html tempdir/
    useradd vicomp
    su vicomp -c "$update_log_stats accesslog tempdir" > output.txt 2>&1
    userdel vicomp
    echo "Error: unable to create log statistics file tempdir/logstats.txt: Permission denied" > exp_output.txt 
    check_results exp_output.txt output.txt
    rm -f accesslog* logstats* *stats.html 
    rm -rf tempdir/
    rm -f exp_output.txt output.txt
    echo "Pass"
fi

############################################################
echo_noline "test no template.html........................."
# Check that update_log_stats works prints an error if the
# template.html file is missing.
############################################################

rm -f accesslog* logstats* *stats.html 
cp apache.accesslog accesslog
mv template.html template.html.orig
$update_log_stats accesslog . > output.txt 2>&1
echo "Error: unable to open file ./template.html, No such file or directory" > exp_output.txt
check_results exp_output.txt output.txt
mv template.html.orig template.html
rm -f exp_output.txt output.txt
rm -f accesslog* logstats* *stats.html 
echo "Pass"

###########################################################

rm -f apache.* squid.* *.html
