#! /usr/bin/python

# Copyright (c) 2003-2007 Visolve
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License version 2.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.



# Vicompress Functional Test Suite
#
# To run this test, simply do:
#
# ./vicompress_test.py <path to vicompress>
# ./vicompress_test.py /usr/local/vicompress/bin/vicompress
#
# The test cases and Pass/Fail result will be printed to the console.
#
# The HttpClient class simulates a client web browser.  It sends a request
# and returns the http reply header and reply body.
#
# The HttpServer class simulates an origin web server.  It spawns a separate
# thread to accept a new client, and send an http reply.
# 
# The VicompressTest class contains the actual test cases.  Each method
# name that starts with "test_" represents a single test case.  The
# run() method executes all the test cases.

import sys, os, os.path, socket, string, thread, gzip, zlib
import time, re, random, signal, StringIO

VICOMPRESS_PATH = "/usr/local/vicompress/bin/vicompress"
PROXY_PORT = 3128
SERVER_PORT = 3181

def myassert(value, msg):
    if (not value):
        raise AssertionError(msg)

def myprint(msg):
    sys.stdout.write(msg)
    sys.stdout.flush()

# The HttpClient class.  The methods are
# HttpClient(port) - Create a new client, and pass the vicompress port.
# connect()            - Connect to the vicompress proxy server
# do_request()         - Send an http request, and return the reply
# do_request_keepalive() - Send an http request using the same connection
# disconnect_request() - Disconnect while sending the http request
# disconnect_reply()   - Disconnect while reading the http reply
# timeout_request()    - Sleep/timeout while sending the http request
# timeout_reply()      - Sleep/timeout while reading the http reply

class HttpClient:

    # init(int port)
    def __init__(self, port):
        self.ip = "127.0.0.1"
        self.port = port
        self.proxy = None

    # bool is_full_header(string data)
    # Return 1 if the data contains a full HTTP header, or 0 otherwise
    def is_full_header(self, data):
        i = string.find(data, "\r\n\r\n")
        j = string.find(data, "\n\n")
        if (i > 0):
            return 1
        elif (j > 0):
            return 1
        else:
            return 0

    # Send the specified data to the proxy server
    def send(self, data):
        total = len(data)
        amount = 0
        while (amount < total):
            n = self.proxy.send(data[amount:])
            amount = amount + n

    # int get_content_length(string header)
    # Return the value of the "Content-Length:" header,
    # or -1 if the header doesn't exist.
    def get_content_length(self, header):
        m = re.search("(Content-length:) (\d+)", header, re.I)
        if (m == None):
            return -1
        return int(m.group(2))

    # string read_until_close(socket proxy)
    # Keep reading data from the socket until the
    # proxy closes the connection 
    def read_until_close(self):
        body = ""
        try:
            data = self.proxy.recv(4096)
            while (data != ""):
                body = body + data
                data = self.proxy.recv(4096)
        except socket.error:
            pass
        return body 


    # string read_amount(socket proxy, int amount)
    # Read the specified amount (in bytes) from the proxy
    # socket, and return a string containing the data read
    def read_amount(self, amount):
        body = ""
        try:
            data = self.proxy.recv(amount)
            body = body + data
            while (data != "" and len(body) < amount):
                data = self.proxy.recv(amount - len(body))
                body = body + data
        except socket.error:
            pass
        return body 

    # string read_chunked(socket proxy, string body)
    # Read a chunked-encoded http reply from the proxy socket.
    # The initial body that was already read is given as argument.
    # A chunked reply ends with "0   \r\n\r\n", so just look for
    # the \r\n\r\n.
    def read_chunked(self, body):
        try:
            while (body.find("\r\n\r\n") == -1):
                data = self.proxy.recv(8192)
                if data == "":
                    return body
                else:
                    body += data
            return body
        except socket.error:
            return body


    # string,string read_http_header(socket proxy)
    # Read the http reply header from the proxy.
    # Return the header string, and any extra data.
    def read_http_header(self):
        header = ""
        body = ""
        offset = 0
        try:
            header = self.proxy.recv(4096)
            if (header == ""):
                return header,body
            while (not self.is_full_header(header)):
                data = self.proxy.recv(4096)
                if (data == ""):
                    return header,body
                header = header + data
        except socket.error:
            pass

        if (string.find(header, "\r\n\r\n") > 0):
            offset = string.find(header, "\r\n\r\n") + 4
        elif (string.find(header, "\n\n") > 0): 
            offset = string.find(header, "\n\n") + 2
        else:
            return header,body
        body = header[offset:]
        header = header[0:offset]
        return header,body

    # string,string read_http_reply(socket proxy)
    # Read the http reply header and reply body from
    # the proxy
    def read_http_reply(self):
        header,body = self.read_http_header()
        cl = self.get_content_length(header)
        chunked = (header.find("Transfer-Encoding: chunked") > 0)
        continue100 = (header.find("HTTP/1.1 100 Continue") >= 0)
        if continue100:
            return header,""
        elif cl >= 0 and cl <= len(body):
            return header,body
        elif cl >= 0:
            body = body + self.read_amount(cl - len(body)) 
        elif chunked:
            body = self.read_chunked(body)
        else:
            body = body + self.read_until_close()
        return header,body 


    # Connect to the proxy server.  The IP address and port were
    # given in the constructor, __init__().
    def connect(self):
        assert(self.proxy == None)
        self.proxy = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.proxy.connect((self.ip, self.port))

    # Close the connection with the proxy
    def close(self):
        if (self.proxy != None):
            self.proxy.close()
            self.proxy = None

    # header,body do_request(string request)
    # - Close any previous connection
    # - Connect to the proxy server
    # - Send the specified http request
    # - Return the http reply header and reply body (as strings)
    # - Close the connection
    def do_request(self, request):
        assert(self.proxy == None)
        self.connect()
        self.send(request)
        reply,body = self.read_http_reply()
        self.proxy.close()
        self.proxy = None
        return reply,body

    # header,body do_request(string request)
    # - Send the specified http request
    # - Return the http reply header and reply body (as strings)
    # - Keep the connection alive
    def do_request_keepalive(self, request, first):
        if first:
            self.connect()
        self.send(request)
        reply,body = self.read_http_reply()
        return reply,body


    # void disconnect_request(string request)
    # - Connect to the proxy server
    # - Send half of the http request
    # - Disconnect from the proxy
    def disconnect_request(self, request):
        self.connect()
        self.send( request[0:len(request)/2] )
        time.sleep(0.05)
        self.proxy.close()
        self.proxy = None

    # void disconnect_reply(string request, int amount)
    # - Connect to the proxy server
    # - Send the http request
    # - Read only the given amount of bytes from the proxy
    # - Disconnect from the proxy
    def disconnect_reply(self, request, amount):
        self.connect()
        self.send(request)
        data = self.proxy.recv(amount)
        self.proxy.close()
        self.proxy = None
        time.sleep(1)


    # void timeout_request(string request)
    # - Connect to the proxy server
    # - Send half of the http request
    # - Sleep for 60 seconds.  The proxy should timeout
    #   and close the connection.
    # - Disconnect from the proxy
    def timeout_request(self, request):
        self.connect()
        self.send( request[0:len(request)/2] )
        time.sleep(60)
        self.proxy.close()
        self.proxy = None


    # void disconnect_reply(string request, int amount)
    # - Connect to the proxy server
    # - Send the http request
    # - Read only the given amount of bytes from the proxy
    # - Sleep for 60 seconds.  The proxy should timeout
    #   and close the connection.
    # - Disconnect from the proxy
    def timeout_reply(self, request, amount):
        self.connect()
        self.send(request)
        data = self.proxy.recv(amount)
        time.sleep(60)
        self.proxy.close()
        self.proxy = None


## End class HttpClient


# The HttpServer class.  The methods are
# HttpServer(port)     - Create a new http server.  Spawn a thread that
#                        listens on the given port, and accepts clients.
# set_reply(string)    - Set the reply data (http header and body) to
#                        send back to any client requests.  Close the 
#                        connection once the data is sent.
# set_timeout(bool)    - If true, the server will sleep and timeout
#                        after sending the reply data, for a single request.
#                        On subsequent requests, the timeout goes back to
#                        false
class HttpServer:

    # bool is_full_header(string data)
    # Return 1 if the data contains a full HTTP header, or 0 otherwise
    def is_full_header(self, data):
        i = string.find(data, "\r\n\r\n")
        j = string.find(data, "\n\n")
        if (i > 0):
            return 1
        elif (j > 0):
            return 1
        else:
            return 0

    # int get_content_length(string header)
    # Return the value of the "Content-Length:" header,
    # or -1 if the header doesn't exist.
    def get_content_length(self, header):
        m = re.search("(Content-length:) (\d+)", header, re.I)
        if (m == None):
            return -1
        return int(m.group(2))


    # string get_url(string request)
    # Given the http request, return the URL requested
    def get_url(self, request):
        i = string.find(request, "\n")
        line = request[0:i-1]
        parts = string.split(line)
        url = parts[1]
        url = url[1:]
        return url


    # string read_request_body(int total)
    # Read the specified number of bytes from the proxy
    def read_request_body(self, proxy, total):
        if (total <= 0):
            return ""
        body = proxy.recv(total)
        while (len(body) < total):
            data = proxy.recv(total - len(body))
            body = body + data
        return body


    # string read_request()
    # Read the http request from the proxy
    def read_request(self, proxy):
        req = proxy.recv(4096)
        if (req == ""):
            return ""
        while (not self.is_full_header(req)): 
            data = proxy.recv(4096)
            if (data == ""):
                return req
            req = req + data
        n = string.find(req, "\r\n\r\n") + 4
        if n < 4:
            n = string.find(req, "\n\n") + 2
        if n < 2:
            n = len(req)
        cl = self.get_content_length(req)
        if cl == 0:
            return req

        if string.find(req, "Expect: 100-continue") > 0:
            proxy.send("HTTP/1.1 100 Continue\r\n\r\n")

        req = req + self.read_request_body(proxy, n + cl - len(req))
        return req


    # void write_reply(string data)
    # Write the given reply back to the proxy.  Vicompress may
    # disconnect prematurely if the http client also disconnects
    # prematurely.  So catch any socket exceptions while writing.
    # Write in small 1K chunks to mimic real-world network behavior.
    def write_reply(self, proxy, data):
        try:
            total = len(data)
            offset = 0
            while (offset < total):
                amount = 1000
                if (total - offset < amount):
                    amount = total - offset
                n = proxy.send(data[offset : offset+amount])
                offset = offset + n
        except socket.error:
            pass

    # init(int port)
    def __init__(self, port):
        self.port = port
        self.listen = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.listen.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.listen.bind(("", port))
        self.listen.listen(20)
        self.reply = ""    # the reply to send back to clients
        self.timeout = 0   # whether to sleep/timeout
        self.do_close = 0  # Set to 1 to stop the HttpServer
        self.keepalive = 0 # Set to 1 to use keep-alive connections
        thread.start_new_thread(self.server_thread, ())


    # Set the reply data (http header and body) to send 
    # back to clients
    def set_reply(self, reply):
        self.reply = reply

    # Set the value of the timeout to true/false.  If true,
    # then wait 60 seconds after sending the client data
    # before closing the connection.
    def set_timeout(self, value):
        self.timeout = value

    # void server_thread()
    # The main thread which handles client requests.
    # - Accept a new client request 
    # - Read the full http request
    # - Write self.reply back to the client
    # - If self.timeout is true, sleep for 60 seconds.
    #   Else, close the connection.

    def server_thread(self):
        while (1):
            client,addr = self.listen.accept()
            if self.do_close == 1:
                client.close()
                self.listen.close()
                self.listen = None
                return
            req = self.read_request(client)
            while req != "":
                url = self.get_url(req)
                self.write_reply(client, self.reply)
                if self.timeout == 1:
                    time.sleep(60)
                    self.timeout = 0
                if self.keepalive == 1:
                    req = self.read_request(client)
                else:
                    req = ""
            client.close()


    # Close this server.  We wake up the server_thread
    # by connecting to the server.  The server_thread will
    # see that do_close is true, and will exit.
    def close(self):
        self.do_close = 1
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.port))
        time.sleep(1)
        s.close()

## End class HttpServer

# The VicompressTest class is the main class which executes the
# functional tests.

class VicompressTest:

    # void start_vicompress(string config)
    # Start the vicompress server.
    # config is the vicompress.conf data to use.
    def start_vicompress(self, config):
        f = open("config.temp", "w")
        f.write(config)
        f.close()
        ret = 0
        if (os.uname()[0][0:6] == "CYGWIN"):
            ret = os.system(VICOMPRESS_PATH + " config.temp &")
        else:
            ret = os.system(VICOMPRESS_PATH + " config.temp")
        if (ret != 0):
            myprint("ERROR: Unable to start vicompress\n")
            sys.exit(1)
        time.sleep(2.0)


    # void stop_vicompress()
    # Stop the vicompress server. 
    def stop_vicompress(self):
        os.system("ps -ef | grep " + VICOMPRESS_PATH + " | grep -v grep | grep -v test > pids.txt")
        f = open("pids.txt", "r")
        line = f.readline()
        f.close()
        os.unlink("pids.txt")
        if line != "" and line != "\n":
            pid = int( line.split()[1] )
            if (os.uname()[0][0:6] == "CYGWIN"):
                os.kill(pid, signal.SIGKILL)
            else:
                os.kill(pid, signal.SIGINT)
        time.sleep(2)
        if os.path.exists("config.temp"):
            os.unlink("config.temp")

    # string get_random_body(int amount)
    # Return a string "amount" bytes long containing random
    # ascii numbers.  The amount must be a multiple of 50.
    def get_random_body(self, amount):
        myassert((amount % 50) == 0, "get_random_body failed 1")
        result = ""
        for linenum in range(0, amount/50):
            line = str( random.randint(100000000, 999999999) ) + " " + \
                   str( random.randint(100000000, 999999999) ) + " " + \
                   str( random.randint(100000000, 999999999) ) + " " + \
                   str( random.randint(100000000, 999999999) ) + " " + \
                   str( random.randint(100000000, 999999999) ) + "\n"
            result = result + line
        myassert(len(result) == amount, "get_random_body failed 2")
        return result



    # string gunzip(string gzipped_data)
    # Return a string containing the gunzipped data
    def gunzip(self, gzipped_data):
        try:
            file = StringIO.StringIO(gzipped_data)
            gzfile = gzip.GzipFile(None, "r", 9, file)
            data = gzfile.read(1024 * 1024) 
            gzfile.close()
            return data
        except IOError, e:
            return "Data not in gzip format: " + str(e)
        except zlib.error, e:
            return "Data not in gzip format: " + str(e)

    # string chunk_encode(string data)
    # Split the data into two parts.  Chunk encode the parts, and
    # return the full chunk-encoded reply, including the 0\r\n footer.
    def chunk_encode(self, data):
        len1 = len(data) / 3
        len2 = len(data) - len1
        data1 = data[0:len1]
        data2 = data[len1:]
        return hex(len1)[2:] + "\r\n" + data1 + "\r\n" + \
               hex(len2)[2:] + "\r\n" + data2 + "\r\n" + \
               "0\r\n\r\n"

    # string chunk_decode(string chunkeddata)
    # Decode the chunked data passed as argument.
    # The chunked data has the format
    # <size>\r\n
    # <data>\r\n
    # <size>\r\n
    # <data>\r\n
    #  ...
    def chunk_decode(self, data):
        parts = data.split("\r\n")
        result = ""
        try:
            i = 0
            while i < len(parts):
                chunksize = int(parts[i], 16)
                if chunksize == 0:
                    return result
                i += 1
                result += parts[i]
                chunksize -= len(parts[i])
                i += 1
                while chunksize > 0:
                    result += "\r\n"
                    chunksize -= 2
                    result += parts[i]
                    chunksize -= len(parts[i])
                    i += 1

            return result
        except ValueError:
            return "Bad chunk-encoding"
        except IndexError:
            return "Bad chunk-encoding"


    # (pass,fail) run_testnames(string[] testnames)
    def run_testnames(self, testnames):
        totalpass = 0
        totalfail = 0
        for name in testnames:
            method = getattr(self, name)
            displayname = name.replace("_", " ")
            dots = "." * (55 - len(displayname))
            myprint(displayname + dots)
            try:
                method()
                myprint("Pass\n")
                totalpass += 1
            except AssertionError, err:
                myprint("Fail\n")
                totalfail += 1
                myprint(("-" * 55) + "\n")
                myprint(str(err) + "\n")
                myprint(("-" * 55) + "\n")
        return (totalpass, totalfail)

    def run(self, fast):

        squid_config = """
webserver 127.0.0.1 3181
listen 127.0.0.1 3128
outgoingip 127.0.0.1
enable_compression yes
enable_caching yes
cache_memory 50
max_cacheditem_size 100
cache_expires 240
enable_dns_caching no
dns_expires 0
hostheader vicompresstest.com
logformat squid
accesslog accesslog 
errorlog errorlog
enable_debug yes
"""

        apache_config = """
webserver 127.0.0.1 3181
listen 127.0.0.1 3128
outgoingip 127.0.0.1
enable_compression yes
enable_caching yes
cache_memory 50
max_cacheditem_size 100
cache_expires 240
enable_dns_caching no
dns_expires 0
hostheader vicompresstest.com
logformat apache
accesslog accesslog 
errorlog errorlog
enable_debug yes
"""

        nocache_config = """
webserver 127.0.0.1 3181
listen 127.0.0.1 3128
enable_compression yes
enable_caching no
hostheader domain.com
logformat apache
accesslog accesslog
errorlog errorlog
enable_debug yes
"""
        nogzip_config = """
webserver 127.0.0.1 3181
listen 127.0.0.1 3128
enable_compression no
enable_caching yes
cache_memory 50
max_cacheditem_size 100
cache_expires 240
enable_dns_caching no
dns_expires 0
hostheader vicompresstest.com
logformat apache
accesslog accesslog
errorlog errorlog
enable_debug yes
"""

        withdns_apache_config = """
listen 0.0.0.0 3128
enable_compression yes
enable_caching yes
cache_memory 50
max_cacheditem_size 100
cache_expires 240
enable_dns_caching yes
dns_expires 48
logformat apache
accesslog accesslog
errorlog errorlog
enable_debug yes
"""

        withdns_squid_config = """
listen 0.0.0.0 3128
enable_compression yes
enable_caching yes
cache_memory 50
max_cacheditem_size 100
cache_expires 240
enable_dns_caching yes
dns_expires 48
logformat squid
accesslog accesslog
errorlog errorlog
enable_debug yes
"""


        if (os.path.exists("accesslog")):
            os.unlink("accesslog")
        if (os.path.exists("errorlog")):
            os.unlink("errorlog")

        self.httpclient = HttpClient(PROXY_PORT)
        self.httpserver = HttpServer(SERVER_PORT)

        testnames = dir(VicompressTest)
        testnames = [ name for name in testnames \
                      if name[0:4] == "test" and \
                         name.find("withdns") == -1 and \
                         name.find("loadbalance") == -1 ]
        if fast:
            testnames = [ name for name in testnames if name.find("timeout") == -1 ]

        totalpass = 0
        totalfail = 0

        myprint("\nTesting with logformat squid\n")
        myprint(  "============================\n")
        self.stop_vicompress()
        self.start_vicompress(squid_config)
        (ppass, fail) = self.run_testnames(testnames)
        totalpass += ppass
        totalfail += fail

        myprint("\nTesting with logformat apache\n")
        myprint(  "=============================\n")
        self.stop_vicompress()
        self.start_vicompress(apache_config)
        (ppass, fail) = self.run_testnames(testnames)
        totalpass += ppass
        totalfail += fail

        myprint("\nTesting with enable_caching=no\n")
        myprint(  "=============================\n")
        self.stop_vicompress()
        self.start_vicompress(nocache_config)
        nocache_names = [ name for name in testnames if name.find("cach") == -1 ]
        (ppass, fail) = self.run_testnames(nocache_names)
        totalpass += ppass
        totalfail += fail

        myprint("\nTesting with enable_compression=no\n")
        myprint(  "===================================\n")
        self.stop_vicompress()
        self.start_vicompress(nogzip_config)
        nogzip_names = [ name for name in testnames if name.find("gzip") == -1 ]
        (ppass, fail) = self.run_testnames(nogzip_names)
        totalpass += ppass
        totalfail += fail

        dns_tests = [ name for name in dir(VicompressTest) \
                      if name[0:4] == "test" and \
                         name.find("withdns") != -1 ]

        myprint("\nTesting with dns, logformat apache\n")
        myprint(  "==================================\n")
        self.stop_vicompress()
        self.start_vicompress(withdns_apache_config)
        (ppass, fail) = self.run_testnames(dns_tests)
        totalpass += ppass
        totalfail += fail


        myprint("\nTesting with dns, logformat squid\n")
        myprint(  "=================================\n")
        self.stop_vicompress()
        self.start_vicompress(withdns_squid_config)
        (ppass, fail) = self.run_testnames(dns_tests)
        totalpass += ppass
        totalfail += fail

        self.stop_vicompress()

        myprint("\nTesting load balancing features\n")
        myprint(  "===============================\n")
        testnames = [ name for name in dir(VicompressTest) \
                      if name[0:4] == "test" and \
                         name.find("loadbalance") != -1 ]
        if fast:
            testnames = [name for name in testnames if name.find("failover") == -1 ]

        self.httpserver2 = HttpServer(3182)
        (ppass, fail) = self.run_testnames(testnames)
        totalpass += ppass
        totalfail += fail


        myprint("\n================\n")
        myprint(  "Total Pass: " + str(totalpass) + "\n")
        myprint(  "Total Fail: " + str(totalfail) + "\n")


    ##########################################
    #  Test Cases
    ##########################################


    # Test a request with no content length in the reply.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_get_request_no_contentlength(self):
        request = \
"""GET /test/get/request/no/contentlength HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(20000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a request with a "Content-Length" in the http reply
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_get_request_with_contentlength(self):
        request = \
"""GET /test/get/request/with/contentlength HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a request where the content-length of the reply is 0.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_get_request_zero_contentlength(self):
        request = \
"""GET /test/get/request/zero/contentlength HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 0

"""

        body = ""
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a request where the URL is greater than LARGE_URL_SIZE (128).
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_get_request_bigurl(self):
        request = \
"""GET /test/get/request/bigurl/longggggggggg123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789 HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a proxy http request, where the first line contains
    # http://hostname/path instead of just /path.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_proxy_request(self):
        request = \
"""GET http://somedomain.com/proxy/request HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)

    # Test a request which doesn't contain a "Host:" header.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_no_host(self):
        request = \
"""GET /test/no/host HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)

    # Test a request where the reply body is large, greater than 100K.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_large_reply(self):
        request = \
"""GET /test/large/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 123450

"""

        body = self.get_random_body(123450)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a POST http request with a small request body.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_small_post_request(self):
        request = \
"""POST /test/small/post/request HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Content-length: 66
Host: vicompresstest.com

send=submit&reply=submit&user=person&someother=value=anothervalue
"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test an http POST request with a large request body, greater than 2K.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_large_post_request(self):
        request = \
"""POST /test/large/post/request HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Content-length: 5000

"""
        request = request + "a" * 5000

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a PUT http request that expects a 100 Continue body.
    # Check that vicompress forwards the "HTTP/1.1 100 Continue" reply header.
    # Check that vicompress continues reading the request body, and then
    # proceeds with reading the http reply.
    def test_100_continue_request(self):
        request = \
"""PUT /test/100/continue/request HTTP/1.1
Expect: 100-continue
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Content-length: 40
Host: vicompresstest.com

"""
        requestbody = "1234567890123456789012345678901234567890"

        reply100 = "HTTP/1.1 100 Continue\r\n\r\n"

        reply = \
"""HTTP/1.1 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body1 = self.httpclient.do_request_keepalive(request, 1)
        myassert(header == reply100, "== 1st expected header ==\n" + reply100 + \
                 "\n== 1st actual header ==\n" + header)
        header,body2 = self.httpclient.do_request_keepalive(requestbody, 0)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)
        self.httpclient.close()


    # Send two requests, where the connection is kept alive between
    # the two requests.  Check that the reply body returned by vicompress 
    # matches the reply body passed to the httpserver.
    def test_keepalive_request(self):
        request1 = \
"""GET /test/keepalive/request1 HTTP/1.0
Connection: keep-alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""
        request2 = \
"""GET /test/keepalive/request2 HTTP/1.0
Connection: Keep-Alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply1 = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""
        reply2 = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 6150

"""

        body1 = self.get_random_body(5000)
        body2 = self.get_random_body(6150)
        self.httpserver.set_reply(reply1 + body1)
        header,body = self.httpclient.do_request_keepalive(request1, 1)
        myassert(body == body1, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body1)
        self.httpserver.set_reply(reply2 + body2)
        header,body = self.httpclient.do_request_keepalive(request2, 0)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)
        self.httpclient.close()


    # Send two requests, where the connection is kept alive between
    # the two requests.  Use an http proxy request rather than a
    # standard http request. Check that the reply body returned by 
    # vicompress matches the reply body passed to the httpserver.
    def test_keepalive_request_proxy(self):
        request1 = \
"""GET http://somedomain.com/test/keepalive/request1 HTTP/1.0
Proxy-Connection: keep-alive
Date: Wed, 30 Oct 2002 12:43:03 GMT

"""
        request2 = \
"""GET http://somedomain.com/test/keepalive/request2 HTTP/1.0
Proxy-Connection: Keep-Alive
Date: Wed, 30 Oct 2002 12:43:03 GMT

"""

        reply1 = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""
        reply2 = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 6150

"""

        body1 = self.get_random_body(5000)
        body2 = self.get_random_body(6150)
        self.httpserver.set_reply(reply1 + body1)
        header,body = self.httpclient.do_request_keepalive(request1, 1)
        myassert(body == body1, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body1)
        self.httpserver.set_reply(reply2 + body2)
        header,body = self.httpclient.do_request_keepalive(request2, 0)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)
        self.httpclient.close()


    # Send two requests, where the connection is kept alive between
    # the two requests.  The server-side connection will also be kept alive.
    # Check that the reply body returned by vicompress 
    # matches the reply body passed to the httpserver.
    def test_server_keepalive_request(self):
        request1 = \
"""GET /test/server/keepalive/request1 HTTP/1.0
Connection: keep-alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""
        request2 = \
"""GET /test/server/keepalive/request2 HTTP/1.0
Connection: Keep-Alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply1 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Content-type: text/html
Content-length: 5000

"""
        reply2 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Content-type: text/html
Content-length: 6150

"""

        body1 = self.get_random_body(5000)
        body2 = self.get_random_body(6150)
        self.httpserver.keepalive = 1
        self.httpserver.set_reply(reply1 + body1)
        header,body = self.httpclient.do_request_keepalive(request1, 1)
        myassert(body == body1, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body1)
        self.httpserver.set_reply(reply2 + body2)
        header,body = self.httpclient.do_request_keepalive(request2, 0)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)
        self.httpclient.close()
        self.httpserver.keepalive = 0



    # Send two requests, where the connection is kept alive between
    # the two requests.  The server-side connection will also be kept alive.
    # Use an http proxy request rather than a
    # standard http request. Check that the reply body returned by 
    # vicompress matches the reply body passed to the httpserver.
    def test_server_keepalive_request_proxy(self):
        request1 = \
"""GET http://somedomain.com/test/server/keepalive/request1 HTTP/1.0
Proxy-Connection: keep-alive
Date: Wed, 30 Oct 2002 12:43:03 GMT

"""
        request2 = \
"""GET http://somedomain.com/test/server/keepalive/request2 HTTP/1.0
Proxy-Connection: Keep-Alive
Date: Wed, 30 Oct 2002 12:43:03 GMT

"""

        reply1 = \
"""HTTP/1.1 200 OK
Connection: keep-alive
Content-type: text/html
Content-length: 5000

"""
        reply2 = \
"""HTTP/1.1 200 OK
Connection: keep=alive
Content-type: text/html
Content-length: 6150

"""

        body1 = self.get_random_body(5000)
        body2 = self.get_random_body(6150)
        self.httpserver.keepalive=1
        self.httpserver.set_reply(reply1 + body1)
        header,body = self.httpclient.do_request_keepalive(request1, 1)
        myassert(body == body1, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body1)
        self.httpserver.set_reply(reply2 + body2)
        header,body = self.httpclient.do_request_keepalive(request2, 0)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)
        self.httpclient.close()
        self.httpserver.keepalive=0


    # Test that vicompress recognizes the end of chunked replies.
    # Send two requests, where the connection is kept alive between
    # the two requests.  The server-side connection will also be kept alive.
    # The reply will have transfer-encoding chunked.
    # Check that the reply body returned by vicompress 
    # matches the reply body passed to the httpserver.
    def test_chunked_reply(self):
        request1 = \
"""GET /test/chunked/reply HTTP/1.1
Connection: keep-alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""
        request2 = \
"""GET /test/chunked/reply2 HTTP/1.1
Connection: Keep-Alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply1 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Transfer-Encoding: chunked
Content-type: text/html

"""
        reply2 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Transfer-Encoding: chunked
Content-type: text/html

"""

        body1 = self.chunk_encode( self.get_random_body(5000) )

        body2 = self.chunk_encode( self.get_random_body(6000) )

        self.httpserver.keepalive = 1
        self.httpserver.set_reply(reply1 + body1)
        header,body = self.httpclient.do_request_keepalive(request1, 1)
        myassert(body == body1, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body1)
        self.httpserver.set_reply(reply2 + body2)
        header,body = self.httpclient.do_request_keepalive(request2, 0)
        myassert(body == body2, "== 2nd expected body ==\n" + body2 + \
                 "\n== 2nd actual body ==\n" + body)
        self.httpclient.close()
        self.httpserver.keepalive = 0


    # Test that vicompress correctly gzips chunk-encoded replies.
    # Send two requests, where the connection is kept alive between
    # the two requests.  The server-side connection will also be kept alive.
    # The reply will have transfer-encoding chunked.  
    # Check that the gunzipped reply body returned by vicompress 
    # matches the reply body passed to the httpserver.
    def test_gzip_chunked_reply(self):
        request1 = \
"""GET /test/gzip/chunked/reply HTTP/1.1
Connection: keep-alive
Accept-Encoding: gzip
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""
        request2 = \
"""GET /test/gzip/chunked/reply2 HTTP/1.1
Accept-Encoding: gzip
Connection: Keep-Alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply1 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Transfer-Encoding: chunked
Content-type: text/html

"""
        reply2 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Transfer-Encoding: chunked
Content-type: text/html

"""

        body1 = self.get_random_body(50000)
        body2 = self.get_random_body(60000)
        chunkedbody1 = self.chunk_encode(body1)
        chunkedbody2 = self.chunk_encode(body2)

        self.httpserver.keepalive = 1
        self.httpserver.set_reply(reply1 + chunkedbody1)
        header,body = self.httpclient.do_request_keepalive(request1, 1)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body = self.chunk_decode(body)
        body = self.gunzip(body)
        myassert(body == body1, "== 1st expected body ==\n" + body1 + \
                 "\n== 1st actual body ==\n" + body)
        self.httpserver.set_reply(reply2 + chunkedbody2)
        header,body = self.httpclient.do_request_keepalive(request2, 0)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body = self.chunk_decode(body)
        body = self.gunzip(body)
        myassert(body == body2, "== 2nd expected body ==\n" + body2 + \
                 "\n== 2nd actual body ==\n" + body)
        self.httpclient.close()
        self.httpserver.keepalive = 0


    # Test that vicompress chunk-encodes replies that have no content-length.
    # Send two requests, where the connection is kept alive between
    # the two requests.  The reply will have no content length.
    # Check that the reply sent by vicompress is chunk-encoded.
    # Check that the reply body returned by vicompress 
    # matches the reply body passed to the httpserver.
    def test_chunk_encode_reply(self):
        request1 = \
"""GET /test/chunk/encode/reply HTTP/1.1
Connection: keep-alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""
        request2 = \
"""GET /test/chunk/encode/reply2 HTTP/1.1
Connection: Keep-Alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply1 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Content-type: text/html

"""
        reply2 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Content-type: text/html

"""

        body1 = self.get_random_body(5000)
        body2 = self.get_random_body(6000)

        self.httpserver.set_reply(reply1 + body1)
        header,body = self.httpclient.do_request_keepalive(request1, 1)
        myassert(header.find("Transfer-Encoding: chunked") >= 0,
                 "== Bad header ==\n" + header)
        body = self.chunk_decode(body)
        myassert(body == body1, "== 1st expected body ==\n" + body1 + \
                 "\n== 1st actual body ==\n" + body)
        self.httpserver.set_reply(reply2 + body2)
        header,body = self.httpclient.do_request_keepalive(request2, 0)
        myassert(header.find("Transfer-Encoding: chunked") >= 0,
                 "== Bad header ==\n" + header)
        body = self.chunk_decode(body)
        myassert(body == body2, "== 2nd expected body ==\n" + body2 + \
                 "\n== 2nd actual body ==\n" + body)
        self.httpclient.close()
        self.httpserver.keepalive = 0


    # Test that vicompress gzips and chunk-encodes the reply correctly.
    # Send two requests, where the connection is kept alive between
    # the two requests.  The reply will have no content length.
    # Check that the reply sent by vicompress is chunk-encoded.
    # Check that the reply sent by vicompress is gzipped.
    # Check that the reply body returned by vicompress 
    # matches the reply body passed to the httpserver.
    def test_gzip_and_chunk_encode_reply(self):
        request1 = \
"""GET /test/gzip/and/chunk/encode/reply HTTP/1.1
Connection: keep-alive
Accept-Encoding: gzip
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""
        request2 = \
"""GET /test/gzip/and/chunk/encode/reply2 HTTP/1.1
Accept-Encoding: gzip
Connection: Keep-Alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply1 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Content-type: text/html

"""
        reply2 = \
"""HTTP/1.1 200 OK
Connection: Keep-alive
Content-type: text/html

"""
        body1 = self.get_random_body(50000)
        body2 = self.get_random_body(60000)

        self.httpserver.set_reply(reply1 + body1)
        header,body = self.httpclient.do_request_keepalive(request1, 1)
        myassert(header.find("Transfer-Encoding: chunked") >= 0,
                 "== Bad header ==\n" + header)
        myassert(header.find("Content-Encoding: gzip") >= 0,
                 "== Bad header ==\n" + header)
        body = self.chunk_decode(body)
        body = self.gunzip(body)
        myassert(body == body1, "== 1st expected body ==\n" + body1 + \
                 "\n== 1st actual body ==\n" + body + "\n==header==\n" + header + "== end ==\n")
        self.httpserver.set_reply(reply2 + body2)
        header,body = self.httpclient.do_request_keepalive(request2, 0)
        myassert(header.find("Transfer-Encoding: chunked") >= 0,
                 "== Bad header ==\n" + header)
        myassert(header.find("Content-Encoding: gzip") >= 0,
                 "== Bad header ==\n" + header)
        body = self.chunk_decode(body)
        body = self.gunzip(body)
        myassert(body == body2, "== 2nd expected body ==\n" + body2 + \
                 "\n== 2nd actual body ==\n" + body)
        self.httpclient.close()
        self.httpserver.keepalive = 0


    # Send an http request with a bad method.  Vicompress will still allow
    # the request to go through.  Check that the reply body returned by 
    # vicompress matches the reply body passed to the httpserver.
    def test_bad_request_method(self):
        request = \
"""BAD /test/bad/request/method HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Send a bad http request with no "HTTP/1.x" version.  Check that
    # vicompress returns the proper error reply.
    def test_bad_request_no_version(self):
        request = \
"""GET /test/bad/request/no/version
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        header = "HTTP/1.1 400 Bad Request"
        header2,body = self.httpclient.do_request(request)
        header2 = header2.split("\r\n")[0]
        myassert(header == header2, "== expected header ==\n" + header + \
                 "\n== actual header ==\n" + header2)


    # Send a bad http request that is too big (> 4K).  Check that
    # vicompress returns the proper error reply.
    def test_bad_request_too_big(self):
        request = \
"""GET /test/bad/request/too/big
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
"""

        header = "SomeHeader: large value " + ("0" * 500) + "\n"
        header = header * 20
        request = request + header + "\n"

        header = "HTTP/1.1 400 Bad Request"
        header2,body = self.httpclient.do_request(request)
        header2 = header2.split("\r\n")[0]
        myassert(header == header2, "== expected header ==\n" + header + \
                 "\n== actual header ==\n" + header2)


    # Send a bad http request that has too many lines.  Check that
    # vicompress returns the proper error reply.
    def test_bad_request_too_many_lines(self):
        request = \
"""GET /test/bad/request/too/many/lines HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
"""

        header = "myheader: value\n"
        header = header * 100
        request = request + header + "\n"

        header = "HTTP/1.1 400 Bad Request"
        header2,body = self.httpclient.do_request(request)
        header2 = header2.split("\r\n")[0]
        myassert(header == header2, "== expected header ==\n" + header + \
                 "\n== actual header ==\n" + header2)


    # Send a bad http request that has header lines with mising colons.  
    # Vicompress will still forward the request on.  Check that the reply
    # body returned by vicompress matches the reply body passed to the
    # httpserver.
    def test_bad_request_no_colon(self):
        request = \
"""GET /test/bad/request/no/colon HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
bad header no colon
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Send a bad http reply missing the "HTTP/1.x" version.
    # Vicompress will always forward http replies, even invalid ones. 
    # Check that the reply body returned by vicompress matches the reply 
    # body passed to the httpserver.
    def test_bad_reply_no_version(self):
        request = \
"""GET /test/bad/reply/no/version HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""not a valid reply
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)

    # Send a bad http reply containing header lines without a colon.
    # Vicompress will always forward http replies, even invalid ones. 
    # Check that the reply body returned by vicompress matches the reply 
    # body passed to the httpserver.
    def test_bad_reply_no_colon(self):
        request = \
"""GET /test/bad/reply/no/colon HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
some header no colon
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)

    # Send a bad http reply where the reply header is > 4K.
    # Vicompress will always forward http replies, even invalid ones. 
    # Check that the reply body returned by vicompress matches the reply 
    # body passed to the httpserver.
    def test_bad_reply_toobig(self):
        request = \
"""GET /test/bad/reply/too/big HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000
"""

        header = "SomeHeader: large value " + ("0" * 500) + "\n"
        header = header * 20
        reply = reply + header + "\n"

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Send a bad http reply where the reply has > 100 lines.
    # Vicompress will always forward http replies, even invalid ones. 
    # Check that the reply body returned by vicompress matches the reply 
    # body passed to the httpserver.
    def test_bad_reply_too_many_lines(self):
        request = \
"""GET /test/bad/reply/too/many/lines HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000
"""

        header = "header: value\n" 
        header = header * 100
        reply = reply + header + "\n"

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)

    # Test that vicompress still works after a client disconnects while
    # sending the http request.  Disconnect the client while sending 
    # the request.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_disconnect_request(self):
        request = \
"""GET /test/client/disconnect/request HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpclient.disconnect_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)

    # Test that vicompress still works after a client disconnects while
    # reading the http reply.  Disconnect the client while reading 
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_disconnect_reply(self):
        request = \
"""GET /test/client/disconnect/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        self.httpclient.disconnect_reply(request, 20)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client disconnects while
    # reading a cached reply.  Disconnect the client while reading 
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_disconnect_cached_reply(self):
        request = \
"""GET /test/client/disconnect/cached/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-type: image/png
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpclient.disconnect_reply(request, 20)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that vicompress still works after a client disconnects while
    # reading a gzipped reply.  Disconnect the client while reading 
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_disconnect_gzipped_reply(self):
        request = \
"""GET /test/client/disconnect/gzipped/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Accept-Encoding: compress, gzip;
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 24000

"""

        body = self.get_random_body(24000)
        self.httpserver.set_reply(reply + body)
        self.httpclient.disconnect_reply(request, 20)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client disconnects while
    # reading a chunked http reply.  Disconnect the client while reading 
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_disconnect_chunked_reply(self):
        request = \
"""GET /test/client/disconnect/chunked/reply HTTP/1.1
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Transfer-Encoding: chunked

"""

        body = self.chunk_encode(self.get_random_body(50000))
        self.httpserver.set_reply(reply + body)
        self.httpclient.disconnect_reply(request, 400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client disconnects while
    # reading an http reply that vicompress is chunk-encoding.  Disconnect 
    # the client while reading the reply.  Then, send a normal http request.  
    # Check that the reply body returned by vicompress matches the reply body 
    # passed to the http server.
    def test_client_disconnect_chunk_encode_reply(self):
        request = \
"""GET /test/client/disconnect/chunk/encode/reply HTTP/1.1
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        self.httpclient.disconnect_reply(request, 400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client disconnects while
    # reading a chunked http reply that vicompress is gzipping.  Disconnect 
    # the client while reading the reply.  Then, send a normal http request.  
    # Check that the reply body returned by vicompress matches the reply body 
    # passed to the http server.
    def test_client_disconnect_chunked_gzipped_reply(self):
        request = \
"""GET /test/client/disconnect/chunked/gzipped/reply HTTP/1.1
Accept-Encoding: gzip
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.1 200 OK
Connection: Close
Content-type: text/html
Transfer-Encoding: chunked

"""

        body = self.get_random_body(50000)
        chunkedbody = self.chunk_encode(body)
        self.httpserver.set_reply(reply + chunkedbody)
        self.httpclient.disconnect_reply(request, 400)
        self.httpserver.set_reply(reply + chunkedbody)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client disconnects while
    # reading an http reply that vicompress is gzipping and chunk-encoding.  
    # Disconnect the client while reading the reply.  Then, send a normal http request.  
    # Check that the reply body returned by vicompress matches the reply body 
    # passed to the http server.
    def test_client_disconnect_chunk_encode_gzipped_reply(self):
        request = \
"""GET /test/client/disconnect/chunk/encode/gzipped/reply HTTP/1.1
Accept-Encoding: gzip
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.1 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        self.httpclient.disconnect_reply(request, 400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body[0:20] + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server disconnects while
    # sending the http reply.  Disconnect the server while sending
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_reply(self):
        request = \
"""GET /test/server/disconnect/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply[0:len(reply)/2] )
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server disconnects while
    # sending the reply body.  Disconnect the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_body(self):
        request = \
"""GET /test/server/disconnect/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)




    # Test that vicompress still works after the server disconnects while
    # sending a reply to be cached.  Disconnect the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_cached_body(self):
        request = \
"""GET /test/server/disconnect/cached/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Date: Thu, 24 Jul 2003  14:07:33 GMT
Last-Modified: Wed, 19 Feb 2003 20:13:10 GMT
Content-type: image/png
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpclient.do_request(request)

        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that vicompress still works after the server disconnects while
    # sending a reply to be cached/gzipped. Disconnect the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_cached_gzipped_body(self):
        request = \
"""GET /test/server/disconnect/cached/gzipped/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Date: Thu, 24 Jul 2003  14:07:33 GMT
Last-Modified: Wed, 19 Feb 2003 20:13:10 GMT
Content-type: text/html
Content-length: 67000

"""

        body = self.get_random_body(67000)
        self.httpserver.set_reply( reply + body[0:100] )
        self.httpclient.do_request(request)

        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that vicompress still works after the server disconnects while
    # sending a reply to be gzipped.  Disconnect the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_gzipped_body(self):
        request = \
"""GET /test/server/disconnect/gzipped/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: gzip

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 45000

"""

        body = self.get_random_body(45000)
        self.httpserver.set_reply( reply + body[0:50] )
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server disconnects while
    # sending a chunked reply.  Disconnect the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_chunked_reply(self):
        request = \
"""GET /test/server/disconnect/chunked/reply HTTP/1.1
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.1 200 OK
Connection: Close
Content-type: text/html
Transfer-Encoding: chunked

"""

        body = self.chunk_encode(self.get_random_body(5000))
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server disconnects while
    # vicompress is chunk-encoding the reply.  Disconnect the server while 
    # sending the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_chunk_encode_reply(self):
        request = \
"""GET /test/server/disconnect/chunk/encode/reply HTTP/1.1
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server disconnects while
    # vicompress is gzipping a chunked reply.  Disconnect the server while 
    # sending the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_chunked_gzipped_reply(self):
        request = \
"""GET /test/server/disconnect/chunked/gzipped HTTP/1.1
Accept-Encoding: gzip
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Transfer-Encoding: chunked

"""

        body = self.get_random_body(5000)
        chunkedbody = self.chunk_encode(body)
        self.httpserver.set_reply( reply + chunkedbody[0:20] )
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server disconnects while
    # vicompress is gzipping and chunk-encoding the reply.  Disconnect the server 
    # while sending the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_disconnect_chunk_encode_gzipped_reply(self):
        request = \
"""GET /test/server/disconnect/chunk/encode/gzipped/reply HTTP/1.1
Accept-Encoding: gzip
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.1 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test retrieving a small cached reply from vicompress.
    # Make the same request twice.  For the second request, don't accept
    # any connections on the httpserver, to verify the reply is cached.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_cache_small_body(self):
        request = \
"""GET /test/cache/small/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpserver.set_reply("")
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test retrieving a large cached reply from vicompress.
    # Make the same request twice.  For the second request, don't accept
    # any connections on the httpserver, to verify the reply is cached.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_cache_large_body(self):
        request = \
"""GET /test/cache/large/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 23000

"""

        body = self.get_random_body(23000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpserver.set_reply("")
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test retrieving a cached reply from vicompress, using the "Expires" header.
    # Make the same request twice.  For the second request, don't accept
    # any connections on the httpserver, to verify the reply is cached.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_cache_with_expires_header(self):
        request = \
"""GET /test/cache/with/expires/header HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Expires: Tue, 28 Sep 2004 19:23:10 GMT
Date: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpserver.set_reply("")
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)



    # Test retrieving a cached reply, keeping the connection alive, and 
    # sending another request on the connection.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for all requests.
    def test_cache_keepalive(self):
        request1 = \
"""GET /test/cache/keepalive/request1 HTTP/1.0
Connection: keep-alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        request2 = \
"""GET /test/cache/keepalive/request2 HTTP/1.0
Connection: keep-alive
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""


        reply1 = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 3000

"""
        reply2 = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 4000

"""


        body = self.get_random_body(3000)
        self.httpserver.set_reply(reply1 + body)
        header,body2 = self.httpclient.do_request_keepalive(request1, 1)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpserver.set_reply("")
        header,body2 = self.httpclient.do_request_keepalive(request1, 0)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)

        body = self.get_random_body(4000)
        self.httpserver.set_reply(reply2 + body)
        header,body2 = self.httpclient.do_request_keepalive(request2, 0)
        self.httpclient.close()
        myassert(body == body2, "== 3rd expected body ==\n" + body + \
                 "\n== 3rd actual body ==\n" + body2)


    # Test that a "Pragma: no-cache" header will bypass the vicompress cache.
    # Also, test that vicompress removes the cached page if it has changed.
    # This test does the following:
    # - Request a page
    # - Request the page again, make sure its cached
    # - Change the page contents
    # - Request the page with "Pragma: no-cache". 
    #   Verify the page returned is the new page, not the cached page
    # - Request the page again
    #   Verify that the new page is returned again. 
    def test_cache_forced_expire(self):
        request1 = \
"""GET /test/cache/forced/expire HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        request2 = \
"""GET /test/cache/forced/expire HTTP/1.0
Connection: close
Pragma: no-cache
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply1 = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
content-length: 5000

"""
        reply2 = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Thu, 30 Sep 2004 13:04:00 GMT
Last-Modified: Tue, 28 Sep 2004 22:48:10 GMT
content-length: 6000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply1 + body)
        header,body2 = self.httpclient.do_request(request1)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)
        self.httpserver.set_reply("")
        header,body2 = self.httpclient.do_request(request1)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)

        self.httpserver.set_reply(reply2 + body)
        header,body2 = self.httpclient.do_request(request2)
        myassert(body == body2, "== 3rd expected body ==\n" + body + \
                 "\n== 3rd actual body ==\n" + body2)

        self.httpserver.set_reply(reply2 + body)
        header,body2 = self.httpclient.do_request(request1)
        myassert(body == body2, "== 4th expected body ==\n" + body + \
                 "\n== 4th actual body ==\n" + body2)


    # Test retrieving a cached reply from vicompress with a "Range: start-end" header.
    # Make the same request twice.  For the second request, use a "Range: start-end"
    # header, so that only a portion of the cached reply is returned.
    # Check that the reply body returned by vicompress matches what is expected.
    def test_cache_range_request(self):
        request1 = \
"""GET /test/cache/range/request HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        request2 = \
"""GET /test/cache/range/request HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Range: 3000-8499

"""


        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 9000

"""

        body = self.get_random_body(9000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request1)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpserver.set_reply("")
        header,body2 = self.httpclient.do_request(request2)
        myassert(header.find("Content-Range: 3000-8499") > 0,
                 "== Invalid header ==\n" + header)
        myassert(body[3000:8500] == body2, 
                 "== 2nd expected body ==\n" + body[3000:8499] + \
                 "\n== 2nd actual body ==\n" + body2)



    # Test retrieving a cached reply from vicompress with a "Range: start-" header.
    # Make the same request twice.  For the second request, use a "Range: start-"
    # header, so that only a portion of the cached reply is returned.
    # Check that the reply body returned by vicompress matches what is expected.
    def test_cache_range_request_no_end(self):
        request1 = \
"""GET /test/cache/range/request/no/end HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        request2 = \
"""GET /test/cache/range/request/no/end HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Range: 3000-

"""


        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 9000

"""

        body = self.get_random_body(9000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request1)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpserver.set_reply("")
        header,body2 = self.httpclient.do_request(request2)
        myassert(header.find("Content-Range: 3000-9000") > 0,
                 "== Invalid header ==\n" + header)
        myassert(body[3000:] == body2, 
                 "== 2nd expected body ==\n" + body[3000:] + \
                 "\n== 2nd actual body ==\n" + body2)



    # Test retrieving the list of all urls in the cache, using
    # the http://hostname/_viewcache_ url.
    # Verify that the body contains http://vicompersstest.com/
    def test_viewcache(self):
        request = \
"""GET /_viewcache_ HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        header,body = self.httpclient.do_request(request)
        url = "http://vicompresstest.com/test/cache/small/body"
        myassert(body.find(url) >= 0, "Incorrect cache url list:\n" + body)



    # Test that requests with "Authorization" header are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_authorization(self):
        request = \
"""GET /test/notcached/authorization HTTP/1.0
Connection: close
Authorization: some-person
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)

    # Test that requests with "Pragma: no-cache" are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_pragma_nocache(self):
        request = \
"""GET /test/notcached/pragma/nocache HTTP/1.0
Connection: close
Pragma: no-cache
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that replies without a content-length are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_no_contentlength(self):
        request = \
"""GET /test/notcache/no/contentlength HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that post requests are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_post_request(self):
        request = \
"""POST /test/notcached/post/request HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Content-length: 10

abcdefghi
"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that non-200 replies are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_status_not_200(self):
        request = \
"""GET /test/notcached/status/not/200 HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 404 Not Found
Connection: Close
Content-type: text/html
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that requests with "Cache-Control: no-cache" are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_cachecontrol_nocache(self):
        request = \
"""GET /test/notcached/cachecontrol/nocache HTTP/1.0
Connection: close
Cache-Control: no-cache
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that replies with "Cache-Control: no-cache" are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_cachecontrol_private(self):
        request = \
"""GET /test/notcached/cachecontrol/private HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400
cache-control: private

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)



    # Test that replies with "Content-Range" are not cached
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_content_range(self):
        request = \
"""GET /test/notcached/content/range HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400
Content-Range: 5-405 bytes

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that requests with "Range:" are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_range(self):
        request = \
"""GET /test/notcached/range HTTP/1.0
Connection: close
Range: 2000-2400 bytes
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that replies with "maxage=0" are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_range(self):
        request = \
"""GET /test/notcached/range HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400
Cache-Control: max-age=0

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that replies with "Cache-control: no-cache" are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_reply_cachecontrol_nocache(self):
        request = \
"""GET /test/notcached/reply/cachecontrol/nocache HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400
Cache-Control: no-cache

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that requests with "If-Modified-Since" are not cached.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_if_modified_since(self):
        request = \
"""GET /test/notcached/if/modified/since HTTP/1.0
Connection: close
If-Modified-Since: Wed, 30, Oct 2002 05:23:06 GMT
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 400

"""

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(400)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that a client that doesn't support gzip will not receive
    # a gzipped cached reply.
    # Make two requests with different reply content.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver, for both requests.
    def test_notcached_request_cant_gzip(self):
        request1 = \
"""GET /test/notcached/request/cant/gzip HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: deflate, gzip, compress;

"""
        request2 = \
"""GET /test/notcached/request/cant/gzip HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""


        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 50000

"""

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request1)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request2)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test a gzipped reply where the reply header has no content length.
    # Check that the gunzipped reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_gzip_no_contentlength(self):
        request = \
"""GET /test/gzip/no/contentlength HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Date: Tue, 28 Sep 2004 19:23:10 GMT

"""

        body = self.get_random_body(2000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a gzipped reply where the reply header has a content length.
    # Check that the gunzipped reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_gzip_with_contentlength(self):
        request = \
"""GET /test/gzip/with/contentlength HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/plain
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 65000

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a gzipped reply where the content type is javascript.
    # Check that the gunzipped reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_gzip_javascript(self):
        request = \
"""GET /test/gzip/javascript HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: application/javascript
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 65000

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a gzipped reply where the content type is x-javascript.
    # Check that the gunzipped reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_gzip_x_javascript(self):
        request = \
"""GET /test/gzip/x-javascript HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: application/x-javascript
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 65000

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a gzipped reply where the content type is pdf.
    # Check that the gunzipped reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_gzip_pdf(self):
        request = \
"""GET /test/gzip/pdf HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: application/pdf
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 65000

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test a gzipped reply where the content type is msword
    # Check that the gunzipped reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_gzip_msword(self):
        request = \
"""GET /test/gzip/msword HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: application/msword
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 65000

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test retrieving a reply that is both cached and gzipped.
    # Make the same request twice.  For the second request, don't accept
    # any connections on the httpserver, to verify the reply is cached.
    # Check that the gunzipped reply body returned by vicompress matches 
    # the reply body passed to the httpserver, for both requests.
    def test_cache_gzip_reply(self):
        request = \
"""GET /test/cache/gzip/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip, *;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-length: 65000

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpserver.set_reply("")
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that a reply that's already compressed won't be gzipped.
    # Check that the reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_cantgzip_already_compressed(self):
        request = \
"""GET /test/cantgzip/already/compressed HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 65000
Content-encoding: compress

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that a reply won't be gzipped if the client doesn't support gzip.
    # Check that the reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_cantgzip_client_wont_accept(self):
        request = \
"""GET /test/cantgzip/client/wont/accept HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 65000

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)

    # Test that a reply won't be gzipped if its not a text file.
    # Check that the reply body returned by vicompress matches 
    # the reply body passed to the httpserver.
    def test_cantgzip_not_text(self):
        request = \
"""GET /test/cantgzip/not/text HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, deflate, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: img/png
Date: Tue, 28 Sep 2004 19:23:10 GMT
Content-length: 65000

"""

        body = self.get_random_body(65000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client times out while
    # sending the http request.  Time out the client while sending 
    # the request.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_timeout_request(self):
        request = \
"""GET /test/client/timeout/request HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpclient.timeout_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client times out while
    # reading the http reply.  Timeout the client while reading 
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_timeout_reply(self):
        request = \
"""GET /test/client/timeout/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 50000

"""

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        self.httpclient.timeout_reply(request, 20)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)



    # Test that vicompress still works after a client times out while
    # reading a cached reply.  Timeout the client while reading 
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_timeout_cached_reply(self):
        request = \
"""GET /test/client/timeout/cached/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Date: Tue, 28 Sep 2004 19:23:10 GMT
Last-Modified: Sun, 26 Sep 2004 19:23:10 GMT
Content-type: image/png
Content-length: 50000

"""

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        self.httpclient.timeout_reply(request, 20)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that vicompress still works after a client times out while
    # reading a gzipped reply.  Timeout the client while reading 
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_timeout_gzipped_reply(self):
        request = \
"""GET /test/client/timeout/gzipped/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Accept-Encoding: compress, gzip;
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 50000

"""

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        self.httpclient.timeout_reply(request, 20)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client times out while
    # reading a chunked http reply.  Timeout the client while reading 
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_timeout_chunked_reply(self):
        request = \
"""GET /test/client/timeout/chunked/reply HTTP/1.1
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Transfer-Encoding: chunked
Content-type: text/html

"""

        body = self.chunk_encode(self.get_random_body(50000))
        self.httpserver.set_reply(reply + body)
        self.httpclient.timeout_reply(request, 20)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client times out while
    # reading a reply that vicompress is chunk-encoding.  Timeout the client 
    # while reading the reply.  Then, send a normal http request.  Check that 
    # the reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_timeout_chunk_encode_reply(self):
        request = \
"""GET /test/client/timeout/chunk/encode/reply HTTP/1.1
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        self.httpclient.timeout_reply(request, 20)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client times out while
    # reading a chunked http reply that vicompress is gzipping.  Timeout the 
    # client while reading the reply.  Then, send a normal http request.  Check 
    # that the reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_client_timeout_chunked_gzipped_reply(self):
        request = \
"""GET /test/client/timeout/chunked/gzipped/reply HTTP/1.1
Accept-Encoding: gzip;
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Transfer-Encoding: chunked

"""

        body = self.get_random_body(50000)
        chunkedbody = self.chunk_encode(body)
        self.httpserver.set_reply(reply + chunkedbody)
        self.httpclient.timeout_reply(request, 20)
        self.httpserver.set_reply(reply + chunkedbody)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after a client times out while
    # reading an http reply that vicompress is gzipping and chunk-encoding.  
    # Timeout the client while reading the reply.  Then, send a normal http 
    # request.  Check that the reply body returned by vicompress matches the 
    # reply body passed to the http server.
    def test_client_timeout_chunk_encode_gzipped_reply(self):
        request = \
"""GET /test/client/timeout/chunk/encode/gzipped/reply HTTP/1.1
Accept-Encoding: gzip
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(50000)
        self.httpserver.set_reply(reply + body)
        self.httpclient.timeout_reply(request, 20)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server times out while
    # sending the http reply.  Timeout the server while sending
    # the reply.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_reply(self):
        request = \
"""GET /test/server/timeout/reply HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply[0:len(reply)/2] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server times out while
    # sending the reply body.  Timeout the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_body(self):
        request = \
"""GET /test/server/timeout/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)




    # Test that vicompress still works after the server times out while
    # sending a reply to be cached.  Timeout the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_cached_body(self):
        request = \
"""GET /test/server/timeout/cached/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Date: Thu, 24 Jul 2003  14:07:33 GMT
Last-Modified: Wed, 19 Feb 2003 20:13:10 GMT
Content-type: image/png
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)

        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that vicompress still works after the server times out while
    # sending a reply to be cached/gzipped. Timeout the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_cached_gzipped_body(self):
        request = \
"""GET /test/server/timeout/cached/gzipped/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: compress, gzip;

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Date: Thu, 24 Jul 2003  14:07:33 GMT
Last-Modified: Wed, 19 Feb 2003 20:13:10 GMT
Content-type: text/html
Content-length: 67000

"""

        body = self.get_random_body(67000)
        self.httpserver.set_reply( reply + body[0:100] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)

        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== 1st expected body ==\n" + body + \
                 "\n== 1st actual body ==\n" + body2)

        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== 2nd expected body ==\n" + body + \
                 "\n== 2nd actual body ==\n" + body2)


    # Test that vicompress still works after the server times out while
    # sending a reply to be gzipped.  Timeout the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_gzipped_body(self):
        request = \
"""GET /test/server/timeout/gzipped/body HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com
Accept-Encoding: gzip

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Content-length: 45000

"""

        body = self.get_random_body(45000)
        self.httpserver.set_reply( reply + body[0:50] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("Content-Encoding: gzip") > 0,
                 "== Bad header ==\n" + header)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server times out while
    # sending a chunked reply.  Timeout the server while sending
    # the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_chunked_reply(self):
        request = \
"""GET /test/server/timeout/chunked/reply HTTP/1.1
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
Transfer-Encoding: chunked

"""

        body = self.chunk_encode(self.get_random_body(5000))
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server times out while
    # vicompress is chunk-encoding the reply.  Timeout the server while 
    # sending the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_chunk_encode_reply(self):
        request = \
"""GET /test/server/timeout/chunk/encode/reply HTTP/1.1
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.1 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server times out while
    # vicompress is gzipping a chunked reply.  Timeout the server while 
    # sending the body.  Then, send a normal http request.  Check that the
    # reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_chunked_gzipped_reply(self):
        request = \
"""GET /test/server/timeout/chunked/gzipped/reply HTTP/1.1
Accept-Encoding: gzip
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.1 200 OK
Connection: Close
Content-type: text/html
Transfer-Encoding: chunked

"""

        body = self.get_random_body(5000)
        chunkedbody = self.chunk_encode(body)
        self.httpserver.set_reply( reply + chunkedbody[0:20] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + chunkedbody)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress still works after the server times out while
    # vicompress is gzipping and chunk-encoding the reply.  Timeout the server 
    # while sending the body.  Then, send a normal http request.  Check that 
    # the reply body returned by vicompress matches the reply body passed
    # to the http server.
    def test_server_timeout_chunk_encode_gzip_reply(self):
        request = \
"""GET /test/server/timeout/chunk/encode/gzip/reply HTTP/1.1
Accept-Encoding: gzip
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.1 200 OK
Connection: Close
Content-type: text/html

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply( reply + body[0:20] )
        self.httpserver.set_timeout(1)
        self.httpclient.do_request(request)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        body2 = self.chunk_decode(body2)
        body2 = self.gunzip(body2)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress works as a forward proxy, where the request
    # contains the origin server hostname in the "Host: " header.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_withdns_hostname(self):
        request = \
"""GET /test/withdns/hostname HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: HOSTNAME:3181

"""
        request = request.replace("HOSTNAME", socket.gethostname())

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that vicompress works as a forward proxy, where the request
    # contains the origin server IP address in the "Host: " header.
    # Check that the reply body returned by vicompress matches the reply body
    # passed to the httpserver.
    def test_withdns_ipaddress(self):
        request = \
"""GET /test/withdns/ipaddress HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: HOSTNAME:3181

"""
        ip = socket.gethostbyname( socket.gethostname() )
        request = request.replace("HOSTNAME", ip)

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: image/png
Content-length: 5000

"""

        body = self.get_random_body(5000)
        self.httpserver.set_reply(reply + body)
        header,body2 = self.httpclient.do_request(request)
        myassert(body == body2, "== expected body ==\n" + body + \
                 "\n== actual body ==\n" + body2)


    # Test that when vicompress works as a forward proxy, requests without
    # a "Host: " header are rejected.
    def test_withdns_no_hostname(self):
        request = \
"""GET /test/withdns/no/hostname HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT

"""

        header,body2 = self.httpclient.do_request(request)
        myassert(header.find("HTTP/1.1 400 Bad Request") >= 0,
                 "== incorrect header ==\n" + header)

    # Test that when vicompress works as a forward proxy, requests for
    # localhost are rejected
    def test_withdns_localhost(self):
        request = \
"""GET /test/withdns/localhost HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: localhost:3181

"""

        header = "HTTP/1.1 503 Service Unavailable"
        header2,body = self.httpclient.do_request(request)
        header2 = header2.split("\r\n")[0]
        myassert(header == header2, "== expected header ==\n" + header + \
                 "\n== actual header ==\n" + header2)

    # Test that when vicompress works as a forward proxy, requests for
    # 127.0.0.1 are rejected
    def test_withdns_localhost_ip(self):
        request = \
"""GET /test/withdns/localhost_ip HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: 127.0.0.1:3181

"""
        header = "HTTP/1.1 503 Service Unavailable"
        header2,body = self.httpclient.do_request(request)
        header2 = header2.split("\r\n")[0]
        myassert(header == header2, "== expected header ==\n" + header + \
                 "\n== actual header ==\n" + header2)


    # Test that when vicompress works as a forward proxy, requests for
    # unknown hostnames result in an error reply
    def test_withdns_unknown_hostname(self):
        request = \
"""GET /test/withdns/unknown/hostname HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT
Host: somerandomhostthatdoesntexist.com:3181

"""
        header = "HTTP/1.1 404 Not Found"
        header2,body = self.httpclient.do_request(request)
        header2 = header2.split("\r\n")[0]
        myassert(header == header2, "== expected header ==\n" + header + \
                 "\n== actual header ==\n" + header2)
 

    # Test that if sessions are disabled, vicompress will distribute
    # requests in a round-robin fashion between the two backend webservers.
    # The two httpservers will send back different replies, so we can
    # verify which webserver the reply came from
    def test_loadbalance_no_sessions(self):
        conf = \
"""
listen 127.0.0.1 3128
webserver 127.0.0.1 3181
webserver 127.0.0.1 3182
enable_sessions no
enable_debug yes
accesslog accesslog
errorlog errorlog
"""

        request = \
"""GET /test/loadbalance/no/sessions HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT 
Host: vicompresstest.com
Cookie: vicompressid=1

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
content-length: 2000

"""

        body1 = self.get_random_body(2000)
        body2 = self.get_random_body(2000)

        self.start_vicompress(conf)
        time.sleep(1)
       
        self.httpserver.set_reply(reply + body1)
        header,body = self.httpclient.do_request(request)
        myassert(body == body1, "== 1st expected body ==\n" + body1 + \
                 "\n== 1st actual body ==\n" + body)
        time.sleep(1)

        self.httpserver2.set_reply(reply + body2)
        header,body = self.httpclient.do_request(request)
        myassert(body == body2, "== 2nd expected body ==\n" + body2 + \
                 "\n== 2nd actual body ==\n" + body)
        time.sleep(1)

        self.httpserver.set_reply(reply + body1)
        header,body = self.httpclient.do_request(request)
        myassert(body == body1, "== 3rd expected body ==\n" + body1 + \
                 "\n== 3rd actual body ==\n" + body)

        self.stop_vicompress()
        time.sleep(1)


    # Test that if sessions are enabled, vicompress will use the backend
    # webserver specified by the cookie vicompressid=<webserverid>.
    # The two httpservers will send back different replies, so we can
    # verify which webserver the reply came from
    def test_loadbalance_with_sessions(self):
        conf = \
"""
listen 127.0.0.1 3128
webserver 127.0.0.1 3181
webserver 127.0.0.1 3182
enable_sessions yes
enable_debug yes
accesslog accesslog
errorlog errorlog
"""

        request1 = \
"""GET /test/loadbalance/with/sessions HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT 
Host: vicompresstest.com
Cookie: vicompressid=1

"""
        request2 = \
"""GET /test/loadbalance/with/sessions HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT 
Host: vicompresstest.com
Cookie: vicompressid=2

"""


        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
content-length: 2000

"""

        body1 = self.get_random_body(2000)
        body2 = self.get_random_body(2000)

        self.start_vicompress(conf)
        time.sleep(1)
       
        self.httpserver.set_reply(reply + body1) 
        header,body = self.httpclient.do_request(request1)
        myassert(body == body1, "== 1st expected body ==\n" + body1 + \
                 "\n== 1st actual body ==\n" + body)

        self.httpserver.set_reply(reply + body1) 
        header,body = self.httpclient.do_request(request1)
        myassert(body == body1, "== 2nd expected body ==\n" + body1 + \
                 "\n== 2nd actual body ==\n" + body)


        self.httpserver2.set_reply(reply + body2) 
        header,body = self.httpclient.do_request(request2)
        myassert(body == body2, "== 3rd expected body ==\n" + body2 + \
                 "\n== 3rd actual body ==\n" + body)

        self.httpserver2.set_reply(reply + body2) 
        header,body = self.httpclient.do_request(request2)
        myassert(body == body2, "== 4th expected body ==\n" + body2 + \
                 "\n== 4th actual body ==\n" + body)

        self.stop_vicompress()
        time.sleep(1)


    # Test that if a webserver is not up, vicompress will detect this
    # and forward requests to the remaining webservers.  Then bring the
    # webserver back up, and wait for 3 minutes.  Vicompress should detect
    # that the webserver is now up.  Then verify that the next reply
    # comes from the newly up webserver.  The two httpservers will send 
    # back different replies, so we can verify which webserver the reply 
    # came from
    def test_loadbalance_failover(self):
        conf = \
"""
listen 127.0.0.1 3128
webserver 127.0.0.1 3181
webserver 127.0.0.1 3183
enable_sessions no
enable_debug yes
accesslog accesslog
errorlog errorlog
"""

        request = \
"""GET /test/loadbalance/failover HTTP/1.0
Connection: close
Date: Wed, 30 Oct 2002 12:43:03 GMT 
Host: vicompresstest.com

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
content-length: 2000

"""

        body1 = self.get_random_body(2000)
        body2 = self.get_random_body(2000)
        self.start_vicompress(conf)
        time.sleep(1)
        
        self.httpserver.set_reply(reply + body1) 
        header,body = self.httpclient.do_request(request)
        myassert(body == body1, "== 1st expected body ==\n" + body1 + \
                 "\n== 1st actual body ==\n" + body)
        time.sleep(1)

        self.httpserver.set_reply(reply + body1) 
        header,body = self.httpclient.do_request(request)
        myassert(body == body1, "== 2nd expected body ==\n" + body1 + \
                 "\n== 2nd actual body ==\n" + body)
        time.sleep(1)

        self.httpserver.set_reply(reply + body1) 
        header,body = self.httpclient.do_request(request)
        myassert(body == body1, "== 3rd expected body ==\n" + body1 + \
                 "\n== 3rd actual body ==\n" + body)

        httpserver3 = HttpServer(3183)
        httpserver3.set_reply("")
        time.sleep(60 * 3 + 5)
         
        httpserver3.set_reply(reply + body2)
        header,body = self.httpclient.do_request(request)
        myassert(body == body2, "== 4th expected body ==\n" + body2 + \
                 "\n== 4th actual body ==\n" + body)

        self.stop_vicompress()
        httpserver3.close()
        time.sleep(1)


    # Test that if webservers are specified with specific hosts, vicompress will
    # forward requests to the backend webserver that matches the "Host:" header
    # in the request.  The two httpservers will send back different replies, so
    # we can verify which webserver the reply came from
    def test_loadbalance_with_hosts(self):
        conf = \
"""
listen 127.0.0.1 3128
webserver 127.0.0.1 3181 test1.com
webserver 127.0.0.1 3182 test2.com test3.com test4.com
enable_sessions no
enable_debug yes
accesslog accesslog
errorlog errorlog
"""

        request = \
"""GET /test/loadbalance/with/hosts HTTP/1.1
Connection: close
Host: <host>

"""

        reply = \
"""HTTP/1.0 200 OK
Connection: Close
Content-type: text/html
content-length: 2000

"""

        body1 = self.get_random_body(2000)

        self.start_vicompress(conf)
        time.sleep(1)

        self.httpserver.set_reply(reply + body1)

        req = request.replace("<host>", "test1.com")
        header,body = self.httpclient.do_request(req)
        myassert(body == body1, "== 1st expected body ==\n" + body1 + \
                 "\n== 1st actual body ==\n" + body)

        body2 = self.get_random_body(2000)
        self.httpserver2.set_reply(reply + body2)
        req = request.replace("<host>", "test2.com")
        header,body = self.httpclient.do_request(req)
        myassert(body == body2, "== 2nd expected body ==\n" + body2 + \
                 "\n== 2nd actual body ==\n" + body)

        body2 = self.get_random_body(2000)
        self.httpserver2.set_reply(reply + body2)
        req = request.replace("<host>", "test3.com")
        header,body = self.httpclient.do_request(req)
        myassert(body == body2, "== 3rd expected body ==\n" + body2 + \
                 "\n== 3rd actual body ==\n" + body)

        body2 = self.get_random_body(2000)
        self.httpserver2.set_reply(reply + body2)
        req = request.replace("<host>", "test4.com")
        header,body = self.httpclient.do_request(req)
        myassert(body == body2, "== 4th expected body ==\n" + body2 + \
                 "\n== 4th actual body ==\n" + body)

        self.stop_vicompress()
        time.sleep(1)

# End class VicompressTest



# void main(string path to vicompress)
def main(path, fast):
    global VICOMPRESS_PATH
    VICOMPRESS_PATH = path
    if not os.path.exists(VICOMPRESS_PATH):
        print("Error: the path " + VICOMPRESS_PATH + " doesnt exist")
        sys.exit(1)

    print("Vicompress functional test")
    print("==========================")
    print("Note: The timeout tests take 1 minute each to run.")
    print("  The total testsuite takes about 1 hour to run.")
    print("\n")
    time.sleep(3)
    v = VicompressTest()
    v.run(fast)


if len(sys.argv) < 2:
    print("""\
Usage: vicompress_test.py <path to vicompress> [fast]
       vicompress_test.py /usr/local/vicompress/bin/vicompress
       vicompress_test.py /usr/local/vicompress/bin/vicompress fast

If "fast" is specified, then the timeout and failover tests will be
skipped, and the tests will finish ine one minute.

If "fast" is not specified, then the full test suite is run, including
timeout and failover tests.  The full test suite takes 1 hour to run.
""")
    sys.exit(0)

fast = 0
if len(sys.argv) == 3 and sys.argv[2] == "fast":
    fast = 1

main(sys.argv[1], fast)

