/* 
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2. 
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * The Buffer structure is used to hold http header and body data.
 * The Buffer has two features:
 * - It is reference counted, so that multiple clients can share
 *   it without worrying about when to free it.
 * - The buffer is allocated from a pool, so allocation is fast.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include "buffer.h"
#include "pool.h"

/* Initialize the buffer pool. 
 * size is the capacity (in bytes) that each buffer will have
 * count is the number of total buffers to hold. 
 */ 
void bufferpool_init(BufferPool *bp, int size, long count)
{
    assert(size > 0);
    pool_init(&bp->bufpool, size, count);
    pool_init(&bp->buffer, sizeof(Buffer), count);
    bp->size = size;
}

/* Return true if the buffer pool has no free entries. */
int bufferpool_is_empty(BufferPool *bp)
{
    return pool_is_empty(&bp->bufpool);
}

/* Allocate a buffer from the pool, and return it.
 * The initial reference count is 1.
 */
Buffer* buffer_alloc(BufferPool *bp)
{
    Buffer *buffer;
    char *buf;

    buffer = (Buffer *) pool_alloc(&bp->buffer);
    buf = pool_alloc(&bp->bufpool);
    buffer->buf = buf;
    buffer->buf[0] = 0;
    buffer->size = 0;
    buffer->capacity = bp->size;
    buffer->refcount = 1;
    buffer->pool = bp;
    return buffer;
} 

/* Free the buffer back to the pool. Make sure the
 * reference count is 0.
 */
static void buffer_free(BufferPool *bp, Buffer *b)
{
    int capacity;

    if (b == NULL)
        return;
    assert(b->refcount == 0);
    capacity = b->capacity;
    if (capacity == bp->size) 
        pool_free(&bp->bufpool, b->buf);
    else 
        free(b->buf);
    
    b->buf = NULL;
    b->size = b->capacity = 0;
    b->pool = NULL;
    pool_free(&bp->buffer, b);
}

/* Increment the reference count on the buffer, and return it. */
Buffer* buffer_ref(Buffer *b)
{
    b->refcount++;
    return b;
}

/* Decrement the reference count on the buffer. If
 * the count goes to 0, free the buffer.
 */
void buffer_unref(Buffer *b)
{
    BufferPool *bp;
    if (b == NULL)
        return;
    b->refcount--;
    if (b->refcount == 0) {
        bp = b->pool; b->pool = NULL;
        buffer_free(bp, b); 
    }
}

/* Increase the buffer capacity to the given newsize (in bytes). */
void buffer_grow(Buffer *buffer, int newsize)
{
    BufferPool *bp;
    char *buf, *newbuf;
    int size, capacity;

    bp = buffer->pool;
    buf = buffer->buf;
    capacity = buffer->capacity;
    size = buffer->size;
    assert(newsize > 0 && newsize > capacity);
    
    newbuf = malloc(newsize);
    buffer->capacity = newsize;
    memcpy(newbuf, buf, size);
    if (capacity == bp->size)
        pool_free(&bp->bufpool, buf);
    else
        free(buf);

    buffer->buf = newbuf;
}

/* Append the given data to the buffer. Grow the buf if needed.
 * data -  the raw binary data to append.
 * size -  the size of the data in bytes.
 */
void buffer_add(Buffer *buffer, const char *data, int size)
{
    while (buffer->size + size > buffer->capacity)
        buffer_grow(buffer, buffer->capacity * 2);
    
    memcpy(&buffer->buf[buffer->size], data, size);
    buffer->size += size;
    if (buffer->size < buffer->capacity)
        buffer->buf[buffer->size] = 0;
}

/* Reset the buffer size to 0. The capacity is not changed */ 
void buffer_reset(Buffer *buffer)
{
    buffer->size = 0;
    buffer->buf[0] = 0;
}

/* Append the string representation of the given 
 * integer to the buffer. 
 */
void buffer_add_int(Buffer *buffer, int x)
{
    char tmp[32];

    tmp[0] = 0;
    sprintf(tmp, "%d", x);
    buffer_add(buffer, tmp, strlen(tmp));
}

