/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * The MemoryCache is used to store and retrieve CachedItems.
 * A CachedItem is essentially an Http reply header and body.
 */


#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "cache.h"
#include "sockets.h"
#include "conf.h"
#include "currenttime.h"


/* Should match BODY_SIZE in vicompress.c */
#define BODY_SIZE  4096

static void cache_delete_item(MemoryCache *cache, CachedItem *item);
static int cache_delete_amount(MemoryCache *cache, long amount);

/** Hash Table routines **/

/* Hash the given url, and return the hash value. Ignore the leading
 * "www." if present. Note that the hash is case sensitive, so
 * google.com/index.html and GOOGLE.COM/INDEX.HTML will not match.
 * tablesize is the number of buckets in the hashtable.
 */
static unsigned int hash_item(Buffer *url, long tablesize)
{
    unsigned int value = 53;
    int i = 0;

    if (strncasecmp(url->buf, "www.", 4) == 0)
        i = 4; 

    while (i < url->size) {
        value = (value * 31 + url->buf[i]) % tablesize;
        i++;
    }
    return value;
}

/* Return a number >=x that is "somewhat" prime */
static long findprime(long x)
{
    while (1) {
        if ((x % 2 != 0) && (x % 3 != 0) && (x % 5 != 0)) {
            return x;
        }
        else {
            x += 1;
        }
    }
}


/* Initialize the hash table used to lookup cached items.
 * maxitems is the maximum number of items the cache will hold.
 * Note that the hashtable bucket size is fixed at creation, 
 * and does not grow during runtime.
 */ 
static void hashtable_init(HashTable *table, long maxitems)
{
    int i;
    table->bucketcount = findprime(maxitems);
    table->buckets = (CachedItem **) 
      calloc(table->bucketcount, sizeof(CachedItem*));
    for (i = 0; i < table->bucketcount; i++) {
        table->buckets[i] = NULL;
    }
}

/* Add the given CachedItem to the hash table.
 * Use the CachedItem url field as the key.
 */
static void hashtable_add(HashTable *table, CachedItem *item)
{
    CachedItem *prev;
    unsigned int hash;
    
    item->hashdata.next = NULL;
    hash = hash_item(item->url, table->bucketcount);
    if (table->buckets[hash] == NULL) {
        table->buckets[hash] = item;
        return;
    }
    prev = table->buckets[hash];
    while (prev->hashdata.next != NULL) {
        prev = prev->hashdata.next;
    }
    prev->hashdata.next = item;
}

/* Remove the given item from the hash table.
 * Use the CachedItem url field as the key.
 */
static void hashtable_remove(HashTable *table, CachedItem *item)
{
    unsigned int hash;
    CachedItem *prev;
    
    hash = hash_item(item->url, table->bucketcount);
    if (table->buckets[hash] == item) {
        table->buckets[hash] = item->hashdata.next;
        item->hashdata.next = NULL;
        return;
    }
    prev = table->buckets[hash];
    while (prev != NULL && prev->hashdata.next != item) {
        prev = prev->hashdata.next;
    }
    if (prev == NULL) {
        return;
    }
    prev->hashdata.next = item->hashdata.next;
    item->hashdata.next = NULL;
}

/* Given two URLs, return 1 if they match, 0 otherwise. 
 * Ignore any leading "www." in the url.
 */
static int url_equals(Buffer *url1, Buffer *url2)
{
    int starturl1, starturl2;
    starturl1 = starturl2 = 0;

    if (strncasecmp(url1->buf, "www.", 4) == 0) {
        starturl1 = 4;
    }
    if (strncasecmp(url2->buf, "www.", 4) == 0) {
        starturl2 = 4;
    }

    if ((url1->size - starturl1) == (url2->size - starturl2)) {
        if (strncmp(&(url1->buf[starturl1]), 
                    &(url2->buf[starturl2]), 
                    url1->size - starturl1) == 0) {
            return 1;
        }
    }
    return 0;
}

/* Lookup the given url in the hash table. Return the CachedItem
 * if the item exists, or NULL if its not in the table. 
 */
static CachedItem* hashtable_lookup(HashTable *table, Buffer *url)
{
    CachedItem *item;
    unsigned int hash;

    hash = hash_item(url, table->bucketcount);
    item = table->buckets[hash];
    while (item != NULL) {
        if (url_equals(url, item->url)) {
            return item;
        }
        else {
            item = item->hashdata.next;
        }
    }
    return NULL;
}

/* Initialize the least recently used item list. Create dummy
 * items to act as the oldest and newest cached items.
 */
static void lru_init(LRUList *lru, MemoryCache *cache)
{
    lru->oldestItem = pool_alloc(&cache->cachedItemPool);
    lru->newestItem = pool_alloc(&cache->cachedItemPool);
    memset(lru->oldestItem, 0, sizeof(CachedItem));
    memset(lru->newestItem, 0, sizeof(CachedItem));
    lru->oldestItem->lrudata.newer = lru->newestItem;
    lru->newestItem->lrudata.older = lru->oldestItem;
}

/* Add an item to the LRU item list. The item will be placed
 * just before the newest item in the LRU list.
 */
static void lru_addItem(LRUList *lru, CachedItem *item)
{
    CachedItem *prevNewest;
    item->lastreferenced = get_current_time();
    prevNewest = lru->newestItem->lrudata.older;
    
    prevNewest->lrudata.newer = item;
    item->lrudata.older = prevNewest;
    item->lrudata.newer = lru->newestItem;
    lru->newestItem->lrudata.older = item;
}

/* Remove an item from the LRU item list. */
static void lru_removeItem(LRUList *lru, CachedItem *item)
{
    CachedItem *prev, *next;
    prev = item->lrudata.older;
    next = item->lrudata.newer;
    prev->lrudata.newer = next;
    next->lrudata.older = prev;
}

/* Update an item in the LRU list. The item is removed
 * from its current position in the list, and inserted again,
 * at the newest slot.
 */
static void lru_updateItem(LRUList *lru, CachedItem *item)
{
    lru_removeItem(lru, item);
    item->lrudata.older = item->lrudata.newer = NULL;
    lru_addItem(lru, item);
}

/* Retrieve the oldest item in the LRU list (not counting
 * the dummy item "oldestItem").  Return NULL if the LRU
 * list is empty.
 */
static CachedItem* lru_getOldestReplacement(LRUList *lru)
{
    CachedItem *item;
    item = lru->oldestItem->lrudata.newer;
    while (item != lru->newestItem && item->lockcount > 0) {
        item = item->lrudata.newer;
    }
    if (item == lru->newestItem) {
        return NULL;
    }
    lru_removeItem(lru, item);
    return item;
}

/*
 * A CachedItem consists of an array of buffers.
 * Read a buffer from the given cached item at the given index.
 * Update the index to the next buffer. Return NULL if
 * we've reached the end of the item. 
 */
Buffer* cacheditem_read(CachedItem *item, int* index)
{
    int pos = *index;
    if (pos >= 0 && pos < item->bodycount) {
        *index = *index + 1;
        return item->body[pos];
    }
    else {
        return NULL;
    }
}

/*
 * Add a buffer to the cached item. The reference count of
 * the buffer is incremented. This is called when storing
 * a new item into the memory cache.
 */
void cacheditem_write(CachedItem *item, Buffer *buf)
{
    Buffer **newbody;
    int i;
    if (item->bodycount >= item->bodycapacity) {
        newbody = (Buffer **)calloc(item->bodycapacity * 2, sizeof(Buffer *));
        for (i = 0; i < item->bodycount; i++) {
            newbody[i] = item->body[i];
        }
        free(item->body);
        item->body = newbody;
        item->bodycapacity = item->bodycapacity * 2;
    } 
    item->body[item->bodycount] = buffer_ref(buf);
    item->bodycount += 1;
}

/*
 * Copy the HttpReply information from the CachedItem to
 * the given HttpReply structure.  The reference count of
 * the reply buffer is incremented.  When done using the
 * reply, the caller must call http_free_reply() to 
 * decrement the ref count.  Note that multiple connections
 * may use the same cached HttpReply at the same time.
 */
void cacheditem_httpreply(CachedItem *item, HttpReply *reply)
{
    memcpy(reply, &item->reply, sizeof(HttpReply));
    reply->buffer = buffer_ref(item->reply.buffer);
}

/*
 * Create a memory cache with the given cache size (in bytes),
 * and given maximum size for a single reply (in bytes).
 */
void cache_init(MemoryCache *cache, long cachesize, int max_cacheditem_size)
{
    /* Assume typical page is 8kb, and http reply headers + url = 1kb */
    int average_item_size = 9*1024;
    cache->maxsize = cachesize;
    cache->maxitems = cachesize / average_item_size;
    cache->size = 0;
    cache->itemcount = 0;
    cache->max_cacheditem_size = max_cacheditem_size;
    pool_init(&cache->cachedItemPool, sizeof(CachedItem), cache->maxitems*2/3);
    hashtable_init(&cache->table, cache->maxitems);
    lru_init(&cache->lru, cache);
}

/* Return the amount of memory (in bytes) needed to store the
 * given http reply and url. Include the content, reply header,
 * url, and CachedItem data structure.  Since data is stored in
 * chunks of 4096, round contentlength up to nearest chunk value.
 */
static int get_memory_size(HttpReply *reply, Buffer *url, int contentlength)
{
    int amount;
    amount = (contentlength + BODY_SIZE-1)/BODY_SIZE * BODY_SIZE;
    amount = amount + reply->buffer->capacity +
             url->capacity + sizeof(CachedItem);
    return amount;
}

/*
 * Create an item in the memory cache, with the given http reply and url.
 * The body will be added later. If the same url is already in the cache,
 * return NULL, regardless of whether it's completely downloaded or not.
 * The cache may delete older items if it has reached the maximum size.
 *
 * You must call cache_create_done() or cache_create_aborted() after
 * the body has been written to the cached item.
 */
CachedItem* cache_create(MemoryCache *cache, Buffer *url, HttpReply *reply)
{
    CachedItem *item;
    int amount;
    int result;
    int estimatedsize;

    /* If reply is being gzipped, we don't know the final content-length yet.
     * Estimate it to be contentlength/4
     */
    if (reply->newlygzipped) {
        estimatedsize = reply->contentlength / 4;
    } else {
        estimatedsize = reply->contentlength;
    }

    if (estimatedsize > cache->max_cacheditem_size) {
        return NULL;
    }
 
    item = hashtable_lookup(&cache->table, url);
    if (item != NULL) {
        return NULL;
    }

    amount = get_memory_size(reply, url, estimatedsize);

    /* If we've reached the cache size limit, delete older items */
    if (cache->size + amount > cache->maxsize) {
        result = cache_delete_amount(cache, amount);
        if (result != 1) {
            return NULL;
        }
    }

    /* If we've reached the maximum number of cached items,
     * delete an older item.
     */
    if (cache->itemcount >= cache->maxitems) {
        CachedItem *old = lru_getOldestReplacement(&cache->lru);
        if (old == NULL) {
            return NULL; 
        } else {
            cache_delete_item(cache, old);
        }
    }

    cache->size += amount;
    cache->itemcount++;

    /* Create a new CachedItem.  Add it to the hashtable and the
     * least recently used (LRU) list.  The body and bodycount
     * will be filled in as the page is downloaded and saved to
     * the cache.
     */
    item = (CachedItem *) pool_alloc(&cache->cachedItemPool);
    item->url = buffer_ref(url);
    memcpy(&item->reply, reply, sizeof(HttpReply));
    item->reply.buffer = buffer_ref(reply->buffer);
    item->orig_contentlength = reply->contentlength;
    item->bodycapacity = (estimatedsize + BODY_SIZE-1)/BODY_SIZE; 
    item->bodycount = 0;
    item->body = (Buffer **) calloc(item->bodycapacity, sizeof(Buffer *));
    item->lastreferenced = get_current_time();
    item->hashdata.next = NULL;
    item->lrudata.older = item->lrudata.newer = NULL;
    item->lockcount = 1;
    item->iscomplete = 0;

    hashtable_add(&cache->table, item);
    return item;
}

/*
 * Mark a cached item as complete.  Update the new content length 
 * if the item has been gzipped.  Now other connections can lookup
 * and use this cached item.
 */
void cache_create_done(MemoryCache *cache, CachedItem *item, 
                       int newcontentlength)
{
    item->iscomplete = 1;
    item->lockcount--;
    lru_addItem(&cache->lru, item);

    if (item->reply.newlygzipped) {
        int estamount, amount;
        estamount = get_memory_size(&item->reply, item->url, 
                                    item->orig_contentlength/4);
        amount = get_memory_size(&item->reply, item->url, newcontentlength);
        cache->size += (amount - estamount);
        if (cache->size > cache->maxsize) {
            cache_delete_amount(cache, cache->maxsize - cache->size);
        }
        item->reply.contentlength = newcontentlength;
    }
}

/* Abort the creation of this item. Delete any memory it uses. */
void cache_create_aborted(MemoryCache *cache, CachedItem *item)
{
    item->lockcount--;
    assert(item->lockcount == 0);

    /* The contentlength is used to update the cache size. When we
     * created the gzipped item, we estimated the final size to be 
     * contentlength/4, and we increased the cache->size by that amount. 
     * Now we must decrease the cache->size by that same amount.
     */
    if (item->reply.newlygzipped) {
        item->reply.contentlength = item->orig_contentlength/4;
    }
    cache_delete_item(cache, item);
}

/* Delete an item from the memory cache. Remove it from the 
 * hashtable and lru list. Free the url, HttpReply, and body buffers.
 * Free the actual cached item. Update the size and itemcount
 * in the cache.
 */
static void cache_delete_item(MemoryCache *cache, CachedItem *item)
{
    int i;

    assert(item->lockcount == 0);
    cache->size -= get_memory_size(&item->reply, item->url, 
                                   item->reply.contentlength);
    cache->itemcount--;

    if (item->iscomplete) {
        lru_removeItem(&cache->lru, item);
    }
    hashtable_remove(&cache->table, item);
    buffer_unref(item->url); item->url = NULL;
    buffer_unref(item->reply.buffer); item->reply.buffer = NULL;
    for (i = 0; i < item->bodycount; i++) {
        if (item->body[i] != NULL) {
            buffer_unref(item->body[i]);
            item->body[i] = NULL;
        }
    }
    free(item->body);
    pool_free(&cache->cachedItemPool, item);
}

/* Delete "numbytes" bytes from the Memory Cache. Return 1 if succesful, 
 * or 0 if not.  If too many cached items are in use, we may not be able 
 * to delete the requested amount of data.
 */
static int cache_delete_amount(MemoryCache *cache, long numbytes)
{
    CachedItem *item;
    LRUList *lru;
    
    lru = &cache->lru;
    while (numbytes > 0) {
        item = lru_getOldestReplacement(lru);
        if (item == NULL) {
            return 0;
        }
        numbytes = numbytes - get_memory_size(&item->reply, item->url, 
                                              item->reply.contentlength);
        cache_delete_item(cache, item);
    }
    return 1;
}

/*
 * Open the cached item with the given url.
 * Lock and return the cached item if it exists, or return NULL.
 * Update the item's position in the LRU (least recently used)
 * list.  If the item has expired, delete it and return NULL.
 *
 * If the item is compressed, it can only be returned if the
 * client accepts gzip encoding (i.e. if acceptgzip is true).
 */
CachedItem* cache_get(MemoryCache *cache, Buffer *url, int acceptgzip)
{
    CachedItem *item;
 
    if (cache->maxitems == 0) {
        return NULL;
    }
 
    item = hashtable_lookup(&cache->table, url);
    if (item == NULL) {
        return NULL;
    }
    if (!item->iscomplete) {
        return NULL;
    }
    if (item->reply.expires < get_current_time()) {
        if (item->lockcount == 0)
            cache_delete_item(cache, item);
        return NULL;
    }
    if (!acceptgzip && (item->reply.newlygzipped || item->reply.alreadygzipped))
        return NULL;

    lru_updateItem(&cache->lru, item);
    item->lockcount++;
    return item;
}

/* Close the given cached item, by unlocking it. If the item 
 * has expired, and no one else is using it, delete it.
 */
void cache_close(MemoryCache *cache, CachedItem *item)
{
    item->lockcount--;
    if (item->lockcount == 0 &&
        item->reply.expires < get_current_time()) {

        cache_delete_item(cache, item);
    }
}

/* Compare two buffers containing URLs. Return
 *  0 if they're equal
 * -1 if a1 < a2
 *  1 if a1 > a2
 * Note that a1 and a2 are Buffer** pointers.  This
 * routine is used by qsort() for comparison.
 */
static int url_compare(const void *a1, const void *a2)
{
    Buffer *b1, *b2;
    Buffer **ptr;
    int minlen;

    ptr = (Buffer**) a1;
    b1 = *ptr;
    ptr = (Buffer**) a2;
    b2 = *ptr;
    ptr = NULL;

    if (b1 == b2)
        return 0;
    if (b1 == NULL)
        return 1;
    if (b2 == NULL)
        return -1;

    if (b1->size < b2->size)
        minlen = b1->size;
    else
        minlen = b2->size;

    return strncmp(b1->buf, b2->buf, minlen);
}

/* Return a list of all the completed, non-expired URLs in
 * the cache.  The refcount of all the urls are incremented
 * so that they are not freed.  The CachedItems are not
 * locked, so normal adding/deleting of cached items can
 * continue.  Also, the URL list is sorted.
 */
CacheUrls* cacheurls_get(MemoryCache *cache)
{
    CacheUrls *list;
    CachedItem *item;
    int i, count;
    time_t now;

    list = (CacheUrls*)malloc(sizeof(CacheUrls));
    list->size = cache->itemcount;
    list->urls = (Buffer **) calloc(list->size, sizeof(Buffer *));
    list->conn = NULL;

    now = get_current_time(); 
    count = 0;
    for (i = 0; i < cache->table.bucketcount; i++) {
        item = cache->table.buckets[i];
        while (item != NULL) {
            if (item->iscomplete && count < list->size && 
                item->reply.expires > now) {

                list->urls[count] = buffer_ref(item->url);
                count++;
            }
            item = item->hashdata.next;
        }
    }
    list->size = count;
    qsort(list->urls, count, sizeof(Buffer*), url_compare);
    return list;
}

/* Free the list of URLs by decrementing the refcount
 * of each URL buffer.
 */
void cacheurls_free(CacheUrls *cacheurls)
{
    int i;
    for (i = 0; i < cacheurls->size; i++) {
        buffer_unref(cacheurls->urls[i]);
        cacheurls->urls[i] = NULL;
    }
    free(cacheurls->urls);
    free(cacheurls);
}

