/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#ifndef _CACHE_H_
#define _CACHE_H_

#include <unistd.h>
#include "http.h"
#include "buffer.h"
#include "conf.h"

typedef struct _CachedItem CachedItem;

struct _HashData
{
    CachedItem *next;
};
typedef struct _HashData HashData;

/* Hash table to lookup a CachedItem, using URL as key */
struct _HashTable
{
    long bucketcount;
    CachedItem **buckets;
};
typedef struct _HashTable HashTable;

/* The linked list of least recently used cache items.
 * The doubly-linked list is sorted from older to newer.
 */
struct _LRUList
{
    CachedItem *oldestItem;
    CachedItem *newestItem;
};
typedef struct _LRUList LRUList;

struct _LruData
{
    CachedItem *older;
    CachedItem *newer;
};
typedef struct _LruData LruData;

struct _CachedItem         /* Data stored for a cached item */
{
    Buffer *url;            /* The url of the item */
    HttpReply reply;        /* The Http reply */
    int orig_contentlength; /* Content-length before compression */
    int bodycapacity;       /* Capacity of body array */
    int bodycount;          /* The number of Buffers in the body */
    Buffer** body;          /* The Buffers of data for the body */
    time_t lastreferenced;  /* The time the item was last referenced */
    HashData hashdata;      /* Link to the next hash item in the lookup table */
    LruData lrudata;        /* Link to the next older/newer items in the cache */
    int lockcount;          /* Lock count on the item. When in use, the 
                             * item is locked. */
    boolean iscomplete;     /* True if the body has been completely downloaded. 
                             * False if not */
};

struct _MemoryCache      /* The Memory Cache used to store cached items */
{
    long size;             /* The current size of the cache, in bytes */
    long maxsize;          /* The maximum size the cache can hold */
    long itemcount;        /* The current number of items in the cache */
    long maxitems;         /* The maximum number of items the cache can hold */
    int max_cacheditem_size;  /* The maximum size of a single reply, in bytes */
    Pool cachedItemPool;   /* A pool of CachedItem structures */
    HashTable table;       /* A hash table to map method/URL -> Cached Items */
    LRUList lru;           /* The Least recently used item list */
};
typedef struct _MemoryCache MemoryCache;


struct _CacheUrls { /* List of all the Urls currently in the cache */
    Buffer **urls;  /* Array of the urls */
    long size;      /* Array length. */
    void *conn;     /* Pointer to the Connection structure */
};
typedef struct _CacheUrls CacheUrls;


void        cache_init(MemoryCache *cache, long maxsize, 
                       int max_cachedreply_size);
CachedItem* cache_create(MemoryCache *cache, Buffer *url, HttpReply *reply);
void        cache_create_done(MemoryCache *cache, CachedItem *item, 
                              int newcontentlength);
void        cache_create_aborted(MemoryCache *cache, CachedItem *item);
CachedItem* cache_get(MemoryCache *cache, Buffer *url, int acceptgzip);
void        cache_close(MemoryCache *cache, CachedItem *item);
void        cacheditem_httpreply(CachedItem *item, HttpReply *reply);
Buffer*     cacheditem_read(CachedItem *item, int* index);
void        cacheditem_write(CachedItem *item, Buffer *buf);
CacheUrls*  cacheurls_get(MemoryCache *cache);
void        cacheurls_free(CacheUrls *urls);

#endif

