/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**
  The cachetest does the following:

  - Initialize the MemoryCache. Set maxitems to 20.
  - Create 10 sample Http replies with
    url = www.domain.com/<id>
    body = line 0 <id>
           line 1 <id>
           ...
           line <BODY_COUNT> <id>
  - Add the 10 replies to the cache
  - Lookup the 10 replies from the cache, and verify
    that the url, body, and contentlength are correct.
  - Add 90 more replies to the cache.
  - Verify that the older items 0-10 have been deleted.
  - Verify that the newer items 90-100 are still there.
  - Retrieve the cache url list. Verify that items 90-100
    are included.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "cache.h"
#include "conf.h"

#define BODY_COUNT  10
#define BUFLEN      512

struct _Reply {
    int id;
    Buffer *url;
    HttpReply reply;
    Buffer **body;
    int bodycount;
};
typedef struct _Reply Reply;

BufferPool *bp;

Buffer* make_buffer(const char *s)
{
    Buffer *b = buffer_alloc(bp);
    buffer_add(b, s, strlen(s));
    return b;
}

void create_reply(int id, Reply *reply)
{
    Parsedata parse;
    Buffer *replybuf;
    char buf[BUFLEN];
    int contentlength, i, n;

    const char *replydata = 
"HTTP/1.0 200 OK\r\n\
Date: Mon, 28 Oct 2002 08:03:59 GMT\r\n\
Last-Modified: Wed, 24 Oct 2002 07:03:59 GMT\r\n\
Content-Type: text/html\r\n\
Connection: close\r\n\
Content-Length: %d\r\n\r\n";

    memset(reply, 0, sizeof(Reply));
    reply->id = id;
    contentlength = 0;
    reply->bodycount = BODY_COUNT;
    reply->body = (Buffer **) calloc(BODY_COUNT, sizeof(Buffer *)); 
    for (i = 0; i < BODY_COUNT; i++) {
        sprintf(buf, "line %d %.8d\n", i, id);
        reply->body[i] = make_buffer(buf);
        contentlength += strlen(buf);
    } 

    sprintf(buf, "www.domain.com/%.8d", id);
    reply->url = make_buffer(buf);

    sprintf(buf, replydata, contentlength);
    replybuf = make_buffer(buf);
    http_init_reply(&(reply->reply), replybuf, &parse);
    n = http_parse_reply(&(reply->reply), &parse);
    assert(n == PARSE_DONE); 
    assert(reply->reply.contentlength == contentlength);
}

void free_reply(Reply *r)
{
    int i;
    for (i = 0; i < BODY_COUNT; i++) {
        buffer_unref(r->body[i]);
    }
    buffer_unref(r->url);
    buffer_unref(r->reply.buffer);
    memset(r, 0, sizeof(r));
}

void check_cached_reply(CachedItem *item, int id)
{
    int i, n, pos, len;
    char buf[512];
    Buffer *data;

    sprintf(buf, "www.domain.com/%.8d", id);
    n = strncmp(item->url->buf, buf, item->url->size);
    assert(n == 0);

    pos = 0;
    i = 0;
    len = 0;
    data = cacheditem_read(item, &pos);
    while (data != NULL) {
        sprintf(buf, "line %d %.8d\n", i, id);
        n = strncmp(data->buf, buf, data->size);
        assert(n == 0);
        len += data->size;

        data = cacheditem_read(item, &pos);
        i++;
    }
    assert(i == BODY_COUNT);
    assert(len == item->reply.contentlength);
}



int main()
{
    MemoryCache *cache; 
    CachedItem *item;
    CacheUrls *cacheurls;
    Buffer *url;
    Reply reply;
    int MB = 1024 * 1024;
    int i, id;
    char buf[BUFLEN];

    printf("cachetest.........");

    Config.cache_expires = 14 * (24 * 60 * 60); /* Used by http_parse_reply */
    bp = malloc(sizeof(BufferPool));
    bufferpool_init(bp, 512, 30);

    cache = (MemoryCache*) malloc(sizeof(MemoryCache)); 
    cache_init(cache, 5 * MB, MB);

    /* Force maximum items in cache to be 20 */
    cache->maxitems = 20;

    /* Create 10 replies, and add them to the cache */
    for (id = 0; id < 10; id++) {
        create_reply(id, &reply);
        item = cache_create(cache, reply.url, &(reply.reply));
        assert(item != NULL);
        for (i = 0; i < BODY_COUNT; i++) {
            cacheditem_write(item, reply.body[i]);
        }
        cache_create_done(cache, item, reply.reply.contentlength);
        free_reply(&reply);
    }

    /* Retrieve the 10 replies, make sure they're correct */
    for (id = 0; id < 10; id++) {
        sprintf(buf, "www.domain.com/%.8d", id);
        url = make_buffer(buf);
        item = cache_get(cache, url, 1);
        assert(item != NULL);
        check_cached_reply(item, id);
        buffer_unref(url);
        cache_close(cache, item);
    }

    /* Create 90 replies and add to the cache. */
    for (id = 11; id < 100; id++) {
        create_reply(id, &reply);
        item = cache_create(cache, reply.url, &(reply.reply));
        assert(item != NULL);
        for (i = 0; i < BODY_COUNT; i++) {
            cacheditem_write(item, reply.body[i]);
        }
        cache_create_done(cache, item, reply.reply.contentlength);
        free_reply(&reply);
    }

    /* Check that older items (0-10) have been deleted */
    for (id = 0; id < 10; id++) {
        sprintf(buf, "www.domain.com/%.8d", id);
        url = make_buffer(buf);
        item = cache_get(cache, url, 1);
        assert(item == NULL);
        buffer_unref(url);
    }

    /* Check that newer items (90-100) are still there */
    for (id = 90; id < 100; id++) {
        sprintf(buf, "www.domain.com/%.8d", id);
        url = make_buffer(buf);
        item = cache_get(cache, url, 1);
        assert(item != NULL);
        cache_close(cache, item);
        buffer_unref(url);
    }

    cacheurls = cacheurls_get(cache);
    i = 1;
    /* The cacheurls list is sorted. Verify that the last
     * entries contain urls 90-99. 
     */

    for (id = 99; id > 90; id--) {
        url = cacheurls->urls[cacheurls->size - i];
        sprintf(buf, "www.domain.com/%.8d", id);
        assert(strncmp(url->buf, buf, url->size) == 0);
        i++;
    }
    cacheurls_free(cacheurls);
    printf("Pass\n");
    return 0; 
}


