/*
 * Copyright (c) 2009 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * The chunked_xxx functions are used to decode chunked-encoded HTTP transfers.
 * Chunked encoding is used to maintain persistent connections with the
 * client without knowing the content-length of the reply.
 * Chunked encoding has the following format
 *
 * <size of chunk in hex>; extra header\r\n
 * <data of size chunksize>\r\n
 *
 * The chunked transfer ends when the chunk size is 0.
 * An example chunked transfer is shown below.
 *
 * 1a; header info\r\n
 * abcdefghijklmnopqrstuvwxyz\r\n
 * 9\r\n
 * 123456789\r\n
 * 0\r\n
 * some-footer\r\n
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "chunked.h"
#include "win32.h"

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/* Find char c in buffer buf, between indexes offset and end.
 * Return the offset where it occurs, or -1 if not found.
 */
int findchar(const char *buf, int offset, int end, char c)
{
    if (offset < 0)
        return -1;
    while (offset < end) {
        if (buf[offset] == c) {
            return offset;
        }
        offset++;
    }
    return -1;
}

/* In the string buffer->buf[offset:end], parse a hex string
 * into an integer and return it.  Return -1 on error.
 */
static int
chunk_parse_hexsize(Buffer *buffer, int offset, int end)
{
    int i ;
    char c;

    /* Verify that all the characters are hex digits */
    for (i = offset; i < end; i++) {
        c = buffer->buf[i];
        if (c >= '0' && c <= '9')  {
        }
        else if (c >= 'a' && c <= 'f') {
        }
        else if (c >= 'A' && c <= 'F') {
        }
        else if (c == ' ' || c == ';' || c == '\r' || c == '\n') {
            break;
        }
        else {
            return -1;
        }
    }
    return strtoul(&buffer->buf[offset], NULL, 16);
}

/* Initialize the parser for parsing a chunked transfer encoding.
 * The initial state is "read size", which means we need to
 * read the chunk size.
 */
void chunked_init(ChunkedTransfer *c)
{
    c->state = CHUNKED_STATE_READ_SIZE;
    c->chunk_size = 0;
    c->chunk_left = 0;
    c->buffer_end = 0;
}

/* Return 1 if the chunked transfer is done, else 0.
 * Do not update the ChunkedTransfer structure.
 */
int chunked_is_done(ChunkedTransfer *c, Buffer *buffer)
{
    ChunkedTransfer copy;
    int ret;

    (void)memcpy(&copy, c, sizeof(ChunkedTransfer));
    ret = chunked_decode(&copy, buffer, NULL);
    if (ret == CHUNKED_DONE)
        return 1;
    else
        return 0;
}

/* Parse the given chunked-encoded data in the input buffer.
 *
 * If the output buffer is not NULL, then copy the raw data (without the
 * chunk size and \r\n) into the output buffer.
 *
 * The possible results are
 *   CHUNKED_ERROR    - Error parsing the data
 *   CHUNKED_DONE     - The chunked transfer is done.  The exact end of the data
 *                      is given in c->buffer_end.
 *   CHUNKED_READSIZE - We can't parse the chunk size until we read more data.
 *   CHUNKED_PARTIAL  - We're in the middle of a chunk.
 */
int chunked_decode(ChunkedTransfer *c, Buffer *input, Buffer *output)
{
    int amount, offset, newoffset;
    int orig_state = c->state;
    int orig_size = c->chunk_size;
    int orig_left = c->chunk_left;

    offset = 0;
    while (offset < input->size) {
        if (c->state == CHUNKED_STATE_READ_SIZE) {
            newoffset = findchar(input->buf, offset, input->size, '\n');
            if (newoffset == -1) {
                c->state = orig_state;
                c->chunk_size = orig_size;
                c->chunk_left = orig_left;
                return CHUNKED_READSIZE;
            }
            c->chunk_size = chunk_parse_hexsize(input, offset, newoffset);
            if (c->chunk_size == -1) {
                return CHUNKED_ERROR;
            }
            c->chunk_left = c->chunk_size;
            offset = newoffset + 1;
            c->state = CHUNKED_STATE_READ_BODY;
        }
        else if (c->state == CHUNKED_STATE_READ_BODY) {
            amount = MIN((input->size - offset), c->chunk_left);
            c->chunk_left -= amount;
            if (output != NULL) {
                buffer_add(output, &input->buf[offset], amount);
            }
            offset += amount;
            if (c->chunk_left == 0) {
                c->state = CHUNKED_STATE_READ_NEWLINE;
            }
        }
        else if (c->state == CHUNKED_STATE_READ_NEWLINE) {
            while (offset < input->size) {
                if (input->buf[offset] == '\r') {
                    offset++;
                    continue;
                }
                else if (input->buf[offset] == '\n') {
                    offset++;
                    c->state = CHUNKED_STATE_READ_SIZE;
                    if (c->chunk_size == 0) {
                        c->buffer_end = offset;
                        return CHUNKED_DONE;
                    }
                    else {
                        c->chunk_size = 0;
                        c->chunk_left = 0;
                        break;
                    }
                }
                else if (c->chunk_size == 0) {
                    /* A footer can appear at the end of the chunked transfer. */
                    offset++;
                }
                else {
                    return CHUNKED_ERROR;
                }
            }
        }
        else {
            assert(0);  /* Unknown state */
        }
    }
    return CHUNKED_PARTIAL;
}


/* Prepare a buffer for chunked encoding.
 * The first 6 bytes will be the chunk size (in hex) followed by \r\n.
 * In addition, the last two bytes must be \r\n.  Reserve the first 6
 * bytes and last 2 bytes by modifying the size and capacity.
 */
void chunk_encode_start(Buffer *b)
{
    int i;
    assert(b->size == 0);
    for (i = 0; i < CHUNKED_HEADERSIZE; i++) {
        b->buf[i] = ' ';
    }
    b->buf[CHUNKED_HEADERSIZE] = 0;
    b->size = CHUNKED_HEADERSIZE;
    b->capacity = b->capacity - 2;
}

/* Chunk encode a buffer.
 * In the first line of the buffer, add the chunk size (in hex) following
 * by \r\n.  Then, add \r\n to the end of the data.
 */
void chunk_encode_end(Buffer *b)
{
    int i;

    assert(b->size >= CHUNKED_HEADERSIZE && b->size <= 0x0000FFFF);
    for (i = 0; i < CHUNKED_HEADERSIZE; i++) {
        assert(b->buf[i] == ' ');
    }
    sprintf(b->buf, "%x", b->size - CHUNKED_HEADERSIZE);
    b->buf[ strlen(b->buf) ] = ' ';
    b->buf[CHUNKED_HEADERSIZE-2] = '\r';
    b->buf[CHUNKED_HEADERSIZE-1] = '\n';
    b->capacity += 2;
    b->buf[b->size] = '\r';
    b->buf[b->size+1] = '\n';
    if (b->size + 2 <  b->capacity) {
        b->buf[b->size+2] = 0;
    }
    b->size += 2;
}


