/*
 * Copyright (c) 2009 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/* The chunkedtest program parses different chunk-encoded strings, and
 * verifies that the chunk size, chunk state, and chunk bytes left
 * are all correct.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "buffer.h"
#include "chunked.h"

BufferPool *bp;
ChunkedTransfer c;
Buffer *buf;
Buffer *output;

Buffer* make_buf(const char *s) {
    buffer_reset(buf);
    buffer_add(buf, s, strlen(s));
    return buf;
}

/* Test a chunk that only has the chunksize */
void test_chunk_sizeonly()
{
    ChunkedTransfer c;
    Buffer *b;
    int ret;

    b = make_buf("1a\r\n");
    chunked_init(&c);
    ret = chunked_decode(&c, b, NULL);
    assert(ret == CHUNKED_PARTIAL);
    assert(c.state == CHUNKED_STATE_READ_BODY);
    assert(0x1a == 26);
    assert(c.chunk_size == 26);
    assert(c.chunk_left == 26);
}

/* Test a chunk with both a chunksize and body */
void test_chunk_body() {
    ChunkedTransfer c;
    Buffer *b;
    int ret;

    b = make_buf("1a\r\nabcdefghijklmnopqrstuvwxyz\r\n34\r\nabcdefg");
    chunked_init(&c);
    buffer_reset(output);
    ret = chunked_decode(&c, b, output);
    assert(ret == CHUNKED_PARTIAL);
    assert(c.state == CHUNKED_STATE_READ_BODY);
    assert(0x34 == 52);
    assert(c.chunk_size == 52);
    assert(c.chunk_left == (52 - 7));
    assert(output->size = 26 + 7);
    assert(strcmp(output->buf, "abcdefghijklmnopqrstuvwxyzabcdefg") == 0);
}

/* Test a chunk with a chunksize, body, and footer */
void test_chunk_with_footer() {
    ChunkedTransfer c;
    Buffer *b;
    const char *str = "1a\r\nabcdefghijklmnopqrstuvwxyz\r\n0\r\nfooter\r\n";
    int ret;

    b = make_buf(str);
    chunked_init(&c);
    buffer_reset(output);
    ret = chunked_decode(&c, b, output);
    assert(ret == CHUNKED_DONE);
    assert(c.buffer_end == strlen(str));
    assert(output->size == 26);
    assert(strcmp(output->buf, "abcdefghijklmnopqrstuvwxyz") == 0);
}

/* Test multiple chunks */
void test_multiple_chunks() {
    ChunkedTransfer c;
    Buffer *b;
    int ret;

    b = make_buf("1a\r\nabc");
    chunked_init(&c);
    ret = chunked_decode(&c, b, NULL);
    assert(ret == CHUNKED_PARTIAL);
    assert(c.state == CHUNKED_STATE_READ_BODY);
    assert(0x1a == 26);
    assert(c.chunk_size == 26);
    assert(c.chunk_left == (26 - 3));

    b = make_buf("defghijklmnopqrstuvwxyz");
    ret = chunked_decode(&c, b, NULL);
    assert(ret == CHUNKED_PARTIAL);
    assert(c.state == CHUNKED_STATE_READ_NEWLINE);
    assert(c.chunk_size == 26);
    assert(c.chunk_left == 0);
}

/* Test chunk-encoding a buffer */
void test_chunk_encode() {
    Buffer *b;
    b = make_buf("");
    chunk_encode_start(b);
    assert(b->capacity == (4096 - 2));
    buffer_add(b, "abcdefghijklmnopqrstuvwxyz", 26);
    chunk_encode_end(b);
    assert(strcmp(b->buf, "1a  \r\nabcdefghijklmnopqrstuvwxyz\r\n") == 0);
    assert(b->capacity == 4096);
    assert(b->size == 6 + 26 + 2);
}

int main()
{
    printf("chunkedtest.......");
    bp = malloc(sizeof(BufferPool));
    bufferpool_init(bp, 4096, 5);
    buf = buffer_alloc(bp);
    output = buffer_alloc(bp);
    test_chunk_sizeonly();
    test_chunk_body();
    test_chunk_with_footer();
    test_multiple_chunks();
    test_chunk_encode();
    printf("Pass\n");
    return 0;
}


