/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <pwd.h>
#include <errno.h>
#include <sys/resource.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "conf.h"
#include "win32.h"


/* The line number of the current directive being parsed in vicompress.conf */
static int linenumber = 0;

/* Set to 1 if any error is found while parsing vicompress.conf. */
static int configerror = 0;

/* Return the maximum file descriptor number
 * this process can open.
 */
static int get_total_fd()
{
    int count, error, i, result;
    int *fd;
    int maxfd = 50000; 

    fd = malloc(sizeof(int) * maxfd);
    memset(fd, 0, sizeof(int) * maxfd);

    count = 0;
    while (count < maxfd) {
        fd[count] = socket(AF_INET, SOCK_STREAM, 0);
        error = errno;
        if (fd[count] > 0) {
            count++;
        }
        else if (fd[count] < 0 && error != EINTR) {
            break;
        }
    }
    for (i = 0; i < count; i++) {
        if (fd[i] > 0) {
#ifdef WIN32
            win_close(fd[i]);
#else
            close(fd[i]);
#endif
        }
    }
    result = count+5;
    free(fd);
    return result;
}


/* Return the number of cpus on the system.
 * For Windows, call win_get_num_cpus() in win32.c
 * For Linux, read /proc/cpuinfo, and count the number of "processor" lines.
 * For other systems, just return 1.
 */
static int get_num_cpus()
{
#ifdef WIN32
    return win_get_num_cpus();
#else
    FILE *fp;
    char buffer[256];
    int num = 0;

    fp = fopen("/proc/cpuinfo", "r");
    if (fp == NULL) {
        return 1;
    }
    while (fgets(buffer, 256, fp) != NULL) {
        if (strncmp(buffer, "processor", strlen("processor")) == 0) {
            num++;
        }
    }
    fclose(fp);
    if (num == 0) {
        num = 1;
    }
    return num;
#endif
}

#ifdef HAVE_IPV6
/* Return 1 if the string has a colon */
static int has_colon(char *str, int len)
{
    int i;
    for (i = 0; i < len; i++) {
        if (str[i] == ':') {
            return 1;
        }
    }
    return 0;
}
#endif

/* Initialize default values for configuration options. 
 * Make sure the defaults shown here match those in the 
 * vicompress manual.
 */
static void config_init_defaults()
{
    Config.num_listen = 0;
    memset(&Config.listenaddr[0], 0, sizeof(sockaddr_all) * MAX_LISTEN);
    Config.num_webservers = 0;
    memset(&Config.webservers[0], 0, sizeof(Webserver)*MAX_WEBSERVERS);
    memset(&Config.outgoingip, 0, sizeof(sockaddr_all));
    Config.outgoingip.ipv4.sin_family = AF_INET;
    Config.enable_compression = 1;
    Config.enable_caching = 1;
    Config.cache_memory = 100 * MEGABYTE;
    Config.max_cacheditem_size = 512 * KILOBYTE;
    Config.cache_expires = 10 * DAY;
    Config.enable_dns_caching = 1;
    Config.dns_expires = 24 * HOUR;
    Config.user = getuid();
    Config.group = getgid();
    Config.max_total_fd = get_total_fd();
    Config.hostheader = NULL;
    Config.logformat = LOG_FORMAT_SQUID;
    Config.accesslog = NULL;
    Config.errorlog = NULL;
    Config.logstats = NULL;
    Config.errorpage = NULL;
    Config.errorpage_size = 0;
    Config.rotatesize = 100 * MEGABYTE;
    Config.enable_debug = 0;
    Config.enable_sessions = 1;
    Config.num_cpus = get_num_cpus();
}

/* Given a pathname, like 
 * "C:\Program Files\ViCompress"
 * remove the quotes, so that it becomes
 * C:\Program Files\ViCompress
 */
static char* stripquotes(char *token)
{
    if (token[0] == '"') {
        token[strlen(token) - 1] = 0;
        return &token[1];
    }
    else {
        return token;
    }
}

/* Give a buffer, split it into tokens using whitespace as delimiter. 
 * Return the number of tokens.
 */
static int splittokens(char *buf, char **result, int maxitems)
{
    int i, item, len, nexttospace, inquote;
    nexttospace = 1;
    inquote = 0;
    len = strlen(buf);
    item = 0;

    memset(result, 0, sizeof(char *) * maxitems);
    for (i = 0; i < len; i++) {
        if (buf[i] == ' ' || buf[i] == '\t' || 
            buf[i] == '\r' || buf[i] == '\n') {

            if (!inquote) {
                buf[i] = 0; nexttospace = 1;
            }
        }
        else {
            if (buf[i] == '"' || buf[i] == '\'') {
                inquote = !inquote;
            }
            if (nexttospace && item < maxitems) {
                result[item] = &buf[i];
                item++;
                nexttospace = 0;
            }
        }
    }
    for (i = 0; i < item; i++) {
        result[i] = stripquotes(result[i]);
    }
    return item;
}

/* Store the character string IP address and port into the
 * struct sockaddr_in/in6.  Return 0 on  success, -1 on error.
 */
int parse_ip(char *ip, int len, sockaddr_all *sock, int port)
{
    unsigned short nport;
    in_addr_t addr;

#ifdef HAVE_IPV6
    if (has_colon(ip, len)) {
        if (ip[0] == '[') {
            ip++;
        }
        if (inet_pton(AF_INET6, ip, &(sock->ipv6.sin6_addr)) <= 0) {
            return -1;
        }
        sock->ipv6.sin6_family = AF_INET6;
        sock->ipv6.sin6_port = htons((unsigned short)port);
        return 0;
    }
#endif

    addr = inet_addr(ip);
    if (addr == INADDR_NONE) {
        return -1;
    }

    sock->ipv4.sin_family = AF_INET;
    nport = htons((unsigned short) port);
    sock->ipv4.sin_port = nport;
    sock->ipv4.sin_addr.s_addr = addr;
    return 0;
}

/* Each store_xxx function below is used to store the configuration
 * options from vicompress.conf into the Config structure.
 */

static void store_listen(char *ip, char *portstr)
{
    int port;

    if (ip == NULL || portstr == NULL) {
        fprintf(stderr, "Config ERROR (line %d): listen takes 2 argument\n",
                linenumber);
        configerror = 1; 
        return;
    }
    port = atoi(portstr);
    if (port <= 0 || port > 65535) {
        fprintf(stderr, "Config ERROR (line %d): invalid port %s\n", 
                linenumber, portstr);
        configerror = 1;
        return;
    }
    if (Config.num_listen == MAX_LISTEN) {
        fprintf(stderr, "Config ERROR (line %d): Too many listen declarations, maximum is %d\n", linenumber, MAX_LISTEN);
        configerror = 1;
        return;
    }

    if (parse_ip(ip, strlen(ip), &(Config.listenaddr[Config.num_listen]), port) != 0) {
        fprintf(stderr, "Config ERROR (line %d): invalid listen IP address %s\n",
                linenumber, ip);
        configerror = 1;
        return;
    }
    Config.num_listen++;
}

static void store_outgoingip(char *ip)
{
    if (ip == NULL) {
        fprintf(stderr, "Config ERROR (line %d): outgoingip takes 1 argument\n",
                linenumber);
        configerror = 1; 
        return;
    }

    if (parse_ip(ip, strlen(ip), &(Config.outgoingip), 0) != 0) {
        fprintf(stderr, "Config ERROR (line %d): invalid outgoingip address %s\n",
                linenumber, ip);
        configerror = 1;
        return;
    }
}


static void store_enable_compression(const char *token)
{
    int len;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): enable_compression takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }

    len = strlen(token);
    if (len == 3 && strncasecmp(token, "yes", 3) == 0)
        Config.enable_compression = 1; 
    else if (len == 2 && strncasecmp(token, "no", 2) == 0)
        Config.enable_compression = 0;
    else {
        fprintf(stderr, 
                "Config ERROR (line %d): enable_compression must be yes or no, not %s\n",
                linenumber, token);
        configerror = 1;
        return;
    }
}


static void store_enable_caching(const char *token)
{
    int len;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): enable_caching takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }

    len = strlen(token);
    if (len == 3 && strncasecmp(token, "yes", 3) == 0)
        Config.enable_caching = 1; 
    else if (len == 2 && strncasecmp(token, "no", 2) == 0)
        Config.enable_caching = 0;
    else {
        fprintf(stderr, 
                "Config ERROR (line %d): enable_caching must be yes or no, not %s\n",
                linenumber, token);
        configerror = 1;
        return;
    }
}

static void store_enable_dns_caching(const char *token)
{
    int len;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): enable_dns_caching takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }

    len = strlen(token);
    if (len == 3 && strncasecmp(token, "yes", 3) == 0)
        Config.enable_dns_caching = 1; 
    else if (len == 2 && strncasecmp(token, "no", 2) == 0)
        Config.enable_dns_caching = 0;
    else {
        fprintf(stderr, 
                "Config ERROR (line %d): enable_dns_caching must be yes or no, not %s\n",
                linenumber, token);
        configerror = 1;
        return;
    }
}


static void store_user(char *token)
{
    struct passwd *p;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): user takes one argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    p = getpwnam(token);
    if (p == NULL) {
        fprintf(stderr, "Config ERROR (line %d): user %s does not exist\n",
                linenumber, token);
        configerror = 1;
        return;
    }
    Config.user = p->pw_uid;
    Config.group = p->pw_gid;
}

static void store_cache_memory(char *token)
{
    long size;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): cache_memory takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }

    size = atoi(token);

    if (!isdigit(token[0]) || size < 0) {
        fprintf(stderr, "Config ERROR (line %d): invalid cache_memory size %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
    else if (sizeof(long) == 4 && size >= 2048) {
        fprintf(stderr, "Config ERROR (line %d): cache_memory %s is too big\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
    else {
        Config.cache_memory = size * MEGABYTE;
    }
}

static void store_max_cacheditem_size(char *token)
{
    int size;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): max_cacheditem_size takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    size = atoi(token);
    if (size < 0 || size >= 2*1024*1024  || !isdigit(token[0]) ) {
        fprintf(stderr, "Config ERROR (line %d): invalid max_cacheditem_size %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
    else {
        Config.max_cacheditem_size = size * KILOBYTE;
    }
}

static void store_cache_expires(char *token)
{
    int hours;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): cache_expires takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    hours = atoi(token);
    if (hours < 0 || !isdigit(token[0]) ) {
        fprintf(stderr, "Config ERROR (line %d): invalid cache_expires %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
    else {
        Config.cache_expires = hours * HOUR;
    }
}


static void store_dns_expires(char *token)
{
    int hours;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): dns_expires takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    hours = atoi(token);
    if (hours < 0 || !isdigit(token[0]) ) {
        fprintf(stderr, "Config ERROR (line %d): invalid dns_expires %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
    else {
        Config.dns_expires = hours * HOUR;
    }
}

static void store_hostheader(const char *token)
{
    int len;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): hostheader takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    len = strlen(token);
    if (len > 256) {
        fprintf(stderr, "Config ERROR (line %d): hostheader too long: %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
    Config.hostheader = malloc(len+1);
    strcpy(Config.hostheader, token);
}


static void store_logformat(const char *token)
{
    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): logformat takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    if (strncasecmp(token, "squid", strlen("squid")) == 0)
        Config.logformat = LOG_FORMAT_SQUID;
    else if (strncasecmp(token, "apache", strlen("apache")) == 0)
        Config.logformat = LOG_FORMAT_APACHE;
    else {
        fprintf(stderr, "Config ERROR (line %d): invalid logformat %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
}

static void store_accesslog(const char *token)
{
    int len;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): accesslog takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    len = strlen(token);
    if (len > 256) {
        fprintf(stderr, "Config ERROR (line %d): invalid accesslog path %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
    Config.accesslog = malloc(len+1);
    strcpy(Config.accesslog, token);
}

static void store_errorlog(const char *token)
{
    int len;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): errorlog takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    len = strlen(token);
    if (len > 256) {
        fprintf(stderr, "Config ERROR (line %d): invalid errorlog path %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
    Config.errorlog = malloc(len+1);
    strcpy(Config.errorlog, token);
}

static void store_logstats(const char *token)
{
    struct stat info;
    int ret;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): logstats takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    memset(&info, 0, sizeof(struct stat));
    ret = stat(token, &info);
    if (ret != 0 || (info.st_mode & S_IFDIR) == 0) {
        fprintf(stderr, "Config ERROR (line %d): invalid logstats directory %s\n",
                linenumber, token);
        configerror = 1;
        return;
    }
    Config.logstats = malloc(strlen(token)+1);
    strcpy(Config.logstats, token);
}

static void store_errorpage(const char *token)
{
    struct stat info;
    int len, total, fd, ret;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): errorpage takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }

    ret = stat(token, &info);
    fd = open(token, O_RDONLY);
    if (ret != 0 || fd == -1) {
        fprintf(stderr, "Config ERROR (line %d): Unable to open errorpage %s, %s\n", 
                linenumber, token, strerror(errno));
        configerror = 1;
        return;
    }
    Config.errorpage_size = info.st_size;
    Config.errorpage = malloc(Config.errorpage_size);
    total = 0;

    do {
        len = read(fd, &(Config.errorpage[total]), Config.errorpage_size - total);
        total += len;
    }
    while (len > 0 && total < Config.errorpage_size);

    if (total != Config.errorpage_size) {
        fprintf(stderr, "Config ERROR (line %d): Unable to read errorpage %s\n", 
                linenumber, token);
        configerror = 1;
        return;
    }
}

static void store_rotatesize(const char *token)
{
    long size;
    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): rotatesize takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }
    size = atoi(token);
    if (size >= 2048) {
        fprintf(stderr, "Config ERROR (line %d): rotatesize %s must be < 2048\n", 
                linenumber, token);
        configerror = 1;
    } 
    if (size <= 0) {
        fprintf(stderr, "Config ERROR (line %d): invalid rotatesize %s\n", 
                linenumber, token);
        configerror = 1;
    }
    else {
        Config.rotatesize = size * MEGABYTE;
    }
}

static void store_enable_debug(const char *token)
{
    int len;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): enable_debug takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }

    len = strlen(token);
    if (len == 3 && strncasecmp(token, "yes", 3) == 0)
        Config.enable_debug = 1; 
    else if (len == 2 && strncasecmp(token, "no", 2) == 0)
        Config.enable_debug = 0;
    else {
        fprintf(stderr, 
                "Config ERROR (line %d): enable_debug must be yes or no, not %s\n",
                linenumber, token);
        configerror = 1;
        return;
    }
}


static void store_enable_sessions(const char *token)
{
    int len;

    if (token == NULL) {
        fprintf(stderr, "Config ERROR (line %d): enable_sessions takes 1 argument\n",
                linenumber);
        configerror = 1;
        return;
    }

    len = strlen(token);
    if (len == 3 && strncasecmp(token, "yes", 3) == 0)
        Config.enable_sessions = 1; 
    else if (len == 2 && strncasecmp(token, "no", 2) == 0)
        Config.enable_sessions = 0;
    else {
        fprintf(stderr, 
                "Config ERROR (line %d): enable_sessions must be yes or no, not %s\n",
                linenumber, token);
        configerror = 1;
        return;
    }
}


static void store_webserver(char **tokens, int numtokens)
{
    Webserver *server;
    char *ip, *portstr;
    int port, index, i;

    ip = tokens[1];
    portstr = tokens[2];

    if (ip == NULL || portstr == NULL) {
        fprintf(stderr, "Config ERROR (line %d): webserver takes 2 arguments\n",
                linenumber);
        configerror = 1;
        return;
    }
    port = atoi(portstr);
    if (port <= 0 || port > 65535) {
        fprintf(stderr, "Config ERROR (line %d): Invalid webserver port %s\n", 
                linenumber, portstr);
        configerror = 1;
        return;
    }

    index = Config.num_webservers;
    if (index >= MAX_WEBSERVERS) {
        fprintf(stderr, 
                "Config ERROR (line %d): too many webserver directives, maximum is %d\n", 
                linenumber, MAX_WEBSERVERS);
        configerror = 1;
        return;
    }

    server = &(Config.webservers[Config.num_webservers]);
    if (parse_ip(ip, strlen(ip), &(server->addr), port) != 0) {
        fprintf(stderr, "Config ERROR (line %d): Invalid webserver IP address %s\n", 
                linenumber, ip);
        configerror = 1;
        return;
    }
    server->num_hosts = 0;
    for (i = 3; i < numtokens; i++) {
        if (tokens[i][0] == '#') {
            break;
        }
        if (server->num_hosts == MAX_HOSTS) {
            fprintf(stderr, "Config ERROR (line %d): Too many webserver host names specified\n",
                    linenumber);
            configerror = 1;
            return;
        }
        server->hosts[server->num_hosts] = malloc(strlen(tokens[i]) + 1);
        strcpy(server->hosts[server->num_hosts], tokens[i]);
        server->num_hosts++;
    }

    Config.num_webservers++;
}

/* Read in the configfile (vicompress.conf) and store the settings
 * in the global Config variable.  Return 0 on success, -1 on error.
 */
int config_init(const char *configfile)
{
    FILE *fp;
    char buffer[512];
    char *tokens[5 + MAX_HOSTS];
    int i, n;

    config_init_defaults();
    linenumber = 0;
    configerror = 0;
    fp = fopen(configfile, "r");
    if (fp == NULL) {
        fprintf(stderr, "ERROR: cannot open config file %s\n", configfile);
        return -1;
    }
    while (fgets(buffer, 256, fp) != NULL) {
        if (buffer[0] == '#' || buffer[0] == '\r' || buffer[0] == '\n') {
            linenumber++;
            continue;
        }
        for (i = 0; i < MAX_HOSTS + 5; i++) {
            tokens[i] = NULL;
        }
        n = splittokens(buffer, tokens, MAX_HOSTS + 5);
        if (n <= 0) {
            linenumber++;
            continue;
        }

        if (tokens[0][0] == '#')
            continue;
        else if (strcmp(tokens[0], "listen") == 0)
            store_listen(tokens[1], tokens[2]);
        else if (strcmp(tokens[0], "outgoingip") == 0)
            store_outgoingip(tokens[1]);
        else if (strcmp(tokens[0], "webserver") == 0)
            store_webserver(tokens, n);
        else if (strcmp(tokens[0], "enable_compression") == 0)
            store_enable_compression(tokens[1]);
        else if (strcmp(tokens[0], "enable_caching") == 0)
            store_enable_caching(tokens[1]);
        else if (strcmp(tokens[0], "cache_memory") == 0)
            store_cache_memory(tokens[1]);
        else if (strcmp(tokens[0], "max_cacheditem_size") == 0)
            store_max_cacheditem_size(tokens[1]);
        else if (strcmp(tokens[0], "cache_expires") == 0)
            store_cache_expires(tokens[1]);
        else if (strcmp(tokens[0], "enable_dns_caching") == 0)
            store_enable_dns_caching(tokens[1]);
        else if (strcmp(tokens[0], "dns_expires") == 0)
            store_dns_expires(tokens[1]);
        else if (strcmp(tokens[0], "user") == 0)
            store_user(tokens[1]);
        else if (strcmp(tokens[0], "hostheader") == 0)
            store_hostheader(tokens[1]);
        else if (strcmp(tokens[0], "logformat") == 0)
            store_logformat(tokens[1]);
        else if (strcmp(tokens[0], "accesslog") == 0)
            store_accesslog(tokens[1]);
        else if (strcmp(tokens[0], "errorlog") == 0)
            store_errorlog(tokens[1]);
        else if (strcmp(tokens[0], "logstats") == 0)
            store_logstats(tokens[1]);
        else if (strcmp(tokens[0], "errorpage") == 0)
            store_errorpage(tokens[1]);
        else if (strcmp(tokens[0], "rotatesize") == 0)
            store_rotatesize(tokens[1]);
        else if (strcmp(tokens[0], "enable_debug") == 0)
            store_enable_debug(tokens[1]);
        else if (strcmp(tokens[0], "enable_sessions") == 0)
            store_enable_sessions(tokens[1]);
        else if (strcmp(tokens[0], "listenip") == 0) {
            fprintf(stderr, "Config ERROR (line %d): The listenip directive has been deprecated. Use the listen directive instead\n", linenumber);
            configerror = 1;
        } 
        else if (strcmp(tokens[0], "listenport") == 0) {
            fprintf(stderr, "Config ERROR (line %d): The listenport directive has been deprecated. Use the listen directive instead\n", linenumber);
            configerror = 1;
        } 
        else if (tokens[0] != NULL) {
            fprintf(stderr, "Config ERROR (line %d): invalid directive %s\n", 
                    linenumber, tokens[0]);
            configerror = 1;
        }
        linenumber++;
    }
    if (configerror) {
        return -1;
    }


    /* If no "listen" option is specified in vicompress.conf, the default
     * value is all IPv4 interfaces (0.0.0.0) port 80.
     */
    if (Config.num_listen == 0) {
        Config.num_listen = 1;
        Config.listenaddr[0].ipv4.sin_family = AF_INET;
        Config.listenaddr[0].ipv4.sin_port = htons(80);
    }

    if (!Config.enable_caching)
        Config.cache_memory = 0;

    /* If we're accelerating a set of backend webservers, then
     * DNS lookups are not needed.
     */
    if (Config.num_webservers > 0)
        Config.enable_dns_caching = 0;

    if (Config.num_webservers == 0)
        Config.hostheader = NULL;

    /* Sticky sessions are only needed if we are load balancing
     * among several backend webservers.
     */
    if (Config.num_webservers <= 1) 
        Config.enable_sessions = 0;

    memset(Config.myhostname, 0, 80);
    gethostname(Config.myhostname, 80);
    return 0;
}

