/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#ifndef _CONF_H_
#define _CONF_H_

#include <netinet/in.h>
#include <unistd.h>
#include "config.h"

/*
 * The Config structure is a global read-only structure that 
 * represents the vicompress.conf config file.  It is initialized 
 * by reading the vicompress.conf file.
 */

/* Common constants for size and time */

#define MAX_LONG    2147483647  /* Max long is pow(2,31) -1 */
#define GIGABYTE    1073741824  /* 1 GB = 1024 * 1024 * 1024 */
#define MEGABYTE    1048576
#define KILOBYTE    1024
#define HOUR        3600    /* 3600 sec = 1 hour */
#define DAY         86400   /* 86400 sec = 1 day */

/* Log formats */
#define LOG_FORMAT_SQUID   1
#define LOG_FORMAT_APACHE  2

/* Maximum number of backend webservers */
#define MAX_WEBSERVERS  64

/* Maximum number of hosts per webserver */
#define MAX_HOSTS       32

/* Maximum number of listening sockets */
#define MAX_LISTEN      16

/* Define a boolean type, for better readability */
typedef char boolean;


/* structure to hold both IPv4 and IPv6 addresses */
typedef union {
    struct sockaddr_in  ipv4;
#ifdef HAVE_IPV6
    struct sockaddr_in6 ipv6;
#endif
} sockaddr_all;

/* webserver entry */
struct _Webserver {
    sockaddr_all addr;      /* IP address and port */
    char *hosts[MAX_HOSTS]; /* Virtual hosts for this webserver */
    int num_hosts;          /* Number of virtual hosts */
    int active_conn;        /* Number of active connections */
    time_t last_used;       /* Time this webserver was last used */
    boolean is_down;           /* True if the webserver is down */
};
typedef struct _Webserver Webserver;


/* The list of Configuration options from vicompress.conf */
struct _Config {
    sockaddr_all listenaddr[MAX_LISTEN]; /* IP/port to listen on */
    int num_listen;             /* Number of listening sockets */
    Webserver webservers[MAX_WEBSERVERS];  /* Webservers to accelerate */
    int num_webservers;         /* Number of webservers specified.  If 0,
                                 * then use the hostname in the http request
                                 * to determine the origin web server.
                                 */
    sockaddr_all outgoingip;    /* Outgoing IP address */
    boolean enable_compression; /* True if gzip compression is enabled */
    boolean enable_caching;     /* True if caching is enabled. */
    long cache_memory;          /* cache memory (bytes) */
    int max_cacheditem_size;    /* Maximum item size that can be cached (bytes) */
    int cache_expires;          /* Time (seconds) when a cached item expires */
    boolean enable_dns_caching; /* True if dns lookups are cached */
    time_t dns_expires;         /* Time when DNS lookup expires (in seconds) */
    uid_t user;                 /* User to run as */
    gid_t group;                /* Group to run as */
    char *hostheader;           /* HTTP Host header to send to webserver */
    char logformat;             /* Apache or Squid */
    char *accesslog;            /* Path to the access log */
    char *errorlog;             /* Path to the error log */
    char *logstats;             /* Directory to store log statistics */
    char *errorpage;            /* The HTML page to use for errors */
    int errorpage_size;         /* The size of the HTML errorpage in bytes */
    long rotatesize;            /* Size (in bytes) at which to do log rotation */
    boolean enable_debug;       /* If yes, write debug info to the errorlog. */
    boolean enable_sessions;    /* If true, enable sticky sessions */
    int max_total_fd;           /* Total number of sockets that can be opened */
    int num_cpus;               /* Number of CPUs on the machine */
    char myhostname[80];        /* The hostname of this vicompress server */
} Config;

int  config_init(const char *configfile);
int  parse_ip(char *ip, int len, sockaddr_all *addr, int port);

#endif

