/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * Used to parse a date from one of the formats below:
 *   Wed, 06 Oct 2002 12:43:03 GMT
 *   Wednesday, 06-Oct-02 12:43:03 GMT
 *   Wed Oct 6 12:43:03 2002
 *
 */

#include <string.h>
#include <stdlib.h>
#include "date.h"

static const char* month[] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};


static int daysinmonth[] = {
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

/* Parse the given 3-letter month, and return the
 * integer value (0 thru 11)
 */
static int getmonth(const char *s, int offset)
{
    int i;
    for (i = 0; i < 12; i++) {
        if (strncmp(month[i], &s[offset], strlen(month[i])) == 0) {
            return i;
        }
    }
    return -1;
}

/* Find the offset where char c occurs in buf, between indexes
 * offset and end. Return -1 if not found.
 */
static int findchar(const char *buf, int offset, int end, char c)
{
    if (offset < 0)
        return -1;
    while (offset < end) {
        if (buf[offset] == c) {
            return offset;
        }
        offset++;
    }
    return -1;
}

/* Skip the given char in buf, between offset and end.
 * Return the new offset after skipping the char.
 */
static int skipchar(const char *buf, int offset, int end, char c)
{
    if (offset < 0)
        return -1;
    while (offset < end && buf[offset] == c)
        offset++;
    return offset;
}

/* Convert a struct tm to time_t (seconds).
 * Same as mktime(), but faster.
 */
static time_t do_mktime(struct tm *tmval)
{
    int leapdays;
    time_t result;
    int i;

    result = 365*24*60*60*(tmval->tm_year - 70);
    result = result + tmval->tm_sec + tmval->tm_min*60 + tmval->tm_hour*3600;
 
    for (i = 0; i < tmval->tm_mon; i++) {
        result = result + daysinmonth[i]*24*60*60;
    }
    result = result + (tmval->tm_mday - 1)*24*60*60;
 
    leapdays = (tmval->tm_year - 69) / 4;
    if ((tmval->tm_year % 4) == 0) {
        if (tmval->tm_mon > 1) {
            leapdays++;
        }
    }
    result = result + leapdays*24*60*60;
    return result;
}

/* Parse a date with the given format: 
 * Wed, 30 Oct 2002 12:43:03 GMT 
 * Return the time_t time of the date, or -1 on error.
 */
static time_t parse_date1(const char *string, int len)
{ 
    struct tm tmval;
    int offset = 0;  

    offset = findchar(string, offset, len, ',');
    offset += 2;
    offset = skipchar(string, offset, len, ' '); 
    if (offset == len) 
        return -1;
    tmval.tm_mday = atoi(&string[offset]);
    offset = findchar(string, offset, len, ' ');
    offset = skipchar(string, offset, len, ' ');
    if (offset == len) 
        return -1;
    tmval.tm_mon = getmonth(string, offset);
    if (tmval.tm_mon == -1)
         return -1;
    offset += 4;
    offset = skipchar(string, offset, len, ' ');
    if (offset == len) 
        return -1;
    tmval.tm_year = 1000*(string[offset] - '0') + 100*(string[offset+1] - '0')
                    + 10 * (string[offset+2] - '0') + (string[offset+3] - '0');
    tmval.tm_year -= 1900;
    offset += 5;
    offset = skipchar(string, offset, len, ' ');
    if (offset == len) 
        return -1;
    tmval.tm_hour = 10 * (string[offset] - '0') + string[offset+1] - '0';
    offset += 3;
    tmval.tm_min = 10 * (string[offset] - '0') + string[offset+1] - '0';
    offset += 3;
    tmval.tm_sec = 10 * (string[offset] - '0') + string[offset+1] - '0';
    return do_mktime(&tmval); 
}

/* Parse a date with the given format:
 * Wednesday, 30-Oct-02 12:43:03 GMT 
 * Return the time_t time of the date, or -1 on error.
 */
time_t parse_date2(const char *string, int len)
{
    struct tm tmval;
    int offset = 0;

    offset = findchar(string, offset, len, ',');
    offset += 2;
    offset = skipchar(string, offset, len, ' ');
    if (offset == len) 
        return -1;

    tmval.tm_mday = 10 * (string[offset] - '0') + string[offset+1] - '0'; 
    offset += 3;
    tmval.tm_mon = getmonth(string, offset);
    if (tmval.tm_mon == -1)
        return -1;
    offset += 4;
    tmval.tm_year = 10 * (string[offset] - '0') + string[offset+1] -'0';
    if (tmval.tm_year < 70)
        tmval.tm_year += 100;

    offset += 3;
    offset = skipchar(string, offset, len, ' ');
    if (offset == len) 
        return -1;
    tmval.tm_hour = 10 * (string[offset] - '0') + string[offset+1] - '0';
    offset += 3;
    tmval.tm_min = 10 * (string[offset] - '0') + string[offset+1] - '0';
    offset += 3;
    tmval.tm_sec = 10 * (string[offset] -'0') + string[offset+1] - '0';
    return do_mktime(&tmval);
}

/* Parse a date with the given format:
 * Wed Oct 30 12:43:03  2002
 * Return the time_t time of the date, or -1 on error.
 */
time_t parse_date3(const char *string, int len)
{
    struct tm tmval;
    int offset = 0;
    
    offset = skipchar(string, 0, len, ' ');
    offset += 4;

    offset = skipchar(string, offset, len, ' ');
    if (offset == len) 
        return -1;

    tmval.tm_mon = getmonth(string, offset);
    if (tmval.tm_mon == -1)
        return -1;

    offset += 4;

    offset = skipchar(string, offset, len, ' ');
    if (offset == len) 
        return -1;

    tmval.tm_mday = atoi(&string[offset]); 
    offset = findchar(string, offset, len, ' ');
    offset = skipchar(string, offset, len, ' ');
    if (offset == len) 
        return -1;

    tmval.tm_hour = 10 * (string[offset] - '0') + string[offset+1] - '0';
    offset += 3;
    tmval.tm_min = 10 * (string[offset] - '0') + string[offset+1] - '0';
    offset += 3;
    tmval.tm_sec = 10 * (string[offset] - '0') + string[offset+1] - '0';
    offset += 3;
    tmval.tm_year = 1000*(string[offset] - '0') + 100*(string[offset+1] - '0')
                    + 10*(string[offset+2] - '0') + string[offset+3] - '0';
    tmval.tm_year -= 1900;
    return do_mktime(&tmval);
}

/* Parse the given date string. Return the time_t
 * value, or -1 on error.
 */
time_t parse_date(const char *string, int len)
{
    int offset = 0;
    int start = 0;
    start = skipchar(string, 0, len, ' ');
    offset = findchar(string, 0, len, ',');
    if (offset == -1)
         return parse_date3(string, len);
    else if (offset-start == 3)
         return parse_date1(string, len);
    else
         return parse_date2(string, len); 
}

