/*
 * Copyright (c) 2003-2010 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * The DnsCache is used to perform dns lookups asynchronously.
 * The DnsCache also caches hostname to ip mappings,
 * so that the dns lookups are faster.
 */

#include <stdlib.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <pthread.h>
#include <stdio.h>
#include "dns.h"
#include "conf.h"
#include "currenttime.h"
#include "win32.h"

static void* dns_thread(void *arg);
static void dns_remove_entry(DnsCache *cache, DnsEntry *e);

/* Hash the given hostname. len is the length of the name.
 * The tablesize is the number of buckets in the hashtable.
 * The hash value is case sensitive.
 */
static unsigned int hash_item(const char *name, int len, int tablesize)
{
    unsigned int value = 53;
    int i;

    for (i = 0; i < len; i++) {
        value = (value * 31 + name[i]) % tablesize;
    }
    return value;
}

/* Find a number >=x that is somewhat prime */
static long findprime(long x)
{
    while (1) {
        if ((x % 2 != 0) && (x % 3 != 0) && (x % 5 != 0))
            return x;
        else
            x += 1;
    }
}

/* Initialize the hashtable used for cached dns lookups.
 * maxitems is the expected maximum number of items the
 * cache can hold.  Note that the bucket size is fixed
 * at creation, and does not grow during runtime.
 */ 
static void hashtable_init(DnsHashtable *table, int maxitems)
{
    int i;
    table->bucketcount = findprime(maxitems);
    table->buckets = 
      (DnsEntry **) calloc(table->bucketcount, sizeof(DnsEntry*));
    for (i = 0; i < table->bucketcount; i++) {
        table->buckets[i] = NULL;
    }
}

/* Add the given DnsEntry to the hash table.
 * The key is the hostname.
 */
static void hashtable_add(DnsHashtable *table, DnsEntry *item)
{
    DnsEntry *prev;
    unsigned int hash;

    item->hashdata.next = NULL;
    hash = hash_item(item->name, item->namelen, table->bucketcount);
    hash = hash % table->bucketcount;
    if (table->buckets[hash] == NULL) {
        table->buckets[hash] = item;
        return;
    }
    prev = table->buckets[hash];
    while (prev->hashdata.next != NULL) {
        prev = prev->hashdata.next;
    }
    prev->hashdata.next = item;
}

/* Remove the given DnsEntry from the hash table.
 * The key is the hostname.
 */
static void hashtable_remove(DnsHashtable *table, DnsEntry *item)
{
    unsigned int hash;
    DnsEntry *prev;

    hash = hash_item(item->name, item->namelen, table->bucketcount);
    hash = hash % table->bucketcount;
    if (table->buckets[hash] == item) {
        table->buckets[hash] = item->hashdata.next;
        item->hashdata.next = NULL;
        return;
    }
    prev = table->buckets[hash];
    while (prev != NULL && prev->hashdata.next != item)
        prev = prev->hashdata.next;
    if (prev == NULL)
        return;
    prev->hashdata.next = item->hashdata.next;
    item->hashdata.next = NULL;
}

/* Lookup the given hostname in the hash table. Return the 
 * DnsEntry if the item exists, or NULL if its not in the table.
 */
static DnsEntry* 
hashtable_lookup(DnsHashtable *table, const char *name, int len)
{
    DnsEntry *item;
    unsigned int hash;

    hash = hash_item(name, len, table->bucketcount);
    hash = hash % table->bucketcount;
    item = table->buckets[hash];
    while (item != NULL) {
        if (item->namelen == len && strncasecmp(name, item->name, len) == 0) {
            return item;
        }
        else {
            item = item->hashdata.next;
        }
    }
    return NULL;
}

/* Add an item to the LRU item list. The item will be placed
 * just before the newest item in the LRU list.
 */
static void lru_addItem(DnsCache *dnscache, DnsEntry *item)
{
    DnsEntry *prevNewest;
    prevNewest = dnscache->newest->lrudata.older;

    item->lrudata.older = prevNewest;
    item->lrudata.newer = dnscache->newest;
    prevNewest->lrudata.newer = item;
    dnscache->newest->lrudata.older = item;
}

/* Remove an item from the LRU item list. */
static void lru_removeItem(DnsCache *dnscache, DnsEntry *item)
{
    DnsEntry *prev, *next;
    prev = item->lrudata.older;
    next = item->lrudata.newer;
    prev->lrudata.newer = next;
    next->lrudata.older = prev;
}

/* Update an item in the LRU list. The item is removed
 * from its current position in the list, and inserted again,
 * at the newest slot.
 */
static void lru_updateItem(DnsCache *dnscache, DnsEntry *item)
{
    lru_removeItem(dnscache, item);
    item->lrudata.older = item->lrudata.newer = NULL;
    lru_addItem(dnscache, item);
}

/* Retrieve the oldest item in the LRU list (not counting
 * the dummy item "oldestItem").
 */
static DnsEntry* lru_getOldestReplacement(DnsCache *dnscache)
{
    DnsEntry *item;
    item = dnscache->oldest->lrudata.newer;
    if (item == dnscache->newest)
        return NULL;
    return item;
}


/* If valid, add the given request to the Dns cache.
 * The hostname must be less than 39 bytes, and must
 * not already exist in the cache.
 */
static void dns_add(DnsCache *dns, DnsRequest *req)
{
    DnsEntry *entry;

    if (req->namelen > MAX_CACHED_HOSTNAME_LEN)
        return;

    if (hashtable_lookup(&dns->table, req->name, req->namelen) != NULL)
        return;

    if (dns->currentitems == dns->maxitems) {
        entry = lru_getOldestReplacement(dns);
        assert(entry != NULL);
        dns_remove_entry(dns, entry);
        memset(entry, 0, sizeof(DnsEntry));
    }
    else {
        entry  = (DnsEntry *) pool_alloc(&dns->entrypool);
    }

    strncpy(entry->name, req->name, req->namelen);
    memcpy(entry->ip, req->ip, 4);
    entry->namelen = req->namelen;
    entry->expires = get_current_time() + dns->expires;
    hashtable_add(&dns->table, entry);
    lru_addItem(dns, entry);
    dns->currentitems++; 
}

/* Remove an entry from the Dns cache */
static void dns_remove_entry(DnsCache *dns, DnsEntry *entry)
{
    lru_removeItem(dns, entry);
    hashtable_remove(&dns->table, entry);
    dns->currentitems--;
}

/* Thread function which does DNS lookups asynchronously.
 * It retrieves requests from the dns->in queue, and places
 * the results on the dns->out queue.
 */
static void* dns_thread(void *arg)
{
    DnsCache *dns = (DnsCache *)arg;
    DnsRequest *request;
    struct hostent hostinfo;
    struct hostent *result;
    char *buf;
    int err, n, bufsize;

#ifdef HAVE_REALTIME_SIGNALS
    sigset_t signalset;
    sigemptyset(&signalset);
    sigaddset(&signalset, SIGRTMIN + 3);
    sigaddset(&signalset, SIGIO);
    sigprocmask(SIG_BLOCK, &signalset, NULL);
#endif

    signal(SIGPIPE, SIG_IGN);
    bufsize = 2048;
    buf = malloc(bufsize);
    while (1) {
        request = (DnsRequest *)queue_remove_wait(&dns->in);
        memset(&hostinfo, 0, sizeof(struct hostent));
        result = NULL;
        err = 0;
#ifdef __linux__
        /* For Linux, we use gethostbyname_r, which is thread-safe,
         * so we can spawn multiple dns threads.
         */
        memset(buf, 0, bufsize);
        n = gethostbyname_r(request->name, &hostinfo, buf, 
                            bufsize, &result, &err);
        if (n == 0) {
            memcpy(request->ip, hostinfo.h_addr_list[0], 4);
        }
#else
        result = gethostbyname(request->name);
        if (result != NULL) {
            memcpy(request->ip, result->h_addr_list[0], 4);
        }
#endif 
        else {
            memset(request->ip, 0, 4);
            request->error = 1;
        }
        if (request->ip[0] == 0) {
            request->error = 1;
        }
        queue_add(&dns->out, (void*) request);
    }
}

/*
 * Initialize the Dns cache, with the given maximum entires.
 * expires is the time in seconds the cached entry is valid.
 */
void dns_init(DnsCache *dns, int maxitems, int expires)
{
    pthread_attr_t attr;
    pthread_t id;
    int ret;

    dns->currentitems = 0;
    dns->expires = expires;
    dns->maxitems = maxitems;

    queue_init(&dns->in, Config.max_total_fd / 2);
    queue_init(&dns->out, Config.max_total_fd / 2);

    if (dns->maxitems > 0) {
        pool_init(&dns->entrypool, sizeof(DnsEntry), dns->maxitems*2/3);
        pool_init(&dns->reqpool, sizeof(DnsRequest), Config.max_total_fd/3);
        hashtable_init(&dns->table, dns->maxitems); 
        dns->oldest = pool_alloc(&dns->entrypool);
        dns->newest = pool_alloc(&dns->entrypool);
        memset(dns->oldest, 0, sizeof(DnsEntry));
        memset(dns->newest, 0, sizeof(DnsEntry));
        dns->oldest->lrudata.newer = dns->newest;
        dns->newest->lrudata.older = dns->oldest;
    }

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    ret = pthread_create(&id, &attr, dns_thread, dns); assert(ret == 0);
#ifdef __linux__
    /* We use gethostbyname_r for linux, so multiple dns lookup
     * threads may be spawned.
     */
    ret = pthread_create(&id, &attr, dns_thread, dns); assert(ret == 0);
    ret = pthread_create(&id, &attr, dns_thread, dns); assert(ret == 0);
#endif
}

/* Allocate a dns request structure */
DnsRequest* dns_alloc_request(DnsCache *dns) {
    return (DnsRequest*) pool_alloc(&dns->reqpool);
}

/* Free a dns request structure */
void dns_free_request(DnsCache *dns, DnsRequest *req) {
    pool_free(&dns->reqpool, req);
}

/* Process the given dns request asynchronously. The request
 * will have the name, namelen, callback, and data fields 
 * filled in.  This function will do a DNS lookup and fill
 * in the error and ip fields.  If the IP address is cached,
 * place the request directly in the dns->out queue.  Otherwise,
 * place the request in the dns->in queue, where it will be
 * processed by the dns_thread.
 */
void dns_lookup(DnsCache *dns, DnsRequest *request)
{
    DnsEntry *entry = NULL;

    request->error = 0; request->iscached = 0;

    if (dns->maxitems > 0) 
        entry = hashtable_lookup(&dns->table, request->name, request->namelen);

    if (entry != NULL && entry->expires >= get_current_time()) {
        lru_updateItem(dns, entry);
        memcpy(request->ip, entry->ip, 4);
        request->iscached = 1;
        queue_add(&dns->out, (void*) request);
        return;
    }

    if (entry != NULL) {
        dns_remove_entry(dns, entry);
        pool_free(&dns->entrypool, entry);
        entry = NULL;
    }
    queue_add(&dns->in, (void *)request);
}

/* If a DNS cache entry exists for the given hostname,
 * remove it.  This function is called when a connect()
 * fails, to ensure that we're not caching an invalid or
 * stale IP for that hostname.
 */
void dns_remove(DnsCache *dns, char *hostname, int namelen)
{
    DnsEntry *entry = NULL;
    if (dns->maxitems > 0) {
        entry = hashtable_lookup(&dns->table, hostname, namelen);
    }
    if (entry != NULL) {
        dns_remove_entry(dns, entry);
        pool_free(&dns->entrypool, entry);
        entry = NULL;
    }
}


/* Retrieve the DNS requests that have finished.  The ip and
 * error fields will be filled in.  The num is the size of
 * the results array.  Returns the number of finished requests.
 * Add successful DNS lookups to the DNS cache.
 */
int dns_getresults(DnsCache *dns, DnsRequest **results, int num)
{
    int i, n;

    n = queue_remove_many(&dns->out, (void **)results, num);

    if (dns->maxitems > 0) {
        for (i = 0; i < n; i++) {
            if (results[i]->error == 0 && results[i]->iscached == 0)
                dns_add(dns, results[i]);
        }
    }
    return n;
}


