/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#ifndef _DNS_CACHE_H_
#define _DNS_CACHE_H_

#include <sys/param.h>
#include "pool.h"
#include "queue.h"
#include "conf.h"

/* Only cache dns lookups for hostnames <= 39  bytes.
 * This keeps the dns cache small.
 */ 
#define MAX_CACHED_HOSTNAME_LEN 39

typedef struct _DnsEntry DnsEntry;

/* Callback used when dns lookup is done. ip is the 4 byte IP address.
 *  error == 0 on success, or nonzero on error.
 */
typedef void DnsCallback(unsigned char *ip, void *data, int error);

/* A Dns request issued by the client */
struct _DnsRequest
{
    char name[MAXHOSTNAMELEN];  /* hostname */
    char namelen;               /* length of hostname */
    unsigned char ip[4];        /* ip address */
    DnsCallback *callback;      /* callback function */
    boolean error;              /* true if an error occured */
    boolean iscached;           /* True if the lookup was cached */
    void *data;                 /* callback data */
};
typedef struct _DnsRequest DnsRequest;

struct _DnsHash
{
    DnsEntry *next;
};
typedef struct _DnsHash DnsHash;

struct _DnsHashtable
{
    int bucketcount;
    DnsEntry **buckets;
};
typedef struct _DnsHashtable DnsHashtable;

struct _DnsLru
{
    DnsEntry *older;
    DnsEntry *newer;
};
typedef struct _DnsLru DnsLru;

/* Entry for a cached Dns mapping */
struct _DnsEntry
{
    char name[MAX_CACHED_HOSTNAME_LEN + 1]; /* hostname */
    char namelen;         /* Length of hostname */
    unsigned char ip[4];  /* IP address */
    time_t expires;       /* Expiration time */
    DnsHash hashdata; 
    DnsLru lrudata;
};

/* The Dns cache structure */
struct _DnsCache
{
    int currentitems;    /* number of items in cache */
    int maxitems;        /* max items allowed in cache */
    Pool entrypool;      /* Pool of Dns entries */
    DnsHashtable table;  /* Hashtable of Dns entries */
    DnsEntry *oldest; 
    DnsEntry *newest; 
    Queue in;            /* Input request queue */
    Queue out;           /* Output result queue */
    Pool reqpool;        /* Pool of Dns requests */
    int expires;         /* Time a cached entry is valid (seconds) */
};
typedef struct _DnsCache DnsCache;

void dns_init(DnsCache *dns, int maxentries, int expires);
DnsRequest* dns_alloc_request(DnsCache *dns);
void dns_free_request(DnsCache *dns, DnsRequest *req);
void dns_lookup(DnsCache *dns, DnsRequest *req);
void dns_remove(DnsCache *dns, char *hostname, int namelen);
int dns_getresults(DnsCache *dns, DnsRequest **results, int num);

#endif

