/*
 * Copyright (c) 2003-2007 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**
  The dnstest does the following:

  1. The gethostbyname_r() system call is overridden. Our new gethostbyname_r()
     function takes hostnames of the form  host_zzz.zzz.zzz.zzz and returns
     the IP address zzz.zzz.zzz.zzz.

     Our gethostbyname_r() can be enabled/disabled.  When disabled, it always
     returns an error.

  2. The test will
     * Create a DNS cache with capacity for 1000 entries
     * Enable gethostbyname_r()
     * Lookup the IP address of 100 hostnames.
     * Disable gethostbyname_r()
     * Lookup the IP address of the same 100 hostnames.
     * Enable gethostbyname_r()
     * Lookup the IP address of 10000 hostnames.  

**/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <netdb.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "dns.h"
#include "pool.h"
#include "conf.h"
#include "win32.h"

#ifdef WIN32
#undef gethostbyname
#endif

#define MAX_ENTRIES   10000
#define DNS_CACHE_SIZE  101


void getip(const char *host, char *ip)
{
  in_addr_t s;
  assert(strncmp(host, "host_", 5) == 0);
  s = inet_addr(&(host[5]));
  assert(s != INADDR_NONE);
  memcpy(ip, &s, 4);
}

static int enable_gethostbyname = 0;

#ifdef __linux__
int gethostbyname_r(const char *name, struct hostent *ret, char *buf,
                    size_t buflen, struct hostent **result, int *err)
{
  if (!enable_gethostbyname) {
    *result = NULL;
    *err = HOST_NOT_FOUND;
    return -1;
  }

  ret->h_addr_list = (char **)&buf[0];
  getip(name, &buf[8]);
  ret->h_addr_list[0] = &buf[8];
  ret->h_length = 4;
  *result = ret;
  *err = 0;
  return 0;
}

#else /* If not linux, use gethostbyname() */

static struct hostent *myent = NULL;

struct hostent *gethostbyname(const char *name)
{
  if (!enable_gethostbyname) {
    return NULL;
  }

  if (myent == NULL) {
    myent = (struct hostent *) calloc(1, sizeof(struct hostent));
    myent->h_length = 4;
    myent->h_addr_list = malloc(sizeof(char *) * 8);
    myent->h_addr_list[0] = malloc(sizeof(char *) * 8);
  }
  getip(name, myent->h_addr_list[0]);
  return myent;
}

#endif /* linux */

void run(DnsCache *cache, int numentries)
{
  DnsRequest *req[MAX_ENTRIES];
  DnsRequest *resp[MAX_ENTRIES];
  char host[64];
  char ip[4]; 
  int i, n, total;

  for (i = 0; i < numentries; i++) {
    req[i] = dns_alloc_request(cache);
    memset(req[i], 0, sizeof(DnsRequest));
    sprintf(req[i]->name, "host_192.168.%d.%d", i / 256, i % 256);
    req[i]->namelen = strlen(req[i]->name);
    dns_lookup(cache, req[i]);
  }

  total = 0; 
  (void) usleep(10);
  while (total < numentries) {
    n = dns_getresults(cache, resp, numentries);
    total += n;
  }

  for (i = 0; i < numentries; i++) {
    sprintf(host, "host_192.168.%d.%d", i / 256, i % 256);
    getip(host, ip);
    assert(req[i]->error == 0);
    assert(memcmp(req[i]->ip, ip, 4) == 0);
    dns_free_request(cache, req[i]);
  }
}

int main() {
  DnsCache *cache;

  printf("dnstest...........");
#ifdef WIN32
  printf("skip\n");
  return 0;
#endif

  Config.max_total_fd = 50;
  Config.enable_dns_caching = 1;
  cache = (DnsCache *)malloc(sizeof(DnsCache));
  dns_init(cache, DNS_CACHE_SIZE, 60 * 60);

  enable_gethostbyname = 1;
  run(cache, 100);
  enable_gethostbyname = 0;
  run(cache, 100);
  enable_gethostbyname = 1;
  dns_remove(cache, "host_192.168.0.0", strlen("host_192.168.0.0"));
  run(cache, MAX_ENTRIES);
  printf("Pass\n");

  return 0;
}


