/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#include "config.h"

#ifndef HAVE_EPOLL
#include "epoll.h"
#include "conf.h"

/* This file contains a fake epoll() implementation for platforms that 
 * don't have epoll().  Currently there are 4 fake implementations:
 * - kqueue (for FreeBSD and Mac OS X)
 * - /dev/poll (for Solaris and HPUX)
 * - realtime signals (for Linux 2.4)
 * - poll (for others)
 *
 * libevent now provides the same thing (a common API for epoll ,kqueue, 
 * /dev/poll, etc), but libevent wasn't available when this was written.
 * Still, we'll continue to use this method.
 */
#include <sys/poll.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>


#ifdef HAVE_KQUEUE
#include <sys/event.h>
#include <sys/time.h>

static struct epoll_event *epollevents = NULL;
static int arraysize = 0;
static struct kevent *kevents;

int epoll_create(int size)
{
    int i, fd;

    signal(SIGPIPE, SIG_IGN);
    fd = kqueue();
    arraysize = size+1;
    epollevents = 
      (struct epoll_event*) calloc(size, sizeof(struct epoll_event));
    kevents = (struct kevent*) calloc(size, sizeof(struct kevent));

    for (i = 0; i < size; i++) {
        epollevents[i].data.fd = -1;
        kevents[i].udata = (void*) -1;
    }
    return fd;

} 

int epoll_ctl(int epfd, int op, int fd, struct epoll_event *event)
{
    struct kevent kev;
    struct epoll_event *ep;

    if (fd < 0 || fd >= arraysize) {
        return -1;
    }

    memset(&kev, 0, sizeof(struct kevent));
    kev.ident = fd;
    kev.udata = (void*) fd;
    if (event->events & EPOLLIN) {
        kev.filter = EVFILT_READ;
    }
    else if (event->events & EPOLLOUT) {
        kev.filter = EVFILT_WRITE;
    }
    else {
        kev.filter = 0; 
    }

    ep = &epollevents[fd];
    if (op == EPOLL_CTL_ADD) {
        ep->events = event->events & (EPOLLIN|EPOLLOUT);
        ep->data = event->data;
        kev.flags = EV_ADD | EV_ENABLE;
        kevent(epfd, &kev, 1, NULL, 0, NULL);
    }
    else if (op == EPOLL_CTL_MOD) {
        ep->events = event->events & (EPOLLIN | EPOLLOUT);
        ep->data = event->data;
        kev.flags = EV_ADD | EV_ENABLE;
        kevent(epfd, &kev, 1, NULL, 0, NULL);
    }
    else if (op == EPOLL_CTL_DEL) {
        ep->data.fd = -1;
        ep->events = 0;
        /* kqueue automatically deletes events when a fd is closed */
    }
    return 0;
}

int epoll_wait(int epfd, struct epoll_event *events,
               int maxevents, int msec) {

    struct timespec timeout;
    int fd, nevents, index, i;
    struct epoll_event *ep;
   
    index = 0; 
    memset(kevents, 0, sizeof(struct kevent) * arraysize);
    timeout.tv_sec = msec / 1000;
    timeout.tv_nsec = (msec % 1000) * 1000 * 1000;
    nevents = kevent(epfd, NULL, 0, kevents, arraysize, &timeout);
    for (i = 0; i < nevents; i++)  {
        fd = (int)kevents[i].udata;
        if (fd < 0 || fd >= arraysize) {
            continue;
        }
        ep = &epollevents[fd];
        events[index].events = epollevents[fd].events;
        events[index].data.fd = fd;
        index++;
    }
    return index;
}

#endif /* HAVE_KQUEUE */





#ifdef HAVE_DEVPOLL
#include <sys/devpoll.h>
#include <sys/poll.h>

static struct epoll_event *epollevents = NULL;
static int arraysize = 0;
static struct pollfd *devpollfds;

int epoll_create(int size)
{
    int i, fd;

    signal(SIGPIPE, SIG_IGN);
    fd = open("/dev/poll", O_RDWR);
    if (fd < 0) {
        fprintf(stderr, "ERROR: Unable to open /dev/poll\n");
        exit(1);
    }
    arraysize = size+1;
    epollevents = 
      (struct epoll_event*) calloc(size, sizeof(struct epoll_event));
    devpollfds = (struct pollfd*) calloc(size, sizeof(struct pollfd));

    for (i = 0; i < size; i++) {
        epollevents[i].data.fd = -1;
    }
    return fd;
} 

int epoll_ctl(int epfd, int op, int fd, struct epoll_event *event)
{
    struct pollfd pfd;
    struct epoll_event *ep;
    int n;

    if (fd < 0 || fd >= arraysize) {
        return -1;
    }

    pfd.fd = fd;
    pfd.revents = 0;
    if (event->events & POLLIN) {
        pfd.events = POLLIN | POLLHUP | POLLERR;
    }
    else if (event->events & EPOLLOUT) {
        pfd.events = POLLOUT | POLLHUP | POLLERR;
    }
    else {
        pfd.events = 0;
    }

    ep = &epollevents[fd];
    if (op == EPOLL_CTL_ADD) {
        ep->events = event->events & (EPOLLIN|EPOLLOUT);
        ep->data = event->data;
        n = write(epfd, &pfd, sizeof(struct pollfd));
    }
    else if (op == EPOLL_CTL_MOD) {
        ep->events = event->events & (EPOLLIN | EPOLLOUT);
        ep->data = event->data;
        n = write(epfd, &pfd, sizeof(struct pollfd));
    }
    else if (op == EPOLL_CTL_DEL) {
        ep->data.fd = -1;
        ep->events = 0;
        pfd.events = POLLREMOVE;
        n = write(epfd, &pfd, sizeof(struct pollfd));
    }
    return 0;
}

int epoll_wait(int epfd, struct epoll_event *events,
               int maxevents, int msec) {

    struct timespec timeout;
    int fd, nevents, index, n, i;
    struct epoll_event *ep;
    struct dvpoll dv;   
   
    index = 0; 
    dv.dp_timeout = msec;
    dv.dp_nfds = arraysize;
    dv.dp_fds = &(devpollfds[0]);
    n = ioctl(epfd, DP_POLL, &dv);
    if (n <= 0) {
        return 0;
    }

    for (i = 0; i < n; i++) {
        fd = devpollfds[i].fd;
        if (fd < 0 || fd >= arraysize) {
            continue;
        }
        ep = &epollevents[fd];
        events[index].events = epollevents[fd].events;
        events[index].data.fd = fd;
        index++;
    }
    return index;
}

#endif /* HAVE_DEVPOLL */






#ifdef HAVE_REALTIME_SIGNALS

#include <sys/poll.h>
#include "sockets.h"
#define EVENT_SIGNAL  SIGRTMIN+3

#ifndef F_SETSIG
#define F_SETSIG 10
#endif

static struct epoll_event *epollevents = NULL;
static int arraysize = 0;
static int maxfd = 0;
static int signal_overflow = 0;

/* Retrieve the maximum number of real-time signals
 * that a process may have queued, or -1 on error.
 */
static int get_max_rt_signals()
{
    FILE *fp;
    char buf[32];
    int n;

    fp = fopen("/proc/sys/kernel/rtsig-max", "r");
    if (fp != NULL) {
        buf[0] = 0;
        fgets(buf, 32, fp);
        n = atoi(buf);
        fclose(fp);
        return n;
    }
    return -1;
}

/* Set the maximum number of real-time signals
 * that a process may have queued.
 */
static void set_max_rt_signals(int amount)
{
    FILE *fp;

    fp = fopen("/proc/sys/kernel/rtsig-max", "w");
    if (fp != NULL) {
        fprintf(fp, "%d\n", amount);
        fclose(fp);
        return;
    }
}

static void enable_realtime_signal(int fd)
{
    int flags, err;

    err = 0;
    flags = fcntl(fd, F_GETFL, err);
    err = fcntl(fd, F_SETFL, flags | O_ASYNC);
    err = fcntl(fd, F_SETSIG, EVENT_SIGNAL);
    err = fcntl(fd, F_SETOWN, getpid());
}

static void disable_realtime_signal(int fd)
{
    int flags, err;
    err = 0;
    flags = fcntl(fd, F_GETFL, err);
    err = fcntl(fd, F_SETFL, flags & (~O_ASYNC));
}

static void init_realtime_signals() {
    sigset_t signalset;
    int maxsignals = 0;
 
    maxsignals = get_max_rt_signals();
    if (maxsignals > 0 && maxsignals < 32768) {
        set_max_rt_signals(32768);
    }

    sigemptyset(&signalset);
    sigaddset(&signalset, EVENT_SIGNAL);
    sigaddset(&signalset, SIGIO);
    sigprocmask(SIG_BLOCK, &signalset, NULL);
}


/* Clear all events from the real-time signal queue */
static void clearsigqueue()
{
    sigset_t signalset;
    int signum, counter;
    siginfo_t siginfo;
    struct timespec ts;

    sigemptyset(&signalset);
    sigaddset(&signalset, EVENT_SIGNAL);
    sigaddset(&signalset, SIGIO);
    sigprocmask(SIG_BLOCK, &signalset, NULL);

    ts.tv_sec = 0;
    ts.tv_nsec = 0;

    counter = 0;
    signum = sigtimedwait(&signalset, &siginfo, &ts);
    while (signum != -1 && counter < 32768) {
      signum = sigtimedwait(&signalset, &siginfo, &ts);
      counter++;
    }
}

/* Disable realtime signals on all sockets */
void disable_all_realtime_signals()
{
  int i;
  for (i = 0; i <= maxfd; i++) {
    if (epollevents[i].data.fd > 0) {
      assert(i == epollevents[i].data.fd);
      disable_realtime_signal(i);
    }
  }
}

/* Enable realtime signals on all sockets */
void enable_all_realtime_signals()
{
  int i;
  for (i = 0; i <= maxfd; i++) {
    if (epollevents[i].data.fd > 0) {
      assert(i == epollevents[i].data.fd);
      enable_realtime_signal(i);
    }
  }
}

int epoll_create(int size)
{
    int i;

    signal(SIGPIPE, SIG_IGN);
    arraysize = size+1;
    epollevents = (struct epoll_event*) calloc(size, sizeof(struct epoll_event));

    for (i = 0; i < size; i++) {
        epollevents[i].data.fd = -1;
    }
    init_realtime_signals();
    return 1;
}

int epoll_ctl(int epfd, int op, int fd, struct epoll_event *event)
{
    struct epoll_event *ep;

    if (fd < 0 || fd >= arraysize) {
        return -1;
    }
    ep  = &epollevents[fd];
    ep->events = event->events & (EPOLLIN | EPOLLOUT);
    ep->data = event->data;

    if (op == EPOLL_CTL_ADD) {
        enable_realtime_signal(fd);
        if (fd > maxfd) {
            maxfd = fd;
        }
    }
    else if (op == EPOLL_CTL_MOD) {
    }
    else if (op == EPOLL_CTL_DEL) {
        ep->data.fd = -1;
        ep->events = 0;
        disable_realtime_signal(fd);

        if (fd == maxfd) {
            maxfd--;
            while (maxfd > 0 && epollevents[maxfd].data.fd == -1) {
                maxfd--;
            }
        }

    }
    return 0;
}

int epoll_wait(int epfd, struct epoll_event *events,
                      int maxevents, int msec)
{
    int counter;
    struct timespec ts;
    int signum, fd, index;
    siginfo_t siginfo;
    sigset_t signalset;
    short revents;

    if (signal_overflow) {
        index = 0;
        for (fd = 0; fd <= maxfd; fd++) {
            if (epollevents[fd].events != 0) {
                events[index].data.fd = fd;
                events[index].events = epollevents[fd].events;
                index++;
            }
        }
        enable_all_realtime_signals();
        signal_overflow = 0;
        return index;
    }

    sigemptyset(&signalset);
    sigaddset(&signalset, EVENT_SIGNAL);
    sigaddset(&signalset, SIGIO);
    sigprocmask(SIG_BLOCK, &signalset, NULL);

    ts.tv_sec = msec/1000;
    ts.tv_nsec = (msec % 1000) * 1000;

    counter = index = 0;

    while (counter < 4096 && index < maxevents) {
        memset(&siginfo, 0, sizeof(siginfo_t));
        signum = sigtimedwait(&signalset, &siginfo, &ts);
        counter++;
        ts.tv_sec = 0;
        ts.tv_nsec = 0;

        if (signum == -1 && errno == EINTR)
            continue;
        if (signum != EVENT_SIGNAL && signum != SIGIO)
            break;

        if (signum == SIGIO) {
            disable_all_realtime_signals();
            clearsigqueue();
            signal_overflow = 1;
            return index;
        }

        fd = siginfo.si_fd;
        revents = siginfo.si_band;
        events[index].data.fd = fd;
        events[index].events = epollevents[fd].events;
        index++;
    }

    return index;
}

#endif /* HAVE_REALTIME_SIGNALS */





#ifdef HAVE_POLL 

#include <sys/poll.h>
static struct epoll_event *epollevents = NULL;
static int arraysize = 0;
static struct pollfd *pollevents;
static struct pollfd *polleventscopy;
static int maxfd = 0;

int epoll_create(int size)
{
    int i;

    signal(SIGPIPE, SIG_IGN);
    arraysize = size+1;
    epollevents = (struct epoll_event*) calloc(size, sizeof(struct epoll_event));
    pollevents     = (struct pollfd *) calloc(size, sizeof(struct pollfd));
    polleventscopy = (struct pollfd *) calloc(size, sizeof(struct pollfd));

    for (i = 0; i < size; i++) {
        epollevents[i].data.fd = -1;
        epollevents[i].events = 0;
        pollevents[i].fd = -1;
        pollevents[i].events = 0;
        pollevents[i].revents = 0;
        polleventscopy[i].fd = -1;
        polleventscopy[i].events = 0;
        polleventscopy[i].revents = 0;
    }

    return 1;
}

int epoll_ctl(int epfd, int op, int fd, struct epoll_event *event)
{
    struct pollfd *pfd;
    struct epoll_event *ep;

    if (fd < 0 || fd >= arraysize) {
        return -1;
    }

    pfd = &pollevents[fd];
    ep  = &epollevents[fd];
    ep->events = event->events & (EPOLLIN | EPOLLOUT);
    ep->data = event->data;

    pfd->fd = fd;
    pfd->revents = 0;
    if ((ep->events & EPOLLIN) && (ep->events & EPOLLOUT))
        pfd->events = POLLIN | POLLOUT| POLLHUP | POLLERR;
    else if (ep->events & EPOLLIN)
        pfd->events = POLLIN | POLLHUP | POLLERR;
    else if (ep->events & EPOLLOUT)
        pfd->events = POLLOUT | POLLHUP | POLLERR;
    else 
        pfd->events = 0;

    if (op == EPOLL_CTL_ADD) {
        if (fd > maxfd) {
            maxfd = fd;
        }
    }
    else if (op == EPOLL_CTL_MOD) {
    }
    else if (op == EPOLL_CTL_DEL) {
        ep->data.fd = -1;
        ep->events = 0;
        pfd->fd = -1;
        pfd->events = pfd->revents = 0;

        if (fd == maxfd) {
            maxfd--;
            while (maxfd > 0 && epollevents[maxfd].data.fd == -1) {
                maxfd--;
            }
        }
    }
    return 0;
}

int epoll_wait(int epfd, struct epoll_event *events,
               int maxevents, int msec)
{
    int index, fd, n, len;
    struct epoll_event *ep;
    struct pollfd *pfd;

    memcpy(polleventscopy, pollevents, sizeof(struct pollfd) * arraysize);
    len = maxfd + 1;

    n = poll(polleventscopy, len, msec);
    if (n < 0)
        return 0;

    index = 0;
    for (fd = 3; fd < len; fd++) {
        if (index >= maxevents) {
            return index;
        }
        pfd = &polleventscopy[fd];
        if (pfd->revents == 0) {
            continue;
        }

        ep = &(epollevents[pfd->fd]);
        events[index].events = ep->events;
        events[index].data.fd = pfd->fd;
        index++;
    }

    return index;
}

#endif /* HAVE_POLL */


#endif /* !HAVE_EPOLL */

