/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * The GzipEncoder is used to compress data using the gzip algorithm.
 * A GzipEncoder is allocated from a pool.  Input data is fed using
 * gzip_add(), and output is checked using gzip_has_output().  After
 * output has been used, the application calls gzip_set_output().
 *
 * The gzip_add() function does the actual compression in a separate
 * thread, to take advantage of machines with multiple processors.
 */

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <signal.h>
#include "gzipencoder.h"
#include "conf.h"
#include "win32.h"


#define WINDOW_LEVEL       13
#define MEM_LEVEL           5
#define COMPRESSION_LEVEL   7

/* Maximum number of items we will compress concurrently */
#define MAX_CONCURRENT_COMPRESS  200

static int window_size[] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 512,
    1024, 2048, 4096, 8192, 16384, 32768
};

static int memlevel_size[] = {
    256, 512, 1024, 2048, 4096, 8192,
    16384, 32768
};

#define CHUNKED_HEADERSIZE  6  /* Size reserved for chunk-encoded headers */

static void* gzip_thread(void *arg);
static void do_gzip_add(GzipEncoder *gz);

/* Allocate memory for the given number of items of the
 * given size.  This is a callback function registered
 * with zlib, which allows us to define our own memory
 * allocator.  We allocate memory from gz->memory, for
 * better performance than malloc.  If gz->memory runs 
 * out (it shouldn't), then fall back to malloc().
 */
static void* zalloc(void *v, uInt items, uInt size)
{
    GzipEncoder *gz = (GzipEncoder *)v;
    void *mem;

    if (gz->currsize + items * size < gz->total) {
        mem = &(gz->memory[gz->currsize]);
        gz->currsize += items * size;
    }
    else {
        mem = malloc(items * size);
    }
    return mem;
}

/* Free the given memory. */
static void zfree(void *v, void *mem)
{
    GzipEncoder *gz = (GzipEncoder *) v;
    char *m = mem;

    if (m < &gz->memory[0] || m > &gz->memory[gz->total-1]) {
        free(m);
    }
}

/* Create a pool of gzip encoders.  We limit the number of
 * concurrent requests that can be gzipped to 
 * MAX_CONCURRENT_COMPRESS (200), so that vicompress doesn't
 * get overwhelmed CPU wise under heavy load.
 *
 * The window size and memlevel size determine how much 
 * memory is required for zlib compression.  With our current
 * values of window=13 and memlevel=5, its about 54K of memory.
 * 54K * 200 = 10 MB, so its not too much.  We create a pool
 * of 56KB gzipmemory, so that allocation is fast.
 *
 * The actual compression is done in a separate thread (gzip_thread).
 * We spawn the gzip threads here.
 */
void gzip_init(GzipPool *pool)
{
    pthread_attr_t attr;
    pthread_t id;
    int ret, i;

    pool->gzmemorysize = window_size[WINDOW_LEVEL] * 4 + 
                         memlevel_size[MEM_LEVEL] * 2 +
                         4096 + 2048;

    pool->maxitems = MAX_CONCURRENT_COMPRESS;
    pool->freeitems = MAX_CONCURRENT_COMPRESS;
    pool_init(&pool->gzencoders, sizeof(GzipEncoder), pool->maxitems);
    pool_init(&pool->gzmemory, pool->gzmemorysize, pool->maxitems); 
    queue_init(&pool->inq, pool->maxitems);
    queue_init(&pool->outq, pool->maxitems);

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    for (i = 0; i < Config.num_cpus; i++) {
        ret = pthread_create(&id, &attr, gzip_thread, pool);
        assert(ret == 0);
    }
}

/* Allocate a gzip encoder from the pool.  If none are
 * available (if we've reached MAX_CONCURRENT_COMPRESS),
 * then return NULL.
 */
GzipEncoder* gzip_alloc(GzipPool *pool)
{
    GzipEncoder *gz;
    z_stream *zstream;
    if (pool->freeitems == 0)
        return NULL;
    pool->freeitems--;
    gz = (GzipEncoder *) pool_alloc(&pool->gzencoders);
    memset(gz, 0, sizeof(GzipEncoder));
    gz->gzippool = pool;
    gz->total = pool->gzmemorysize;
    gz->memory = pool_alloc(&pool->gzmemory);
    gz->crc = crc32(0, NULL, 0);
    zstream = &gz->zstream;
    zstream->zalloc = zalloc;
    zstream->zfree = zfree;
    zstream->opaque = gz;
    zstream->data_type = Z_ASCII;
    if (deflateInit2(zstream, COMPRESSION_LEVEL, Z_DEFLATED, 
                     -WINDOW_LEVEL, MEM_LEVEL, Z_DEFAULT_STRATEGY) != Z_OK) {
        gzip_free(pool, gz);
        return NULL;
    }
    return gz;
}

/* Free the given GzipEncoder back into the pool. */
void gzip_free(GzipPool *pool, GzipEncoder *gz)
{
    if (gz == NULL)
        return;
    gz->out = NULL;
    deflateEnd(&gz->zstream); 
    pool_free(&pool->gzmemory, gz->memory);
    gz->memory = NULL;
    pool_free(&pool->gzencoders, gz);
    pool->freeitems++;
}

/* Add the gzip header to the gz->out buffer. 
 * bytes 0-1 = magic number
 * byte  2   = compression method
 * byte  3   = flags
 * bytes 4-7 = timestamp
 * byte  8   = extra flags
 * byte  9   = OS (Unix=3) 
 */
static void add_gzip_header(GzipEncoder *gz)
{
    Buffer *out;
    char header[10];

    memset(header, 0, 10);
    header[0] = 0x1f;
    header[1] = 0x8b;
    header[2] = Z_DEFLATED;
    header[9] = 3;
    out = gz->out;
    assert(out->size == 0 || out->size == CHUNKED_HEADERSIZE);
    buffer_add(out, header, 10);
    gz->zstream.next_out = (unsigned char*) &out->buf[out->size];
    gz->zstream.avail_out = out->capacity - out->size;
}

/* Add the gzip footer to the gz->out buffer.
 * This is the 32-bit CRC of the uncompressed data,
 * and the 32-bit total length of the uncompressed data.
 */
static void add_gzip_footer(GzipEncoder *gz)
{
    Buffer *out = gz->out;
    int origcapacity = out->capacity;
    int i, size;
    uLong crc, totalin;

    if (gz->zstream.avail_out < 8) {
        /* The "out" buffer may (or may not) be chunk-encoded.
         * If it is chunk-encoded, the capacity needs to be reduced
         * by 2, so that the last 2 bytes are used for \r\n.
         * See function chunk_encode_start().
         *
         * Although we only need 8 extra bytes for the footer,
         * we increase the actual capacity by 10, and set the
         * out->capacity field to 8, in case the buffer is
         * chunk-encoded.
         */
        buffer_grow(out, out->capacity + 10);
        out->capacity -= 2;
        gz->zstream.next_out = 
          (unsigned char*) &out->buf[origcapacity - gz->zstream.avail_out];
        gz->zstream.avail_out += (out->capacity - origcapacity);
    }
    gz->zstream.avail_out -= 8; 
    crc = gz->crc; size = out->size;
    for (i = size; i < size + 4; i++) {
        out->buf[i] = (crc & 0xff);
        crc = crc >> 8;
    }
    size += 4;
    totalin = gz->zstream.total_in;
    for (i = size; i < size + 4; i++) {
        out->buf[i] = (totalin & 0xff);
        totalin = totalin >> 8;
    }
    out->size += 8; 
    gz->wrotefooter = 1;
}

/* Return true if the gzip encoder has output data ready. */
int gzip_has_output(GzipEncoder *gz)
{
    if (gz->zstream.avail_in != 0)
        return 1;
    else if (gz->out != NULL && gz->zstream.avail_out == 0)
        return 1;
    else if (gz->out != NULL && gz->zstream.avail_out < gz->out->capacity 
             && gz->flushed)
        return 1;
    else
        return 0;
}

/* Compress the input buffer data. The compression is done in 
 * a separate thread, to take advantage of multiprocessors.  
 * To retrieve the results, you must call gzip_get_finished().
 * The connection is stored in gz->connection, so that the
 * caller can easily retrieve the Connection when the compression
 * is finishsed.
 */
void gzip_add(GzipEncoder *gz, Buffer *in, void *connection)
{
    GzipPool *gzippool = gz->gzippool;

    assert(gz->zstream.avail_in == 0);
    assert(gz->out != NULL);
    assert(gz->zstream.next_out != NULL);
    assert(gz->zstream.avail_out > 0);

    gz->connection = connection;
    gz->in = in;

    queue_add(&(gzippool->inq), gz);
}

/* Perform the actual compression of the data. */
static void do_gzip_add(GzipEncoder *gz) {
    int result;
    Buffer *in = gz->in;

    if (in->size == 0) {
        gz->in = NULL;
        return;
    }   
 
    gz->crc = crc32(gz->crc, (unsigned char *)in->buf, in->size);
    gz->zstream.next_in = (unsigned char *)in->buf;
    gz->zstream.avail_in = in->size;
    
    result = deflate(&gz->zstream, Z_NO_FLUSH);
    if (gz->zstream.avail_out > 0 && gz->zstream.avail_in > 0) {
        result = deflate(&gz->zstream, Z_SYNC_FLUSH);
    }
    gz->out->size = gz->out->capacity - gz->zstream.avail_out;
    if (result != Z_OK)
        gz->error = 1;
    gz->in = NULL;
}

/* Retrieve the gzip encoders that have finished compressing
 * their data.  Return the number of encoders. The len is
 * the size of the array. The caller can retrieve the connection
 * associated with the gzipencoder using GzipEncoder->connection.
 */
int gzip_get_finished(GzipPool *pool, GzipEncoder **array, int len)
{
    if (pool->freeitems == pool->maxitems)
        return 0;
    else
        return queue_remove_many(&pool->outq, (void**)array, len);
}

/* Set the out buffer where compressed data should be placed.
 * The gzip encoder will immediately compress any leftover input
 * data it has, and place it in the out buffer. So its possible 
 * that output data can be ready right after calling this function.
 */
#define CHUNKED_HEADERSIZE  6
void gzip_set_output(GzipEncoder *gz, Buffer *out)
{
    int result;

    gz->zstream.next_out = (unsigned char *)(&out->buf[out->size]);
    gz->zstream.avail_out = out->capacity - out->size;

    assert(out->size == 0 || out->size == CHUNKED_HEADERSIZE);
    gz->out = out;

    /* If we already wrote the footer, return */
    if (gz->wrotefooter)
        return;

    /* If its the first time we set the output,
     * write the gzip header, and return.
     */
    if (!gz->wroteheader) {
        gz->wroteheader = 1;
        add_gzip_header(gz);
        return;
    }

    if (gz->flushed) {
        result = deflate(&gz->zstream, Z_FINISH);
        gz->out->size = out->capacity - gz->zstream.avail_out;
        if (result == Z_STREAM_END) {
            add_gzip_footer(gz);
        }
    }
    else {
        result = deflate(&gz->zstream, Z_NO_FLUSH);
        gz->out->size = out->capacity - gz->zstream.avail_out;
    }

    if (gz->zstream.avail_in == 0)
        gz->zstream.next_in = NULL;

}

/* The input data stream has ended.  Compress any
 * remaining input data, and write the gzip footer
 * to the gz->out buffer.
 */
void gzip_flush(GzipEncoder *gz)
{
    int hasoutput, result;
    hasoutput = gzip_has_output(gz);
    gz->flushed = 1;
    if (!hasoutput) {
        gz->zstream.next_in = NULL; gz->zstream.avail_in = 0;
        result = deflate(&gz->zstream, Z_FINISH);
        gz->out->size = gz->out->capacity - gz->zstream.avail_out;
        if (result == Z_STREAM_END) {
            add_gzip_footer(gz);
        }
    }
}

/* The thread which performs the gzip compression.
 * It reads GzipEncoders from pool->inq and places
 * the results in pool->outq.
 */
static void* gzip_thread(void *v)
{
    GzipPool *gzippool;
    GzipEncoder *gz;

#ifdef HAVE_REALTIME_SIGNALS
    sigset_t signalset;
    sigemptyset(&signalset);
    sigaddset(&signalset, SIGRTMIN+3);
    sigaddset(&signalset, SIGIO);
    sigprocmask(SIG_BLOCK, &signalset, NULL);
#endif

    signal(SIGPIPE, SIG_IGN);

    gzippool = v;
    while (1) {
        gz = (GzipEncoder *) queue_remove_wait(&gzippool->inq);
        do_gzip_add(gz);
        queue_add(&gzippool->outq, gz);
    }
    return NULL;
}


