/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**
    The gziptest will do the following:

    1. Create a random 20 kilobyte file gziptest.out
    2. Gzip the file using the gzipencoder, and save to gziptest.out.gz
    3. gunzip the gziptest.out.gz file
    4. Make sure gziptest.out equals gziptest.out

**/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <string.h>
#include "conf.h"
#include "gzipencoder.h"
#include "win32.h"

void writebuf(int fd, char *buf, int size);

void create_data_file(const char *file)
{
    int fd, i, n;
    const char *data = "The open system call is used to convert a pathname\
 into a file descriptor. When the call is successful, the file descriptor\
 returned will be the lowest file descriptor not currently open for the process.\n";

    fd = open(file, O_WRONLY | O_CREAT, 0666);
    assert(fd > 0);
    n = strlen(data);

    for (i = 0; i < 100; i++)
        writebuf(fd, (char *)data, n);

    close(fd);
}

int readBuf(int fd, Buffer *b)
{
    int n;

    b->size = 0;
    while (b->size < b->capacity) {
        n = read(fd, &b->buf[b->size], b->capacity - b->size);
        if (n < 0) {
            fprintf(stderr, "Error in readBuf\n");
            exit(1);
        }
        if (n == 0) {
            return 1;
        }
        b->size += n;
    }
    return 0;
}

void writeBuf(int fd, Buffer *b)
{
    writebuf(fd, b->buf, b->size);
}

void writebuf(int fd, char *buf, int size)
{
    int amount = 0;
    int n;

    while (amount < size) {
        n = write(fd, &buf[amount], size - amount);
        if (n <= 0) {
            fprintf(stderr, "Error in writebuf\n");
            exit(1);
        }
        amount += n;
    }
}

void checkoutput()
{
    int result;
    result = system("gunzip gziptest.out2.gz");
    assert(result == 0);
    result = system("diff gziptest.out gziptest.out2");
    if (result == 0) {
        printf("Pass\n");
        remove("gziptest.out2");
    }
    else {
        printf("Fail\n");
    }
}

void wait_gzip_finished(GzipPool *pool)
{
    GzipEncoder *result[1];
    int n;
    while (1) {
        n = gzip_get_finished(pool, result, 1);
        if (n == 1)
            return;
        (void)usleep(100);
    } 
}

int main()
{
    BufferPool bp;
    GzipPool gzpool;
    GzipEncoder *gz;
    Buffer *in;
    Buffer *out;
    int infd, outfd, eof;

    printf("gziptest..........");
    Config.num_cpus = 1;

    remove("gziptest.out.gz");
    remove("gziptest.out"); 
    create_data_file("gziptest.out"); 
    infd = open("gziptest.out", O_RDONLY);
    outfd = open("gziptest.out2.gz", O_WRONLY | O_CREAT | O_EXCL, 0666);
    assert(infd > 0);
    assert(outfd > 0);

    bufferpool_init(&bp, 4096, 10);
    in = buffer_alloc(&bp);
    out = buffer_alloc(&bp);
    gzip_init(&gzpool);
    gz = gzip_alloc(&gzpool);

    gzip_set_output(gz, out);
    eof = readBuf(infd, in);
    while (!eof) {
        gzip_add(gz, in, NULL);
        wait_gzip_finished(&gzpool);
        while (gzip_has_output(gz)) {
            writeBuf(outfd, out);
            buffer_reset(out);
            gzip_set_output(gz, out);
        }
        eof = readBuf(infd, in);
    }
    gzip_add(gz, in, NULL);
    wait_gzip_finished(&gzpool);
    gzip_flush(gz);
    while (gzip_has_output(gz)) {
        writeBuf(outfd, out);
        buffer_reset(out); 
        gzip_set_output(gz, out);
    }
    close(infd);
    close(outfd);
    gzip_free(&gzpool, gz);
    checkoutput();
    remove("gziptest.out");
    return 0;
}


