/*
 * Copyright (c) 2003-2010 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * The HttpRequest and HttpReply structures store information 
 * parsed from the Http request and reply headers. The structures 
 * are initialized using http_parse_request() and http_parse_reply(). 
 * Once initialized, the structure fields should be considered 
 * read-only.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <limits.h>
#include "http.h"
#include "buffer.h"
#include "date.h"
#include "conf.h"
#include "currenttime.h"
#include "version.h"

/* Only compress pages larger than this size. */
#define MIN_GZIP_SIZE  1024

/* Http Headers we care about */
const char *HeaderStrings[] =
{   "",
    "Accept-Encoding",
    "Authorization",
    "Cache-Control",
    "Connection",
    "Content-Encoding",
    "Content-Length",
    "Content-Range",
    "Content-Type",
    "Cookie",
    "Date",
    "Expect",
    "Expires",
    "Host",
    "If-Modified-Since",
    "If-Range",
    "Last-Modified",
    "Pragma",
    "Proxy-Connection",
    "Range",
    "Referer",
    "Set-Cookie",
    "TE",
    "Transfer-Encoding",
    "User-Agent",
    "Vary"
};

/* Enum values of Http Headers we care about */
enum {
    HDR_UNKNOWN = 0,
    HDR_ACCEPT_ENCODING,
    HDR_AUTHORIZATION,
    HDR_CACHE_CONTROL,
    HDR_CONNECTION,
    HDR_CONTENT_ENCODING,
    HDR_CONTENT_LENGTH,
    HDR_CONTENT_RANGE,
    HDR_CONTENT_TYPE,
    HDR_COOKIE,
    HDR_DATE,
    HDR_EXPECT,
    HDR_EXPIRES,
    HDR_HOST,
    HDR_IF_MODIFIED_SINCE,
    HDR_IF_RANGE,
    HDR_LAST_MODIFIED,
    HDR_PRAGMA,
    HDR_PROXY_CONNECTION,
    HDR_RANGE,
    HDR_REFERER,
    HDR_SET_COOKIE,
    HDR_TE,
    HDR_TRANSFER_ENCODING,
    HDR_USER_AGENT,
    HDR_VARY,
    HDR_ENUM_END
};


#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/***  Parsing helper functions  ***/

/* Find char c in buffer buf, between indexes offset and end.
 * Return the offset where it occurs, or -1 if not found.
 */
static int findchar(const char *buf, int offset, int end, char c)
{
    if (offset < 0)
        return -1;
    while (offset < end) {
        if (buf[offset] == c) {
            return offset;
        }
        offset++;
    }
    return -1;
}

/* Skip char c in the buf, between offset and end. Return the offset
 * where char c ends.
 */
static int skipchar(const char *buf, int offset, int end, char c)
{
    if (offset < 0)
        return -1;
    while (offset < end) {
        if (buf[offset] != c) {
            return offset;
        }
        offset++;
    }
    return offset;
}

/* Find the null-terminated string in buf, between indexes offset and end.
 * Return the offset where the string starts, or -1 if not found.
 */
static int findstring(const char *buf, int offset, int end, const char *string)
{
    int stringlen;
    stringlen = strlen(string);

    while (stringlen <= end - offset) {
        if (strncasecmp(&buf[offset], string, stringlen) == 0) {
            return offset;
        }
        offset++;
    }
    return -1;
}

/* Return true if the given buf ends with "string".
 * "end" is the length of the buf.
 */
/* Not currently used
static int findend(const char *buf, int end, const char *string)
{
    int stringlen;
    stringlen = strlen(string);
    if (end < stringlen)
        return 0;
    if (strncasecmp(&buf[end-stringlen], string, stringlen) == 0)
        return 1;
    else
        return 0;
}
*/


/* The HeaderMap is used to help map http header
 * strings to the HDR_ enum values above.
 */
#define HEADER_MAP_SIZE 101
struct _HeaderMap {
    struct {
        int id;
        const char *name;
        int namelen;
    } data[HEADER_MAP_SIZE]; 
} HeaderMap;

/* Hash function for http header names */
static int 
hash_header(const char *name, int len)
{
    unsigned int result = 5;
    int i;
    char c[3];
    if (len < 3) { return 0; }

    c[0] = name[0];
    c[1] = name[len-2];
    c[2] = name[len-1];
    
    for (i = 0; i < 3; i++) {
        if (c[i] >= 'A' && c[i] <= 'Z') {
            c[i] = (c[i] - 'A') + 'a';
        }
        result = (result * 13 + c[i]) % 0x7fffffff;
    }
    result = (result * 13 + len * 31) % 0x7fffffff;
    return (int)(result % HEADER_MAP_SIZE);
}

/* Initialize the HeaderMap with the header names given
 * above in the HeaderStrings array.
 */
static void init_header_map()
{
    int id;
    const char *name;
    int hash, len;

    memset(&HeaderMap, 0, sizeof(struct _HeaderMap));
    for (id = 1; id < HDR_ENUM_END; id++) {
        name = HeaderStrings[id];
        len = strlen(name); 
        hash = hash_header(name, len);
        /* There should be no hash collisions */
        assert(HeaderMap.data[hash].name == NULL);
        HeaderMap.data[hash].namelen = len;
        HeaderMap.data[hash].name = name;
        HeaderMap.data[hash].id = id;
    }
}

/* Given an Http Header string, return the corresponding
 * header id. If there is no matching header id, return
 * HDR_UNKNOWN (0).
 */ 
static int http_get_headerid(const char *name, int len)
{
    static int init = 0;
    int hash;
    const char *headername;
    int id, namelen;
    
    if (!init) {
        init = 1;
        init_header_map();
    }
    if (len <= 0)
        return 0;
    
    hash = hash_header(name, len);
    headername = HeaderMap.data[hash].name;
    namelen = HeaderMap.data[hash].namelen; 
    id = HeaderMap.data[hash].id;
    if (headername != NULL && namelen == len && 
        strncasecmp(name, headername, len) == 0) {

        return id;
    }
    return HDR_UNKNOWN;
}

/* Given raw character data containing http headers,
 * parse the data into the given HttpHeader array.
 * Parsedata contains the offsets where line breaks (\n)
 * occur. len is the length of the array.
 * If an error occurs while parsing, be forgiving, and 
 * just return the number of headers we were able to 
 * parse correctly.
 */
static int getheaders(char *buf, int size, Parsedata *parse, 
                                            HttpHeader *array, int len)
{
    int offset, nameend, end, i, count;
    HttpHeader *h;

    count = parse->numlines - 1; /* First line isn't header */
    if (count > len)
        count = len;

    for (i = 0; i < count; i++) { 
        offset = parse->lineoffsets[i] + 1;
        end = parse->lineoffsets[i+1];
        h = &array[i];
        nameend = findchar(buf, offset, end, ':');
        if (nameend == -1)
            return i;

        h->id = http_get_headerid(&buf[offset], nameend - offset);
        h->name.buf = &buf[offset];
        h->name.size = nameend - offset;
        nameend++;
        while (buf[nameend] == ' ')
            nameend++;
        h->value.buf = &buf[nameend];
        h->value.size = end - nameend;
        if (buf[end-1] == '\r')
            h->value.size--;
    }
    return count;
}

/* Return true if the given HDR_ enum id exists
 * in the HttpHeader array. len is the length
 * of the array.
 */
static int http_hasheader(HttpHeader *array, int len, int id)
{
    int i;
    for (i = 0; i < len; i++) {
        if (id == array[i].id) {
            return 1;
        }
    }
    return 0;
}

/*
 * Find the given HDR_ enum id in the HttpHeader array, and
 * return the header value. If the header is not found, the
 * Substring size is 0
 */
static Substring http_getheader(HttpHeader *array, int len, int id)
{
    Substring blank;
    int i;

    blank.buf = NULL; blank.size = 0;
    for (i = 0; i < len; i++) {
        if (id == array[i].id) {
            return array[i].value;
        }
    }
    return blank;
}

/*
 * Find the given HDR_ enum id in the HttpHeader array, and
 * return the integer value. If the header is not found, the
 * result is 0. If the value is not an integer, return 0.
 */
static unsigned long http_getheader_int(HttpHeader *array, int len, int id)
{
    Substring s;
    s = http_getheader(array, len, id);
    if (s.size <= 0)
        return 0;
    return strtoul(s.buf, NULL, 10);
}

/*
 * Find the given HDR_ enum id in the HttpHeader array, and
 * return the time_t value. If the header is not found, the
 * result is 0.
 */
static time_t http_getheader_date(HttpHeader *array, int len, int id)
{
    Substring s;

    s = http_getheader(array, len, id);
    if (s.size <= 0)
        return 0;
    return parse_date(s.buf, s.size); 
}

/* Return the id in the header "Cookie: vicompressid=<id>", or 0
 * if the vicompressid is not found. 
 */
static int http_get_vicompressid(HttpHeader *array, int len)
{
    Substring s;
    char *value;
    int pos, i;

    for (i = 0; i < len; i++) {
        if (HDR_COOKIE == array[i].id) {
            s = array[i].value;
            pos = findstring(s.buf, 0, s.size, "vicompressid=");
            if (pos >= 0) {
                value = (char *) &(s.buf[pos + sizeof("vicompressid=") - 1]);
                return atoi(value);
            }
        }
    }
    return 0;
}

/***   Parsing Request   ***/

/* Return 1 if the buffer has all the requests, and 0 if not.
 * To make parsing faster, store the end-of-line offsets and
 * the last offset reached in the parse data.
 */
static int request_is_complete(HttpRequest *request, Parsedata *parse)
{
    int lastoffset = parse->lastoffset;
    int i, j;
    Buffer *buffer = request->buffer;

    for (i = lastoffset; i < buffer->size; i++) {
        if (buffer->buf[i] == '\n') {
            if (parse->numlines == MAX_PARSE_LINES) {
                parse->numlines = -1;
                return -1;
            }
            j = parse->numlines;
            parse->lineoffsets[j] = i;
            parse->numlines++;
            if (i >= 2 && (buffer->buf[i-1] == '\n' || 
                buffer->buf[i-2] == '\n')) {

                request->requestlength = i+1;
                parse->numlines--;
                return 1;
            }
        }
    }
    parse->lastoffset = buffer->size;
    return 0;
}

/* Parse the first line in the request. Return 1 if successful, or 0 if error. 
 * Store the method, url, path, host, port, and http version.
 * in the request structure. 
 */
static int request_parse_firstline(HttpRequest *request, Parsedata *parse)
{ 
    int endline, offset, prevoffset, offset2;
    Buffer *buffer = request->buffer;
    char *buf = buffer->buf;
    
    endline = parse->lineoffsets[0];
    prevoffset = skipchar(buf, 0, endline, ' ');
    if (prevoffset == endline)
        return 0;

    /* Parse the method (GET, POST, etc) */
    request->method.size = 0;
    request->method.buf = &(buf[prevoffset]);
    offset = findchar(buf, prevoffset, endline, ' ');
    if (offset == -1)
        return 0;
    request->method.size = offset - prevoffset;

    offset = skipchar(buf, offset, endline, ' ');
    if (offset == endline)
        return 0;

    if (strncmp(&buffer->buf[offset], "http://", strlen("http://")) == 0) {
        /* This is a proxy http request with the full URL in the request:
         * GET http://hostname:port/path HTTP/1.0
         */
        offset += 7;

        request->host.buf = &buffer->buf[offset];
        request->host.size = 0;
        prevoffset = offset;

        /* IPv6 addresses in URLs are delimited by [ ] like
         * http://[fe80::xxxx:1]:port/path
         * http://[fe80::xxxx:1]/path
         */
        if (request->host.buf[0] == '[') {
            offset = findchar(buf, prevoffset, endline, ']');
            if (offset == -1)
                return 0;
            offset = findchar(buf, offset, endline, ':');
        }
        else {
            offset = findchar(buf, prevoffset, endline, ':');
        }
        offset2 = findchar(buf, prevoffset, endline, '/');
        if (offset != -1 && offset < offset2) {
            request->port = atoi(&buf[offset+1]);
            if (request->port == 0) {
                return 0;
            }
            request->host.size = offset - prevoffset;
        }
        else {
            request->port = 80;
        }
        offset = findchar(buf, prevoffset, endline, '/');
        if (offset == -1)
            offset = findchar(buf, prevoffset, endline, ' ');
        if (offset == -1)
            return 0;
        if (request->host.size == 0)
            request->host.size = offset - prevoffset;
    }
    else {
        request->port = 80;
        request->host.size = 0;
    }

    if (buffer->buf[offset] == '/') {
        request->path.buf = &buffer->buf[offset];
        prevoffset = offset;
        offset = findchar(buf, offset, endline, ' ');
        if (offset == -1)
            return 0;
        request->path.size = offset - prevoffset;
    }
    else {
        request->path.buf = "/";
        request->path.size = 1;
    }
    
    offset = skipchar(buf, offset, endline, ' ');
    if (offset == -1)
        return 0;
    
    if (strncasecmp(&buf[offset], "HTTP/1.0", strlen("HTTP/1.0")) == 0)
        request->version = 0;
    else if (strncasecmp(&buf[offset], "HTTP/1.1", strlen("HTTP/1.1")) == 0)
        request->version = 1;
    else
        return 0;

    return 1;
}

/* Given an HttpRequest, return 1 if the request is
 * cachable, and 0 if not.
 */
static char 
request_is_cachable(HttpRequest *request, HttpHeader *array, int len)
{
    Substring value;
 
    if (request->method.size != 3 || 
        strncasecmp(request->method.buf, "GET", 3) != 0) {
        return 0;
    }

    if (http_hasheader(array, len, HDR_AUTHORIZATION)) {
        return 0;
    }
    if (http_hasheader(array, len, HDR_IF_MODIFIED_SINCE))
        return 0;
    if (http_hasheader(array, len, HDR_IF_RANGE))
        return 0;

    value = http_getheader(array, len, HDR_PRAGMA);
    if (findstring(value.buf, 0, value.size, "no-cache") != -1)
        return 0;

    value = http_getheader(array, len, HDR_CACHE_CONTROL);
    if (findstring(value.buf, 0, value.size, "no-cache") != -1 ||
        findstring(value.buf, 0, value.size, "no-store") != -1 ||
        findstring(value.buf, 0, value.size, "max-age=0") != -1) {

        return 0;
    }

    return 1;
}

/* Free the buffer used by HttpRequest.
 * Free the buffer used by the url.
 * zero out the struct values.
 */
void http_free_request(HttpRequest *request)
{
    if (request->url != NULL) {
        buffer_unref(request->url);
        request->url = NULL;
    }
    if (request->buffer != NULL) {
        buffer_unref(request->buffer);
        request->buffer = NULL;
    }
    memset(request, 0, sizeof(HttpRequest));
}

/* Return true if the request has a keep-alive connection. */
static int http_can_keepalive(HttpHeader *array, int len, int version)
{
    Substring value;

    value = http_getheader(array, len, HDR_CONNECTION);
    if (value.size > 0 && 
        strncasecmp(value.buf, "Keep-Alive", strlen("Keep-Alive")) == 0) {
        return 1;
    }
    else if (value.size > 0 &&
        strncasecmp(value.buf, "close", strlen("close")) == 0) {
        return 0;
    }
    value = http_getheader(array, len, HDR_PROXY_CONNECTION);
    if (value.size > 0 &&
        strncasecmp(value.buf, "Keep-Alive", strlen("Keep-Alive")) == 0) {
        return 1;
    }
    else if (value.size > 0 &&
        strncasecmp(value.buf, "close", strlen("close")) == 0) {
        return 0;
    }
    /* If no "Connection:" or "Proxy-Connection:" header is given,
     * return 1 for HTTP/1.1 and 0 for HTTP/1.0.
     */
    else if (version == 1) {
        return 1;
    }
    else {
        return 0;
    }
}

/* Return true if this request accepts gzip encoding */
static char request_acceptgzip(HttpHeader *array, int len)
{
    Substring value;
    value = http_getheader(array, len, HDR_ACCEPT_ENCODING);
    if (value.size > 0) {
        if (findstring(value.buf, 0, value.size, "gzip") != -1) {
            return 1;
        }
    }
    return 0;
}


/* Parse the Host and port from the http request "Host" header. */
static void 
store_host_port(HttpHeader *array, int len, Substring *host, int *port)
{
    Substring s;
    int p, offset;

    p = 0;
    s = http_getheader(array, len, HDR_HOST);

    if (s.size == 0)
        return;

    /* Host headers with IPv6 addresses use [] brackets, like
     * Host: [fe80::xxx:1]:80 
     * Host: [fe80::xxx:1]
     */
    if (s.buf[0] == '[') {
        offset = findchar(s.buf, 0, s.size, ']');
        if (offset == -1)
            return;
        offset = findchar(s.buf, offset, s.size, ':');
    }
    else {
        offset = findchar(s.buf, 0, s.size, ':');
    }
    if (offset != -1) {
        s.size = offset;
        p = atoi(&s.buf[offset+1]);
    }
    if (p == 0)
        p = 80;
    if (s.size > 0) {
        host->buf = s.buf;
        host->size = s.size;
    }
    *port = p;
}

/* Check for a "Range: start-end" header.  If found:
 * - Set is_range = true
 * - Store the start and end values in HttpRequest.range_start and range_end
 * - If there is no "end" value, set range_end = 0
 */
static void http_parse_range(HttpRequest *request)
{
    Substring value;
    int hyphen, end;

    value = http_getheader(request->headers, request->headercount, HDR_RANGE);
    if (value.size > 0) {
        request->is_range = 1;
        hyphen = findchar(value.buf, 0, value.size, '-');
        if (hyphen > 0) {
            request->range_start = atoi(value.buf);
            if (request->range_start < 0) {
                request->is_range = 0;
                return;
            }
            end = skipchar(value.buf, hyphen+1, value.size, ' ');
            if (end < value.size) {
                request->range_end = atoi( &(value.buf[end]) );
                if (request->range_end <= request->range_start) {
                    request->is_range = 0;
                    return;
                }
            }
            else { 
                /* Range requests can have the form
                 *  Range: start-
                 */
                request->range_end = 0;
            }
        }
    }
}

/*
 * Before calling this, you must first call http_init_request()
 * to assign a data buffer to the HttpRequest.  Parse the data 
 * in the buffer, and fill in the values in the HttpRequest structure.
 * Store the last offset parsed and the line offsets in Parsedata.
 * 
 * Return PARSE_PARTIAL if more data is needed.
 * Return PARSE_ERROR if the request is invalid.
 * Return PARSE_DONE if the request is successfully parsed.
 */
int http_parse_request(HttpRequest *request, Parsedata *parse)
{
    int result;
    HttpHeader *hdrs;
    int hdrcount;
    Substring value;

    assert(request->buffer != NULL);
    assert(parse->state == PARSE_PARTIAL);
 
    result = request_is_complete(request, parse);
    if (result == -1) {
        parse->state = PARSE_ERROR;
        return PARSE_ERROR;
    }
    else if (result == 0) 
        return PARSE_PARTIAL;

    /* state is ERROR until we reach end of function */
    parse->state = PARSE_ERROR;

    request->firstline = parse->lineoffsets[0] + 1;
    result = request_parse_firstline(request, parse);
    if (result == 0) 
        return PARSE_ERROR;

    hdrs = &(request->headers[0]);
    hdrcount = getheaders(request->buffer->buf, request->buffer->size,
                          parse, hdrs, MAX_REQUEST_HEADERS);
    request->headercount = hdrcount;

    request->contentlength = 
      http_getheader_int(hdrs, hdrcount, HDR_CONTENT_LENGTH);
    request->is_cachable = request_is_cachable(request, hdrs, hdrcount);
    request->keepalive = http_can_keepalive(hdrs, hdrcount, request->version);
    request->acceptgzip = request_acceptgzip(hdrs, hdrcount);
    request->webserverid = http_get_vicompressid(hdrs, hdrcount);
    value = http_getheader(hdrs, hdrcount, HDR_PROXY_CONNECTION);
    request->is_proxy = (value.size > 0);
    value = http_getheader(hdrs, hdrcount, HDR_EXPECT);
    if (value.size > 0 && 
        findstring(value.buf, 0, value.size, "100-continue") >= 0) {
        request->expect100 = 1;
    }

    /* Check for the header "Range: start-end" */
    http_parse_range(request);

    /* If the hostname wasn't in the URL, get it from the Host header */
    if (request->host.size == 0) {
        store_host_port(hdrs, hdrcount, &request->host, &request->port);
    }

    /* For IPv6 addresses like [fe80::1], remove the brackets */
    if (request->host.size >= 2 && request->host.buf[0] == '[') {
        request->host.buf++;
        request->host.size -= 2;
    }

    /* If this is a forward proxy, the http request must have
     * the destination host to connect to.
     */
    if (request->host.size == 0 && Config.num_webservers == 0) {
        if (request->path.size != 12 || 
            strncmp(request->path.buf, "/_viewcache_", 12) != 0) {
            return PARSE_ERROR;
        }
    }

    parse->state = PARSE_DONE;
    return PARSE_DONE;
}

/* Initialize the HttpRequest structure. Assign the given
 * buffer to this HttpRequest.  Initialize the Parsedata,
 * which keeps track of the last offset parsed and the
 * offsets of newlines (\n) in the data.
 */
void http_init_request(HttpRequest *req, Buffer *b, Parsedata *parse)
{
    memset(req, 0, sizeof(HttpRequest));
    req->buffer = b;
    parse->state = PARSE_PARTIAL;
    parse->lastoffset = 0;
    parse->numlines = 0;
    memset(parse->lineoffsets, 0, sizeof(int) * MAX_PARSE_LINES);
}

/* Given the original HttpRequest received from the client, create
 * a request suitable for sending to an origin server, and write
 * the request data to the given out Buffer.
 * 
 * Ignore the original Connection/Proxy-Connection headers.
 * Remove any TE (transfer-encoding) headers.  vicompress does not
 * recognize chunked transfer encoding.  Write our own Host header.
 * If the hostheader option was given in the config file, use that 
 * as the host.  Write our own Cache-Control header so that the
 * origin server knows we want to cache the reply.  Add an
 * X-Forwarded-For header if we are 
 */
void http_write_request(HttpRequest *request, Buffer *out, 
                        sockaddr_all *clientip)
{
    int i, id;
    Substring name, value;
    char ipstring[50];
    unsigned char *ip;

    buffer_add(out, request->method.buf, request->method.size);
    buffer_add(out, " ", 1);
    buffer_add(out, request->path.buf, request->path.size);
    if (request->version == 1) {
        buffer_add(out, " HTTP/1.1\r\n", strlen(" HTTP/1.1\r\n"));
    }
    else {
        buffer_add(out, " HTTP/1.0\r\n", strlen(" HTTP/1.0\r\n"));
    }

    for (i = 0; i < request->headercount; i++) {
        id = request->headers[i].id;
        name = request->headers[i].name;
        value = request->headers[i].value;
 
        if (id == HDR_CONNECTION || id == HDR_PROXY_CONNECTION ||
            id == HDR_CACHE_CONTROL || id == HDR_HOST) {
            /* skip header */
        }
        else {
            buffer_add(out, name.buf, name.size);
            buffer_add(out, ": ", 2);
            buffer_add(out, value.buf, value.size);
            buffer_add(out, "\r\n", 2); 
        }
    }
    if (request->keepalive) {
        buffer_add(out, "Connection: keep-alive\r\n", sizeof("Connection: keep-alive\r\n")-1);
    }
    else {
        buffer_add(out, "Connection: close\r\n", sizeof("Connection: close\r\n")-1);
    }

    if (Config.num_webservers > 0 && Config.hostheader != NULL) {
        buffer_add(out, "Host: ", sizeof("Host: ")-1);
        buffer_add(out, Config.hostheader, strlen(Config.hostheader));
        buffer_add(out, "\r\n", sizeof("\r\n")-1);
    }
    else if (request->host.size > 0) {
        buffer_add(out, "Host: ", sizeof("Host: ")-1);
        buffer_add(out, request->host.buf, request->host.size);
        if (request->port != 80 && Config.num_webservers == 0) {
            buffer_add(out, ":", 1);
            buffer_add_int(out, request->port);
        }
        buffer_add(out, "\r\n", sizeof("\r\n")-1);
    }

    if (request->is_cachable) {
        buffer_add(out, "Cache-Control: max-age=25200\r\n",
                   sizeof("Cache-Control: max-age=25200\r\n") - 1);
    }

    /* Write the X-Forwarded-For header */
    ipstring[0] = ipstring[47] = 0;
    buffer_add(out, "X-Forwarded-For: ", sizeof("X-Forwarded-For: ") - 1);
#ifdef HAVE_IPV6
    if (clientip->ipv4.sin_family == AF_INET6) {
        inet_ntop(AF_INET6, &(clientip->ipv6.sin6_addr), ipstring, 47);
        strcat(ipstring, "\r\n");
    }
#endif
    if (clientip->ipv4.sin_family == AF_INET) {
        ip = (unsigned char*) &(clientip->ipv4.sin_addr.s_addr);
        sprintf(ipstring, "%u.%u.%u.%u\r\n", ip[0], ip[1], ip[2], ip[3]);
    }
    buffer_add(out, ipstring, strlen(ipstring));
    buffer_add(out, "\r\n", sizeof("\r\n")-1);
}


/*** Parse Reply ***/

/* Return 1 if we have read a complete reply header, or 0 otherwise.
 * To make parsing faster, store the index of the end of the first
 * line, the number of headers in the reply, and the last offset
 * reached in the parse.
 */
static int reply_is_complete(HttpReply *reply, Parsedata *parse)
{
    int lastoffset = parse->lastoffset;
    int i, j;
    Buffer *buffer = reply->buffer;

    for (i = lastoffset; i < buffer->size; i++) {
        if (buffer->buf[i] == '\n') {
            if (parse->numlines == MAX_PARSE_LINES) {
                parse->numlines = -1;
                return -1;
            }
            j = parse->numlines;
            parse->lineoffsets[j] = i;
            parse->numlines++;
            if (i >= 2 && (buffer->buf[i-1] == '\n' || 
                buffer->buf[i-2] == '\n')) {

                reply->replylength = i+1;
                parse->numlines--;
                return 1;
            }
        }
    }
    parse->lastoffset = buffer->size;
    return 0;
}

/* Parse the first line in the reply. Return 1 if successful or 0 if error. 
 * Store the http version and status code in the HttpReply structure.
 */
static int reply_parse_firstline(HttpReply *reply, Parsedata *parse)
{
    int endline, offset;
    Buffer *buffer = reply->buffer;
    char *buf = buffer->buf;
    
    endline = parse->lineoffsets[0];
    offset = skipchar(buf, 0, endline, ' ');
    if (offset == endline)
        return 0;
    
    if (strncmp(&buf[offset], "HTTP/1.0", strlen("HTTP/1.0")) == 0)
        reply->version = 0;
    else if (strncmp(&buf[offset], "HTTP/1.1", strlen("HTTP/1.1")) == 0)
        reply->version = 1;
    else
        return 0;
 
    offset = findchar(buf, offset, endline, ' ');
    if (offset == -1)
        return 0;
    offset = skipchar(buf, offset, endline, ' ');
    if (offset == endline)
        return 0;
    reply->status = atoi(&buf[offset]);
    if (reply->status <= 0 || reply->status > 700)
        return 0;

    return 1;
}

/* Return true if the reply is cachable */
static char reply_is_cachable(HttpReply *reply, HttpHeader *array, int len)
{
    Substring cachecontrol;
    const char *buf;
    int end, offset;
    int maxage;

    if (reply->status != 200)
        return 0;
    if (!reply->has_contentlength || reply->is_chunked)
        return 0;
    if (http_hasheader(array, len, HDR_CONTENT_RANGE))
        return 0;
    if (!http_hasheader(array, len, HDR_DATE))
        return 0;
    if (!http_hasheader(array, len, HDR_LAST_MODIFIED) &&
        !http_hasheader(array, len, HDR_EXPIRES))
        return 0;

    cachecontrol = http_getheader(array, len, HDR_CACHE_CONTROL);
    buf = cachecontrol.buf;
    end = cachecontrol.size;
    if (cachecontrol.size > 0) {
        if (findstring(buf, 0, end, "private") != -1 ||
            findstring(buf, 0, end, "no-cache") != -1 ||
            findstring(buf, 0, end, "no-store") != -1) {

            return 0;
        }
        if ((offset = findstring(buf, 0, end, "max-age=")) != -1) {
            offset += strlen("max-age=");
            maxage = atoi(&buf[offset]);
            if (maxage <= 0)
                return 0;
        }
        else if ((offset = findstring(buf, 0, end, "s-maxage=")) != -1) {
            offset += strlen("s-maxage=");
            maxage = atoi(&buf[offset]);
            if (maxage <= 0)
                return 0;
        }
    }
    return 1;
}

/* Return 1 if the http reply has "Transfer-Encoding: chunked".
 * Else, return 0.
 */
static int reply_is_chunked(HttpReply *reply)
{
    Substring s;

    s = http_getheader(reply->headers, reply->headercount, HDR_TRANSFER_ENCODING);
    if (s.size == 0) {
        return 0;
    }
    if (findstring(s.buf, 0, s.size, "chunked") != -1) {
        return 1;
    }
    else {
        return 0;
    }
}


/* Free the buffer used by HttpRequest.
 * zero out the struct values.
 */
void http_free_reply(HttpReply *reply)
{
    if (reply->buffer != NULL) {
        buffer_unref(reply->buffer);
        reply->buffer = NULL;
    }
    memset(reply, 0, sizeof(HttpReply));
}

/* Return true if the http reply is already compressed */
static char reply_alreadygzipped(HttpHeader *array, int len) {
    Substring value;
    
    value = http_getheader(array, len, HDR_CONTENT_ENCODING);
    if (value.size > 0) {
        if (findstring(value.buf, 0, value.size, "identity") == -1)
            return 1;
    }
    return 0;
}

/* Return true if this http reply can be gzipped.
 * The reply must be > MIN_GZIP_SIZE bytes, and must be html,
 * plain text, PDF, or MS Word.
 */
static char reply_cangzip(HttpReply *reply, HttpHeader *array, int len) {
    Substring value;

    if (reply->has_contentlength && reply->contentlength <= MIN_GZIP_SIZE)
        return 0;

    if (reply->alreadygzipped)
        return 0;

    value = http_getheader(array, len, HDR_CONTENT_TYPE);
    if (value.size > 0) {
        if (findstring(value.buf, 0, value.size, "text/") != -1)
            return 1;
        else if (findstring(value.buf, 0, value.size, "application/x-javascript") != -1)
            return 1;
        else if (findstring(value.buf, 0, value.size, "application/javascript") != -1)
            return 1;
        else if (findstring(value.buf, 0, value.size, "application/pdf") != -1)
            return 1;
        else if (findstring(value.buf, 0, value.size, "application/msword") != -1)
            return 1;
        else
            return 0; 
    }
    return 0;
}
 
/*
 * Before calling this, you must first call http_init_reply()
 * to assign a data buffer to the HttpReply.  Parse the data
 * in the buffer, and fill in the values in the HttpReply structure.
 * Store the last offset parsed and the line offsets in Parsedata.
 *
 * Return PARSE_PARTIAL if more data is needed.
 * Return PARSE_ERROR if the request is invalid.
 * Return PARSE_DONE if the request is successfully parsed.
 */
int http_parse_reply(HttpReply *reply, Parsedata *parse)
{
    int result;
    HttpHeader *hdrs;
    int hdrcount;
    time_t replydate, expiredate, expires;

    assert(reply->buffer != NULL); 
    assert(parse->state == PARSE_PARTIAL);

    result = reply_is_complete(reply, parse);
    if (result == -1) {
        parse->state = PARSE_ERROR; 
        return PARSE_ERROR;
    }
    else if (result == 0)
        return PARSE_PARTIAL;

    /* Assume state is ERROR until we reach end of function */
    parse->state = PARSE_ERROR;

    reply->firstline = parse->lineoffsets[0] + 1;
    result = reply_parse_firstline(reply, parse);
    if (result == 0)
        return PARSE_ERROR;
    hdrs = &(reply->headers[0]);
    hdrcount = getheaders(reply->buffer->buf, reply->buffer->size, parse,
                          hdrs, MAX_REPLY_HEADERS);
    reply->headercount = hdrcount;

    if (http_hasheader(hdrs, hdrcount, HDR_CONTENT_LENGTH)) {
        reply->has_contentlength = 1;
        reply->contentlength = 
          http_getheader_int(hdrs, hdrcount, HDR_CONTENT_LENGTH);
        if (reply->contentlength == ULONG_MAX) {
            reply->has_contentlength = 0;
            reply->contentlength = 0;
        }
    }
    else {
        reply->has_contentlength = 0;
        reply->contentlength = 0;
    }

    reply->lastmodified = 
      http_getheader_date(hdrs, hdrcount, HDR_LAST_MODIFIED);
    reply->alreadygzipped = reply_alreadygzipped(hdrs, hdrcount);
    reply->is_chunked = reply_is_chunked(reply);
    reply->keepalive = http_can_keepalive(hdrs, hdrcount, reply->version);
    reply->is_cachable = reply_is_cachable(reply, hdrs, hdrcount);
    reply->cangzip = reply_cangzip(reply, hdrs, hdrcount);

    if (reply->headercount == MAX_REPLY_HEADERS) {
        return PARSE_ERROR;
    }

    replydate = http_getheader_date(hdrs, hdrcount, HDR_DATE);
    expiredate = http_getheader_date(hdrs, hdrcount, HDR_EXPIRES);
    if (replydate == 0)
        replydate = get_current_time();
    if (expiredate > 0)
        expires = expiredate - replydate;
    else if (reply->lastmodified > 0)
        expires = (replydate - reply->lastmodified)/2;
    else
        expires = 0;

    if (expires > Config.cache_expires)
        expires = Config.cache_expires;
    if (expires < HOUR)
        reply->is_cachable = 0;

    if (reply->is_cachable)
        reply->expires = get_current_time() + expires;

    parse->state = PARSE_DONE;
    return PARSE_DONE;
}

/* Initialize the HttpReply structure. Assign the given
 * buffer to this HttpReply.  Initialize the Parsedata,
 * which keeps track of the last offset parsed and the
 * offsets of newlines (\n) in the data.
 */
void http_init_reply(HttpReply *reply, Buffer *b, Parsedata *parse)
{
    memset(reply, 0, sizeof(HttpReply));
    assert(reply->buffer == NULL);
    reply->buffer = b;
    parse->state = PARSE_PARTIAL;
    parse->lastoffset = 0;
    parse->numlines = 0;
    memset(parse->lineoffsets, 0, sizeof(int) * MAX_PARSE_LINES);
}

/* Given the original HttpReply received from the origin server,
 * create a reply suitable for sending back to the client, and
 * write the data to the given out Buffer.
 *
 * - iscached    : whether or not this reply was retrieved from the cache
 * - chunk_encode: whether or not vicompress is chunk-encoding the reply
 * - webserverid : the backend webserver the reply came from. 
 *                 If sessions are enabled, then a Set-Cookie header will be 
 *                 added with the webserverid.
 * 
 * Skip any Connection headers.  If the reply is from the cache,
 * remove any Cookie headers.
 *
 * If the reply is being gzipped, remove the content-length,
 * content-encoding, and vary headers.  Then, write our own
 * content-encoding and vary headers.
 * 
 * If the request has a "Range: start-end" header, and this is a cached reply,
 * then add a "Content-Range: start-end" header.
 */

void http_write_reply(HttpReply *reply, HttpRequest *req, Buffer *out, 
                      boolean iscached, boolean chunk_encode_reply, int webserverid)
{
    int i, id;
    Substring name;
    Substring value;
    int keepalive;

    buffer_add(out, reply->buffer->buf, reply->firstline);

    /* Change HTTP/1.0 to HTTP/1.1 */
    out->buf[7] = '1';

    for (i = 0; i < reply->headercount; i++) {
        id = reply->headers[i].id;
        name = reply->headers[i].name;
        value = reply->headers[i].value;

        if (id == HDR_CONNECTION || id == HDR_PROXY_CONNECTION) {
            /* skip headers */
        }
        else if (iscached && id == HDR_SET_COOKIE) {
            /* skip header */
        }
        else if (reply->newlygzipped && 
                 (id == HDR_CONTENT_ENCODING ||
                  id == HDR_TRANSFER_ENCODING ||
                  id == HDR_CONTENT_LENGTH ||
                  id == HDR_VARY)) {
            /* skip headers */
        }
        else if (chunk_encode_reply && (id == HDR_TRANSFER_ENCODING)) {
            /* skip headers */
        }
        else {
            buffer_add(out, name.buf, name.size);
            buffer_add(out, ": ", 2);
            buffer_add(out, value.buf, value.size);
            buffer_add(out, "\r\n", 2); 
        } 
    }

    /* Add the Connection header */
    keepalive = (req->keepalive && 
                 (reply->has_contentlength || reply->is_chunked) &&
                 (!reply->newlygzipped));

    if (req->is_proxy)
        buffer_add(out, "Proxy-Connection: ", sizeof("Proxy-Connection: ")-1);
    else 
        buffer_add(out, "Connection: ", sizeof("Connection: ")-1);

    if (keepalive)
        buffer_add(out, "Keep-Alive\r\n", sizeof("Keep-Alive\r\n")-1);
    else
        buffer_add(out, "Close\r\n", sizeof("Close\r\n")-1);

    /* If we are gzipping server reply, the final content length is not
     * known at this point, so don't include a content-length header
     */
    if (!reply->has_contentlength || (reply->newlygzipped && !iscached)) {
    }
    else {
        buffer_add(out, "Content-Length: ", sizeof("Content-Length: ")-1);
        buffer_add_int(out, reply->contentlength);
        buffer_add(out, "\r\n", 2);
    }
    if (reply->newlygzipped) {
        buffer_add(out, "Content-Encoding: gzip\r\n", 
                   sizeof("Content-Encoding: gzip\r\n")-1);
        buffer_add(out, "Vary: Content-Encoding\r\n", 
                   sizeof("Vary: Content-Encoding\r\n")-1);
    }

    /* If we are chunk-encoding the reply, add the chunked header */
    if (chunk_encode_reply) {
        buffer_add(out, "Transfer-Encoding: chunked\r\n",
                   sizeof("Transfer-Encoding: chunked\r\n")-1);
    }

    /* If this is a "Range: start-end" request, then
     * - Change reply status from 200 to 206 (partial content)
     * - Add "Content-Range: start-end" header
     */
    if (req->is_range && iscached) {
        int offset = findstring(out->buf, 0, reply->firstline, "200");
        if (offset > 0) {
            out->buf[offset+2] = '6';
        }
        buffer_add(out, "Content-Range: ", sizeof("Content-Range: ") - 1);
        buffer_add_int(out, req->range_start);
        buffer_add(out, "-", 1);
        buffer_add_int(out, req->range_end);
        buffer_add(out, "\r\n", 2);
    }

    /* Via: 1.0 <hostname> (vicompress/<version>) */
    if (Config.num_webservers == 0) {   
        buffer_add(out, "Via: 1.1 ", sizeof("Via: 1.1 ")-1);
        buffer_add(out, Config.myhostname, strlen(Config.myhostname));
        buffer_add(out, " (vicompress/", sizeof(" (vicompress/")-1);
        buffer_add(out, VICOMPRESS_VERSION, strlen(VICOMPRESS_VERSION));
        buffer_add(out, ")\r\n", 3);
    }
    if (iscached) {
        buffer_add(out, "X-Cache: HIT\r\n", sizeof("X-Cache: HIT\r\n")-1);
    }

    if (Config.enable_sessions == 1 && webserverid != 0) {
        buffer_add(out, "Set-Cookie: vicompressid=", 
                   sizeof("Set-Cookie: vicompressid=") - 1);
        buffer_add_int(out, webserverid);
        buffer_add(out, "\r\n", 2);
    }

    buffer_add(out, "\r\n", 2);
}

/* Given the HttpRequest parsed from the client, find the
 * given header name, and return the corresponding value.
 * Return a substring of len 0 if the header is not found.
 */
Substring http_get_request_header(HttpRequest *request, const char *name)
{
    Substring s;
    int id;

    s.buf = NULL; s.size = 0;
    id = http_get_headerid(name, strlen(name));
    if (id == HDR_UNKNOWN) {
        return s;
    }
    return http_getheader(request->headers, request->headercount, id);
}

/* Given the HttpReply parsed from the server, find the
 * given header name, and return the corresponding value.
 * Return a substring of len 0 if the header is not found.
 */
Substring http_get_reply_header(HttpReply *reply, const char *name)
{
    Substring s;
    int id;

    s.buf = NULL; s.size = 0;
    id = http_get_headerid(name, strlen(name));
    if (id == HDR_UNKNOWN) {
        return s;
    }
    return http_getheader(reply->headers, reply->headercount, id);
}


/* Return the base64 value of the given character */
static unsigned char base64(unsigned char c)
{
    unsigned char x = 0;
    if (c >= 'A' && c <= 'Z')
        x = c - 'A';
    else if (c >= 'a' && c <= 'z')
        x = c - 'a' + 26;
    else if (c >= '0' && c <= '9')
        x = c - '0' + 52;
    else if (c == '+')
        x = 62;
    else if (c == '/')
        x = 63;
    else if (c == '=')
        x = 0;
    return x;
}

/* Decode the base64 data in "base" into the 
 * regular data in "dest". Return the size
 * of the decoded data.
 */
static int 
base64decode(const unsigned char *base, int baselen, 
             unsigned char *dest, int destlen)
{
    int i, basepos, destpos;
    unsigned char b0, b1, b2, b3;
    unsigned char d0, d1, d2;

    i = 0; basepos = destpos = 0;

    while (basepos + 4 <= baselen && destpos + 3 <= destlen) {
        b0 = base64(base[basepos]);
        b1 = base64(base[basepos+1]);
        b2 = base64(base[basepos+2]);
        b3 = base64(base[basepos+3]);

        d0 = (unsigned char) ((b0 << 2) | (b1 >> 4));
        d1 = (unsigned char) ((b1 << 4) | (b2 >> 2));
        d2 = (unsigned char) ((b2 << 6) | (b3));

        dest[destpos] = d0;
        dest[destpos+1] = d1;
        dest[destpos+2] = d2;

        basepos += 4;
        destpos += 3;
    }
    dest[destpos] = 0;
    return destpos;
}

/* Retrieve the username given in the "Authorization" header,
 * and store it in the name buffer. Store "" if no Authorization
 * header exists. The authorization header format is
 * Authorization: Basic <username:password>
 * where the <username:password> is encoded in base64.
 */
void http_get_username(HttpRequest *req, unsigned char *name, int namelen)
{
    Substring s;
    int buflen, n;
    char buf[128];

    name[0] = 0;
    buf[0] = 0;

    s = http_getheader(req->headers, req->headercount, HDR_AUTHORIZATION); 
    if (s.size < 6)
        return;

    if (strncasecmp(s.buf, "Basic", 5) != 0)
        return;

    n = skipchar(s.buf, 6, s.size, ' '); 
    if (n == s.size)
        return;

    buflen = base64decode((unsigned char *)&(s.buf[n]), s.size-n, 
                          (unsigned char *)buf, sizeof(buf)); 
    if (buflen == -1)
        return;
    n = findchar(buf, 0, buflen, ':');
    if (n == -1)
        return;
    if (n > namelen-1)
        n = namelen-1;
    memcpy(name, buf, n);
    name[n] = 0;
    return;
}


