/*
 * Copyright (c) 2003-2010 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#ifndef _HTTP_H_
#define _HTTP_H_

#include <time.h>
#include <netinet/in.h>
#include "buffer.h"
#include "conf.h"

/* Return codes for http_parse_request() and http_parse_reply() */
#define PARSE_DONE      1
#define PARSE_ERROR     2
#define PARSE_PARTIAL   3


/* Http Header */
struct _HttpHeader {
    int id;           /* One of the HDR_ ids listed in http.c */
    Substring name;   /* The name of the header */
    Substring value;  /* The value of the header */
};
typedef struct _HttpHeader HttpHeader;


#define MAX_PARSE_LINES 100 
struct _Parsedata {
    char state;              /* Parsing states (DONE, ERROR, PARTIAL) */
    int lastoffset;          /* Offset where parsing last ended */
    char numlines;           /* Number of lines in request/reply */
    int lineoffsets[MAX_PARSE_LINES];  /* Offsets where lines end */
};
typedef struct _Parsedata Parsedata;


/* Because vicompress caches thousands of HTTP replies in memory,
 * we want the HttpReply structure to be fairly small.  Therefore, we
 * only allow 24 headers max for replies.  If the reply has more than
 * 24 headers, we simply write a direct reply (conn_writeDirectReply)
 * and do not attempt to parse the Http reply header.
 *
 * The contentlength is an unsigned long, which 32 bits for 32-bit
 * executables, and 64 bits for 64-bit executables.  So a 32-bit
 * vicompress can support downloads/uploads only up to 4 GB.
 */
#define MAX_REQUEST_HEADERS 100
#define MAX_REPLY_HEADERS    32


/* THe Http Request */
struct _HttpRequest         /* The HTTP request */
{
    Buffer *buffer;           /* Buffer holding entire request */
    int firstline;            /* Length of first line */
    int requestlength;        /* Length of request */
    unsigned long contentlength; /* Content-length of POST data */
    int port;                 /* The destination port of the request */
    Substring method;         /* The method of the request (GET, POST, etc) */
    Substring host;           /* The destination hostname (not including port) */ 
    Substring path;           /* The urlpath of the url */
    Buffer *url;              /* A copy of the url of this request */ 
    int webserverid;          /* The value of "Cookie: vicompressid=<id> */
    char headercount;         /* Number of HTTP headers */
    HttpHeader headers[MAX_REQUEST_HEADERS];  /* List of Http Headers */
    int range_start;          /* Values of header "Range: start-end" */
    int range_end;
    char version;             /* Http version. Either 1 or 0 */
    boolean is_cachable;      /* True if request is cachable, false otherwise */
    boolean acceptgzip;       /* True if the request accepts gzip-encoding */
    boolean is_range;         /* True if the request has a "Range: " header */
    boolean is_proxy;         /* True if the request has a "Proxy-Connection" header */
    boolean keepalive;        /* True if the header "Connection: keep-alive" is set */
    boolean expect100;        /* True if the request expects a "100 Continue" reply */
    int flags;                /* One of the HTTP_FLAG_xx values above */
};
typedef struct _HttpRequest HttpRequest;

/* The Http reply */
struct _HttpReply
{
    Buffer *buffer;          /* Buffer holding entire reply header (not body) */
    int firstline;           /* Length of the first line of Http reply */
    int replylength;         /* Length of reply header */
    int status;              /* Http status code */
    unsigned long contentlength; /* Content-Length of body, 0 if none given */
    boolean has_contentlength;  /* True if a Content-Length header was given. */
    char version;            /* Http version. Either 1 or 0 */
    time_t expires;          /* Expiration date of reply */
    time_t lastmodified;     /* Date when item was last modified */
    char headercount;        /* Number of HTTP headers */
    HttpHeader headers[MAX_REPLY_HEADERS];  /* List of Http Headers */
    boolean is_cachable;     /* True if the reply is cachable */
    boolean cangzip;         /* True if the reply can be gzipped */
    boolean alreadygzipped;  /* True if the reply is already gzipped */
    boolean newlygzipped;    /* True if vicompress is gzipping the reply. */
    boolean is_chunked;      /* True if the reply is chunked-encoded */
    boolean keepalive;        /* True if the header "Connection: keep-alive" is set */
};
typedef struct _HttpReply HttpReply;


void http_init_request(HttpRequest *req, Buffer *buf, Parsedata *parse);
void http_init_reply(HttpReply *reply, Buffer *buf, Parsedata *parse);
int  http_parse_request(HttpRequest *req, Parsedata *parse);
int  http_parse_reply(HttpReply *reply, Parsedata *parse);
void http_free_request(HttpRequest *request);
void http_free_reply(HttpReply *reply);
void http_write_request(HttpRequest *request, Buffer *out, 
                        sockaddr_all *clientip);
void http_write_reply(HttpReply *reply, HttpRequest *req, Buffer *out, 
                      boolean iscached, boolean chunk_encode, int webserverid);
void http_get_username(HttpRequest *request, unsigned char *name, int namelen);
Substring http_get_request_header(HttpRequest *request, const char *name);
Substring http_get_reply_header(HttpReply *reply, const char *name);

#endif

