/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */


/**
  The httptest does the following:

  For each http request and reply, parse it, and make sure the
  fields in the HttpRequest/Reply structures are correct.

**/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include "buffer.h"
#include "http.h"
#include "conf.h"
#include "currenttime.h"

BufferPool *bp;

int parse_req(HttpRequest *req, const char *s)
{
    Buffer *b;
    Parsedata parse;
    int len, ret;

    memset(req, 0, sizeof(HttpRequest));
    memset(&parse, 0, sizeof(Parsedata)); 
    b = buffer_alloc(bp);
    len = strlen(s);
    http_init_request(req, b, &parse);
    buffer_add(b, s, len/2);
    ret = http_parse_request(req, &parse);
    assert(ret == PARSE_PARTIAL);
    buffer_add(b, &s[len/2], len - len/2);
    return http_parse_request(req, &parse); 
}

int parse_reply(HttpReply *reply, const char *s)
{
    Buffer *b;
    Parsedata parse;
    int len, ret;

    memset(reply, 0, sizeof(HttpReply));
    memset(&parse, 0, sizeof(Parsedata)); 
    b = buffer_alloc(bp);
    len = strlen(s);
    http_init_reply(reply, b, &parse);
    buffer_add(b, s, len/2);
    ret = http_parse_reply(reply, &parse);
    assert(ret == PARSE_PARTIAL);
    buffer_add(b, &s[len/2], len - len/2);
    return http_parse_reply(reply, &parse); 
}


void test_req_get()
{
    HttpRequest req;
    Substring str;
    int ret;
    const char *s;
    const char *text = \
"GET /index.html HTTP/1.1\r\n\
Host: localhost\r\n\
Connection: keep-alive\r\n\
\r\n";

    ret = parse_req(&req, text); 
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text));
    assert(req.contentlength == 0);
    assert(strncmp(req.method.buf, "GET", 3) == 0);
    assert(req.version == 1);
    assert(req.is_cachable == 1);
    assert(req.keepalive == 1);
    assert(req.acceptgzip == 0);
    assert(req.port == 80);
    assert(req.webserverid == 0);

    s = "localhost";
    assert(strncmp(req.host.buf, s, strlen(s)) == 0);
    s= "/index.html";
    assert(strncmp(req.path.buf, s, strlen(s)) == 0);

    str = http_get_request_header(&req, "connection");
    assert(strncmp(str.buf, "keep-alive", str.size) == 0);
    str = http_get_request_header(&req, "user-agent");
    assert(str.size == 0);

    http_free_request(&req);
    assert(req.buffer == NULL);
}


void test_req_post()
{
    HttpRequest req;
    const char *s;
    int ret;
    const char *text = \
"POST /cgi-bin/test.php HTTP/1.0\r\n\
Host: myhost:3128\r\n\
Proxy-Connection: keep-alive\r\n\
Content-Length: 37\r\n\r\n\
name=john&age=20&item=car&color=blue\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text) - 37);
    assert(req.contentlength == 37);
    assert(strncmp(req.method.buf, "POST", 4) == 0);
    assert(req.version == 0);
    assert(req.is_cachable == 0);
    assert(req.keepalive == 1);
    assert(req.acceptgzip == 0);
    assert(req.port == 3128);
    assert(req.webserverid == 0);

    s = "myhost";
    assert(strncmp(req.host.buf, s, strlen(s)) == 0);
    s= "/cgi-bin/test.php";
    assert(strncmp(req.path.buf, s, strlen(s)) == 0);

    http_free_request(&req);
    assert(req.buffer == NULL);
}


void test_req_no_cache1()
{
    HttpRequest req;
    const char *s;
    int ret;
    const char *text = \
"GET /index.html HTTP/1.1\r\n\
Host: localhost\r\n\
Connection: close\r\n\
Pragma: no-cache\r\n\
\r\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text));
    assert(req.contentlength == 0);
    assert(strncmp(req.method.buf, "GET", 3) == 0);
    assert(req.version == 1);
    assert(req.is_cachable == 0);
    assert(req.keepalive == 0);
    assert(req.acceptgzip == 0);
    assert(req.port == 80);
    assert(req.webserverid == 0);

    s = "localhost";
    assert(strncmp(req.host.buf, s, strlen(s)) == 0);
    s= "/index.html";
    assert(strncmp(req.path.buf, s, strlen(s)) == 0);

    http_free_request(&req);
    assert(req.buffer == NULL);
}


void test_req_no_cache2()
{
    HttpRequest req;
    const char *s;
    int ret;
    const char *text = \
"GET /index.html HTTP/1.1\r\n\
Host: localhost\r\n\
Connection: close\r\n\
Cache-Control: max-age=0\r\n\
\r\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text));
    assert(req.contentlength == 0);
    assert(strncmp(req.method.buf, "GET", 3) == 0);
    assert(req.version == 1);
    assert(req.is_cachable == 0);
    assert(req.keepalive == 0);
    assert(req.acceptgzip == 0);
    assert(req.port == 80);
    assert(req.webserverid == 0);

    s = "localhost";
    assert(strncmp(req.host.buf, s, strlen(s)) == 0);
    s= "/index.html";
    assert(strncmp(req.path.buf, s, strlen(s)) == 0);

    http_free_request(&req);
    assert(req.buffer == NULL);
}


void test_req_cangzip()
{
    HttpRequest req;
    const char *s;
    int ret;
    const char *text = \
"GET /index.html HTTP/1.1\r\n\
Host: localhost\r\n\
Connection: close\r\n\
Accept-Encoding: deflate, gzip\r\n\
\r\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text));
    assert(req.contentlength == 0);
    assert(strncmp(req.method.buf, "GET", 3) == 0);
    assert(req.version == 1);
    assert(req.is_cachable == 1);
    assert(req.keepalive == 0);
    assert(req.acceptgzip == 1);
    assert(req.port == 80);
    assert(req.webserverid == 0);

    s = "localhost";
    assert(strncmp(req.host.buf, s, strlen(s)) == 0);
    s= "/index.html";
    assert(strncmp(req.path.buf, s, strlen(s)) == 0);

    http_free_request(&req);
    assert(req.buffer == NULL);
}


void test_req_auth()
{
    HttpRequest req;
    int ret;
    const char *s;
    const char *text = \
"GET /index.html HTTP/1.1\r\n\
Host: localhost\r\n\
Connection: close\r\n\
Authorization: Basic bXluYW1lOm15cGFzcw==\r\n\
\r\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text));
    assert(req.contentlength == 0);
    assert(strncmp(req.method.buf, "GET", 3) == 0);
    assert(req.version == 1);
    assert(req.is_cachable == 0);
    assert(req.keepalive == 0);
    assert(req.acceptgzip == 0);
    assert(req.port == 80);
    assert(req.webserverid == 0);

    s = "localhost";
    assert(strncmp(req.host.buf, s, strlen(s)) == 0);
    s= "/index.html";
    assert(strncmp(req.path.buf, s, strlen(s)) == 0);

    http_free_request(&req);
    assert(req.buffer == NULL);
}


void test_req_vicompressid()
{
    HttpRequest req;
    int ret;
    const char *text = \
"GET /index.html HTTP/1.1\r\n\
Host: localhost\r\n\
Cookie: vicompressid=5\r\n\
Connection: keep-alive\r\n\
\r\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text));
    assert(req.contentlength == 0);
    assert(strncmp(req.method.buf, "GET", 3) == 0);
    assert(req.version == 1);
    assert(req.is_cachable == 1);
    assert(req.keepalive == 1);
    assert(req.acceptgzip == 0);
    assert(req.port == 80);
    assert(req.webserverid == 5);

    http_free_request(&req);
    assert(req.buffer == NULL);
}

void test_req_ipv6_1()
{
    HttpRequest req;
    int ret;
    const char *s;
    const char *text = \
"GET http://[fe80::213:21ff:fe7c:7471]:81/index.html HTTP/1.1\r\n\
Connection: keep-alive\r\n\
\r\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text));
    assert(req.contentlength == 0);
    assert(strncmp(req.method.buf, "GET", 3) == 0);
    assert(req.version == 1);
    assert(req.is_cachable == 1);
    assert(req.keepalive == 1);
    assert(req.acceptgzip == 0);
    assert(req.port == 81);
    assert(req.webserverid == 0);

    s = "fe80::213:21ff:fe7c:7471";
    assert(strncmp(req.host.buf, s, strlen(s)) == 0);
    s= "/index.html";
    assert(strncmp(req.path.buf, s, strlen(s)) == 0);

    http_free_request(&req);
    assert(req.buffer == NULL);
}

void test_req_ipv6_2()
{
    HttpRequest req;
    int ret;
    const char *s;
    const char *text = \
"GET http://[fe80::213:21ff:fe7c:7471]/index.html HTTP/1.1\r\n\
Connection: keep-alive\r\n\
\r\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_DONE);
    assert(req.requestlength == strlen(text));
    assert(req.contentlength == 0);
    assert(strncmp(req.method.buf, "GET", 3) == 0);
    assert(req.version == 1);
    assert(req.is_cachable == 1);
    assert(req.keepalive == 1);
    assert(req.acceptgzip == 0);
    assert(req.port == 80);
    assert(req.webserverid == 0);

    s = "fe80::213:21ff:fe7c:7471";
    assert(strncmp(req.host.buf, s, strlen(s)) == 0);
    s= "/index.html";
    assert(strncmp(req.path.buf, s, strlen(s)) == 0);

    http_free_request(&req);
    assert(req.buffer == NULL);
}


void test_reply()
{
    HttpReply reply;
    int ret;
    Substring str;
    const char *text = \
"HTTP/1.0 200 OK\r\n\
Date: Wed, 30 Oct 2002 12:43:03 GMT\r\n\
Connetion: close\r\n\
\r\n\
<html>";

    ret = parse_reply(&reply, text);
    assert(ret == PARSE_DONE);
    assert(reply.replylength == strlen(text) - strlen("<html>") );
    assert(reply.has_contentlength == 0);
    assert(reply.contentlength == 0);
    assert(reply.version == 0);
    assert(reply.status == 200);
    assert(reply.is_cachable == 0);
    assert(reply.cangzip == 0);
    assert(reply.alreadygzipped == 0);

    str = http_get_reply_header(&reply, "content-type");
    assert(str.size == 0);
    str = http_get_reply_header(&reply, "connection");
    assert(strncmp(str.buf, "close", str.size) == 0);

    http_free_reply(&reply);
    assert(reply.buffer == NULL);
}


void test_reply_cachable()
{
    HttpReply reply;
    int ret;
    const char *text = \
"HTTP/1.1 200 OK\r\n\
Date: Wed, 30 Oct 2002 12:43:03 GMT\r\n\
Last-Modified: Wed, 02 Oct 2002 09:20:45 GMT\r\n\
Content-Length: 8192\r\n\
Connetion: close\r\n\
\r\n";

    ret = parse_reply(&reply, text);
    assert(ret == PARSE_DONE);
    assert(reply.replylength == strlen(text));
    assert(reply.has_contentlength == 1);
    assert(reply.contentlength == 8192);
    assert(reply.version == 1);
    assert(reply.status == 200);
    assert(reply.is_cachable == 1);
    assert(reply.cangzip == 0);
    assert(reply.alreadygzipped == 0);

    http_free_reply(&reply);
    assert(reply.buffer == NULL);
}


void test_reply_gzip()
{
    HttpReply reply;
    int ret;
    const char *text = \
"HTTP/1.1 200 OK\r\n\
Date: Wed, 30 Oct 2002 12:43:03 GMT\r\n\
Content-type: text/html\r\n\
Connetion: close\r\n\
\r\n";

    ret = parse_reply(&reply, text);
    assert(ret == PARSE_DONE);
    assert(reply.replylength == strlen(text));
    assert(reply.has_contentlength == 0);
    assert(reply.contentlength == 0);
    assert(reply.version == 1);
    assert(reply.status == 200);
    assert(reply.is_cachable == 0);
    assert(reply.cangzip == 1);
    assert(reply.alreadygzipped == 0);

    http_free_reply(&reply);
    assert(reply.buffer == NULL);
}

void test_reply_big_length()
{
    HttpReply reply;
    int ret;
    const char *text = \
"HTTP/1.1 200 OK\r\n\
Date: Wed, 30 Oct 2002 12:43:03 GMT\r\n\
Content-Length: 4100200300\r\n\
Connetion: close\r\n\
\r\n";

    ret = parse_reply(&reply, text);
    assert(ret == PARSE_DONE);
    assert(reply.replylength == strlen(text));
    assert(reply.has_contentlength == 1);
    assert(reply.contentlength == 4100200300U && reply.contentlength > 0);
    assert(reply.version == 1);
    assert(reply.status == 200);
    assert(reply.is_cachable == 0);
    assert(reply.cangzip == 0);
    assert(reply.alreadygzipped == 0);

    http_free_reply(&reply);
    assert(reply.buffer == NULL);
}


void test_req_bad()
{
    HttpRequest req;
    int ret;
    const char *text = "This is a bad request\r\n\r\n";

    ret = parse_req(&req, text);
    assert(ret == PARSE_ERROR);
    http_free_request(&req);
}


void test_reply_bad()
{
    HttpReply reply;
    int ret;
    const char *text = "This is a bad reply\r\n\r\n";

    ret = parse_reply(&reply, text);
    assert(ret == PARSE_ERROR);
    http_free_reply(&reply);
}

int main()
{

    printf("httptest..........");
    Config.cache_expires = DAY * 10;
    bp = malloc(sizeof(BufferPool));
    bufferpool_init(bp, 1024, 5);
    set_current_time();
    
    test_req_get();
    test_req_post();
    test_req_no_cache1();
    test_req_no_cache2();
    test_req_cangzip();
    test_req_auth();
    test_req_vicompressid();
    test_reply();
    test_reply_cachable();
    test_reply_gzip();
    test_reply_big_length();
    test_req_bad();
    test_reply_bad();

    printf("Pass\n");
    return 0;
}

