/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */


/* 
 * The LoadBalancer is used to load balance among multiple backend webservers. 
 * Each webserver is assigned an id, starting from 1 (0 is not a valid id).
 *
 * In the vicompress.conf file, a backend webserver is specified using the format
 *
 * webserver <IP address> <port> [host1.com] [host2.com] ...
 *
 * The IP address and port are required.  The hostnames are only required if
 * different websites (virtual hosts), run on different addresses/ports.
 *
 * If vicompress fails to connect to a webserver, the webserver is marked "down"
 * and will not receive any more requests.  Vicompress will immediately send a
 * test connection to the webserver, to see if the server has come back up.
 * If the webserver is still down after that, Vicompress will try to connect to
 * the down webservers every 3 minutes (see CHECK_WEBSERVER_INTERVAL in
 * vicompress.c).
 *
 * If sticky sessions are enabled (enable_sessions yes), when a new client
 * connects, vicompress sends back the webserver id as a cookie:
 * 
 * Set-Cookie: vicompressid=1
 *
 * When the client connects again, they will send the same id:
 *
 * Cookie: vicompressid=1
 *
 * Using cookies, clients are sent to the same webserver, so that sessions 
 * can be preserved. The session ends when the client closes/exits the browser.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "loadbalancer.h"
#include "currenttime.h"
#include "sockets.h"
#include "win32.h"

static void send_test_connection(LoadBalancer *b, int id);


/* Intialize the LoadBalancer with the given list of web servers.
 * The 'Sockets' is used to send test connections to down webservers.
 * The errorlog is used to notify when a webserver goes up or down.
 */
void loadbalancer_init(LoadBalancer *balancer, Sockets *s, ErrorLog *errorlog,
                       Webserver *servers, int count)
{
    int i;

    assert(count > 0);
    balancer->errorlog = errorlog;
    balancer->sockets = s;
    balancer->num_webservers = count;
    balancer->num_down_servers = 0;
    balancer->webservers = servers;

    balancer->has_hosts = 0;
    for (i = 0; i < balancer->num_webservers; i++) {
        if (balancer->webservers[i].num_hosts > 0) {
            balancer->has_hosts = 1;
        }
    } 
}

/* Retrieve the web server IP address/port that corresponds
 * to the given id.  Return NULL if there is no match.
 */
sockaddr_all* loadbalancer_getaddr(LoadBalancer *balancer, int id)
{
    if (id <= 0 || id > balancer->num_webservers)
        return NULL;
    else
        return &(balancer->webservers[id-1].addr);
}

/* Given two webservers, choose which one to use, based on the
 * current load (number of connections), and which server was
 * used more recently.
 */
static int
loadbalancer_choose(LoadBalancer *balancer, int id1, int id2)
{
    Webserver *s1, *s2;

    if (id1 < 0)
        return id2;

    if (id2 < 0)
        return id1;

    s1 = &balancer->webservers[id1];
    s2 = &balancer->webservers[id2];

    if (s1->is_down)
        return id2;

    if (s2->is_down)
        return id1;

    if (s1->active_conn < s2->active_conn)
        return id1;

    if (s2->active_conn < s1->active_conn)
        return id2;

    if (s2->last_used < s1->last_used)
        return id2;

    return id1;
}


/* Retrieve a webserver id from the LoadBalancer. If the webserver
 * entries have "host" options, then look for webservers with hosts
 * that match the given host argument.  Webservers marked as 
 * down will be skipped. However, if all webservers are down, then 
 * a down webserver must be returned.
 *
 * The id returned is incremented by one, so that 0 is not a valid
 * webserver id.
 */
int loadbalancer_get_id(LoadBalancer *balancer, Substring host)
{
    Webserver *server;
    int id, best_id, j;
    char buf[64];

    best_id = -1;     /* ID of matching server, with the least load */

    memset(buf, 0, 64);
    strncpy(buf, host.buf, host.size);

    if (balancer->has_hosts && host.size > 0) {
        /* Find a webserver with a virtual host that matches
         * the given host.
         */
        for (id = 0; id < balancer->num_webservers; id++) {
            server = &(balancer->webservers[id]);
            for (j = 0; j < server->num_hosts; j++) {
                
                if (strncmp(server->hosts[j], host.buf, host.size) == 0) {
                    best_id = loadbalancer_choose(balancer, best_id, id);
                }
            }
        }
        /* Return the id of the matching webserver, if any. */
        if (best_id != -1) {
            return best_id+1;
        }
    }

    /* Find a webserver with the least number of connections */
    for (id = 0; id < balancer->num_webservers; id++) {
        server = &(balancer->webservers[id]);
        if (server->num_hosts == 0) {
            best_id = loadbalancer_choose(balancer, best_id, id);
        }
    }
    if (best_id == -1) {
        best_id = 0;
    }
    return best_id+1;
}


/* Return an id other than the current one. This function is implemented
 * by marking the current_id as "down", calling loadbalancer_get_id(),
 * and then marking current_id back to its original setting.
 */
int loadbalancer_get_another_id(LoadBalancer *balancer, int current_id, Substring host)
{
    Webserver *s;
    int id;
    char is_down;

    assert(current_id >= 1 && current_id <= balancer->num_webservers);
    s = &(balancer->webservers[current_id-1]);
    is_down = s->is_down;
    s->is_down = 1;
    balancer->num_down_servers++;
    id = loadbalancer_get_id(balancer, host);
    s->is_down = is_down;
    balancer->num_down_servers--;
    return id;
}

/* Return 1 if the given id refers to a valid server, else 0 */
boolean loadbalancer_is_valid(LoadBalancer *balancer, int id, Substring host)
{
    Webserver *server;
    int i;
    if (id <= 0 || id > balancer->num_webservers)
        return 0;

    server = &(balancer->webservers[id-1]);
    if (server->num_hosts == 0 || host.size == 0)
        return 1;

    for (i = 0; i < server->num_hosts; i++) {
        if (strncasecmp(server->hosts[i], host.buf, host.size) == 0) {
            return 1;
        }
    }
    return 0;
}

/* Return 1 if the given server is down, else 0 */
boolean loadbalancer_is_server_down(LoadBalancer *balancer, int id)
{
    if (id <= 0 || id > balancer->num_webservers)
        return 1;
    else
        return balancer->webservers[id-1].is_down;
}

/* Note that a new connection has started to the given webserver.
 * This function is needed to fairly load-balance among the
 * webservers.
 */
void loadbalancer_connect_started(LoadBalancer *balancer, int id)
{
    if (id <= 0 || id > balancer->num_webservers)
        return;
    balancer->webservers[id-1].active_conn++; 
    balancer->webservers[id-1].last_used = get_current_time();
}

/* Note that a webserver connection has finished.  If the
 * connection failed, mark the webserver as down, and 
 * immediately kick off a test connection to see if the webserver
 * has come back up.
 */
void loadbalancer_connect_done(LoadBalancer *balancer, int id, int error)
{
    Webserver *s;

    if (id <= 0 || id > balancer->num_webservers) 
        return;

    s = &balancer->webservers[id-1];
    s->active_conn--;
    assert(s->active_conn >= 0);
    if (error == 0) {
        if (s->is_down) {
            balancer->num_down_servers--;
        }
        s->is_down = 0;
        return;
    }

    else {
        if (s->is_down) {
            return;
        }
        s->is_down = 1;

        errorlog_printf(balancer->errorlog, NULL,
                        "Marking webserver %s as down",
                        sockets_to_string(&s->addr));

        balancer->num_down_servers++;
        send_test_connection(balancer, id);
    }
}


static ConnectCallback connect_status;

/* Callback data passed to the sockets_connect() function. 
 * The "id" is the webserver id which we tried to connect to.
 */
struct _ServerStatus {
    LoadBalancer *loadbalancer;
    int id;
};
typedef struct _ServerStatus ServerStatus;

/* Callback function for connecting to web servers. */
static void connect_status(int fd, void *data, int error)
{
    ServerStatus *status;
    LoadBalancer *balancer;
    Webserver *server;
    sockaddr_all *addr;
    int id;

    status = (ServerStatus*) data;
    balancer = status->loadbalancer;
    id = status->id;
    server = &(balancer->webservers[id-1]);

    addr = loadbalancer_getaddr(balancer, id);

    if (error == 0) {
        if (server->is_down) {
            balancer->num_down_servers--;
        }
        server->is_down = 0;
        errorlog_printf(balancer->errorlog, NULL,
                        "Webserver %s is now up",
                        sockets_to_string(addr));
    }
    else {
        DEBUG_PRINTF(balancer->errorlog, NULL,
                     "Webserver %s is still down",
                     sockets_to_string(addr));
    }
    sockets_close(balancer->sockets, fd);
    free(status);
}

/* Send a test connection to the given webserver id.  The connect_status()
 * function above will be invoked when the connection succeeds/fails.
 */
static void send_test_connection(LoadBalancer *balancer, int id)
{
    int sockfd;
    sockaddr_all *addr;
    ServerStatus *status;

    addr = loadbalancer_getaddr(balancer, id);
    assert(addr != NULL);

    sockfd = sockets_new(balancer->sockets, addr->ipv4.sin_family);
    if (sockfd == -1) {
        return;
    }
    status = malloc(sizeof(ServerStatus));
    if (status == NULL) {
        sockets_close(balancer->sockets, sockfd);
        return;
    }

    status->loadbalancer = balancer;
    status->id = id;
    DEBUG_PRINTF(balancer->errorlog, NULL,
                 "Checking if webserver %s is up",
                 sockets_to_string(addr));

    sockets_connect(balancer->sockets, sockfd, addr, 
                    connect_status, (void*)status);
}

/* Check the status of webservers that are currently marked down.
 * Try to connect to the web server.  If the connection
 * succeeds, mark the server as "up".
 */
void loadbalancer_check_webservers(LoadBalancer *balancer)
{
    int id;

    for (id = 1; id <= balancer->num_webservers; id++) {
        if (loadbalancer_is_server_down(balancer, id)) {
            send_test_connection(balancer, id);
        }
    }
}



