/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**
    The loadbalancer_test will do the following:
    - Test that get_id() returns the server with the least number of
      connections, or the server that was last used.
    - Test that multiple failed connections will mark the server as down.
    - Test that get_id() will not return a server marked as down.
    - For 2 of the 3 webservers, create listening sockets, and mark
      all servers as down.  Verify that check_webservers() will mark
      the 2 servers with listening sockets as up.
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "sockets.h"
#include "log.h"
#include "loadbalancer.h"
#include "conf.h"
#include "win32.h"

void init_addr(sockaddr_all *addr, const char *ip, int port)
{
    in_addr_t addr4;
    unsigned short nport;

    addr4 = inet_addr(ip);
    nport = htons(port);

    memset(addr, 0, sizeof(sockaddr_all));
    addr->ipv4.sin_family = AF_INET;
    addr->ipv4.sin_port = nport;
    addr->ipv4.sin_addr.s_addr = addr4;
}

int getport(sockaddr_all *addr) {
    if (addr == NULL)
        return 0;
    else
        return ntohs(addr->ipv4.sin_port);
}

char* getip(sockaddr_all *addr)
{
    unsigned char *ip;
    char *buf;

    if (addr == NULL)
        return NULL;
    buf = malloc(32);
    ip = (unsigned char*) &(addr->ipv4.sin_addr.s_addr);
    sprintf(buf, "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
    return buf;
}


void server_accept(int clientfd, sockaddr_all addr, void *data)
{
    Sockets *s = data;

    sockets_close(s, clientfd);
}

void test_no_hosts() {
    LoadBalancer *balancer;
    Sockets *sockets;
    int sockfd[2];
    Webserver webservers[3];
    char *ip;
    int i, port, id, ret; 
    time_t curr, end;
    Substring host;

    sockets = malloc(sizeof(Sockets));
    Config.max_total_fd = 100;
    Config.num_webservers = 3;
    sockets_init(sockets);

    memset(webservers, 0, sizeof(webservers));
    host.buf = "";
    host.size = 0;

    for (i = 0; i < 3; i++)  {
        init_addr(&webservers[i].addr, "127.0.0.1", 3180 + i);
    }

    balancer = (LoadBalancer *)malloc(sizeof(LoadBalancer));
    loadbalancer_init(balancer, sockets, NULL, webservers, 3);

    for (i = 0; i < 3; i++) {
        ip = getip(loadbalancer_getaddr(balancer, i+1));
        port = getport(loadbalancer_getaddr(balancer, i+1));
        assert(strcmp(ip, "127.0.0.1") == 0);
        assert(port == 3180 + i);
        ret = loadbalancer_is_server_down(balancer, i+1);
        assert(ret == 0);
        ret = loadbalancer_is_valid(balancer, i+1, host);
        assert(ret == 1);
    }
    ret = loadbalancer_is_valid(balancer, 5, host);
    assert(ret == 0);

    for (i = 0; i < 3; i++) {
        id = loadbalancer_get_id(balancer, host);
        loadbalancer_connect_started(balancer, id);
        assert(id == i+1);
    }
    loadbalancer_connect_done(balancer, 3, 0);
    id = loadbalancer_get_id(balancer, host);
    loadbalancer_connect_started(balancer, id);
    assert(id == 3);

    for (i = 0; i < 4; i++) {
        id = loadbalancer_get_another_id(balancer, 1, host);
        loadbalancer_connect_started(balancer, id);
        assert(id != 1);
    }

    loadbalancer_connect_started(balancer, 1);
    loadbalancer_connect_done(balancer, 1, 1);
    ret = loadbalancer_is_server_down(balancer, 1);
    assert(ret == 1);

    for (i = 0; i < 10; i++) {
        sockets_checkevents(sockets, 100);
    }

    for (i = 0; i < 3; i++) {
        id = loadbalancer_get_id(balancer, host); 
        loadbalancer_connect_started(balancer, id);
        assert(id > 1);
    }
    loadbalancer_connect_started(balancer, 1);
    loadbalancer_connect_done(balancer, 1, 0);
    ret = loadbalancer_is_server_down(balancer, 1);
    assert(ret == 0);

    for (i = 0; i < 3; i++) {
        loadbalancer_connect_started(balancer, i+1);
        loadbalancer_connect_done(balancer, i+1, 1);
        ret = loadbalancer_is_server_down(balancer, i+1);
        assert(ret == 1);
    }

    for (i = 0; i < 10; i++) {
        sockets_checkevents(sockets, 100);
    }

    for (i = 0; i < 3; i++) {
        ret = loadbalancer_is_server_down(balancer, i+1);
        assert(ret == 1 );
    }

    for (i = 0; i < 2; i++) {
        struct sockaddr_in addr;

        memset(&addr, 0, sizeof(struct sockaddr_in));
        addr.sin_family = AF_INET;
        addr.sin_addr = webservers[i].addr.ipv4.sin_addr;
        addr.sin_port = webservers[i].addr.ipv4.sin_port;

        sockfd[i] = sockets_new(sockets, AF_INET);
        ret = 1;
        setsockopt(sockfd[i], SOL_SOCKET, SO_REUSEADDR, &ret, sizeof(int));
        ret = bind(sockfd[i], (struct sockaddr*) &addr, 
                   sizeof(struct sockaddr_in));
        assert(ret == 0);
        listen(sockfd[i], 1024);
        sockets_listen(sockets, sockfd[i], server_accept, sockets);
    }

    sleep(1);
    loadbalancer_check_webservers(balancer);
    curr = time(NULL);
    end = curr + 3;
    while (curr < end) {
        sockets_checkevents(sockets, 100);
        curr = time(NULL);
    }

    ret = loadbalancer_is_server_down(balancer, 1); assert(ret == 0);
    ret = loadbalancer_is_server_down(balancer, 2); assert(ret == 0);
    ret = loadbalancer_is_server_down(balancer, 3); assert(ret == 1);

    free(sockets);
    free(balancer);
}


void test_with_hosts() {
    LoadBalancer *balancer;
    Sockets *sockets;
    Webserver webservers[20];
    char *ip;
    int i, port, id, ret; 
    Substring host;

    sockets = malloc(sizeof(Sockets));
    Config.max_total_fd = 100;
    Config.num_webservers = 3;
    sockets_init(sockets);

    memset(webservers, 0, sizeof(webservers));
    for (i = 0; i < 20; i++)  {
        init_addr(&webservers[i].addr, "127.0.0.1", 3180 + i);
    }
    for (i = 0; i < 10; i++) {
        webservers[i].num_hosts = 1;
        webservers[i].hosts[0] = "host1";
    }
    for (i = 10; i < 20; i++) {
        webservers[i].num_hosts = 1;
        webservers[i].hosts[0] = "host2";
    }

    balancer = (LoadBalancer *)malloc(sizeof(LoadBalancer));
    loadbalancer_init(balancer, sockets, NULL, webservers, 20);

    host.buf = ""; host.size = 0;
    for (i = 0; i < 20; i++) {
        ip = getip(loadbalancer_getaddr(balancer, i+1));
        port = getport(loadbalancer_getaddr(balancer, i+1));
        assert(strcmp(ip, "127.0.0.1") == 0);
        assert(port == 3180 + i);
        ret = loadbalancer_is_server_down(balancer, i+1);
        assert(ret == 0);
        ret = loadbalancer_is_valid(balancer, i+1, host);
        assert(ret == 1);
    }
    ret = loadbalancer_is_valid(balancer, 23, host);
    assert(ret == 0);
    host.buf = "sample.com"; host.size = strlen("sample.com");
    ret = loadbalancer_is_valid(balancer, 5, host);
    assert(ret == 0);
    ret = loadbalancer_is_valid(balancer, 15, host);
    assert(ret == 0);

    host.buf = "host1"; host.size = 5;
    for (i = 0; i < 10; i++) {
        id = loadbalancer_get_id(balancer, host);
        loadbalancer_connect_started(balancer, id);
        assert(id == i+1);
    }
    for (i = 0; i < 10; i++) {
        id = loadbalancer_get_id(balancer, host);
        loadbalancer_connect_started(balancer, id);
        assert(id == i+1);
    }
    loadbalancer_connect_done(balancer, 6, 0);
    id = loadbalancer_get_id(balancer, host);
    loadbalancer_connect_started(balancer, id);
    assert(id == 6);

    host.buf = "host2";
    for (i = 0; i < 10; i++) {
        id = loadbalancer_get_id(balancer, host);
        loadbalancer_connect_started(balancer, id);
        assert(id == i+11);
    }
    for (i = 0; i < 10; i++) {
        id = loadbalancer_get_id(balancer, host);
        loadbalancer_connect_started(balancer, id);
        assert(id == i+11);
    }

    for (i = 0; i < 20; i++) {
        id = loadbalancer_get_another_id(balancer, 11, host);
        loadbalancer_connect_started(balancer, id);
        assert(id != 11);
    }

    loadbalancer_connect_done(balancer, 12, 1);
    ret = loadbalancer_is_server_down(balancer, 12);
    assert(ret == 1);

    for (i = 0; i < 20; i++) {
        id = loadbalancer_get_id(balancer, host);
        loadbalancer_connect_started(balancer, id);
        assert(id != 12);
    }
}

int main() {
    printf("loadbalancer......");
    test_no_hosts();
    test_with_hosts();
    printf("Pass\n");
    return 0;
}

