/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**
The ErrorLog is used to log errors or debugging information.
The errorlog has the following format:

  [Dayweek Month day time year] [clientip:port] message
  [Mon Aug 21 17:26:45 2003] [192.10.10.51:2352] A message
  [Mon Aug 21 17:26:52 2003] [192.10.10.51:6532] A debug message

The Accesslog is used to store information about client requests
to the proxy. The accesslog can be in one of two formats:
  The Apache Combined Log Format
  The Squid log format

The Apache Combined Log Format has the following fields:
  clientip hitcompress username date "firstline" status contentlength "referer" "user-agent"

  15.13.130.10 hit24359 - [21/Aug/2003:17:26:45 18000] "GET /index.html HTTP/1.0" 200 1234 "http://www.anothersite.com" "Mozilla/4.0 (compatible; MSIE 6.0)"

  clientip:       IP address of client (%s)
  hit/compress:   Either hit or miss, followed by pre-compression contentlength
  username:       Username sent for Authentication, or '-' if not present.
  date:           Date of response [day/month/year:hour:min:sec -timezone]
  firstline:      method urlpath HTTP/version
  status:         Response code of server reply
  contentlength:  Data delivered to client, excluding headers
  referer:        The website that referred to this URL
  user-agent:     The name and version of the browser

The Squid access log format has the following fields:

 time duration remotehost code/status bytes method URL compress peerstatus/peerhost type

 time:            UTC seconds.millsec  (%9d.%.03d)
 duration:        time between accept and close, in milliseconds (%6d)
 client address:  IP address of client (%s)
 result code:     TCP_HIT, TCP_MISS
 status:          HTTP status (%03d)
 bytes:           %d bytes delivered to client, including headers (%d)
 method:          GET, POST, HEAD, or whatever (%s)
 URL:             Full request url, http://www.site.com/path.html (%s)
 ident:           The contentlength before compression
 peerstatus:      NONE (for hits), DIRECT (for miss)
 peerhost:        IP address of server, or "-" if hit.
 type:            Content-type of http reply. Just "-"

**/


#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include "log.h"
#include "conf.h"
#include "currenttime.h"
#include "win32.h"

#define MAX_LOG_ENTRIES      512
#define ENTRY_BUF_SIZE       512
#define MAX_ENTRY_BUF_SIZE  4096

static void* log_thread(void *arg);

static char* months[] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
    "Aug", "Sep", "Oct", "Nov", "Dec"
};

static char* daysofweek[] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

/* Round this time_t value down to the beginning of the hour. */
static time_t gethour(time_t curr) {
    curr = curr - (curr % HOUR);
    return curr;
}

static struct tm lastdate;
static time_t lasthour = 0;

/* To improve performance, only call localtime_r() when
 * the hour changes. 
 */ 
static void do_localtime(time_t thetime, struct tm *m) {
    time_t currhour;

    currhour = gethour(get_current_time());
    if (currhour != lasthour) {
        lasthour = currhour;
        localtime_r(&lasthour, &lastdate);
    } 

    if (thetime > lasthour && (thetime - lasthour < HOUR)) {
        time_t diff = thetime - lasthour;
        memcpy(m, &lastdate, sizeof(struct tm));
        m->tm_sec = diff % 60;
        m->tm_min = diff / 60; 
        return;
    }
    else {
        localtime_r(&thetime, m);
    }
}

/* Given the logfile path, execute the commands:
 * mv path.1 path.2
 * mv path path.1
 */
static void rotate_log(const char *path)
{
    struct stat info;
    char *oldpath;
    char *newpath;

    oldpath = malloc(256);
    newpath = malloc(256);

    sprintf(oldpath, "%s.1", path);
    sprintf(newpath, "%s.2", path);
    if (stat(oldpath, &info) == 0) {
        rename(oldpath, newpath);
    }

    sprintf(newpath, "%s.1", path);
    if (stat(path, &info) == 0) {
        rename(path, newpath);
    }
    free(oldpath);
    free(newpath);
}

/* Rotate the access log. Execute:
 * mv accesslog.1 accesslog.2
 * mv accesslog accesslog.1
 * log->fd = open(accesslog)
 */ 
static void rotate_access_log(AccessLog *log)
{
    close(log->fd);
    rotate_log(log->path);
    log->currentsize = 0;
    log->fd = open(log->path, O_WRONLY | O_APPEND | O_CREAT, 0644);
}

/* Rotate the error log. Execute:
 * mv errorlog.1 errorlog.2
 * mv errorlog errorlog.1
 * log->fd = open(errorlog)
 */ 
static void rotate_error_log(ErrorLog *log)
{
    close(log->fd);
    rotate_log(log->path);
    log->currentsize = 0;
    log->fd = open(log->path, O_WRONLY | O_APPEND | O_CREAT, 0644);
}

/* Return the file size (in bytes) of the given file path */
static long getfilesize(const char *path)
{
    struct stat info;
    int n;
    n = stat(path, &info);
    if (n != 0)
        return 0;
    return (long) info.st_size;
}

/* Return the offset where char c occurs in buf.
 * c must be between offset and end.
 * If c is not found, return -1.
 */
static int findchar(const char *buf, int offset, int end, char c)
{
    if (offset < 0)
        return -1;
    while (offset < end) {
        if (buf[offset] == c) {
            return offset;
        }
        offset++;
    }
    return -1;
}

/* Return the number of double quotes (") found in the string. */
static int count_quotes(Substring s)
{
    int i;
    int count = 0;
    for (i = 0; i < s.size; i++) {
        if (s.buf[i] == '"') {
            count++;
        }
    }
    return count;
}

/* Copy the Substring s into buffer buf.  When double quotes (")
 * are encountered, replace them with \"
 */
static int copy_and_escape_quotes(char *buf, Substring s)
{
    int src, dest;
    src = dest = 0; 

    for (src = 0; src < s.size; src++) {
        if (s.buf[src] == '"') {
            buf[dest] = '\\';
            dest++;
            buf[dest] = s.buf[src];
        }
        else {
            buf[dest] = s.buf[src];
        }
        dest++;
    }
    return dest;
}

/* Open the access_log at the given path. Spawn a thread to
 * do the actual disk writes asynchronously.
 */
void accesslog_init(AccessLog *log, const char *path)
{
    pthread_attr_t attr;
    pthread_t id;
    int result;

    set_current_time();
    memset(log, 0, sizeof(AccessLog));
    log->path = path;
    queue_init(&log->entries, MAX_LOG_ENTRIES + 1);
    bufferpool_init(&log->entrypool, ENTRY_BUF_SIZE, MAX_LOG_ENTRIES*2/3);

    log->fd = open(log->path, O_WRONLY | O_APPEND | O_CREAT, 0644);
    if (log->fd <= 0) {
        fprintf(stderr, "vicompress ERROR: Cannot open accesslog file %s, %s\n",
                log->path, strerror(errno));
        exit(1);
    }
    log->currentsize = getfilesize(log->path);

    log->queuesize = 0;
    pthread_mutex_init(&log->poollock, NULL);
    pthread_cond_init(&log->sizenotify, NULL);

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    result = pthread_create(&id, &attr, log_thread, (void *) log);
    assert(result == 0);
}

/* Close the accesslog. Send an empty entry to
 * the write thread to wake it up.  The thread will
 * do the actual close of the file descriptor.
 */
void accesslog_close(AccessLog *log)
{
    Buffer *b;

    pthread_mutex_lock(&log->poollock);
    log->queuesize += 1;
    b = buffer_alloc(&log->entrypool);
    pthread_mutex_unlock(&log->poollock);
    log->doclose = 1;
    b->size = 0;
    queue_add(&log->entries, b);
    (void)sleep(1);
}

/* Return an IP address as a string */
static char* ip_to_string(sockaddr_all *addr)
{
    static char str[64];
    unsigned char *ip;

    str[0] = 0;
#ifdef HAVE_IPV6
    if (addr->ipv4.sin_family == AF_INET6) {
        char ipstr[50];
        inet_ntop(AF_INET6, &(addr->ipv6.sin6_addr), ipstr, 50);
        sprintf(str, "%s", ipstr);
        return &str[0];
    }
#endif
    ip = (unsigned char *)&(addr->ipv4.sin_addr.s_addr);
    sprintf(str, "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);
    return &str[0];
}

/* Return the size (in bytes) needed to write a Squid formatted
 * log entry for the given http data.
 */
static int get_squid_size(Substring method, Substring host, Substring path,
                                                    Substring contenttype)
{
    int total = 0;
    total += 15;                        /* time */
    total += 6;                         /* duration */
    total += 41;                        /* client ip */
    total += 11;                        /* TCP_HIT, TCP_MISS, TCP_DENIED */
    total += 4;                         /* HTTP status */
    total += 8;                         /* content length */
    total += method.size;               /* method */
    total += 9 + host.size + path.size; /* URL */
    total += 8;                         /* ident */
    total += 7;                         /* peer status */
    total += 17;                        /* server ip */
    total += contenttype.size;          /* Content-type */
    total += 2;                         /* newline */
    return total;
}

/* Given the following http and network info, create
 * a Squid formatted accesslog line, and write it to
 * buffer b.
 */
static void 
create_squid_entry(Buffer *b, HttpRequest *req, HttpReply *reply,
                   sockaddr_all *client, 
                   sockaddr_all *server,
                   unsigned long replysize, unsigned long pregzip_size, 
                   int iscachehit, int isrequesterr, int duration)
{
    Substring contenttype;
    struct timeval tv;
    int total;
    char *buf;
    int amount, n;
    char hostip[50], pregzip[20];
    char *cachedcode, *peer, *urlprefix;

    if (isrequesterr) {
        req->method.buf = "NONE";
        req->method.size = 4;
    } 

    if (req->host.size == 0 && req->path.size == 0)
        urlprefix = " error:unsupported-request-method";
    else
        urlprefix = " http://";

    contenttype = http_get_reply_header(reply, "Content-type");
    if (contenttype.size == 0) {
        contenttype.buf = "-"; contenttype.size = 1;
    }
    n = findchar(contenttype.buf, 0, contenttype.size, ' ');
    if (n > 0 && n < contenttype.size)
        contenttype.size = n;

    total = get_squid_size(req->method, req->host, req->path, contenttype);

    if (total > MAX_ENTRY_BUF_SIZE)
        return;
    if (total > b->capacity && total < MAX_ENTRY_BUF_SIZE)
        buffer_grow(b, total);

    buf = b->buf;

    hostip[0] = 0;

    if (iscachehit) {
        cachedcode = "TCP_HIT"; peer = "NONE";
        strcpy(hostip, "-");
    }
    else if (isrequesterr) {
        cachedcode = "TCP_DENIED"; peer = "NONE";
        strcpy(hostip, "-");
    }
    else {
        cachedcode = "TCP_MISS"; peer = "DIRECT";
        sprintf(hostip, "%s", ip_to_string(server));
    }

    gettimeofday(&tv, NULL);
    amount = sprintf(buf, 
#ifdef WIN32
                     "%.9d.%.3d %6d %s %s/%d %lu ", 
#else
                     "%.9d.%.3d %6d %s %s/%d %lu ", 
#endif
                     (int)tv.tv_sec,
                     (int)(tv.tv_usec / 1000), 
                     duration,
                     ip_to_string(client),
                     cachedcode, 
                     reply->status,
                     replysize);

    memcpy(&buf[amount], req->method.buf, req->method.size);
    amount += req->method.size;
    memcpy(&buf[amount], urlprefix, strlen(urlprefix));
    amount += strlen(urlprefix);
    memcpy(&buf[amount], req->host.buf, req->host.size);
    amount += req->host.size;
    memcpy(&buf[amount], req->path.buf, req->path.size);
    amount += req->path.size;

    snprintf(pregzip, sizeof(pregzip), " %lu ", pregzip_size);
    memcpy(&buf[amount], pregzip, strlen(pregzip));
    amount += strlen(pregzip);
    memcpy(&buf[amount], peer, strlen(peer));
    amount += strlen(peer);
    memcpy(&buf[amount], "/", 1);
    amount += 1;
    memcpy(&buf[amount], hostip, strlen(hostip));
    amount += strlen(hostip);
    memcpy(&buf[amount], " ", 1);
    amount += 1;
    memcpy(&buf[amount], contenttype.buf, contenttype.size);
    amount += contenttype.size;
     
    buf[amount] = '\n';
    amount++;
    buf[amount] = 0;

    b->size = amount; 
}

/* Return the size (in bytes) needed to write an Apache formatted
 * log entry for the given http data.
 */
static int get_apache_size(Substring firstline, Substring referer,
                           Substring useragent, char *username)
{
    int total = 0;
    total += 41;                      /* client ip */
    total += 12;                      /* ident name */
    total += strlen(username) + 1;    /* username */
    total += 29;                      /* date/time */
    total += 3 + firstline.size + count_quotes(firstline); /* firstline */
    total += 4;                       /* status code */
    total += 8;                       /* content length */
    total += 3 + referer.size + count_quotes(referer);     /* referer */
    total += 3 + useragent.size + count_quotes(useragent); /* user-agent */
    total += 2;                       /* newline */
    return total;
}

/* Given the following http and network info, create
 * an Apache formatted accesslog line, and write it to
 * buffer b.
 */
static void 
create_apache_entry(Buffer *b, HttpRequest *req, HttpReply *reply,
                    sockaddr_all *client, unsigned long replysize,
                    unsigned long pregzip_size, int iscachehit)
{
    Substring firstline, referer, useragent;
    char username[64];
    char *buf;
    int len, amount, total;
    struct tm thedate;
    char tmzone[12];
    const char *hit;
    int thetimezone;

    firstline.buf = req->buffer->buf;
    if (req->firstline > 0) {
        firstline.size = req->firstline - 1;
        if (firstline.buf[firstline.size-1] == '\r') {
            firstline.size--;
        }
    }
    else {
        firstline.size = 0;
    }  
    referer = http_get_request_header(req, "Referer");
    useragent = http_get_request_header(req, "User-Agent");
    if (referer.size == 0) {
        referer.buf = "-"; referer.size = 1;
    }
    if (useragent.size == 0) {
        useragent.buf = "-"; useragent.size = 1;
    }
    if (iscachehit)
        hit = "hit";
    else
        hit = "miss";


    username[0] = 0;
    http_get_username(req, (unsigned char *)username, sizeof(username));
    if (username[0] == 0)
        strcpy(username, "-");

    total = get_apache_size(firstline, referer, useragent, username);
    if (total > MAX_ENTRY_BUF_SIZE)
        return;
    if (total > b->capacity && total < MAX_ENTRY_BUF_SIZE)
        buffer_grow(b, total);

    do_localtime(get_current_time(), &thedate);

#if defined(__CYGWIN__) || defined(WIN32)
    thetimezone = _timezone;
#elif defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
    thetimezone = thedate.tm_gmtoff;
#else
    thetimezone = timezone;
#endif

    sprintf(tmzone, " %.2d%.2d", 
            abs(thetimezone/HOUR), abs(thetimezone % HOUR) / 60);
    if (thetimezone > 0)
        tmzone[0] = '-';
    else
        tmzone[0] = '+';

    buf = b->buf;
    amount = 0;
    amount = sprintf(buf, "%s %s%lu %s ", 
                     ip_to_string(client), hit, pregzip_size, username);

    len = sprintf(&buf[amount], "[%.2d/%s/%.4d:%.2d:%.2d:%.2d %s] \"",
                  thedate.tm_mday, months[thedate.tm_mon], 
                  thedate.tm_year + 1900,
                  thedate.tm_hour, thedate.tm_min, thedate.tm_sec,
                  tmzone); 
    amount += len;

    amount += copy_and_escape_quotes(&buf[amount], firstline);

    len = sprintf(&buf[amount], "\" %3d %lu \"", 
                  reply->status, replysize);
    amount += len;

    amount += copy_and_escape_quotes(&buf[amount], referer);
 
    memcpy(&buf[amount], "\" \"", 3);
    amount += 3;

    amount += copy_and_escape_quotes(&buf[amount], useragent);

    memcpy(&buf[amount], "\"\n", 2);
    amount += 2;
    buf[amount] = 0;
    b->size = amount;
}

/* Given the following http request/reply info, write a log entry
 * to the access log.  Create the entry line, store it in a 
 * buffer, and then queue that buffer to the log thread.  The
 * log thread will do the actual write to disk.
 *
 * log          - The access log
 * req          - The http request received from the client. Data used
 *                includes the url, Referer, User-Agent (browser), and 
 *                Authorization username.
 * reply        - The http reply.  Data used includes the reply status and 
 *                content type.
 * client       - The client IP address and port
 * server       - The origin server IP address and port.  For cache hits,
 *                the values are 0.0.0.0 and port = 0.
 * replysize    - The amount of content written to the client.
 *                Effectively, the content-length.
 * pregzip_size - The original size of the content, before compression.
 *                If no compression was done, then replysize == pregzip_size.
 * iscachehit   - True if this was a cache hit, else false.
 * isrequesterr - True if the http request was invalid, else false.
 * duration     - The duration of the request, from the initial accept()
 *                to the final data written to the client.
 */
void accesslog_write(AccessLog *log, HttpRequest *req, HttpReply *reply,
                     sockaddr_all *client, sockaddr_all *server,
                     unsigned long replysize, unsigned long pregzip_size, 
                     int iscachehit, int isrequesterr, int duration)
{
    Buffer *b;

    /* If we can't write to log, return */
    if (log->fd == -1)
        return;

    /* Make sure log entry queue is < MAX_LOG_ENTRIES. If
     * queue length reaches maximum value, wait until the 
     * queue size shrinks before adding another entry.
     */
    pthread_mutex_lock(&log->poollock); 
    if (log->queuesize >= MAX_LOG_ENTRIES) {
        pthread_cond_wait(&log->sizenotify, &log->poollock);
    }
    log->queuesize += 1;
    b = buffer_alloc(&(log->entrypool));
    pthread_mutex_unlock(&log->poollock);

    if (Config.logformat == LOG_FORMAT_APACHE) {
        create_apache_entry(b, req, reply, client, replysize, 
                            pregzip_size, iscachehit);
    }
    else {
        create_squid_entry(b, req, reply, client, server, replysize, 
                           pregzip_size, iscachehit, isrequesterr, duration);
    }
    queue_add(&log->entries, b);
}

/* Write the given buffer data to the given file. */
static int write_buffer(int fd, char *buffer, int total)
{
    int amount, n, error;

    amount = 0;
    while (amount < total) {
        n = write(fd, &buffer[amount], total - amount);
        error = errno;
        if (n < 0 && errno == EINTR)
            continue; 
        else if (n < 0 && errno != EINTR) {
            /*
            fprintf(stderr, "vicompress: ERROR writing to logfile: %s\n", 
                            strerror(error));
            */
            return amount;
        }
        amount += n;
    } 
    return total;
}

/* Thread which handles the accesslog writes asynchronously. */
static void* log_thread(void *arg)
{
    AccessLog *log;
    Buffer *b;
    int n;

#ifdef HAVE_REALTIME_SIGNALS
    sigset_t signalset;
    sigemptyset(&signalset);
    sigaddset(&signalset, SIGRTMIN+3);
    sigaddset(&signalset, SIGIO);
    sigprocmask(SIG_BLOCK, &signalset, NULL);
#endif

    signal(SIGPIPE, SIG_IGN);

    log = (AccessLog *) arg;
    while (1) {
        b = (Buffer *) queue_remove_wait(&log->entries);
        if (log->doclose) {
            close(log->fd);
            return NULL;
        }

        if (log->currentsize + b->size > Config.rotatesize)
            rotate_access_log(log);

        if (log->fd > 0) {
            n = write_buffer(log->fd, b->buf, b->size);
            log->currentsize += n;
        }
        pthread_mutex_lock(&log->poollock);
        buffer_unref(b);
        log->queuesize -= 1;
        if (log->queuesize == MAX_LOG_ENTRIES-1) {
            pthread_cond_signal(&log->sizenotify);
        }
        pthread_mutex_unlock(&log->poollock);
    }
    close(log->fd);
    return NULL;
}


/*** Errorlog functions ***/

/* Print the current date to the given buffer */
static void printDate(char *buffer, int len)
{
    struct tm thedate;
    time_t curr;
 
    curr = get_current_time();
    do_localtime(curr, &thedate); 
    sprintf(buffer, "[%s %s %.2d %.2d:%.2d:%.2d %.4d] ",
            daysofweek[thedate.tm_wday], months[thedate.tm_mon],
            thedate.tm_mday, thedate.tm_hour, thedate.tm_min,
            thedate.tm_sec, thedate.tm_year + 1900);
}

/* Open the errorlog at the given path */
void errorlog_init(ErrorLog *log, const char *path)
{
    set_current_time();
    memset(log, 0, sizeof(ErrorLog));
    log->path = path;

    log->fd = open(log->path, O_WRONLY | O_APPEND | O_CREAT, 0644);
    if (log->fd <= 0) {
        fprintf(stderr, "vicompress ERROR: Cannot open errorlog file %s, %s\n",
                log->path, strerror(errno));
        exit(1);
    }
    log->currentsize = getfilesize(log->path);
}

/* Close the errorlog */
void errorlog_close(ErrorLog *log)
{ 
    close(log->fd);
}

static int get_error_size(int msglen)
{
    int total = 0;
    total += 28;   /* date/time */
    total += 16;   /* IP address */
    total += 7;    /* port */
    total += msglen;
    return total; 
}

/* Write the given message to the errorlog, in the format:
 *  [Date] [clientip:port] <message>
 *  [Tue May 10 17:18:41 2005] [127.0.0.1:52689] New client accepted.
 * If debug == 1, only print the message if debugging is enabled.
 */
void errorlog_print(ErrorLog *log, sockaddr_all *client,
                    const char *msg, int msglen)
{
    char buffer[64];
    int len;

    if (log == NULL || msg == NULL)
        return;

    if (log->currentsize + get_error_size(msglen) + 5 > Config.rotatesize)
        rotate_error_log(log);

    if (log->fd == -1)
        return; 

    memset(buffer, 0, sizeof(buffer));
    printDate(buffer, 64);
    len = strlen(buffer);
    if (client) {
        snprintf(&buffer[len], 64-len, "[%s:%d] ",
                 ip_to_string(client), ntohs(client->ipv4.sin_port));
    }
    len += strlen(&buffer[len]);

    write_buffer(log->fd, buffer, len);
    write_buffer(log->fd, (char *)msg, msglen); 
    write_buffer(log->fd, "\n", 1);
    log->currentsize += len + msglen + 1;
}

/* Print the given message to the errorlog. 
 * fmt is the printf-like format string to use.
 */
void errorlog_printf(ErrorLog *log, sockaddr_all *client, const char *fmt, ...)
{ 
    char msg[256];
    int msglen;
    va_list args;

    if (log == NULL || log->fd == -1 || fmt == NULL)
        return;

    memset(msg, 0, sizeof(msg));

    va_start(args, fmt);
    vsnprintf(msg, 256, fmt, args);
    va_end(args);

    msglen = strlen(msg);
    errorlog_print(log, client, msg, msglen);
}


