/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**
  The logtest does the following:

  1. Create two sample HTTP requests and replies.  This includes the
     client and server IP address, the reply size before and after
     gzip compression, and whether the reply was gzipped or cached or both.

  2. Write the two entries to the accesslog in apache format. Read back
     the entires from the accesslog file and compare them with the expected 
     Apache format output.  Because the date/time will be different, we 
     replace the date/time with XXXXXXX before doing the comparison.

  3. Write the two entries to the accesslog in squid format. Read back
     the entires from the accesslog file and compare them with the expected 
     Squid format output.  Because the date/time will be different, we 
     replace the date/time with XXXXXXX before doing the comparison.

  4. Write some entries to the errorlog.  Read back the entries from the
     errorlog file and compare them with the expected output.  Because the
     date/time will be different, we replace the date/time with XXXXX before
     doing the comparison.

  5. Test accesslog rotation. We will write the same line to the access log
     in apache format.  Set the rotatesize to 10 * the length of the line.
     Then write 25 entries.  Check that the files accesslog, accesslog.1,
     accesslog.2, are created.  Check the file size of each accesslog.

  6. Test errorlog rotation. We will write the same line to the errorlog.
     Set the rotatesize to 10 * the length of the line. Then write 25 entries.
     Check that the files errorlog, errorlog.1, errorlog.2 are created.
     Check the file size of each errorlog.

**/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/stat.h>
#include "log.h"
#include "http.h"
#include "conf.h"
#include "win32.h"

#ifdef WIN32
#define LINEEND "\r\n"
#else
#define LINEEND "\n"
#endif

#define max(a,b) ((a) > (b) ? (a) : (b))

static time_t currtime;
void set_current_time() { currtime = time(NULL); }
time_t get_current_time() { return currtime; }


/* Read a line from a file at the given line number (which) */
static void readline(const char *file, char *buf, int size, int which)
{
    FILE *fp;
 
    fp = fopen(file, "r");
    assert(fp != NULL);
    while (which >= 0) {
        memset(buf, 0, size);
        fgets(buf, size, fp);
        which--;
    }
    fclose(fp);
}

/* Given an apache log line, replace the date/time with XXXX.. */
static void erase_apache_date(char *buf) {
    int i;
    int len;

    len = strlen(buf);
    i = 0;
    while (buf[i] != '[' && i < len)
        i++;

    i++;
    while (buf[i] != ']' && i < len) {
        buf[i] = 'X'; i++; 
    }
}

/* Given a squid log line, replace the timestamp with XXXX... */
static void erase_squid_date(char *buf) {
    int i;
    for (i = 0; i < 14; i++)
        buf[i] = 'X';
}

/* Given a vicompress errorlog line, replace the date/time with XXX... */
static void erase_errorlog_date(char *buf) {
    int i;
    for (i = 0; i < 26; i++)
        buf[i] = 'X';
}

static void do_accesslog_write(BufferPool *bp, AccessLog *log,
                               const char *request, const char *reply,
                               const char *clientip, const char *serverip,
                               int replysize, int pregzip_size, int is_hit, 
                               int isrequesterr, int duration)
{ 
    Buffer *reqb, *repb;
    HttpRequest req;
    HttpReply rep;
    Parsedata parse;
    sockaddr_all client, server;
    int ret;

    reqb = buffer_alloc(bp);
    repb = buffer_alloc(bp);

    memset(&req, 0, sizeof(req));
    memset(&rep, 0, sizeof(rep));
    http_init_request(&req, reqb, &parse);
    buffer_add(reqb, request, strlen(request));
    ret = http_parse_request(&req, &parse);
    assert(ret == PARSE_DONE);

    http_init_reply(&rep, repb, &parse);
    buffer_add(repb, reply, strlen(reply));
    ret = http_parse_reply(&rep, &parse);
    assert(ret == PARSE_DONE);

    client.ipv4.sin_family = AF_INET;
    server.ipv4.sin_family = AF_INET;
    client.ipv4.sin_addr.s_addr = inet_addr(clientip);
    server.ipv4.sin_addr.s_addr = inet_addr(serverip);

    accesslog_write(log, &req, &rep, &client, &server,
                    replysize, pregzip_size, is_hit, 
                    isrequesterr, duration);

    http_free_request(&req);
    http_free_reply(&rep);
}


void test_accesslog()
{
    AccessLog *log;
    char fileline[512];
    char expectedline[512];

    int i;
    const char *filename = "apachelog";
    BufferPool bp;
    const char *requests[] = {

"GET /path/anotherpath/index.html HTTP/1.0\r\n\
Referer: http://www.yahoo.com/index.html\r\n\
User-Agent: Mozilla 4.0 (\"IE 6.0\" Compatible)\r\n\
Host: www.amazon.com\r\n\
Connection: close\r\n\r\n",

"POST /cgi-bin/account.pl?name=value,bankacct=1239234 HTTP/1.1\r\n\
User-Agent: Opera 5.0 (IE 6.0 compatible)\r\n\
Host: www.amazon.com\r\n\
Content-length: 24\r\n\
Authorization: Basic ZnJlZDpteXBhc3M=\r\n\
Connection: close\r\n\r\n"

    };

    const char *replies[] = {

"HTTP/1.0 200 OK\r\n\
Content-length: 1592\r\n\
Content-type: text/html\r\n\
Connection: close\r\n\r\n",

"HTTP/1.0 200 OK\r\n\
Connection: close\r\n\r\n"

    };

    const char* clientips[] = {"15.13.130.10", "67.123.45.9"};
    const char* serverips[] = {"", "15.40.248.51" };

    const int is_hit[] = { 1, 0 };
    const int replysize[] = { 1592, 94234 };
    const int pregzip_size[] = { 5011, 94234 };
    const int duration[] = { 2717, 23562 };

    const char *apachelines[] = { 
        "15.13.130.10 hit5011 - [21/Aug/2003:17:26:45 -0700] \"GET /path/anotherpath/index.html HTTP/1.0\" 200 1592 \"http://www.yahoo.com/index.html\" \"Mozilla 4.0 (\\\"IE 6.0\\\" Compatible)\"" LINEEND,

        "67.123.45.9 miss94234 fred [21/Aug/2003:17:28:45 -0700] \"POST /cgi-bin/account.pl?name=value,bankacct=1239234 HTTP/1.1\" 200 94234 \"-\" \"Opera 5.0 (IE 6.0 compatible)\"" LINEEND
    };

    const char *squidlines[] = {
       "1118250346.591   2717 15.13.130.10 TCP_HIT/200 1592 GET http://www.amazon.com/path/anotherpath/index.html 5011 NONE/- text/html" LINEEND,

       "1118250552.243  23562 67.123.45.9 TCP_MISS/200 94234 POST http://www.amazon.com/cgi-bin/account.pl?name=value,bankacct=1239234 94234 DIRECT/15.40.248.51 -" LINEEND
    };

    unlink(filename);

    log = (AccessLog *) malloc(sizeof(AccessLog));
    Config.logformat = LOG_FORMAT_APACHE;
    Config.rotatesize = 100 * 1024 * 1024;
    accesslog_init(log, filename);
    bufferpool_init(&bp, 1024, 20);

    for (i = 0; i < 2; i++) {
        do_accesslog_write(&bp, log, requests[i], replies[i], 
                           clientips[i], serverips[i],
                           replysize[i], pregzip_size[i], is_hit[i],
                           0, duration[i]);
    }
    (void)sleep(2);
    accesslog_close(log);

    for (i =0; i < 2; i++) {
        readline(filename, fileline, sizeof(fileline), i);
        strcpy(expectedline, apachelines[i]);
        erase_apache_date(fileline);
        erase_apache_date(expectedline);
        /* printf("fileline: %s\n", fileline);
        printf("exp line: %s\n", expectedline);  */
        assert(strcmp(fileline, expectedline) == 0);
    }
    unlink(filename);
    free(log);

    filename = "squidlog";
    log = (AccessLog *) malloc(sizeof(AccessLog));
    Config.logformat = LOG_FORMAT_SQUID;
    Config.rotatesize = 100 * 1024 * 1024;
    accesslog_init(log, filename);

    for (i = 0; i < 2; i++) {
        do_accesslog_write(&bp, log, requests[i], replies[i], 
                           clientips[i], serverips[i],
                           replysize[i], pregzip_size[i], is_hit[i], 
                           0, duration[i]);
    }
    (void)sleep(2);
    accesslog_close(log);

    for (i =0; i < 2; i++) {
        readline(filename, fileline, sizeof(fileline), i);
        strcpy(expectedline, squidlines[i]);
        erase_squid_date(fileline);
        erase_squid_date(expectedline);
        /* 
        printf("\n");
        printf("fileline: %s\n", fileline);
        printf("exp line: %s\n", expectedline); 
        */
        assert(strcmp(fileline, expectedline) == 0);
    }
    unlink(filename);

}


void test_errorlog()
{
    ErrorLog *log;
    sockaddr_all addr;
    const char *filename = "errorlog";
    int i;
    char fileline[512];

    const char *lines[] = {
        "[Sun Jun 05 13:08:51 2005] [192.168.0.1:52689] New client accepted." LINEEND,
        "[Sun Jun 05 13:08:51 2005] [192.168.0.1:52689] Read http reply from server 192.168.0.101:8080: status=Bad HTTP Reply from server" LINEEND
    };
    const char *msgs[] = {
        "New client accepted.",
        "Read http reply from server 192.168.0.101:8080: status=Bad HTTP Reply from server"
    };

    unlink(filename); 

    log = (ErrorLog *) malloc(sizeof(ErrorLog));
    Config.rotatesize = 10 * 1024 * 1024;
    errorlog_init(log, filename);

    for (i = 0; i < 2; i++) {
        addr.ipv4.sin_family = AF_INET;
        addr.ipv4.sin_addr.s_addr = inet_addr("192.168.0.1");
        addr.ipv4.sin_port = htons(52689);
        errorlog_print(log, &addr, msgs[i], strlen(msgs[i]));
    }

    errorlog_close(log);

    for (i = 0; i < 2; i++) {
        char buf[512];

        readline(filename, fileline, sizeof(fileline), i);
        strcpy(buf, lines[i]);
        erase_errorlog_date(fileline);
        erase_errorlog_date(buf);
        /*
        printf("\n"); 
        printf("fileline: %s\n", fileline);
        printf("exp line: %s\n", buf); 
        */
        assert(strcmp(fileline, buf) == 0);
    }
    unlink(filename);
}


void test_accesslogrotation()
{
    AccessLog *log;
    char path[64];
    struct stat info;
    int i, n;
    const char *filename = "apachelog";
    BufferPool bp;

    const char *request = 
"GET /path/anotherpath/index.html HTTP/1.0\r\n\
Referer: http://www.yahoo.com/index.html\r\n\
User-Agent: Mozilla 4.0 (IE 6.0 Compatible)\r\n\
Host: www.amazon.com\r\n\
Connection: close\r\n\r\n";

    const char *reply = 
"HTTP/1.0 200 OK\r\n\
Content-length: 1592\r\n\
Content-type: text/html\r\n\
Connection: close\r\n\r\n";

    const char* clientip = "15.13.130.10";
    int replysize = 1592;

    const char *line =  
         "15.13.130.10 miss1592 - [21/Aug/2003:17:26:45 -0700] \"GET /path/anotherpath/index.html HTTP/1.0\" 200 1592 \"http://www.yahoo.com/index.html\" \"Mozilla 4.0 (IE 6.0 Compatible)\"" LINEEND;

    unlink(filename);
    for (i = 1; i <= 4; i++) {
        sprintf(path, "%s.%d", filename, i);
        unlink(path);
    }

    log = (AccessLog *) malloc(sizeof(AccessLog));
    Config.logformat = LOG_FORMAT_APACHE;
    Config.rotatesize = strlen(line) * 10;
    accesslog_init(log, filename);
    bufferpool_init(&bp, 1024, 20);

    for (i = 0; i < 25; i++) {
        do_accesslog_write(&bp, log, request, reply, 
                           clientip, "",
                           replysize, replysize, 0, 0, 0);
    }
    (void)sleep(5);
    accesslog_close(log);

    sprintf(path, "%s.2", filename);
    n = stat(path, &info);
    assert(n == 0);
    assert(info.st_size == strlen(line) * 10);
    unlink(path);

    sprintf(path, "%s.1", filename);
    n = stat(path, &info);
    assert(n == 0);
    assert(info.st_size == strlen(line) * 10);
    unlink(path);

    n = stat(filename, &info);
    assert(n == 0);
    assert(info.st_size == strlen(line) * 5);

    unlink(filename); 
}

void test_errorlogrotation()
{
    ErrorLog *log;
    int i, n;
    char buf[512];
    const char *filename = "errorlog";
    struct stat info;
    char path[64];
    sockaddr_all addr;
    const char *line = "[Tue May 10 17:18:41 2005] [127.0.0.1:5000] This is message %2d" LINEEND;
    const char *msg = "This is message %2d";

    unlink(filename);

    log = (ErrorLog *) malloc(sizeof(ErrorLog));
    Config.rotatesize = strlen(line)*10 + 5;
    Config.enable_debug = 1;
    errorlog_init(log, filename);

    addr.ipv4.sin_family = AF_INET;
    addr.ipv4.sin_addr.s_addr = inet_addr("127.0.0.1");
    addr.ipv4.sin_port = htons(5000);

    for (i = 0; i < 25; i++) {
        sprintf(buf, msg, i);
        errorlog_print(log, &addr, buf, strlen(buf));
    }
    errorlog_close(log);

    sprintf(path, "%s.2", filename);
    n = stat(path, &info);
    assert(n == 0);
    assert(info.st_size == (strlen(line)-1) * 10);
    unlink(path);

    sprintf(path, "%s.1", filename);
    n = stat(path, &info);
    assert(n == 0);
    assert(info.st_size == (strlen(line)-1) * 10);
    unlink(path);

    n = stat(filename, &info);
    assert(n == 0);
    assert(info.st_size < strlen(line) * 5 + 5);

    unlink(filename); 
}


int main()
{
    printf("logtest...........");
    currtime = time(NULL);
    test_accesslog();
    test_errorlog();
    test_accesslogrotation();
    test_errorlogrotation();
    printf("Pass\n");
    return 0;
}

