/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * A pool is used to store memory for objects of the same size.
 * One can malloc and free objects from the pool faster than the
 * standard malloc() and free().
 */


#include <stdlib.h>
#include <assert.h>
#include "pool.h"


/*
 * Initialize a pool, with the given objectsize and maximum number of
 * items to store.
 *
 * To avoid wasting memory, set the maxcapacity to less than the
 * actual maximum you need.  For example, if you can have at most 4000
 * client http requests at a time, don't set the maxcapacity to 4000.
 * If the number of clients spikes up to 4000, but is on average 1000,
 * then you're wasting 3000 * 1kb = 3 MB of memory.  If maxcapacity
 * is too small and the pool becomes empty, it will just call malloc(),
 * so there's no harm done.
 */
void pool_init(Pool *p, int objectsize, long maxcapacity)
{
    int i;
    assert(objectsize > 0);
    assert(maxcapacity >= 0);
    p->objectsize = objectsize;
    p->maxitems = maxcapacity;
    p->freeitems = (void**)malloc(sizeof(void *) * maxcapacity);
    p->nfreeitems = 10;
    p->numalloc = p->nfreeitems;
    for (i = 0; i < p->nfreeitems; i++)
        p->freeitems[i] = malloc(objectsize);
}

/* Allocate an object from the pool. */
void* pool_alloc(Pool *p)
{
    void *item;
    if (p->nfreeitems == 0) {
        p->numalloc++;
        item = malloc(p->objectsize);
        assert(item != NULL);
        return item;
    }
    else {
        item = p->freeitems[p->nfreeitems-1];
        p->nfreeitems -= 1;
        return item;
    }
}

/* Free an object back into the pool */
void pool_free(Pool *p, void *item)
{
    if (p->nfreeitems == p->maxitems) {
        p->numalloc--;
        free(item); 
    }
    else {
        p->freeitems[p->nfreeitems] = item;
        p->nfreeitems++;
    }
}

/* Not used.
void pool_destroy(Pool *p)
{
    int i;
    for (i = 0; i < p->nfreeitems; i++) {
        free(p->freeitems[i]);
    }
    free(p->freeitems);
}
*/

/*
 * Return true if this pool is empty,
 * and the maximum number of items has
 * already been allocated.
 */
int pool_is_empty(Pool *p)
{
    if (p->numalloc >= p->maxitems && p->nfreeitems == 0)
        return 1;
    else
        return 0;
}

