/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/* An array-based, thread-safe queue */

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "queue.h"
#include "win32.h"

/* Intitialize the queue to the given capacity */
void queue_init(Queue *q, int size)
{
    assert(size > 0);
    q->front = q->back = 0;
    q->space = size;
    q->len = 0;
    q->buf = (void**) malloc(sizeof(void *) * size);
    memset(q->buf, 0, sizeof(void *) * size);
    pthread_mutex_init(&q->lock, NULL);
    pthread_cond_init(&q->notify, NULL);
}

/* Add the item to the queue. */
void queue_add(Queue *q, void *data)
{
    pthread_mutex_lock(&q->lock);
    queue_add_fast(q, data);
    if (q->len == 1)
        pthread_cond_signal(&q->notify);

    pthread_mutex_unlock(&q->lock);
}

/* Add the item to the queue, without thread locking.  */
void queue_add_fast(Queue *q, void *data)
{
    int i, j, newsize;
    void **newbuf;

    if ((q->back + 1) % q->space == q->front) {
        newsize = q->space * 2;
        newbuf = (void**) malloc(sizeof(void *) * newsize);
        j = 0;
        for (i = q->front; i != q->back; i = (i + 1) % q->space) {
            newbuf[j] = q->buf[i];
            j++;
        }
        free(q->buf);
        q->buf = newbuf;
        q->front = 0;
        q->back = q->space-1;
        q->space = newsize;
    }

    q->buf[q->back] = data;
    q->back = (q->back + 1) % q->space;
    q->len++;
}

/* Remove an item from the queue. If its empty,
 * this function will block until an item is
 * added to the queue.
 */

void* queue_remove_wait(Queue *q)
{
    void *data;
    pthread_mutex_lock(&q->lock);
    while (q->front == q->back) {
        pthread_cond_wait(&q->notify, &q->lock);
    }
    data = q->buf[q->front];
    q->front = (q->front + 1) % q->space;
    q->len--;
    pthread_mutex_unlock(&q->lock);
    return data;
}

/*
    Remove many items from the queue, and store
    them in the items array. nitems is the size of the
    arary. Return the number of items removed.
*/
int queue_remove_many(Queue *q, void** items, int nitems)
{
    int i = 0;

    pthread_mutex_lock(&q->lock);
    if (q->front == q->back) {
        pthread_mutex_unlock(&q->lock);
        return 0;
    }
    if (q->front == 0) {
        int ncopy = q->back;
        if (ncopy > nitems)
            ncopy = nitems;
        memcpy(items, q->buf, ncopy * sizeof(void*));
        q->len -= ncopy;
        if (q->len > 0)
            q->front = q->back - q->len;
        else
            q->front = q->back = 0;
        pthread_mutex_unlock(&q->lock);
        return ncopy;
    }
    while (i < nitems && (q->front != q->back)) {    
        items[i] = q->buf[q->front];
        q->front = (q->front + 1) % q->space;
        q->len--;
        i++;
    }
    if (q->front == q->back) {
        q->front = q->back = 0;
    }
    pthread_mutex_unlock(&q->lock);
    return i;
}

/* Return true if the queue is empty, else false */
int queue_is_empty(Queue *q)
{
    int result;
    result = q->len == 0;
    return result;
}

/* Return the number of items in the queue. */
int queue_length(Queue *q)
{
    return q->len;
}

