/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/*
 * The sockets structure is used to manage non-blocking connect/read/write
 * socket calls.  The socket_ functions take a callback function to execute
 * when the socket call is actually completed.  The data associated with
 * the connect/read/write call is stored in a SocketData structure.  When
 * the call completes (either success or fail), the socket is added to
 * the outgoing queue sockets->outq.  When sockets_checkevents() is invoked,
 * the callback functions for those completed calls in the outq are
 * executed.
 */

#include <string.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/poll.h>
#include <arpa/inet.h>
#include "sockets.h"
#include "conf.h"
#include "currenttime.h"
#include "win32.h"

/* Event types */
#define READ_EVENT    1
#define WRITE_EVENT   2
#define CONNECT_EVENT 3
#define ACCEPT_EVENT  4

/* Timeouts for read and connect */
#define READ_TIMEOUT     40
#define CONNECT_TIMEOUT  80
#define BALANCER_CONNECT_TIMEOUT 10
#define WRITE_TIMEOUT    12       /* Timeout for 1024 byte write */
#define WRITE_TIMEOUT_SIZE  1024 

/* read/write/connect status */
#define STATUS_DONE       1
#define STATUS_MORE       2
#define STATUS_WAIT       3

/* Socket type */
#define CLIENT_SOCKET   1
#define SERVER_SOCKET   2

static void do_connect(Sockets *s, int fd);
static void do_read(Sockets *s, int fd);
static void do_write(Sockets *s, int fd);
static void do_accept(Sockets *s, int fd); 

static void check_timeouts(Sockets *s);


/* Initialize the Sockets structure.  
 *
 * The main structure is the SocketData array, which is indexed 
 * by the socket file descriptor. It contains the buffer to 
 * read/write and the callback functions to invoke.
 *
 * Initialize the out queue (outq). After the read/write/connect 
 * is performed, we don't want to invoke the callback function
 * immediately.  So we queue the sockets, and then invoke the
 * callbacks later, when sockets_checkevents() is called.
 *
 * Initialize the DNS caching if enabled.
 */
void sockets_init(Sockets *s)
{
    memset(s, 0, sizeof(Sockets));
    s->data = 
      (SocketData *) calloc(Config.max_total_fd + 10, sizeof(SocketData)); 
    s->numclients = s->numservers = 0; 
    s->current_time = time(NULL);
    s->lasttimeout = s->current_time;
    s->epollfd = epoll_create(Config.max_total_fd + 10);
    s->epollevents = (struct epoll_event *) 
        calloc(Config.max_total_fd + 10, sizeof(struct epoll_event));
    queue_init(&s->outq, Config.max_total_fd);
    s->maxclients = (Config.max_total_fd - (Config.num_listen + Config.num_webservers + 5)) / 2;

    if (Config.num_webservers == 0) {
        int maxitems = 10000;
        if (!Config.enable_dns_caching)
            maxitems = 0;
        s->dns = (DnsCache *) malloc(sizeof(DnsCache));
        dns_init(s->dns, maxitems, Config.dns_expires);
    }
}


static int ignore_errno(int error)
{
    switch (error) {
    case EINPROGRESS:
#ifdef WIN32
    case EWOULDBLOCK:
#endif
    case EAGAIN:
    case EALREADY:
    case EINTR:
        return 1;
    default:
        return 0;
    }
}

/* The two functions below keep track of the number of
 * client and server sockets allocated.  These numbers
 * currently aren't used anywhere, they were initially
 * used for debugging during development.
 */
static void socket_added(Sockets *s, int fd, int type)
{
    if (type == CLIENT_SOCKET) {
        s->numclients += 1;
        s->data[fd].isclient = 1;
    }
    else if (type == SERVER_SOCKET) {
        s->numservers += 1;
        s->data[fd].isclient = 0;
    }
}

static void socket_removed(Sockets *s, int fd)
{
    if (s->data[fd].isclient) {
        s->numclients -= 1;
        s->data[fd].isclient = 0;
    }
    else {
        s->numservers -= 1;
    }
}

static void epoll_set(Sockets *s, int fd, int op, int events)
{
    struct epoll_event event;
    int n;

    event.events = 0;
    if (events != 0) {
        event.events = events | EPOLLHUP | EPOLLERR;
    }
    event.data.fd = fd;
    n = epoll_ctl(s->epollfd, op, fd, &event);
    assert(n == 0);
}

/* Clear the callback data and poll data for the
 * given socket.
 */
static void clear_event(Sockets *s, int fd, int type)
{
    SocketData *d;
    int opened; 
    char isclient;

    d = &s->data[fd];
    opened = d->opened;
    isclient = d->isclient;
    memset(d, 0, sizeof(SocketData));
    d->opened = opened;
    d->isclient = isclient;
    epoll_set(s, fd, EPOLL_CTL_MOD, 0);
}

static void set_nonblock(Sockets *s, int fd) {
    fcntl(fd, F_SETFL, O_NONBLOCK);
    s->data[fd].opened = 1;

    epoll_set(s, fd, EPOLL_CTL_ADD, EPOLLIN);
    if (s->maxfd < fd)
        s->maxfd = fd;
}

/* Create a new, non-blocking TCP socket */
int sockets_new(Sockets *s, int family)
{
    int fd;

    fd = socket(family, SOCK_STREAM, 0);
    if (fd == -1) {
        return -1;
    }
    set_nonblock(s, fd);
    return fd;
}


/* Wait for a read event on the given socket file descriptor. When 
 * the read event occurs, read as much data as possible into the
 * given buffer.  If any data is read (or a timeout occurs), 
 * invoke the callback function with the given callback data.
 */
void sockets_read(Sockets *s, int fd, Buffer *buffer,
                  ReadCallback *callback, void *data)
{
    sockets_read_timeout(s, fd, buffer, READ_TIMEOUT, callback, data);
}

/* Same as sockets_read, but use a different timeout value 
 * (in seconds) instead of the default (40 seconds).  This is used
 * to reduce the timeout period when reading multiple requests
 * from a persistent connection.
 */
void sockets_read_timeout(Sockets *s, int fd, Buffer *buffer, int timeout,
                          ReadCallback *callback, void *data)
{
    SocketData *d;

    assert(fd > 2 && fd <= Config.max_total_fd);
    d = &s->data[fd];
    assert(d->opened == 1);
    assert(d->eventtype == 0);
    assert(d->data == NULL);
    assert(d->status == 0);
    assert(callback != NULL);
    assert(d->buffer == NULL);
    d->eventtype = READ_EVENT;
    d->readcallback = callback;
    d->data = data;
    d->buffer = buffer;
    d->timeout = s->current_time + timeout;
    do_read(s, fd);
}

/* Read data from the fd. Save the error of eof values in
 * the SocketData. Set the status to
 * DONE - if data was read or an error occured
 * MORE - if no data was read
 */
static void do_read(Sockets *s, int fd)
{
    SocketData *d = &s->data[fd];
    Buffer *buffer = d->buffer;
    int initial_size, len, amount;

    initial_size = buffer->size;
    amount = buffer->capacity - buffer->size;
    assert(amount > 0);
    errno = 0;
    do {
        len = recv(fd, &buffer->buf[buffer->size], amount, 0);
        if (len > 0) {
            buffer->size += len;
            amount -= len;
        }
    }
    while (len > 0 && buffer->size < buffer->capacity);
    d->error = errno;
    d->eof = (len == 0);

    if (ignore_errno(d->error))
        d->error = 0;

    if (buffer->size == initial_size && !d->error && !d->eof) {
        d->status = STATUS_MORE;
        epoll_set(s, fd, EPOLL_CTL_MOD, EPOLLIN);
    }
    else {
        d->status = STATUS_DONE;
    }

    if (buffer->size < buffer->capacity) {
        buffer->buf[buffer->size] = 0;
    }

    queue_add_fast(&s->outq, (void *) fd);
}

static void do_read_callback(Sockets *s, int fd)
{
    SocketData *d = &s->data[fd];
    ReadCallback *callback = d->readcallback;
    Buffer *buffer = d->buffer;
    void *data = d->data;
    int error = d->error;
    int eof = d->eof;
    assert(d->status == STATUS_DONE); d=NULL;
    clear_event(s, fd, READ_EVENT);
    callback(fd, data, buffer, error, eof); 
}

/* Wait for a connection on the given socket. The IP/port to connect to
 * is specified by addr. Invoke the callback function with the given 
 * callback data once the socket connects to the destination, or an
 * error/timeout occurs.
 */
void sockets_connect(Sockets *s, int fd, sockaddr_all *addr, 
                     ConnectCallback *callback, void *data)
{
    SocketData *d;
    int ret;
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
    int socklen;
#else
    socklen_t socklen;
#endif
    assert(fd > 2 && fd <= Config.max_total_fd);
    socket_added(s, fd, SERVER_SOCKET);
    d = &s->data[fd];
    assert(d->opened == 1);
    assert(d->data == NULL);
    assert(d->eventtype == 0);
    assert(d->status == 0);
    d->eventtype = CONNECT_EVENT;
    d->connectcallback = callback;
    d->data = data;
    d->connectaddr = *addr;
    if (Config.num_webservers > 1)
        d->timeout = get_current_time() + BALANCER_CONNECT_TIMEOUT;
    else
        d->timeout = get_current_time() + CONNECT_TIMEOUT;

    socklen = sizeof(struct sockaddr_in);
#ifdef HAVE_IPV6
    if (d->connectaddr.ipv4.sin_family == AF_INET6) {
        socklen = sizeof(struct sockaddr_in6);
    }
#endif
    do {
        errno = 0;
        ret = connect(fd, (struct sockaddr *)&(d->connectaddr.ipv4), socklen);
    } while (ret == -1 && errno == EINTR);

    if (ret == 0 || (ret == -1 && errno == EISCONN))
        d->error = 0;
    else
        d->error = errno;

    if (d->error != 0 && ignore_errno(d->error)) {
        d->status = STATUS_MORE;
        epoll_set(s, fd, EPOLL_CTL_MOD, EPOLLOUT);
    }
    else {
        d->status = STATUS_DONE;
    }

    queue_add_fast(&s->outq, (void *)fd);
}

/* Check the status of the connect() call for the given fd,
 * by calling getsockopt().  Set the status to
 * DONE - if connect succeeded or failed
 * MORE - if still in progress.
 */
static void do_connect(Sockets *s, int fd)
{
    SocketData *d;
    int ret, error;
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
    int len;
#else
    socklen_t len;
#endif
    errno = 0;
    d = &s->data[fd];
    error = 0;
    len = sizeof(int);
    do {
        ret = getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len);
    } while (ret == -1 && errno == EINTR);

    if (ret == -1)
        d->error = EINVAL;
    else
        d->error = error;

    if (d->error != 0 && ignore_errno(d->error)) {
        d->status = STATUS_MORE;
        epoll_set(s, fd, EPOLL_CTL_MOD, EPOLLOUT);
    }
    else {
        d->status = STATUS_DONE;
    }

    queue_add_fast(&s->outq, (void *)fd);
}

static void do_connect_callback(Sockets *s, int fd)
{
    SocketData *d = &s->data[fd];
    ConnectCallback *callback = d->connectcallback;
    int error = d->error;
    void *data = d->data;
    assert(d->status == STATUS_DONE); d = NULL;
    clear_event(s, fd, CONNECT_EVENT);
    callback(fd, data, error); 
}

/* Write the given buffer to the socket fd. Invoke the callback 
 * function with the given callback data when the data has been 
 * completely written, or an error occurs.
 */
void sockets_write(Sockets *s, int fd, Buffer *buffer, 
                   WriteCallback *callback, void *data)
{
    SocketData *d;
 
    assert(fd > 2 && fd <= Config.max_total_fd);
    d = &s->data[fd];
    assert(d->opened == 1);
    assert(d->data == NULL);
    assert(d->eventtype == 0);
    assert(d->status == 0);
    d->eventtype = WRITE_EVENT;
    d->writeoffset = 0;
    d->writecallback = callback;
    d->data = data;
    d->buffer = buffer;
    d->timeout = get_current_time() + 10 + 
                 buffer->size * WRITE_TIMEOUT / WRITE_TIMEOUT_SIZE;
    do_write(s, fd);
}

/* Write the data in the buffer. Save the writeoffset
 * and error. Set the status to
 * DONE - if all data was written, or an error occured
 * MORE - if only some of the data was written
 */
static void do_write(Sockets *s, int fd)
{
    SocketData *d;
    Buffer *buffer;
    int len;

    d = &s->data[fd];
    buffer = d->buffer;

    if (d->writeoffset == buffer->size) {
        d->error = 0;
        d->status = STATUS_DONE;
        queue_add_fast(&s->outq, (void *)fd);
        return;
    }

    d->error = 0;
    do {
        len = send(fd, &(buffer->buf[d->writeoffset]), 
                   buffer->size - d->writeoffset, 0);
        d->error = errno;
        if (len > 0)
            d->writeoffset = d->writeoffset + len;
    }
    while (len > 0 && d->writeoffset < buffer->size);

    if (len == -1 && ignore_errno(d->error)) {
        d->error = 0;
        d->status = STATUS_MORE;
        epoll_set(s, fd, EPOLL_CTL_MOD, EPOLLOUT);
    }
    else {
        d->status = STATUS_DONE; 
    }

    if (len == 0 && d->writeoffset < buffer->size)
        d->error = EPIPE;
    else if (len > 0)
        d->error = 0;

    queue_add_fast(&s->outq, (void *)fd);
}

static void do_write_callback(Sockets *s, int fd)
{
    SocketData *d = &s->data[fd];
    WriteCallback *callback = d->writecallback;
    void *data = d->data;
    int error = d->error;
    Buffer *b = d->buffer;
    assert(d->status == STATUS_DONE); d = NULL;
    clear_event(s, fd, WRITE_EVENT);
    callback(fd, data, b, error);
}
 
/* Wait for client connections on the listening socket, When a 
 * client connects, invoke the callback function with the given 
 * callback data.  The callback function and callback data are kept 
 * until the listening socket is closed.
 */
void sockets_listen(Sockets *s, int fd, AcceptCallback *callback, void *data)
{
    SocketData *d;

    d = &s->data[fd];
    assert(d->opened == 1);
    assert(d->data == NULL);
    assert(d->eventtype == 0);
    d->eventtype = ACCEPT_EVENT;
    d->acceptcallback = callback;
    d->data = data;
    d->timeout = 0;
    assert(s->num_listen < MAX_LISTEN);
    s->listenfd[s->num_listen] = fd;
    s->num_listen++;
}

#if defined(HAVE_REALTIME_SIGNALS) || defined(HAVE_KQUEUE)
/* Check the listening socket for events. */
static void sockets_checklisten(Sockets *s)
{
    struct pollfd pfd[MAX_LISTEN];
    int i;

    if (s->listenfd[0] <= 0) {
        return;
    }
    for (i = 0; i < s->num_listen; i++) {
        pfd[i].fd = s->listenfd[i];
        pfd[i].events = POLLIN | POLLHUP | POLLERR;
        pfd[i].revents = 0;
    }
    if (poll(pfd, s->num_listen, 0) > 0) {
        for (i = 0; i < s->num_listen; i++) {
            if (pfd[i].revents != 0) {
                do_accept(s, pfd[i].fd);
            }
        }
    }
}
#endif

/* Accept up to 80 connections from the listening socket fd.
 * If successful, invoke the callback function directly.
 */
static void do_accept(Sockets *s, int fd)
{
    int clientfd;
    sockaddr_all addr;
    int count, maxfds, maxaccept, err;
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
    int len;
#else
    socklen_t len;
#endif
    AcceptCallback *callback;
    void *data;

    count = 0;
    maxaccept = s->maxclients - s->numclients;
    if (maxaccept > 80)
        maxaccept = 80;
    /* Don't accept connections if no more file descriptors */
    if (maxaccept <= 0) {
        return;
    }

    while (count < maxaccept) {
        len = sizeof(struct sockaddr_in);
#ifdef HAVE_IPV6
        len = sizeof(struct sockaddr_in6);
#endif
        clientfd = accept(fd, (struct sockaddr *)&(addr.ipv4), &len);
        err = errno;
        count++;
        if (clientfd == -1 && err == EINTR)
            continue;
        else if (clientfd == -1)
            return;

        set_nonblock(s, clientfd);
        socket_added(s, clientfd, CLIENT_SOCKET);
        assert(clientfd <= Config.max_total_fd);
        callback = s->data[fd].acceptcallback;
        data = s->data[fd].data;
        callback(clientfd, addr, data); 
        count++;
    }
}

/* Lookup the ip address of the given hostname.  When the
 * lookup completes, invoke the given callback function
 * with the given callback data.
 */
void sockets_dnslookup(Sockets *s, Substring host,
                       DnsCallback *callback, void *data) {
    DnsCache *dns;
    DnsRequest *req;

    dns = s->dns;
    req = dns_alloc_request(dns);
    memset(req, 0, sizeof(DnsRequest));
    if (host.size >= MAXHOSTNAMELEN)
        host.size = MAXHOSTNAMELEN-1;

    memcpy(req->name, host.buf, host.size);
    req->name[host.size] = 0;
    req->namelen = host.size;
    req->callback = callback;
    req->data = data;

    dns_lookup(dns, req);
}


/* Close the given socket and remove associated data
 * from the socket events structure.
 */
void sockets_close(Sockets *s, int fd)
{
    SocketData *d;
    char buffer[1024]; 
    int n;

    d = &(s->data[fd]);
    epoll_set(s, fd, EPOLL_CTL_DEL, 0);
    shutdown(fd, 1); 
    socket_removed(s, fd);
    assert(d->opened == 1);
    assert(d->eventtype == 0);
    assert(d->data == NULL);
    memset(d, 0, sizeof(SocketData));
    if (s->maxfd == fd) {
        while (s->maxfd > 0 && s->data[s->maxfd].opened == 0) {
            s->maxfd--;
        }
    }

    /* Read any excess data before closing */
    n = recv(fd, buffer, sizeof(buffer), 0);
#ifdef WIN32
    win_close(fd);
#else
    close(fd);
#endif
}


/* Retrieve the results of the read/write/connect calls.
 * Invoke the callback function if done.  If an event
 * was missed, invoke the corresponding do_read/do_write/
 * do_connect function.
 */
static void check_out_queue(Sockets *s)
{ 
    void* results[2048];
    SocketData *d;
    int i, n, fd, event, status;
    time_t currtime;

    currtime = get_current_time();
    n = queue_remove_many(&s->outq, results, 2048);
    for (i = 0; i < n; i++) {
        fd = (int) results[i];
        d = &s->data[fd];
        event = d->eventtype;
        status = d->status;

        if ((event == READ_EVENT) && (status == STATUS_DONE)) {
            do_read_callback(s, fd);
        }
        else if ((event == READ_EVENT) && (status == STATUS_MORE)) {
            if (d->timeout > 0 && d->timeout < currtime) {
                d->error = ETIMEDOUT;
                d->status = STATUS_DONE;
                do_read_callback(s, fd);
            }
            else if (d->missedevent) {
                d->missedevent = 0;
                do_read(s, fd);
            }
            else {
                d->status = STATUS_WAIT; 
            }
        }
        else if ((event == WRITE_EVENT) && (status == STATUS_DONE)) {
            do_write_callback(s, fd);
        }
        else if ((event == WRITE_EVENT) && (status == STATUS_MORE)) {
            if (d->timeout > 0 && d->timeout < currtime) {
                d->error = ETIMEDOUT;
                d->status = STATUS_DONE;
                do_write_callback(s, fd);
            }
            else if (d->missedevent) {
                d->missedevent = 0;
                do_write(s, fd);
            }
            else {
                d->status = STATUS_WAIT; 
            }
        }
        else if ((event == CONNECT_EVENT) && (status == STATUS_DONE)) {
            do_connect_callback(s, fd);
        }
        else if ((event == CONNECT_EVENT) && (status == STATUS_MORE)) {
            if (d->timeout > 0 && d->timeout < currtime) {
                d->error = ETIMEDOUT;
                d->status = STATUS_DONE;
                do_connect_callback(s, fd);
            }
            else if (d->missedevent) {
                d->missedevent = 0;
                do_connect(s, fd);
            }
            else {
                d->status = STATUS_WAIT; 
            }
        }
    }
}


/* Retrieve the results of the dns lookups.
 * Invoke the callback function.
 */
static void check_dns(Sockets *s) {
    DnsRequest *results[8192];
    DnsCallback *callback;
    DnsRequest *req;
    void *data;
    int i, n, error;
    unsigned char ip[4];

    n = dns_getresults(s->dns, results, 8192);
    for (i = 0; i < n; i++) {
        req = results[i];
        callback = req->callback;
        data = req->data;
        memcpy(ip, req->ip, 4);
        error = req->error;
        dns_free_request(s->dns, req);
        callback(ip, data, error);
    }
}


/* Scan every open socket for timeouts. */
static void check_timeouts(Sockets *s)
{
    int fd;
    int max = s->maxfd;
    time_t currtime;
    SocketData *d;

    currtime = get_current_time();
    for (fd = 3; fd <= max; fd++) {
        d = &s->data[fd];
        if (!d->opened || d->eventtype == 0 || d->status != STATUS_WAIT)
            continue;
        if (d->timeout > 0 && d->timeout < currtime) {
            d->error = ETIMEDOUT;
            d->status = STATUS_DONE;
            if (d->eventtype == CONNECT_EVENT)
                do_connect_callback(s, fd);
            else if (d->eventtype == READ_EVENT)
                do_read_callback(s, fd);
            else if (d->eventtype == WRITE_EVENT)
                do_write_callback(s, fd);
        }
    }
}

/* Check for socket events using epoll(). If an event
 * occurs, start the read/write/connect operation.
 */
static void sockets_checkevents_epoll(Sockets *s, int msec)
{
    int fd, i, n, maxfd;
    SocketData *d;
    int readevent, writeevent;

    s->current_time = time(NULL);
    maxfd = s->maxfd;

    n = epoll_wait(s->epollfd, s->epollevents, Config.max_total_fd, msec);
    if (n < 0) 
        return;
    
    for (i = 0; i < n; i++) {
        readevent = s->epollevents[i].events & (EPOLLIN | EPOLLHUP | EPOLLERR);
        writeevent = s->epollevents[i].events & (EPOLLOUT | EPOLLHUP | EPOLLERR);
        fd = s->epollevents[i].data.fd;

        d = &s->data[fd];
        if (!d->opened || d->eventtype == 0)
            continue;

        if (d->eventtype == ACCEPT_EVENT && readevent) {
            do_accept(s, fd);
        }
        else if (d->eventtype == READ_EVENT && readevent) {
            if (d->status == STATUS_WAIT) {
                do_read(s, fd);
            }
            else {
                d->missedevent = 1;
            }
        } 
        else if (d->eventtype == ACCEPT_EVENT) {
        }  
        else if (d->eventtype == CONNECT_EVENT && writeevent) {
            if (d->status == STATUS_WAIT) {
                do_connect(s, fd);
            }
            else {
                d->missedevent = 1;
            }
        } 
        else if (d->eventtype == WRITE_EVENT && writeevent) {
            if (d->status == STATUS_WAIT) {
                do_write(s, fd);
            }
            else {
                d->missedevent = 1;
            }
        }
        else if (readevent) {
            d->missedevent = 1;
        }
    }
}

/* Check for socket events, waiting the specified number of 
 * milliseconds. 
 *
 * sockets_checkevents_epoll() will check for events and perform 
 * the corresponding read/write/connect.
 *
 * check_out_queue() will invoke the callback functions for 
 * completed events. 
 *
 * check_dns() will check for completed dns lookups (epoll cannot 
 * check dns lookups)
 *
 * check_timeouts() scans all the opened sockets in SocketData to
 * see if any have timed out.  Since this is expensive, it is only
 * checked every 4 seconds.
 *
 * check_listen() checks if new clients can be accepted, and
 * invokes the accept callback function. New client events cannot
 * be checked by an edge-triggered epoll() or by real-time signals,
 * because only one event is reported when multiple clients connect.
 * Vicompress doesn't currently use edge-triggered epoll(), but it
 * may in the future. 
 */
 
void sockets_checkevents(Sockets *s, int msec)
{
    int i = 0;

    s->current_time = time(NULL);

#if defined(HAVE_REALTIME_SIGNALS) || defined(HAVE_KQUEUE)
    sockets_checklisten(s);
#endif

    if (Config.num_webservers == 0) {
        check_dns(s);
    }

    for (i = 0; i < 3; i++) {
        if (s->outq.len != 0) {
            check_out_queue(s);
        }
        else {
            break;
        }
    }

    /* If we have finished socket events to process in the out queue,
     * i.e. the out queue is not empty:
     *
     * - If we sleep when waiting for new socket events (msec > 0),
     *   the response time will increase.
     * - If we don't sleep (msec=0), the CPU usage becomes quite
     *   high, especially under heavy load.
     *
     * As a compromise, we sleep only if the number of clients is
     * greater than 10.
     */
    if (s->outq.len != 0 && s->numclients <= 10) {
        msec = 0;
    }
    sockets_checkevents_epoll(s, msec);

    if (s->current_time - s->lasttimeout >= 4) {
        s->lasttimeout = s->current_time;
            check_timeouts(s); 
    }
}

/* Convert the IP address and port into a string representation. */
char* sockets_to_string(sockaddr_all *addr)
{
    static char str[64]; 
    unsigned char *ip;

    str[0] = 0;
#ifdef HAVE_IPV6
    if (addr->ipv4.sin_family == AF_INET6) {
        char ipstr[50];
        inet_ntop(AF_INET6, &(addr->ipv6.sin6_addr), ipstr, 50);
        sprintf(str, "%s:%u", ipstr, ntohs(addr->ipv6.sin6_port));
        return &str[0];
    }
#endif
    ip = (unsigned char *) &(addr->ipv4.sin_addr.s_addr);
    sprintf(str, "%u.%u.%u.%u:%u", ip[0], ip[1], ip[2], ip[3],
            ntohs(addr->ipv4.sin_port));
    return &str[0];
}

