/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#ifndef _SOCKET_EVENTS_H_
#define _SOCKET_EVENTS_H_

#include <sys/time.h>
#include <netinet/in.h>
#include <unistd.h>
#include "conf.h"

#ifdef HAVE_EPOLL
#include <sys/epoll.h>
#else
#include "epoll.h"
#endif

#include "buffer.h"
#include "queue.h"
#include "dns.h"

/* Callback function for accept()
    clientfd is the new client socket 
    addr is the address of the new client socket.
    data is the callback data passed.
*/
typedef void AcceptCallback(int clientfd, sockaddr_all addr, void *data);

/* Callback function for connect():
    fd is the socket file descriptor.
    data is the callback data passed.
    error is the errno result of the connect() call.
*/
typedef void ConnectCallback(int fd, void *data, int error);

/* Callback function when a "read" occurs. 
    fd is the socket file descriptor.
    data is the callback data passed.
    buffer is the data that was read.
    error is the errno code of the read() call.
    eof=1 if an end of file was encountered while reading.
*/
typedef void ReadCallback(int fd, void *data, Buffer *buffer, int error, int eof);

/* Callback function for a write.
    fd is the socket file descriptor.
    data is the callback data passed.
    buffer is the data that was written.
    error is the errno code of the write() call.
*/
typedef void WriteCallback(int fd, void *data, Buffer *buffer, int error);


/* Data for each socket */
struct _SocketData
{
    char eventtype;              /* Type of event: READ_EVENT, WRITE_EVENT, 
                                    CONNECT_EVENT, ACCEPT_EVENT */
    char status;                 /* DONE, MORE, INPROGRESS, or WAIT */
    boolean missedevent;         /* True if a read/write/connect event was missed */
    boolean eof;                 /* Whether read() reached eof */
    Buffer *buffer;                    /* Buffer to read or write */
    sockaddr_all connectaddr;          /* Address to connect to */
    ReadCallback *readcallback;        /* Callback for read event */
    ConnectCallback *connectcallback;  /* Callback for connect event */
    WriteCallback *writecallback;      /* Callback for write event */
    AcceptCallback *acceptcallback;    /* Callback for accept event */
    void *data;                        /* Data to pass to callback function */
    time_t timeout;              /* Time at which this socket times out */
    int error;                   /* The error value of the socket call */
    int writeoffset;             /* Offset that's been written so far */ 
    boolean opened;              /* True if socket is opened */
    boolean isclient;            /* True if its a client socket */
};
typedef struct _SocketData SocketData;

/* Data to track socket events */
struct _Sockets
{
    SocketData *data;              /* Array of socket data. The fd is 
                                      index into array */
    int numclients;                /* Current number of client connections */
    int numservers;                /* Current number of server connections */
    int maxclients;                /* The maximum number of concurrent clients */
    int epollfd;                   /* epoll() file descriptor */
    struct epoll_event *epollevents;  /* epoll events */
    int maxfd;                     /* Maximum fd opened so far. */
    int listenfd[MAX_LISTEN];      /* The listening sockets */
    int num_listen;                /* Number of listening sockets */
    Queue outq;                    /* Sockets are queued here after the
                                      read/write/connect is performed. */
    time_t current_time;           /* The current time, in seconds */
    time_t lasttimeout;            /* The time we last checked timeouts */
    DnsCache *dns;                 /* The dns lookup cache */
};
typedef struct _Sockets Sockets;

void sockets_init(Sockets *s);
int sockets_new(Sockets *s, int family);

void sockets_read(Sockets *s, int fd, Buffer *buffer,
                  ReadCallback *callback, void *data);

void sockets_read_timeout(Sockets *s, int fd, Buffer *buffer, int timeout,
                          ReadCallback *callback, void *data);

void sockets_connect(Sockets *s, int fd, sockaddr_all *addr, 
                     ConnectCallback *callback, void *data);

void sockets_write(Sockets *s, int fd, Buffer *buffer, 
                   WriteCallback *callback, void *data);

void sockets_listen(Sockets *s, int fd, AcceptCallback *callback, void *data);

void sockets_dnslookup(Sockets *s, Substring host,
                       DnsCallback *callback, void *data);

void sockets_close(Sockets *s, int fd);

void sockets_checkevents(Sockets *s, int msec);

char* sockets_to_string(sockaddr_all *addr);

#endif
