/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**
  The sockettest will do the following:

  1. Fork a child process.  The server will run in the child process.
       The client will run in the parent process.
  2. The client will connect to the server.
  3. The client will write the CLIENT_MESSAGE, which the server will read.
  4. The server will write the SERVER_MESSAGE, which the client will read.
  5. The server will close the connection, and the client will wait for EOF.

**/
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#include "buffer.h"
#include "sockets.h"
#include "conf.h"
#include "currenttime.h"
#include "win32.h"

#define SERVER_PORT 1099
#define CLIENT_MESSAGE "=CLIENT="
#define SERVER_MESSAGE "=SERVER="

BufferPool *bp;

WriteCallback server_write;
void server_write(int fd, void *data, Buffer *b, int error)
{
    Sockets *s = data;

    assert(error == 0);
    buffer_unref(b);
    sockets_close(s, fd); 
    exit(0);
}

ReadCallback server_read;
void server_read(int fd, void *data, Buffer *b, int error, int eof)
{
    Sockets *s = data;
    Buffer *a;

    assert(error == 0);
    assert(eof == 0);
    assert(b->size == strlen(CLIENT_MESSAGE));
    assert(strcmp(b->buf, CLIENT_MESSAGE) == 0);

    a = buffer_alloc(bp);
    buffer_add(a, SERVER_MESSAGE, strlen(SERVER_MESSAGE));
    sockets_write(s, fd, a, server_write, s);
}

AcceptCallback server_accept;
void server_accept(int clientfd, sockaddr_all addr, void *data)
{
    Buffer *b;
    Sockets *s = data;

    b = buffer_alloc(bp);
    sockets_read(s, clientfd, b, server_read, s);
}

void run_server()
{
    int fd, n;
    sockaddr_all addr;
    Sockets *s;
    time_t end, curr;

    s = malloc(sizeof(Sockets));
    assert(Config.max_total_fd > 0); 
    sockets_init(s);
 
    fd = sockets_new(s, AF_INET);
    n = 1; 
    setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &n, sizeof(int));
    memset(&addr, 0, sizeof(addr));
    addr.ipv4.sin_family = AF_INET;
    addr.ipv4.sin_port = htons(SERVER_PORT);
    addr.ipv4.sin_addr.s_addr = INADDR_ANY;
    n = bind(fd, (struct sockaddr *) &(addr.ipv4), sizeof(struct sockaddr_in));
    assert(n == 0);
    listen(fd, 1024);
    sockets_listen(s, fd, server_accept, s);

    end = time(NULL) + 10;
    curr = time(NULL);
    while (curr < end) {
        sockets_checkevents(s, 100);
        curr = time(NULL);
    }
    assert(0);
}


/***** Client part *****/

void client_close(Sockets *s, int fd)
{
    sockets_close(s, fd);
    (void) sleep(1);
    printf("Pass\n");
    exit(0);
}

ReadCallback client_read_eof;
void client_read_eof(int fd, void *data, Buffer *b, int error, int eof)
{
    Sockets *s = data;
    assert(error == 0);
    assert(eof == 1);
    client_close(s, fd);
}

ReadCallback client_read;
void client_read(int fd, void *data, Buffer *b, int error, int eof)
{
    Sockets *s = data;

    assert(error == 0);
    assert(b->size == strlen(SERVER_MESSAGE));
    assert(strcmp(b->buf, SERVER_MESSAGE) == 0);

    if (eof) {
        buffer_unref(b);
        client_close(s, fd); 
    }
    else {
        buffer_reset(b);
        sockets_read(s, fd, b, client_read_eof, s);
    }
}

void client_write(int fd, void *data, Buffer *b, int error)
{
    Sockets *s = data;

    buffer_unref(b);
    assert(error == 0);
    b = buffer_alloc(bp);
    sockets_read(s, fd, b, client_read, s);
}

void client_connect(int fd, void *data, int error)
{
    Sockets *s = data;
    Buffer *b;

    assert(error == 0);
    b = buffer_alloc(bp);
    buffer_add(b, CLIENT_MESSAGE, strlen(CLIENT_MESSAGE));
    sockets_write(s, fd, b, client_write, s);
}

void run_client()
{
    int fd;
    sockaddr_all addr;
    Sockets *s;
    time_t end, curr;

    s = malloc(sizeof(Sockets));
    sockets_init(s);
    fd = sockets_new(s, AF_INET);
    memset(&addr, 0, sizeof(addr));
    addr.ipv4.sin_family = AF_INET;
    addr.ipv4.sin_port = htons(SERVER_PORT);
    addr.ipv4.sin_addr.s_addr = inet_addr("127.0.0.1");
    sockets_connect(s, fd, &addr, client_connect, s);

    end = time(NULL) + 10;
    curr = time(NULL);
    while (curr < end) {
        sockets_checkevents(s, 100);
        curr = time(NULL);
    }
    assert(0);
}

int main(int argc, char **argv)
{
    char * serverargs[] = { "sockettest", "server", NULL };
    int n;

    bp = malloc(sizeof(BufferPool));
    bufferpool_init(bp, 1024, 5);
    memset(&Config, 0, sizeof(Config));
    Config.max_total_fd = 100; 
    Config.num_webservers = 0;

    if (argc == 2 && strcmp(argv[1], "server") == 0) {
        run_server();
        return 0;
    }
    write(1, "sockettest........", 18);

    n = fork();
    if (n == 0) {
        execv("sockettest", serverargs);
    }
    (void) sleep(2);
    run_client();
    return 0;
}



