/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**
  Usage: ./update_log_stats <accesslog> <logstats directory> [<interval>]

  This is a standalone program, separate from vicompress. This program
  is used to generate the HTML log statistics from the accesslog.
  The program does three main actions:

  1. Parse all the access logs (accesslog.2, accesslog.1, accesslog) into
     a Logstats structure.

  2. Write the Logstats structure to the file <logstats dir>/logstats.txt
     This is a plain text file containing raw statistics about bandwidth,
     requests, caching, and compression.  The format of the file is described
     in the function write_logstatistics() below.

  3. For each year/month, create the file <logstats dir>/YYYYMMstats.html.
     For example, 200502stats.html (Feb 2005). This file is an HTML view of the 
     log statistics for that particular month.

  4. Create the file <logstats dir>/statsindex.html.  This file is an index
     which contains hyperlinks to all the monthly statistics.

  5. If the interval argument is provided, continuously repeat steps 1-3
     above, waiting <interval> hours in between.  This is used to periodically
     generated the log statistics.


  A couple additional notes:

  *  When called multiple times, this program will not re-compute all the log
     statistics it previously read. It will load the previous logstats.txt file,
     and start computing where it left off.  That is, it will fseek() into the 
     accesslog to the place where it last finished, and then start computing 
     from there.

  *  The file <logstats dir>/template.html contains the HTML template. The
     template file contains variables such as $month_beforegzip$, 
     $hour_aftergzip9$, and so on.  This program simply replaces those 
     variables with the actual numerical values.

**/

#include <unistd.h>
#include <signal.h>

#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include "win32.h"

#define LOG_FORMAT_UNKNOWN 0
#define LOG_FORMAT_APACHE  1
#define LOG_FORMAT_SQUID   2

#define MAX_LINE_SIZE  32768
#define MAX_YEAR 2100

#define MAX_TEMPLATE_TOKENS  4096

static const char *monthstrings[] = {
    "", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
    "Aug", "Sep", "Oct", "Nov", "Dec"
};

static int daysinmonth[] = {
    0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

/* 
 * The 5 element arrays in the Monthstats struct below contain
 * the following 5 values:
 * 
 * NON_GZIP - Amount of non-gzipped data written to clients
 * GZIP     - Amount of gzipped data written to clients
 * ORIG     - Original size of gzipped data written to clients
 * MISS     - Amount of data read from origin websites
 * HIT      - Amount of data read from the cache
 */
#define NON_GZIP   0
#define GZIP       1
#define ORIG       2
#define MISS       3
#define HIT        4

/* Monthstats stores the bandwidth and request statistics for
 * a given year and month.
 */
struct _Monthstats {
    int isused;      /* Are there stats for this year/month ? */
    int haschanged;  /* Has the stats changed from the logstats.txt file */
    int year;        /* The year and month these stats are for */
    int month;

    double month_size[5];   /* Bandwidth usage per month */
    double month_count[5];  /* Number of requests per month */

    double day_size[32][5];  /* Bandwidth usage per day */
    double day_count[32][5]; /* Number of requests per day */

    double dayweek_size[8][5];  /* Average banwidth per day of week */
    double dayweek_count[8][5]; /* Average requests per day of week */

    double hour_size[24][5];    /* Average bandwidth per hour */
    double hour_count[24][5];   /* Average requests per hour */

};
typedef struct _Monthstats Monthstats;

/* Lastupdate stores information about where the last update_log_stats
 * run left off while reading the accesslog.  When update_log_stats is
 * called again, it can then skip to line in the accesslog where it
 * left off, making the parsing much faster.
 */
struct _Lastupdate {
    unsigned long filesize;    /* Size of the accesslog */
    long numlines;             /* Number of lines in accesslog */
    time_t startdate, enddate; /* The first and last dates in the accesslog */
    int logformat;             /* The format of the accesslog (squid or apache) */
};
typedef struct _Lastupdate Lastupdate;

/* Logstats contains
 * - the stats of the accesslog during the last update.
 * - the statistics for each year/month.
 */
struct _Logstats {
    Lastupdate last;
    Monthstats monthstats[MAX_YEAR - 2000][13];
};
typedef struct _Logstats Logstats;

/* Contains the parsed information for a single request/line in
 * the accesslog file.
 */
struct _Line {
    struct tm date;   /* The date of the request */
    time_t datetime;
    char iscachehit;  /* True if the request is a cache hit */
    double gzip_size; /* The size of the reply after gzipping */
    double orig_size; /* The size of the reply before gzipping */
};
typedef struct _Line Line;

/* String structure is used to store the template.html file. */
struct _String {
    char *value;   /* String data */
    int len;       /* Length of string */
};
typedef struct _String String;

/* HtmlTemplate contains the parsed template.html file.  After computing 
 * the statistics, this program does a search and replace, replacing all
 * the $variable$ in the template.html with the actual numerical value.
 * There are about 1200 variables in template.html.  We want to speed up
 * the search and replace, so we parse the file into string tokens, where
 * each $variable$ is a separate token.  Therefore, replacing a variable
 * requires scanning only 1200 tokens, instead of scanning 100K of raw
 * string data.
 */ 
struct _HtmlTemplate {
    String tokens[MAX_TEMPLATE_TOKENS];
    int count;
};
typedef struct _HtmlTemplate HtmlTemplate;

static int parse_lastupdate(Lastupdate *last, FILE *fp, 
                            const char *filename, int *linenumber);
static int parse_month_stats(Monthstats *stats, FILE *fp, 
                             const char *filename, int *linenumber);
static int parse_apache_line(Line *line, char *data, int datalen, 
                             const char *filename, int linenumber);
static int parse_squid_line(Line *line, char *data, int datalen, 
                            const char *filename, int linenumber);

static int write_html_month_stats(HtmlTemplate *t, Monthstats *m, 
                                  char *statsdir);

static int gotsignal = 0;


/* String memory routines.  We allocate lots of small strings in this 
 * program when doing the search-and-replace on the template.html file.
 * To avoid memory fragmentation, we allocate a large pool of memory
 * up front for strings.
 */
#define STRING_MEMORY_SIZE  1048576  /* 1 MB */
static int stringindex = 0;
static char* stringmemory = NULL;

static void stringmemory_init() {
    stringmemory = malloc(STRING_MEMORY_SIZE);
    stringindex = 0;
}

/* Allocate memory for the given size.  The memory
 * cannot be freed individually.
 */
static void* stringmemory_alloc(int size) {
    char* result;
    result = &stringmemory[stringindex];
    stringindex += size;

    assert(stringindex < STRING_MEMORY_SIZE);
    /* Make sure index is aligned on 8 byte boundary */
    while (stringindex % 8 != 0) {
        stringindex++;
    }
    return result;
}

/* Free all the string memory allocated */
static void stringmemory_reset() {
    stringindex = 0;
}

static char* stringdup(const char *s) {
    char *result = stringmemory_alloc(strlen(s) + 1);
    strcpy(result, s);
    return result;
}

void handle_signal(int signum)
{
    gotsignal = 1;
}

/* Print an error message. Also print the file and line number given. */
static void printerr(const char *file, int line, const char *fmt, ...)
{
    char msg[256];
    va_list args;

    memset(msg, 0, sizeof(msg));
    sprintf(msg, "Error file %s line %d: ", file, line);

    va_start(args, fmt);
    vsprintf(&msg[strlen(msg)], fmt, args);
    va_end(args);
    fprintf(stderr, "%s", msg);
}

/* Print a warning message. Also print the file and line number given. */
static void printwarn(const char *file, int line, const char *fmt, ...)
{
    char msg[256];
    va_list args;

    memset(msg, 0, sizeof(msg));
    sprintf(msg, "Warning file %s line %d: ", file, line);

    va_start(args, fmt);
    vsprintf(&msg[strlen(msg)], fmt, args);
    va_end(args);
    fprintf(stderr, "%s", msg);
}

/* Redirect stdin, stdout, stderr to /dev/null.  */
static void redirect_stdio_to_devnull()
{
#ifndef WIN32
    int fd;
    fd = open("/dev/null", O_RDWR, 0);
    if (fd < 0)
        return;
    dup2(fd, 0);
    dup2(fd, 1);
    dup2(fd, 2);
    close(fd);
#endif
}


/* Return the integer corresponding to the given 3-letter 
 * month name (Jan, Feb, ..).  Return 1 to 12, or -1 on error.
 */
static int getmonth(const char *s)
{
    int i;
    for (i = 1; i <= 12; i++) {
        if (strncmp(s, monthstrings[i], 3) == 0)
            return i;
    }
    return -1;
}

/* Give a buffer, split it into tokens using whitespace as delimiter. 
 * Store the tokesn in result.  Return the number of tokens.
 */
static int splittokens(char *buf, int len, char **result, int maxitems)
{
    int i, item, nexttospace;
    nexttospace = 1;
    item = 0;

    for (i = 0; i < maxitems; i++)
        result[i] = "";

    result[0] = buf;
    for (i = 0; i < len; i++) {
        if (buf[i] == ' ' || buf[i] == '\t' || 
            buf[i] == '\r' || buf[i] == '\n') {

            buf[i] = 0; nexttospace = 1;
        }
        else if (nexttospace && item < maxitems) {
            result[item] = &buf[i];
            item++;
            nexttospace = 0;
        }
    }
    return item;
}

/* Given a vicompress accesslog file, return the log format of the file: 
 * LOG_FORMAT_APACHE, LOG_FORMAT_SQUID, or LOG_FORMAT_UNKNOWN.
 */
static int get_logformat(const char *filename)
{
    FILE *fp;
    char buf[8192];
    char *tokens[32];
    int n;

    fp = fopen(filename, "r");
    if (fp == NULL) {
        fprintf(stderr, "Error: cannot open file %s: %s\n", filename, strerror(errno));
        return LOG_FORMAT_UNKNOWN;
    }
    fgets(buf, 8192, fp);
    fclose(fp);
    n = splittokens(buf, strlen(buf), tokens, 32);
    if (n < 5)
        return LOG_FORMAT_UNKNOWN;
    else if (tokens[3][0] == '[')
        return LOG_FORMAT_APACHE;
    else if (tokens[3][0] == 'T')
        return LOG_FORMAT_SQUID;
    else
        return LOG_FORMAT_UNKNOWN;
} 

/* Return the size (in bytes) of the given file, or -1 on error. */
static long get_filesize(const char *filename)
{
    struct stat info;
    int n;

    n = stat(filename, &info);
    if (n != 0)
        return -1;

    return info.st_size;
}

/* Return the starting date of the given accesslog file, or -1 on error. */
static time_t get_startdate(const char *filename)
{
    FILE *fp;
    Line line;
    char buf[8192];
    int format, n = -1;

    format = get_logformat(filename);
    if (format == LOG_FORMAT_UNKNOWN) {
        fprintf(stderr, "Error: unknown log format file %s\n", filename);
        return -1;
    }

    fp = fopen(filename, "r");

    if (fp == NULL) {
        fprintf(stderr, "Error: cannot open file %s: %s\n", filename, strerror(errno));
        return -1;
    }
    fgets(buf, 8192, fp);
    fclose(fp);
    if (buf[strlen(buf)-1] != '\n')
        return -1;

    if (format == LOG_FORMAT_APACHE)
        n = parse_apache_line(&line, buf, strlen(buf), filename, 1);
    else if (format == LOG_FORMAT_SQUID)
        n = parse_squid_line(&line, buf, strlen(buf), filename, 1);
    else
        assert(0);

    if (n == -1)
        return -1;
    else
        return line.datetime;
}

/* Convert the given date into a time_t value. */
static time_t make_time(int year, int month, int day, 
                        int hour, int min, int sec, int daylight)
{
    struct tm thedate;
    memset(&thedate, 0, sizeof(thedate));
    thedate.tm_year = year - 1900; 
    thedate.tm_mon = month - 1;
    thedate.tm_mday = day;
    thedate.tm_hour = hour; 
    thedate.tm_min = min; 
    thedate.tm_sec = sec;
    thedate.tm_isdst = daylight;
    return mktime(&thedate);
}

/* Parse the logstats.txt file given in filename into a logstats structure. */
static int parse_logstatistics(Logstats *logstats, const char *filename)
{
    Monthstats *m;
    char buf[256];
    char *tokens[32];
    FILE *fp;
    int n, err, year, month;
    int linenumber = 0;

    memset(logstats, 0, sizeof(Logstats));

    fp = fopen(filename, "r"); err = errno;
    if (fp == NULL) {
        if (err != ENOENT) {
            fprintf(stderr, "Error: unable to open statistics file %s: %s\n", 
                    filename, strerror(err));
            return -1;
        }
        else
            return 0;
    }

    while (fgets(buf, 256, fp) != NULL) {
        linenumber += 1;
        n = splittokens(buf, strlen(buf), tokens, 32);
        if (n == 0 || buf[0] == '#')
            continue;
        if (strcmp(tokens[0], "yearmonth") == 0) {
            if (n != 3) {
                printerr(filename, linenumber, "yearmonth requires 2 arguments\n");
                fclose(fp); return -1;
            }
            year = atoi(tokens[1]);
            month = atoi(tokens[2]);
            if (year < 2000 || year > MAX_YEAR) {
                printerr(filename, linenumber, "Invalid year %s\n", tokens[1]);
                fclose(fp); return -1;
            }
            if (month < 1 || month > 12) {
                printerr(filename, linenumber, "Invalid month %s\n", tokens[2]);
                fclose(fp); return -1;
            }
            m = &(logstats->monthstats[year - 2000][month]);
            m->isused = 1;
            m->year = year;
            m->month = month;
            n = parse_month_stats(m, fp, filename, &linenumber); 
            if (n == -1) {
                fclose(fp); return -1;
            }
        }
        else if (strcmp(tokens[0], "lastupdate") == 0) {
            n = parse_lastupdate(&(logstats->last), fp, filename, &linenumber);
            if (n == -1) {
                fclose(fp); return -1;
            }
        }
        else {
        }
    }
    fclose(fp);
    return 0;
}

/* Parse the "lastupdate" portion of the logstats.txt file in the
 * given Lastupdate structure.
 */
static int parse_lastupdate(Lastupdate *last, FILE *fp, 
                            const char *filename, int *linenumber)
{
    char buf[4096];
    char *tokens[32];
    int n;

    while (fgets(buf, 4096, fp) != NULL) {
     *linenumber = *linenumber + 1;
        n = splittokens(buf, strlen(buf), tokens, 32);
        if (n == 0 || buf[0] == '#')
            continue;
        if (strcmp(tokens[0], "filesize") == 0)
            last->filesize = atol(tokens[1]);
        else if (strcmp(tokens[0], "numlines") == 0)
            last->numlines = atol(tokens[1]);
        else if (strcmp(tokens[0], "startdate") == 0) {
            if (n != 8) {
                printerr(filename, *linenumber, "startdate requires 7 arguments\n");
                return -1;
            }
            last->startdate = 
                make_time(atoi(tokens[1]), atoi(tokens[2]), atoi(tokens[3]),
                          atoi(tokens[4]), atoi(tokens[5]), atoi(tokens[6]),
                          atoi(tokens[7]));
            if (last->startdate < 0) {
                printerr(filename, *linenumber, "Invalid startdate\n");
                return -1;
            }
        }
        else if (strcmp(tokens[0], "enddate") == 0) {
            if (n != 8) {
                printerr(filename, *linenumber, "enddate requires 7 arguments\n");
                return -1;
            }
            last->enddate = 
                make_time(atoi(tokens[1]), atoi(tokens[2]), atoi(tokens[3]),
                          atoi(tokens[4]), atoi(tokens[5]), atoi(tokens[6]),
                          atoi(tokens[7]));
            if (last->enddate < 0) {
                printerr(filename, *linenumber, "Invalid enddate\n");
                return -1;
            }
        } 
        else if (strcmp(tokens[0], "endlastupdate") == 0) {
            return 0;
        }
    }
    printerr(filename, *linenumber, "No endlastupdate entry found\n");
    return -1;
}

/* Parse the statistics for a single month from the logstats.txt file.
 * A single month looks like:
 *   yearmonth 2006 1
 *   ....
 *   endyearmonth
 *
 * This function is called with a "yearmonth" line is encountered, so
 * the current file position will be the next line after that.  When
 * this function returns, the file position will be the line after the
 * endyearmonth line.
 */
static int parse_month_stats(Monthstats *stats, FILE *fp, 
                             const char *filename, int *linenumber)
{
    char buf[8192];
    int i, n;
    char *tokens[32];

    while (fgets(buf, 8192, fp) != NULL) {
        *linenumber = *linenumber + 1;
        n = splittokens(buf, strlen(buf), tokens, 32);
        if (n == 0 || buf[0] == '#')
            continue;

        if (strcmp(tokens[0], "endyearmonth") == 0)
            return 0;

        if (n != 6) {
            printerr(filename, *linenumber, "%s requires 5 arguments\n", tokens[0]);
            return -1;
        }

        if (strcmp(tokens[0], "month_size") == 0) {
            for (i = 0; i < 5; i++) 
                stats->month_size[i] = atof(tokens[i+1]);
        }
        else if (strcmp(tokens[0], "month_count") == 0) {
            for (i = 0; i < 5; i++) 
                stats->month_count[i] = atof(tokens[i+1]);
        }
        else if (strncmp(tokens[0], "day_size", strlen("day_size")) == 0) {
            int day = atoi( &(tokens[0][strlen("day_size")]) );
            if (day < 1 || day > 31) {
                printerr(filename, *linenumber, "Invalid name %s\n", tokens[0]);
                return -1;
            }
            for (i = 0; i < 5; i++) 
                stats->day_size[day][i] = atof(tokens[i+1]);
        }
        else if (strncmp(tokens[0], "day_count", strlen("day_count")) == 0) {
            int day = atoi( &(tokens[0][strlen("day_count")]) );
            if (day < 1 || day > 31) {
                printerr(filename, *linenumber, "Invalid name %s\n", tokens[0]);
                return -1;
            }
            for (i = 0; i < 5; i++) 
                stats->day_count[day][i] = atof(tokens[i+1]);
        }
        else if (strncmp(tokens[0], "dayweek_size", strlen("dayweek_size")) == 0) {
            int dayweek = atoi( &(tokens[0][strlen("dayweek_size")]) );
            if (dayweek < 1 || dayweek > 7) {
                printerr(filename, *linenumber, "Invalid name %s\n", tokens[0]);
                return -1;
            }
            for (i = 0; i < 5; i++) 
                stats->dayweek_size[dayweek][i] = atof(tokens[i+1]);
        }
        else if (strncmp(tokens[0], "dayweek_count", strlen("dayweek_count")) == 0) {
            int dayweek = atoi( &(tokens[0][strlen("dayweek_count")]) );
            if (dayweek < 1 || dayweek > 7) {
                printerr(filename, *linenumber, "Invalid name %s\n", tokens[0]);
                return -1;
            }
            for (i = 0; i < 5; i++) 
                stats->dayweek_count[dayweek][i] = atof(tokens[i+1]);
        }
        else if (strncmp(tokens[0], "hour_size", strlen("hour_size")) == 0) {
            int hour = atoi( &(tokens[0][strlen("hour_size")]) );
            if (hour < 0 || hour > 23) {
                printerr(filename, *linenumber, "Invalid name %s\n", tokens[0]);
                return -1;
            }
            for (i = 0; i < 5; i++) 
                stats->hour_size[hour][i] = atof(tokens[i+1]);
        }
        else if (strncmp(tokens[0], "hour_count", strlen("hour_count")) == 0) {
            int hour = atoi( &(tokens[0][strlen("hour_count")]) );
            if (hour < 0 || hour > 23) {
                printerr(filename, *linenumber, "Invalid name %s\n", tokens[0]);
                return -1;
            }
            for (i = 0; i < 5; i++) 
                stats->hour_count[hour][i] = atof(tokens[i+1]);
        }
        else {
            printerr(filename, *linenumber, "Invalid name %s\n", tokens[0]);
            return -1;
        }
    }
    printerr(filename, *linenumber, "No endyearmonth entry found\n");
    return -1;
}

/* Parse the given data into a Line structure. The data is a line
 * from a squid log file. Return 0 on success, -1 on error.
 */
static int parse_squid_line(Line *line, char *data, int datalen, 
                            const char *filename, int linenumber)
{
    char *tokens[20];
    int year, n;

    n = splittokens(data, datalen, tokens, 20);
    if (n < 10) {
        printwarn(filename, linenumber, "Expected at least 10 fields, got %d\n", n);
        return -1;
    }
    line->datetime = atoi(tokens[0]);
    if (line->datetime == 0) {
        printwarn(filename, linenumber, "Invalid date: %s\n", tokens[0]); 
        return -1;
    }
    localtime_r(&line->datetime, &line->date);
    year = line->date.tm_year + 1900;
    if (year < 2000 || year >= MAX_YEAR) {
        printwarn(filename, linenumber, "Invalid date: %s\n", tokens[0]);
        return -1;
    }

    line->gzip_size = atof(tokens[4]);
    if ( !isdigit(tokens[4][0]) ) {
        printwarn(filename, linenumber, "Invalid content length: %s\n", tokens[4]);
        return -1;
    }

    if ( isdigit(tokens[7][0]) )
        line->orig_size = atof(tokens[7]);
    else
        line->orig_size = line->gzip_size;

    if (strncmp(tokens[3], "TCP_HIT", 7) == 0)
        line->iscachehit = 1;
    else
        line->iscachehit = 0;
    return 0;
}

/* Parse the given data into a Line structure. The data is a line
 * from an apache log file. Return 0 on success, -1 on error.
 */
static int parse_apache_line(Line *line, char *data, int datalen, 
                             const char *filename, int linenumber)
{
    char *tokens[20];
    int i, n, len;
    char *date;
    char *ident;
    char *curr;

    n = splittokens(data, datalen, tokens, 20);
    if (n < 10) {
        printwarn(filename, linenumber, "Expected at least 10 fields, got %d\n", n);
        return -1;
    }
    date = tokens[3];
    if (strlen(date) < 21) {
        printwarn(filename, linenumber, "Invalid date: %s\n", date);
        return -1;
    }
    memset(&line->date, 0, sizeof(struct tm));
    line->date.tm_mday = atoi(&date[1]);
    line->date.tm_mon = getmonth(&date[4]) - 1;
    line->date.tm_year = atoi(&date[8]) - 1900;
    line->date.tm_hour = atoi(&date[13]);
    line->date.tm_min = atoi(&date[16]);
    line->date.tm_sec = atoi(&date[19]);
    line->date.tm_isdst = -1;

    line->datetime = mktime(&line->date);
    if (line->datetime < 0) {
        printwarn(filename, linenumber, "Invalid date: %s\n", date);
        return -1;
    }
    localtime_r(&line->datetime, &line->date);

    /* The URL request is delimited by quotes, such as
     * "GET some path\" or other.html HTTP/1.0". To skip this
     * field, we must find the ending ", and we must skip any
     * escaped quotes, \".
     */
    i = 5;
    while (i < n) {
        curr = tokens[i];
        len = strlen(curr);
        if (len == 1 && curr[len-1] == '"')
            break;
        else if (len >= 2 && curr[len-1] == '"' && curr[len-2] != '\\')
            break;
        else
            i++;
    } 

    if (i >= n) {
        printwarn(filename, linenumber, "Couldn't find content length field\n");
        return -1;
    }
    i += 2;
    if (i >= n) {
        printwarn(filename, linenumber, "Couldn't find content length field\n");
        return -1;
    }
    if (tokens[i][0] == '-')
        line->gzip_size = 0;
    else
        line->gzip_size = atof(tokens[i]);

    if ( !isdigit(tokens[i][0]) && tokens[i][0] != '-' ) {
        printwarn(filename, linenumber, "Invalid content length: %s\n", tokens[i]);
        return -1;
    }

    ident = tokens[1];
    if (strncmp(ident, "hit", 3) == 0) {
        line->iscachehit = 1;
        line->orig_size = atof(&ident[3]);
    }
    else if (strncmp(ident, "miss", 4) == 0) {
        line->iscachehit = 0;
        line->orig_size = atof(&ident[4]);
    }
    else {
        line->iscachehit = 0;
        line->orig_size = line->gzip_size;
    } 
    return 0;
}

/* Return the number of days in the given month and year.
 * month is from 1 to 12.
 */
static int get_days_in_month(int year, int month)
{
    assert(month >= 1 && month <= 12);
    if (month != 2)
        return daysinmonth[month];

    if ((year % 4 == 0) && (year % 100 != 0))
        return 29;
    else
        return 28;
}

/* For the given year/month, store the number of times each
 * day of the week (sunday, monday, .. saturday) occurs.
 * In the dayweek array, in index 1 thru 7, store the values
 * for sunday thru saturday.
 */
static void get_dayweek_in_month(int* dayweek, int year, int month)
{
    struct tm date;
    time_t t;
    int day, i, total;
 
    memset(&date, 0, sizeof(date));
    date.tm_year = year - 1900;
    date.tm_mon = month - 1;

    t = mktime(&date);
    localtime_r(&t, &date);

    total = get_days_in_month(year, month);
    day = date.tm_wday;

    for (i = 1; i <= 7; i++)
        dayweek[i] = 0;

    for (i = 1; i <= total; i++) {
        dayweek[day+1] += 1;
        day = (day + 1) % 7; 
    }
}

/* Given a line from the vicompress log file, update the
 * log statistics structure.
 */
static void update_stats(Logstats *logstats, Line *line)
{
    Monthstats *stats;
    int i;
    int year, month, day, dayweek, hour;
    double sizes[5];
    double counts[5];

    i = 0; 
    year = line->date.tm_year + 1900;
    month = line->date.tm_mon + 1;

    assert(year >= 2000 && year < MAX_YEAR);
    assert(month >= 1 && month <= 12);

    stats = &logstats->monthstats[year - 2000][month];
    stats->year = year;
    stats->month = month;
    stats->isused = 1;
    stats->haschanged = 1;

    memset(sizes, 0, sizeof(sizes));
    memset(counts, 0, sizeof(counts));

    if (line->orig_size == line->gzip_size) {
        sizes[NON_GZIP] = line->orig_size;
        counts[NON_GZIP] = 1;
    }
    else {
        sizes[GZIP] = line->gzip_size;
        sizes[ORIG] = line->orig_size;
        counts[GZIP] = 1;
        counts[ORIG] = 1;
    }
    if (line->iscachehit) {
        sizes[HIT]  = line->orig_size;
        counts[HIT] = 1;
    }
    else {
        sizes[MISS]  = line->orig_size;
        counts[MISS] = 1;
    }

    day = line->date.tm_mday;
    dayweek = line->date.tm_wday + 1;
    hour = line->date.tm_hour;

    for (i = 0; i < 5; i++) {
        stats->month_size[i] += sizes[i];
        stats->month_count[i] += counts[i];
        stats->day_size[day][i] += sizes[i];
        stats->day_count[day][i] += counts[i];
        stats->dayweek_size[dayweek][i] += sizes[i];
        stats->dayweek_count[dayweek][i] += counts[i];
        stats->hour_size[hour][i] += sizes[i];
        stats->hour_count[hour][i] += counts[i];
    }
}

/* Return 1 if the given file exists, else 0. */
static int file_exists(const char *filename)
{
    struct stat info;
    int n;

    n = stat(filename, &info);
    if (n == -1)
        return 0;
    else if (n == 0 && info.st_size == 0)
        return 0;
    else
        return 1;
}

/* Return 1 if
 * - The date of the first line in filename is newer than the lastupdate
 *   date in the Lastupdate structure.
 * - The date of the first line in filename is the same as the lastupdate
 *   date in the Lastupdate structure, and the filename size is bigger
 *   than the lastupdate filesize.
 * Else, return 0.
 */
static int is_accesslog_new(Lastupdate *last, const char *filename)
{
    time_t startdate;
    long filesize;

    startdate = get_startdate(filename);
    if (startdate == -1)
        return 0;

    filesize = get_filesize(filename);

    if (startdate < last->startdate)
        return 0;
    else if (startdate > last->startdate)
        return 1;
    else if (startdate == last->startdate && filesize <= last->filesize)
        return 0;
    else if (startdate == last->startdate && filesize > last->filesize)
        return 1;
    else
        return 0;
}

/* Parse the given logfile and update the logstats structure.
 * Update the monthstats and the lastupdate data.
 * Return 0 on success, -1 on error.
 */
static int parse_accesslog(Logstats *logstats, const char *filename)
{
    FILE *fp;
    char buf[MAX_LINE_SIZE];
    int n, len, linenumber, numerrors;
    Line line;
    int logformat;
    time_t startdate, newenddate;
    long filesize, newfilesize;

    logformat = get_logformat(filename);
    if (logformat == LOG_FORMAT_UNKNOWN) {
        printerr(filename, 1, "Unknown log format (neither apache nor squid)\n",
                 filename);
        return -1;
    }
    startdate = get_startdate(filename);
    newenddate = startdate;
    if (startdate == -1) {
        printerr(filename, 1, "Unable to parse date\n");
        return -1;
    }
    filesize = get_filesize(filename);
    assert(filesize >= 0);

    fp = fopen(filename, "r");
    assert(fp != NULL);

    if (startdate == logstats->last.startdate) {
        newfilesize = logstats->last.filesize;
        linenumber = logstats->last.numlines;
        fseek(fp, newfilesize, SEEK_SET);
    }
    else {
        logstats->last.startdate = startdate;
        newfilesize = 0;
        linenumber = 0;
    }
    numerrors = 0;

    while (fgets(buf, sizeof(buf), fp) != NULL) {
        len = strlen(buf);
        if (buf[len-1] != '\n')
            break;
        linenumber++;
        newfilesize += strlen(buf);
        if (buf[0] == '\n' || buf[1] == '\n')
            continue;
        if (logformat == LOG_FORMAT_APACHE)
            n = parse_apache_line(&line, buf, strlen(buf), filename, linenumber);
        else if (logformat == LOG_FORMAT_SQUID) 
            n = parse_squid_line(&line, buf, strlen(buf), filename, linenumber);
        else
            assert(0); 

        if (n == -1) {
            numerrors++;
            if (numerrors >= 50) {
                fprintf(stderr, "Too many errors parsing accesslog %s, exiting\n", filename);
                fclose(fp);
                return -1;
            }
        }
        else {
            update_stats(logstats, &line);
            newenddate = line.datetime;
        }
    }
    fclose(fp);
    logstats->last.filesize = newfilesize;
    logstats->last.numlines = linenumber;
    logstats->last.enddate = newenddate;
    return 0;
}

/* Print the given name and list of big integers on the same line:
 * <name> <int> <int> <int> ... <int>
 * We use double instead of int in order to avoid integer overflow.
 */ 
static void fprintarray(FILE *fp, const char *name, double *values, int count)
{
    int i;
    fprintf(fp, "%s", name);
    for (i = 0; i < count; i++)
        fprintf(fp, " %.0f", values[i]);
    fprintf(fp, "\n");
}

/* Write the log statistics contained in the given structure to the
 * given "logstats.txt" file (statfile). Return 0 on success, -1 on error.
 */
static int write_logstatistics(Logstats *logstats, const char *statfile)
{
    FILE *fp;
    int i, j, k;
    Lastupdate *last;
    Monthstats *m;
    struct tm thedate;
    int daysinmonth;
    char name[32];
    const char *s;

    fp = fopen(statfile, "w");
    if (fp == NULL) {
        fprintf(stderr, "Error: unable to create log statistics file %s: %s\n",
                statfile, strerror(errno));
        return -1;
    }

s =
"# logstats.txt\n\
#\n\
# This single text file contains caching and compression statistics.\n\
# The format is as follows:\n\
#\n\
# lastupdate\n\
#     information about the accesslog when the last update was done\n\
# filesize\n\
#     The size of the accesslog (in bytes)\n\
# numlines\n\
#     The number of lines in the accesslog\n\
# startdate <year> <month> <day> <hour> <min> <sec> <is daylight savings\n\
#     The date of the first entry in the accesslog\n\
# enddate <year> <month> <day> <hour> <min> <sec> <is daylight savings\n\
#     The date of the last entry in the accesslog\n\
# endlastupdate\n\
#\n\
# yearmonth <year> <month>\n\
#     Statistics for a specific month\n\
# month_size    <non-gzip> <gzip> <orig> <miss> <hit>\n\
# day_sizeN     <non-gzip> <gzip> <orig> <miss> <hit>\n\
# dayweek_sizeN <non-gzip> <gzip> <orig> <miss> <hit>\n\
# hour_sizeN    <non-gzip> <gzip> <orig> <miss> <hit>\n\
#\n\
#     Bandwidth statistics for a particular time period.\n\
#     month   -- The entire month\n\
#     day     -- A specific day in the month. N = 1 to 31\n\
#     dayweek -- A specific day of the week. N = 1 (Sunday) to 7\n\
#     hour    -- A specific hour of the day. N = 0 to 23\n\
#\n\
#     The 5 statistics measured are given below:\n\
#     non-gzip -- Amount of non-gzipped data written to clients\n\
#     gzip     -- Amount of gzipped data written to clients\n\
#     orig     -- Original size of gzipped data written to clients\n\
#     miss     -- Amount of data read from origin websites\n\
#     hit      -- Amount of data read from the cache\n\
#     All measurements are in bytes\n\
#\n\
# month_count    <non-gzip> <gzip> <orig> <miss> <hit>\n\
# day_countN     <non-gzip> <gzip> <orig> <miss> <hit>\n\
# dayweek_countN <non-gzip> <gzip> <orig> <miss> <hit>\n\
# hour_countN    <non-gzip> <gzip> <orig> <miss> <hit>\n\
#\n\
#     These statistics are similar to the ones above, except we are\n\
#     measuring the number of requests, rather than the bandwidth\n\
#\n";

    fprintf(fp, "%s", s);

    last = &(logstats->last);
    fprintf(fp, "lastupdate\n");
    fprintf(fp, "filesize %ld\n", last->filesize);
    fprintf(fp, "numlines %ld\n", last->numlines);
    localtime_r(&last->startdate, &thedate);
    fprintf(fp, "startdate %d %d %d %d %d %d %d\n",
            thedate.tm_year + 1900, thedate.tm_mon + 1, thedate.tm_mday,
            thedate.tm_hour, thedate.tm_min, thedate.tm_sec, thedate.tm_isdst);
 
    localtime_r(&last->enddate, &thedate);
    fprintf(fp, "enddate %d %d %d %d %d %d %d\n",
            thedate.tm_year + 1900, thedate.tm_mon + 1, thedate.tm_mday,
            thedate.tm_hour, thedate.tm_min, thedate.tm_sec, thedate.tm_isdst);
    fprintf(fp, "endlastupdate\n");


    for (i = 2000; i < MAX_YEAR; i++) {
        for (j = 1; j <= 12; j++) {
            m = &logstats->monthstats[i - 2000][j];
            daysinmonth = get_days_in_month(i, j);

            if (m->isused) {
                assert(m->year >= 2000);
                assert(m->month >= 1 && m->month <= 12);
                fprintf(fp, "yearmonth %d %d\n", m->year, m->month);
                fprintarray(fp, "month_size", m->month_size, 5);
                fprintarray(fp, "month_count", m->month_count, 5);

                for (k = 1; k <= daysinmonth; k++) {
                    sprintf(name, "day_size%d", k);
                    fprintarray(fp, name, (double *)m->day_size[k], 5);
                }
                for (k = 1; k <= daysinmonth; k++) {
                    sprintf(name, "day_count%d", k);
                    fprintarray(fp, name, (double *)m->day_count[k], 5);
                }

                for (k = 1; k <= 7; k++) {
                    sprintf(name, "dayweek_size%d", k);
                    fprintarray(fp, name, (double *)m->dayweek_size[k], 5);
                }
                for (k = 1; k <= 7; k++) {
                    sprintf(name, "dayweek_count%d", k);
                    fprintarray(fp, name, (double *)m->dayweek_count[k], 5);
                }

                for (k = 0; k < 24; k++) {
                    sprintf(name, "hour_size%d", k);
                    fprintarray(fp, name, (double *)m->hour_size[k], 5);
                }
                for (k = 0; k < 24; k++) {
                    sprintf(name, "hour_count%d", k);
                    fprintarray(fp, name, (double *)m->hour_count[k], 5);
                }

                fprintf(fp, "endyearmonth\n");
            }
        }
    }
    fclose(fp);
    return 0;
}

/* Write the given data to the given file. */
static void write_data(int fd, const char *data, int total)
{
    int amount = 0;
    int n;

    while (amount < total) {
        n = write(fd, &data[amount], total - amount);
        if (n > 0)
            amount += n;
        else
            return; 
    }
}

/* Return the index where "string" occurs in buf, starting at
 * the given offset. Return -1 if the string is not found.
 */
static int findstring(const char *buf, int offset, int end, char *string)
{
    int stringlen;
    stringlen = strlen(string);

    while (stringlen <= end - offset) {
        if (strncasecmp(&buf[offset], string, stringlen) == 0)
            return offset;
        offset++;
    }
    return -1;
}

/* Read the given file into a char array and return it. 
 * Return NULL if there is an error reading the file.
 */
static char* read_file(char *filename)
{
    int fd;
    char *data;
    int n, amount, filesize;

    filesize = get_filesize(filename);
    if (filesize == -1) {
        fprintf(stderr, "Error: unable to open file %s, %s\n", 
                filename, strerror(errno));
        return NULL;
    }

    fd = open(filename, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "Error: unable to open file %s, %s\n", 
                filename, strerror(errno));
        return NULL;
    }

    data = calloc(1, filesize);
    amount = 0;
    while (amount < filesize) {
        n = read(fd, &data[amount], filesize - amount);
        if (n > 0)
            amount += n;
        else
            break;
    }
    close(fd);
    return data;
}

/* Read the given template.html file, and parse it into an
 * HtmlTemplate.  Return NULL on error.
 */
static HtmlTemplate* read_template(char *filename)
{
    char *buf;
    int len;
    int prev, offset, index, in_variable, n;
    HtmlTemplate *result;

    buf = read_file(filename);
    if (buf == NULL) {
        return NULL;
    }
    len = strlen(buf);

    result = (HtmlTemplate *) calloc(1, sizeof(HtmlTemplate));

    in_variable = 0;
    prev = offset = index = 0;

    while (offset < len && index < MAX_TEMPLATE_TOKENS) {
        prev = offset;
        offset = findstring(buf, prev+1, len, "$");
        if (offset == -1) {
            n = len - prev;
            result->tokens[index].value = &buf[prev];
            result->tokens[index].len = n;
            index++;
            result->count = index;
            return result;
        }

        if (in_variable == 0) {
            n = offset - prev;
            result->tokens[index].value = &buf[prev];
            result->tokens[index].len = n;
            in_variable = 1;
            index++;
        }
        else {
            offset += 1; 
            n = offset - prev;
            result->tokens[index].value = &buf[prev];
            result->tokens[index].len = n;
            in_variable = 0;
            index++;
        }
    }
    result->count = index;
    return result;
}

static void copy_template(HtmlTemplate *orig, HtmlTemplate *copy)
{
    memcpy(copy, orig, sizeof(HtmlTemplate));
}

/* Perform a search-and-replace on the HtmlTemplate structure.  Replace
 * strings containing name with the given value.
 */
static void replace(HtmlTemplate *template, char *orig, char *newvalue)
{
    int i;

    for (i = 0; i < template->count; i++) {
        if (template->tokens[i].len == strlen(orig) &&
                strncmp(template->tokens[i].value, orig, strlen(orig)) == 0) {
            template->tokens[i].value = newvalue;
            template->tokens[i].len = strlen(newvalue);
        }
    }
}

/* Return a formatted string for the given numerical size */
static char* formatsize(double size) {
    char s[64];

    if (size <= 0)
        sprintf(s, "0 KB");
    else if (size >= 0 && size < 1024 * 1024)
        sprintf(s, "%5.2f KB", size / 1024.0);
    else if (size >= 1024*1024 && size < 1024*1024*1024)
        sprintf(s, "%5.2f MB", size / (1024.0*1024));
    else
        sprintf(s, "%7.2f GB", size / (1024.0*1024*1024));

    return stringdup(s);
}

/* Return a formatted string for the given size.
 * Also, show the percent of the original size.
 */
static char* percentsize(double size, double orig) {
    char s[64];
    int percent;

    if (orig == 0)
        orig = 1;
    percent = (int)(100 * size / orig);

    if (size <= 0)
        sprintf(s, "0 KB (%d%%)", percent);
    else if (size >= 0 && size < 1024 * 1024)
        sprintf(s, "%5.2f KB (%d%%)", size / 1024.0, percent);
    else if (size >= 1024*1024 && size < 1024*1024*1024)
        sprintf(s, "%5.2f MB (%d%%)", size / (1024.0*1024), percent);
    else
        sprintf(s, "%7.2f GB (%d%%)", size / (1024.0*1024*1024), percent);

    return stringdup(s);
}

static char* formatint(double d) {
    char s[64];
    sprintf(s, "%.0f", d);
    return stringdup(s);
}

static char* percentint(double d, double orig) {
    char s[64];
    int percent;

    if (orig == 0)
        orig = 1;
    percent = (int)(100 * d / orig);
    sprintf(s, "%.0f (%d%%)", d, percent);
    return stringdup(s);
}



/* For all the year/month dates that have log statistics, store
 * the month and year in the given string buf.  maxlen is the maximum
 * size of the buffer.  This function is used to generate the statsindex.html
 * page.
 */
static void get_all_dates(Logstats *logstats, char *buf, int maxlen)
{
    int i, j, len;
    char msg[256];

    len = 0;
    for (i = MAX_YEAR-1; i >= 2000; i--) {
        for (j = 12; j >= 1; j--) {
            if (logstats->monthstats[i-2000][j].isused) {
                sprintf(msg, 
                        "\n<a href=\"%.4d%.2dstats.html\">%s %.4d</a><br>\n", 
                        i, j, monthstrings[j], i);
                memcpy(&buf[len], msg, strlen(msg));
                len += strlen(msg);
                assert(len < maxlen);
            }
        }
    }
    buf[len] = 0;
}

/* Given all the log statistics, generate the statsindex.html page,
 * and save it into the statsdir directory.
 */
static int write_statsindex(Logstats *logstats, char *statsdir)
{
    char indexfile[256];
    char alldates[8192];
    const char *html;
    int fd;

    sprintf(indexfile, "%s/statsindex.html", statsdir);
    fd = open(indexfile, O_WRONLY | O_CREAT, 0644);
    if (fd == -1) {
        fprintf(stderr, "Error: unable to create file %s: %s\n", 
                indexfile, strerror(errno));
        return -1;
    }

    html = "\
<html>\n\
<head><title>ViCompress Log Statistics</title></head>\n\
<body>\n\
<font face=\"verdana\" size=\"2\">\n\
<p>\n\
<img src=\"visolvelogo.gif\"></img>\n\
<p>\n\
<b>ViCompress Log Statistics</b>\n\
<p>\n\
<table cellspacing=0 cellpadding=2 border=0>\n\
  <tr>\n\
    <td bgcolor=\"#CCCCEE\" nowrap>\n\
      <font face=\"verdana\" size=\"2\">\n\
      <b>&nbsp;Select Date &nbsp;</b>\n\
      </font>\n\
      <table border=0 cellpadding=4 cellspacing=0 width=\"100%\">\n\
        <tr>\n\
          <td bgcolor=\"FFFFFF\" nowrap>\n\
            <font face=\"verdana\" size=\"2\">\n";

    write_data(fd, html, strlen(html));
    get_all_dates(logstats, alldates, sizeof(alldates));
    write_data(fd, alldates, strlen(alldates));

    html = "\n\
            </font>\n\
          </td>\n\
        </tr>\n\
      </table>\n\
    </td>\n\
  </tr>\n\
</table>\n\
<p>&nbsp; \n\
</body></html>\n";

    write_data(fd, html, strlen(html));
 
    close(fd);
    return 0;
}

/* Given the parse template.html file, the log statistics, and the 
 * directory which contains HTML stats files, generating an HTML
 * file for each month that has statistics.  Save the file as
 * <year><month>stats.html (e.g. 200601stats.html) under thst statsdir
 * directory.  Only create the HTML file if the statistics have
 * changed since the last logstats.txt file.
 */
static int write_html_stats(HtmlTemplate *temp, Logstats *logstats, 
                            char *statsdir)
{
    int year, month, n;
    Monthstats *m = NULL;
    char htmlfile[256];
    char statschanged = 0;

    for (year = 2000; year < MAX_YEAR; year++) {
        for (month = 1; month <= 12; month++) {
            m = &logstats->monthstats[year - 2000][month];
            sprintf(htmlfile, "%s/%.4d%.2dstats.html", 
                    statsdir, m->year, m->month);
            if (m->isused && (m->haschanged || !file_exists(htmlfile))) {
                statschanged = 1;
                n = write_html_month_stats(temp, m, htmlfile);
                if (n != 0) {
                    return -1;
                }
            }
        }
    }

    if (statschanged) {
        n = write_statsindex(logstats, statsdir);
        if (n != 0) {
            return -1;
        }
    }
    return 0;
}

/* Write the HTML stats file for a given month and save it to htmlfile.
 * Create a copy of the parsed template.html file, and save it in
 * templatecopy.  Then, do a search-and-replace of all the variables
 * (like $month_beforegzip$), updating the strings in the templatecopy.
 * Finally, write all the templatecopy strings to the htmlfile.
 */
static int write_html_month_stats(HtmlTemplate *temp, Monthstats *m, 
                                  char *htmlfile)
{
    int fd, i;
    double beforegzip, aftergzip, beforecache, aftercache, 
         maxsize, height, totalreq, gzippedreq, cachedreq;
    int daysinmonth;
    int dayweek_in_month[8];
    char tmp[64];
    HtmlTemplate templatecopy;
    int day, dayweek, hour;

    height = 150;

    stringmemory_reset();

    assert(m->year >= 2000);
    assert(m->month >= 1 && m->month <= 12);

    daysinmonth = get_days_in_month(m->year, m->month);
    get_dayweek_in_month(dayweek_in_month, m->year, m->month);

    copy_template(temp, &templatecopy);

    beforegzip  = m->month_size[NON_GZIP] + m->month_size[ORIG];
    aftergzip   = m->month_size[NON_GZIP] + m->month_size[GZIP];
    beforecache = m->month_size[MISS] + m->month_size[HIT];
    aftercache  = m->month_size[MISS];
    totalreq    = m->month_count[NON_GZIP] + m->month_count[GZIP];
    gzippedreq  = m->month_count[GZIP];
    cachedreq   = m->month_count[HIT]; 

    replace(&templatecopy, "$month_beforegzip$", formatsize(beforegzip));
    replace(&templatecopy, "$month_aftergzip$", percentsize(aftergzip, beforegzip));
    replace(&templatecopy, "$month_beforecache$", formatsize(beforecache));
    replace(&templatecopy, "$month_aftercache$", percentsize(aftercache, beforecache));
    replace(&templatecopy, "$month_total_count$", formatint(totalreq));
    replace(&templatecopy, "$month_gzipped_count$", percentint(gzippedreq, totalreq));
    replace(&templatecopy, "$month_cached_count$", percentint(cachedreq, totalreq));

    maxsize = beforegzip;

    replace(&templatecopy, "$month_beforegzip_height$", 
            formatint( (beforegzip / maxsize) * height + 1));
    replace(&templatecopy, "$month_aftergzip_height$", 
            formatint( (aftergzip / maxsize) * height + 1));
    replace(&templatecopy, "$month_beforecache_height$", 
            formatint( (beforecache / maxsize) * height + 1));
    replace(&templatecopy, "$month_aftercache_height$", 
            formatint( (aftercache / maxsize) * height + 1));


    maxsize = 0;
    for (day = 1; day <= 31; day++) {
        beforegzip = m->day_size[day][NON_GZIP] + m->day_size[day][ORIG];
        if (maxsize < beforegzip)
            maxsize = beforegzip;
    }

    for (day = 1; day <= 31; day++) {
        beforegzip  = m->day_size[day][NON_GZIP] + m->day_size[day][ORIG];
        aftergzip   = m->day_size[day][NON_GZIP] + m->day_size[day][GZIP];
        beforecache = m->day_size[day][MISS] + m->day_size[day][HIT];
        aftercache  = m->day_size[day][MISS];
        totalreq    = m->day_count[day][NON_GZIP] + m->day_count[day][GZIP];
        gzippedreq  = m->day_count[day][GZIP];
        cachedreq   = m->day_count[day][HIT];

        sprintf(tmp, "$day_beforegzip%d$", day);
        replace(&templatecopy, tmp, formatsize(beforegzip));
        sprintf(tmp, "$day_aftergzip%d$", day);
        replace(&templatecopy, tmp, percentsize(aftergzip, beforegzip));
        sprintf(tmp, "$day_beforecache%d$", day);
        replace(&templatecopy, tmp, formatsize(beforecache));
        sprintf(tmp, "$day_aftercache%d$", day);
        replace(&templatecopy, tmp, percentsize(aftercache, beforecache));
        sprintf(tmp, "$day_total_count%d$", day);
        replace(&templatecopy, tmp, formatint(totalreq));
        sprintf(tmp, "$day_gzipped_count%d$", day);
        replace(&templatecopy, tmp, percentint(gzippedreq, totalreq));
        sprintf(tmp, "$day_cached_count%d$", day);
        replace(&templatecopy, tmp, percentint(cachedreq, totalreq));

        sprintf(tmp, "$day_beforegzip_height%d$", day);
        replace(&templatecopy, tmp, formatint((beforegzip / maxsize) * height + 1));
        sprintf(tmp, "$day_aftergzip_height%d$", day);
        replace(&templatecopy, tmp, formatint((aftergzip / maxsize) * height + 1));
        sprintf(tmp, "$day_beforecache_height%d$", day);
        replace(&templatecopy, tmp, formatint((beforecache / maxsize) * height + 1));
        sprintf(tmp, "$day_aftercache_height%d$", day);
        replace(&templatecopy, tmp, formatint((aftercache / maxsize) * height + 1));

    }

    maxsize = 0;
    for (dayweek = 1; dayweek <= 7; dayweek++) {
        int num = dayweek_in_month[dayweek];
        beforegzip = (m->dayweek_size[dayweek][NON_GZIP] + 
                      m->dayweek_size[dayweek][ORIG]) / num;
        if (maxsize < beforegzip)
            maxsize = beforegzip;
    }

    for (dayweek = 1; dayweek <= 7; dayweek++) {
        int num = dayweek_in_month[dayweek];

        beforegzip  = (m->dayweek_size[dayweek][NON_GZIP] + 
                       m->dayweek_size[dayweek][ORIG]) / num;
        aftergzip   = (m->dayweek_size[dayweek][NON_GZIP] + 
                       m->dayweek_size[dayweek][GZIP]) / num;
        beforecache = (m->dayweek_size[dayweek][MISS] + 
                       m->dayweek_size[dayweek][HIT]) / num;
        aftercache  =  m->dayweek_size[dayweek][MISS] / num;
        totalreq    = (m->dayweek_count[dayweek][NON_GZIP] + 
                       m->dayweek_count[dayweek][GZIP])/ num;
        gzippedreq  =  m->dayweek_count[dayweek][GZIP] / num;
        cachedreq   =  m->dayweek_count[dayweek][HIT]  / num;

        sprintf(tmp, "$dayweek_beforegzip%d$", dayweek);
        replace(&templatecopy, tmp, formatsize(beforegzip));
        sprintf(tmp, "$dayweek_aftergzip%d$", dayweek);
        replace(&templatecopy, tmp, percentsize(aftergzip, beforegzip));
        sprintf(tmp, "$dayweek_beforecache%d$", dayweek);
        replace(&templatecopy, tmp, formatsize(beforecache));
        sprintf(tmp, "$dayweek_aftercache%d$", dayweek);
        replace(&templatecopy, tmp, percentsize(aftercache, beforecache));
        sprintf(tmp, "$dayweek_total_count%d$", dayweek);
        replace(&templatecopy, tmp, formatint(totalreq));
        sprintf(tmp, "$dayweek_gzipped_count%d$", dayweek);
        replace(&templatecopy, tmp, percentint(gzippedreq, totalreq));
        sprintf(tmp, "$dayweek_cached_count%d$", dayweek);
        replace(&templatecopy, tmp, percentint(cachedreq, totalreq));


        sprintf(tmp, "$dayweek_beforegzip_height%d$", dayweek);
        replace(&templatecopy, tmp, formatint((beforegzip / maxsize) * height + 1));
        sprintf(tmp, "$dayweek_aftergzip_height%d$", dayweek);
        replace(&templatecopy, tmp, formatint((aftergzip / maxsize) * height + 1));
        sprintf(tmp, "$dayweek_beforecache_height%d$", dayweek);
        replace(&templatecopy, tmp, formatint((beforecache / maxsize) * height + 1));
        sprintf(tmp, "$dayweek_aftercache_height%d$", dayweek);
        replace(&templatecopy, tmp, formatint((aftercache / maxsize) * height + 1));
    }

    maxsize = 0;
    for (hour = 0; hour < 24; hour++) {
        beforegzip = (m->hour_size[hour][NON_GZIP] + 
                      m->hour_size[hour][ORIG]) / daysinmonth;
        if (maxsize < beforegzip) {
            maxsize = beforegzip;
        }
    }

    for (hour = 0; hour < 24; hour++) {
        beforegzip  = (m->hour_size[hour][NON_GZIP] + 
                       m->hour_size[hour][ORIG]) / daysinmonth;
        aftergzip   = (m->hour_size[hour][NON_GZIP] + 
                       m->hour_size[hour][GZIP]) / daysinmonth;
        beforecache = (m->hour_size[hour][MISS] + 
                       m->hour_size[hour][HIT]) / daysinmonth;
        aftercache  =  m->hour_size[hour][MISS] / daysinmonth;
        totalreq    = (m->hour_count[hour][NON_GZIP] + 
                       m->hour_count[hour][GZIP]) / daysinmonth;
        gzippedreq  =  m->hour_count[hour][GZIP]  / daysinmonth;
        cachedreq   =  m->hour_count[hour][HIT]   / daysinmonth;

        sprintf(tmp, "$hour_beforegzip%d$", hour);
        replace(&templatecopy, tmp, formatsize(beforegzip));
        sprintf(tmp, "$hour_aftergzip%d$", hour);
        replace(&templatecopy, tmp, percentsize(aftergzip, beforegzip));
        sprintf(tmp, "$hour_beforecache%d$", hour);
        replace(&templatecopy, tmp, formatsize(beforecache));
        sprintf(tmp, "$hour_aftercache%d$", hour);
        replace(&templatecopy, tmp, percentsize(aftercache, beforecache));
        sprintf(tmp, "$hour_total_count%d$", hour);
        replace(&templatecopy, tmp, formatint(totalreq));
        sprintf(tmp, "$hour_gzipped_count%d$", hour);
        replace(&templatecopy, tmp, percentint(gzippedreq, totalreq));
        sprintf(tmp, "$hour_cached_count%d$", hour);
        replace(&templatecopy, tmp, percentint(cachedreq, totalreq));


        sprintf(tmp, "$hour_beforegzip_height%d$", hour);
        replace(&templatecopy, tmp, formatint((beforegzip / maxsize) * height + 1));
        sprintf(tmp, "$hour_aftergzip_height%d$", hour);
        replace(&templatecopy, tmp, formatint((aftergzip / maxsize) * height + 1));
        sprintf(tmp, "$hour_beforecache_height%d$", hour);
        replace(&templatecopy, tmp, formatint((beforecache / maxsize) * height + 1));
        sprintf(tmp, "$hour_aftercache_height%d$", hour);
        replace(&templatecopy, tmp, formatint((aftercache / maxsize) * height + 1));
    }

    sprintf(tmp, "%.4d", m->year);
    replace(&templatecopy, "$year$", stringdup(tmp));
    replace(&templatecopy, "$monthname$", stringdup(monthstrings[m->month]));

    fd = open(htmlfile, O_WRONLY | O_CREAT, 0644);
    if (fd == -1) {
        fprintf(stderr, "Error: unable to create file %s: %s\n", 
                        htmlfile, strerror(errno));
        return -1;
    }
    for (i = 0; i < templatecopy.count; i++) {
        write_data(fd, templatecopy.tokens[i].value, templatecopy.tokens[i].len);
    }
    close(fd);
    return 0;
}


int main(int argc, char **argv) {
    Logstats* logstats;
    HtmlTemplate *template;
    const char *accesslog;
     
    char statsdir[256];
    char templatefile[256];
    char logstatsfile[256];
    char oldaccesslog[256];
    int i, n, interval, updated, inbackground;


    if (argc < 3) {
        printf("Usage: update_log_stats <accesslog> <logstats directory> [<interval>]\n");
        printf("       Create the HTML log statistics from the accesslog data\n");
        printf("       accesslog  -- File path of the accesslog\n");
        printf("       logstats   -- Directory to write the log statistics\n");
        printf("       interval   -- Interval (in hours) to periodically generate the stats\n");
        exit(1);
    }

    accesslog = argv[1];
    n = strlen(argv[2]);
    if ( argv[2][n-1] == '/' )
        n -= 1;
    if (argc == 4)
        interval = atoi(argv[3]) * 3600; 
    else
        interval = 0;

    inbackground = 0;

    stringmemory_init();

    memset(statsdir, 0, sizeof(statsdir));
    memset(logstatsfile, 0, sizeof(logstatsfile));
    memset(oldaccesslog, 0, sizeof(oldaccesslog));
    strncpy(statsdir, argv[2], n);
    sprintf(logstatsfile, "%s/logstats.txt", statsdir);
    logstats = (Logstats *) calloc(1, sizeof(Logstats));

    sprintf(templatefile, "%s/template.html", statsdir);
    template = read_template(templatefile);
    if (template == NULL) {
        exit(1);
    }

    while (1) {
        n = parse_logstatistics(logstats, logstatsfile);
        if (n == -1) {
            exit(-1);
        }
        updated = 0;
        for (i = 2; i >= 1; i--) {
            sprintf(oldaccesslog, "%s.%d", accesslog, i);
            if (file_exists(oldaccesslog) && 
                    is_accesslog_new(&logstats->last, oldaccesslog)) {
                updated = 1;
                n = parse_accesslog(logstats, oldaccesslog);
                if (n == -1) {
                    exit(-1);
                } 
            }
        }

        if (file_exists(accesslog) && 
            is_accesslog_new(&logstats->last, accesslog)) {
            updated = 1;
            n = parse_accesslog(logstats, accesslog);
            if (n == -1) {
                exit(-1);
            }
        }
 
        if (updated) {
            /* Don't interrupt writing the log statistics */
            signal(SIGINT, handle_signal);
            signal(SIGTERM, handle_signal);
            n = write_logstatistics(logstats, logstatsfile);
            if (n == -1) {
                exit(-1);
            }
            signal(SIGINT, SIG_DFL);
            signal(SIGTERM, SIG_DFL);
        }

        n = write_html_stats(template, logstats, statsdir); 
        if (n == -1) {
            exit(-1);
        }

        if (interval == 0 || gotsignal == 1) {
            return 0;
        }
        else {
            /* If process is not in background yet, make it so. */
            if (inbackground == 0) {
                if (fork() > 0) {
                    exit(0);
                }
                redirect_stdio_to_devnull();
                inbackground = 1;
            }
            (void) sleep(interval);
        }
    }

    return 0;
}


