/*
 * Copyright (c) 2003-2010 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/* This file contains the main vicompress code.
 *
 * Vicompress manages the socket connections using non-blocking sockets and
 * polling [sockets.c], as opposed to one thread/process per connection.  When
 * a socket call is made (sockets_read/write/connect) a callback function
 * is invoked when the socket operation completes.
 *
 * The functions http_parse_request() and http_parse_reply() [http.c] are
 * used the parse raw the http request/reply from the raw byte buffers.
 *
 * Vicompress reads data from origin webservers into reference-counted Buffers
 * [buffer.c].  The Buffers are then written to the client, and written 
 * to the in-memory cache.  Vicompress reads from the origin server and writes
 * to the client in parallel (conn_read_server_body_cb and conn_write_server_body).
 *
 * The MemoryCache [cache.c] is a hashtable for storing/retrieving
 * downloaded images/pages/etc.  The MemoryCache uses an LRU (least-recently
 * used) linked-list for deleting older items.  The cache is never written
 * to disk.
 *
 * Vicompress can gzip text and html replies on-the-fly [gzipencoder.c].
 * As Buffers are read from the origin webserver, they are passed to the
 * gzip encoder using gzip_add().  The gzip encoder uses multiple threads
 * to perform the actual gzip compression, thereby scaling to multiple CPUs.
 *
 * Vicompress logs all requests to an accesslog [log.c], using either the
 * Apache or Squid log format.  The log entries are written to disk in a 
 * separate thread, so that the disk I/O does not block the main process.
 * 
 * The pathflow of a single request (non-cached and non-gzipped) is shown below:
 *
 * main()
 * vicompress_start()
 * config_init()                  - Read in the vicompress.conf settings
 * start_listening()              - Start listening for requests
 * sockets_checkevents()
 * conn_accept_cb()               - Accept a new client connection
 * conn_read_request_cb()         - Read in the http request data
 * http_parse_request()           - Parse the http request
 * conn_read_request_done()       - Determine if the request is in the cache or not
 * conn_read_server_reply()       - Lookup the hostname to connect to
 * conn_dnslookup_cb() 
 * conn_connect_to_server()       - Connect to the destination webserver
 * conn_connect_cb()              - Write the http request header to the webserver
 * conn_write_request_cb()
 * conn_read_reply_cb()           - Read the http reply from the webserver
 * http_parse_reply()             - Parse the http reply
 * conn_read_server_reply_done()  - Determine whether to cache, gzip, and/or
 *                                  chunk-encode the http reply body.
 * conn_write_server_reply()      - Write the http reply header to the client
 * conn_write_reply_cb()
 * conn_read_server_body_cb()     - Read some data from the webserver
 * conn_write_server_body()       - Write the data to the client
 * conn_write_server_body_cb()
 * conn_write_server_reply_done() - The reply has been completely written.
 * logaccess()                    - Write an entry to the accesslog
 * conn_close()                   - Close the client and server connection
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/resource.h>
#include <libgen.h>
#include "vicompress.h"
#include "win32.h"

/* Buffer pool sizes */
#define SMALL_URL_SIZE            64
#define LARGE_URL_SIZE           128
#define REPLY_SIZE               512
#define REQUEST_SIZE            1024
#define BODY_SIZE               4096
#define MAX_REQUEST_SIZE        8192

/* Connection status codes */
#define CONN_SUCCESS               0
#define CONN_BAD_REQUEST           1
#define CONN_BAD_REPLY             2
#define CONN_CLIENT_CLOSED         3
#define CONN_SERVER_CLOSED         4
#define CONN_CONNECT_ERROR         5
#define CONN_CLIENT_READ_ERROR     6
#define CONN_CLIENT_WRITE_ERROR    7
#define CONN_SERVER_READ_ERROR     8
#define CONN_SERVER_WRITE_ERROR    9
#define CONN_DNS_ERROR            10

/* HTTP Error status codes */
#define ERROR_BAD_REQUEST    400
#define ERROR_NOT_FOUND      404
#define ERROR_NOT_AVAILABLE  503

#define CHECK_WEBSERVER_INTERVAL 180  /* 3 minutes */
#define LOGSTATS_INTERVAL        3600 /* 1 hour */

/* Socket callback functions (cb = callback) */
static AcceptCallback  conn_accept_cb;
static DnsCallback     conn_dnslookup_cb;
static ConnectCallback conn_connect_cb;
static ReadCallback    conn_read_request_cb;
static ReadCallback    conn_read_requestbody_cb;
static WriteCallback   conn_write_request_cb;
static ReadCallback    conn_read_100continue_cb;
static WriteCallback   conn_write_100continue_cb;
static ReadCallback    conn_read_reply_cb;
static WriteCallback   conn_write_reply_cb;
static ReadCallback    conn_read_server_body_cb;
static WriteCallback   conn_write_server_body_cb;
static WriteCallback   conn_write_gzippedbody_cb;
static WriteCallback   conn_write_cached_reply_cb;
static WriteCallback   conn_write_cached_body_cb;
static WriteCallback   conn_write_error_reply_cb;
static WriteCallback   conn_write_cacheurls_cb;
static WriteCallback   conn_write_cacheurlsbody_cb;


static void conn_close(Connection *conn);
static void conn_read_cached_body(Connection *conn);
static void conn_read_cacheurls(CacheUrls *list, Connection *conn, Buffer *out);
static void conn_connect_to_server(Connection *conn);
static void conn_write_server_body(Connection *conn, Buffer *buffer);
static void conn_write_gzippedbody(Connection *conn);
static void conn_continue_write_gzippedbody(Connection *conn);
static void conn_keepalive_read(Connection *conn);
static void conn_write_cached_reply(Connection *conn, CachedItem *item);
static void conn_write_server_reply(Connection *conn, GzipEncoder *gzip, CachedItem *item);
static void conn_write_direct_reply(Connection *conn);
static void conn_check_finished_gzip(GzipPool *gzippool);
static void conn_write_error_reply(Connection *conn, int status);
static void conn_write_cacheurls(Connection *conn, CacheUrls *list);
static const char *conn_status_string(int status);

static void conn_read_request_done(Connection *conn, int status);
static void conn_read_server_reply_done(Connection *conn, int status);
static void conn_write_cached_reply_done(Connection *conn, int status);
static void conn_write_server_reply_done(Connection *conn, int status);
static void conn_write_error_reply_done(Connection *conn, int status);
static void conn_write_cacheurls_done(CacheUrls *list, Connection *conn, int status);


/* The global vicompress server */
static Vicompress *vicompress;

#ifdef WIN32
/* 1 if vicompress is running as a Win32 service, else 0 */
extern int is_service;
#endif

/* Redirect stdin, stdout, stderr to /dev/null */
static void redirect_stdio_to_devnull()
{
#ifndef WIN32
    int fd;
    fd = open("/dev/null", O_RDWR, 0);
    if (fd < 0)
        return;
    dup2(fd, 0);
    dup2(fd, 1);
    dup2(fd, 2);
    close(fd);
#endif
}

/* Write the process id of vicompress to /var/run/vicompress.pid.
 * Change the owner of the file to Config.user.
 */
static void create_pidfile()
{
#ifndef WIN32
    int fd, error;
    char buf[32];

    fd = open("/var/run/vicompress.pid", O_WRONLY | O_CREAT | O_TRUNC, 0644);
    error = errno;
    if (fd < 0) {
        /* This is not a fatal error, hence only a warning message. */
        /*
        fprintf(stderr, "Warning: unable to create /var/run/vicompress.pid: %s\n",
                strerror(error));
        */
        return;
    }
    snprintf(buf, 32, "%d\n", getpid());
    (void)write(fd, buf, strlen(buf));
    close(fd);
    (void)chown("/var/run/vicompress.pid", Config.user, Config.group);
#endif
}

static void delete_pidfile()
{
#ifndef WIN32
    unlink("/var/run/vicompress.pid");
#endif
}

static int ignore_errno(int error)
{
    switch (error) {
    case EINPROGRESS:
#ifdef WIN32
    case EWOULDBLOCK:
#endif
    case EAGAIN:
    case EALREADY:
    case EINTR:
        return 1;
    default:
        return 0;
    }
}


/* Close the given connection (both client and server),
 * and free any memory used, including the http request,
 * http reply, and read/write buffers.
 */
static void conn_close(Connection *conn)
{
    Sockets *s;

    s = vicompress->sockets;
    buffer_unref(conn->serverread); conn->serverread = NULL;
    buffer_unref(conn->clientwrite); conn->clientwrite = NULL;
    buffer_unref(conn->gzipout); conn->gzipout = NULL;
    buffer_unref(conn->gzip_dechunked); conn->gzip_dechunked = NULL;
    http_free_request(&conn->request);
    http_free_reply(&conn->reply);
    if (conn->serverfd != 0) {
        sockets_close(s, conn->serverfd);
        conn->serverfd = 0;
    }
    if (conn->clientfd != 0) {
        sockets_close(s, conn->clientfd);
        conn->clientfd = 0;
    }
    (void)memset(conn, 0, sizeof(Connection));
}


/* Invoked when a new client is accepted.  Allocate a new Connection
 * structure for this client.  Allocate a buffer to read in the
 * http request.  Call conn_read_request_cb when the http request
 * is read.
 *
 * clientfd - The client socket
 * addr     - The client IP address
 * data     - not used
 */
static void conn_accept_cb(int clientfd, sockaddr_all addr, void *data)
{
    Connection *conn;
    Buffer *b;

    conn = (Connection *) &(vicompress->connpool[clientfd]);
    memset(conn, 0, sizeof(Connection));
    conn->clientfd = clientfd;
    conn->clientip = addr;
    DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, "New client accepted.");

    b = buffer_alloc(&vicompress->reqpool);
    http_init_request(&conn->request, b, &conn->parsedata);

    if (Config.logformat == LOG_FORMAT_SQUID) {
        gettimeofday(&conn->start, NULL);
    }
    sockets_read(vicompress->sockets, conn->clientfd, conn->request.buffer,
                 conn_read_request_cb, conn);

}

/* Start listening on the configured IPs/ports.  Call
 * conn_accept_cb when a new client is accepted.
 * Return 0 on success, or -1 if unable to listen on
 * the configured address.
 */
static int start_listening(Vicompress *v)
{
    int i, fd, ret, option;
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
    int len;
#else
    socklen_t len;
#endif
    sockaddr_all *addr;

    for (i = 0; i < Config.num_listen; i++) {
        addr = &(Config.listenaddr[i]);

        fd = sockets_new(v->sockets, addr->ipv4.sin_family);
        if (fd == -1) {
            fprintf(stderr, "ERROR: Unable to create listening socket, %s\n",
                    strerror(errno));
            return -1;
        }

        option = 1;
        ret = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &option, sizeof(int));
        len = sizeof(struct sockaddr_in);
#ifdef HAVE_IPV6
        if (addr->ipv4.sin_family == AF_INET6) {
            len = sizeof(struct sockaddr_in6);
        }
#endif
        ret = bind(fd, (struct sockaddr *) addr, len);
        if (ret != 0) {
            fprintf(stderr, "ERROR: Cannot bind to %s, %s\n", 
                    sockets_to_string(addr), strerror(errno));
            return -1;
        }

        ret = listen(fd, 4096);
        sockets_listen(v->sockets, fd, conn_accept_cb, v);
    }
    return 0;
}

/* A http request/reply has just ended. We have a "keep-alive"
 * connection with both the client and server.  Clear all the 
 * connection data except for:
 * - The client socket
 * - The server socket
 * - The client ip
 * - The server ip
 * - The http request buffer
 * 
 * If the request buffer has any extra data, it is part of the
 * next http request, so copy it over.  Then wait for the next
 * http request.
 */
static void conn_keepalive_read(Connection *conn)
{
    Sockets *s = vicompress->sockets;
    Buffer *b;
    sockaddr_all clientip, serverip;
    int reqsize, clientfd, serverfd;

    /* If the request buffer contains more data than the first
     * http request, copy that extra data to the new request buffer.
     */
    b = buffer_alloc(&vicompress->reqpool);
    reqsize = conn->request.requestlength + conn->request.contentlength;
    if (conn->request.buffer->size > reqsize) {
        buffer_add(b, &(conn->request.buffer->buf[reqsize]), 
                   conn->request.buffer->size - reqsize);
    }

    /* Check if we're also keeping the server connection alive */
    if (conn->serverfd != 0) {
        if (conn->reply.keepalive && !conn->eof &&
            (conn->reply.has_contentlength || conn->reply.is_chunked)) {
            /* keep server connection */
        }
        else {
            sockets_close(s, conn->serverfd);
            conn->serverfd = 0;
        }
    }

    /* Save the client/server sockets and IP addresses */
    clientfd = conn->clientfd; conn->clientfd = 0;
    serverfd = conn->serverfd; conn->serverfd = 0;
    (void)memcpy(&clientip, &conn->clientip, sizeof(sockaddr_all));
    (void)memcpy(&serverip, &conn->serverip, sizeof(sockaddr_all));

    conn_close(conn);

    conn->clientfd = clientfd;
    conn->serverfd = serverfd;
    (void)memcpy(&conn->clientip, &clientip, sizeof(sockaddr_all));
    (void)memcpy(&conn->serverip, &serverip, sizeof(sockaddr_all));
    (void)memcpy(&conn->serverip_keepalive, &serverip, sizeof(sockaddr_all));

    /* Initialize the new http request */
    http_init_request(&conn->request, b, &conn->parsedata);

    if (conn->request.buffer->size > 0) {
        conn_read_request_cb(conn->clientfd, conn, conn->request.buffer,
                             0, 0);
    }
    else {
        int timeout = 4;
        if (s->numclients < 100) {
            timeout = 20;
        }
        sockets_read_timeout(s, conn->clientfd, conn->request.buffer, 
                             timeout, conn_read_request_cb, conn);
    }
}

/* Return a buffer containing the URL of the request, i.e.
 * www.hostname.com:port/urlpath
 */
static Buffer* create_url(HttpRequest *req)
{
    int size;
    Buffer *b;
    
    size = req->host.size + req->path.size;
    if (req->port != 80)
        size += 7;
    if (size < SMALL_URL_SIZE && !bufferpool_is_empty(&vicompress->smallurlpool))
        b = buffer_alloc(&vicompress->smallurlpool);
    else
        b = buffer_alloc(&vicompress->largeurlpool);
    buffer_add(b, req->host.buf, req->host.size);
    if (req->port != 80) {
        buffer_add(b, ":", 1);
        buffer_add_int(b, req->port);
    }
    buffer_add(b, req->path.buf, req->path.size);
    return b;
}

/* Callback invoked when there is data to read on a client socket.
 * Read and parse an HTTP request header.  When a full request
 * header is read, call conn_read_request_done().
 *
 * fd - The client socket
 * data - The client Connection structure
 * buffer - The data (http request) read from the socket
 * eof - True if the client socket recv() returned end-of-file (EOF)
 */
static void conn_read_request_cb(int fd, void *data, Buffer *buffer, 
                                 int error, int eof)
{
    Connection *conn = data;
    Sockets *s;
    int result;

    s = vicompress->sockets;

    /* For keep-alive client connections, the "start" time is when the
     * request is received.
     */
    if (Config.logformat == LOG_FORMAT_SQUID && conn->start.tv_sec == 0) {
        gettimeofday(&(conn->start), NULL);
    }

    if (error) {
        conn->error = error;
        conn_read_request_done(conn, CONN_CLIENT_READ_ERROR);
        return;
    }

    assert(buffer == conn->request.buffer);
    result = http_parse_request(&conn->request, &conn->parsedata);
    if (result == PARSE_ERROR) {
        conn_read_request_done(conn, CONN_BAD_REQUEST); 
        return;
    }
    else if (result == PARSE_PARTIAL && eof) {
        conn_read_request_done(conn, CONN_CLIENT_CLOSED);
        return;
    }
    else if ((result == PARSE_PARTIAL) && (buffer->size == buffer->capacity)) {
        /* If the request is larger than REQUEST_SIZE (1K), expand 
         * the buffer to BODY_SIZE (4K), then to MAX_REQUEST_SIZE (8K).
         * If its more than that, return an error.
         */
        if (buffer->capacity < BODY_SIZE) {
            buffer_grow(buffer, BODY_SIZE);
            sockets_read(s, fd, buffer, conn_read_request_cb, conn);
            return;
        }
        else if (buffer->capacity < MAX_REQUEST_SIZE) {
            buffer_grow(buffer, MAX_REQUEST_SIZE);
            sockets_read(s, fd, buffer, conn_read_request_cb, conn);
            return;
        }
        else {
            conn_read_request_done(conn, CONN_BAD_REQUEST);
            return;
        }
    }
    else if (result == PARSE_PARTIAL) {
        sockets_read(s, fd, buffer, conn_read_request_cb, conn);
        return;
    }
    else if (result == PARSE_DONE) {
        conn->request.url = create_url(&conn->request);
        conn_read_request_done(conn, CONN_SUCCESS);
        return;
    }
}

/* Return 1 if this is ip is 127.0.0.1 or ::1, else 0. */
static int is_localhost(sockaddr_all *addr)
{
    unsigned char *ip;

    if (addr->ipv4.sin_family == AF_INET) {
        ip = (unsigned char *) &(addr->ipv4.sin_addr);
        if (ip[0] == 127 && ip[1] == 0 && ip[2] == 0 && ip[3] == 1)
            return 1;
        else
            return 0;
    }
#ifdef HAVE_IPV6
    else if (addr->ipv4.sin_family == AF_INET6) {
        int i;
        ip = (unsigned char *) &(addr->ipv6.sin6_addr);
        for (i = 0; i < 15; i++) {
            if (ip[i] != 0) {
                return 0;
            }
        }
        if (ip[15] == 1)
            return 1;
        else
            return 0;
    } 
#endif
    else {
        return 0;
    }
}

/* Connect to the origin webserver and read the http reply header.
 * First, determine whether 
 * - We are reverse proxying a backend webserver, or
 * - The URL contains an IP address, or
 * - We have to do a DNS lookup on that URL
 * Once we have the origin server IP address, call conn_connect_to_server().
 *
 * conn - The client connection we are servicing
 */
static void conn_read_server_reply(Connection *conn)
{
    ErrorLog *errorlog = vicompress->errorlog;
    LoadBalancer *balancer = vicompress->loadbalancer;
    Sockets *s;

    s = vicompress->sockets;

    /* If we have backend webservers, and we still have a
     * previous connection to the backend webserver, use it.
     */
    if (Config.num_webservers > 0 && conn->serverfd != 0) {
        memcpy(&conn->serverip, &conn->serverip_keepalive, sizeof(sockaddr_all));
        conn_connect_to_server(conn);
        return;
    }
    /* If we have backend webservers configured, choose a webserver
     * to connect to.
     */
    else if (Config.num_webservers > 0) {
        sockaddr_all *server;
        int id;

        id = conn->request.webserverid;
        if (!loadbalancer_is_valid(balancer, id, conn->request.host) ||
            loadbalancer_is_server_down(balancer, id) ||
            Config.enable_sessions == 0) {

            id = loadbalancer_get_id(balancer, conn->request.host);
        }
        conn->webserverid = id;
        assert(balancer == vicompress->loadbalancer);
        server = loadbalancer_getaddr(balancer, id);
        loadbalancer_connect_started(balancer, id);
        memcpy(&conn->serverip, server, sizeof(sockaddr_all));
        conn_connect_to_server(conn);
        return;
    }
    /* If the hostname in the URL is an IP address, no
     * DNS lookup is needed.  Do a direct connection.
     */
    else if (parse_ip((char *)conn->request.host.buf, conn->request.host.size,
                      &conn->serverip, conn->request.port) == 0) {

        /* Do not allow connections to localhost/127.0.0.1 unless it's a
         * configured Config.webserver.
         */
        if (is_localhost(&conn->serverip)) {
            conn->error = ECONNREFUSED;
            conn_read_server_reply_done(conn, CONN_CONNECT_ERROR);
            return;
        }
        else {
            conn_connect_to_server(conn);
            return;
        }
    }
    /* We have to do a DNS lookup before connecting to the server */
    else {
        DEBUG_PRINTF(errorlog, &conn->clientip,
                        "Looking up IP address for hostname ");
        DEBUG_PRINT(errorlog, &conn->clientip,
                       conn->request.host.buf, conn->request.host.size);

        sockets_dnslookup(s, conn->request.host, conn_dnslookup_cb, conn);
        return;
    }
}

/* Callback function invoked when a dns lookup has finished.
 * If the dns lookup succeeded, connect to the server.
 * Else, report that connection failed.
 *
 * ip - The IPv4 address
 * data - The connection
 * error - 0 if the dns lookup succeeded, else 1
 */
static void conn_dnslookup_cb(unsigned char *ip, void *data, int error)
{
    Connection *conn = data;

    if (error) {
        conn->error = error;
        conn_read_server_reply_done(conn, CONN_DNS_ERROR); 
        return;
    }
    memset(&conn->serverip, 0, sizeof(sockaddr_all));
    conn->serverip.ipv4.sin_family = AF_INET;
    memcpy(&(conn->serverip.ipv4.sin_addr.s_addr), ip, 4);
    conn->serverip.ipv4.sin_port = htons(conn->request.port);

    /* Don't allow connections to the localhost */
    if (is_localhost(&(conn->serverip)) ) {
        conn->error = ECONNREFUSED;
        conn_read_server_reply_done(conn, CONN_CONNECT_ERROR);
        return;
    }

    conn_connect_to_server(conn);
}

/* Connect to the server given by conn->serverip.
 * Invoke conn_connect_cb() when the connect action finishes.
 */
static void conn_connect_to_server(Connection *conn)
{
    int fd, ret;
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
    int len;
#else
    socklen_t len;
#endif
    char buf[8];
    Sockets *s = vicompress->sockets;
    ErrorLog *errorlog = vicompress->errorlog;

    conn->error = 0;

    /* Check if we already have a persistent server connection. */
    if (conn->serverfd != 0) {
        /* Make sure the connection is valid, by reading from it */
        ret = recv(conn->serverfd, buf, 8, 0);
        if ( !(ret == -1 && ignore_errno(errno)) ) {
            sockets_close(s, conn->serverfd);
            conn->serverfd = 0;
        }
        else {
            /* Make sure the server address of the request matches the
             * server address of the socket connection.
             */
            len = sizeof(sockaddr_all);
            ret = memcmp(&conn->serverip, &conn->serverip_keepalive, len);
            if (ret != 0) {
                sockets_close(s, conn->serverfd);
                conn->serverfd = 0;
            }
        }
    }
    if (conn->serverfd != 0) {
        /* We're already connected to the server */
        DEBUG_PRINTF(errorlog, &conn->clientip,
                     "Re-using persistent server connection to %s",
                     sockets_to_string(&conn->serverip));
        conn_connect_cb(conn->serverfd, conn, 0);
        return;
    }

    fd = sockets_new(s, conn->serverip.ipv4.sin_family);
    if (fd == -1) {
        conn->error = EMFILE;
        conn_read_server_reply_done(conn, CONN_CONNECT_ERROR);
        return;
    }

    Config.outgoingip.ipv4.sin_port = 0;
    len = sizeof(struct sockaddr_in);
#ifdef HAVE_IPV6
    if (Config.outgoingip.ipv4.sin_family == AF_INET6) {
        len = sizeof(struct sockaddr_in6);
    }
#endif

    ret = bind(fd, (struct sockaddr *) &(Config.outgoingip.ipv4), len);
    if (ret != 0) {
        conn->error = errno;
        conn_read_server_reply_done(conn, CONN_CONNECT_ERROR);
        return;
    }
    conn->serverfd = fd;

    DEBUG_PRINTF(errorlog, &conn->clientip,
                  "Connecting to server %s", 
                  sockets_to_string(&conn->serverip));
    sockets_connect(s, fd, &conn->serverip, conn_connect_cb, (void *)conn);   
}

/* Callback function invoked when we've finished connecting to a origin
 * server.  Write the http request to the origin server. 
 *
 * fd - The socket connection with the server
 * data - The Connection structure
 * error - 0 if the connection succeeded, else an errno value
 */
static void conn_connect_cb(int fd, void *data, int error)
{
    Connection *conn = data;
    Buffer *out;
    Sockets *s;
    int contentlen;

    if (error) {
        conn->error = error;

        if (Config.num_webservers == 0 && Config.enable_dns_caching) {
            /* If we failed to connect, the IP address may be incorrect.
             * Delete the IP address in the DNS lookup cache.
             */
            dns_remove(vicompress->sockets->dns, conn->request.host.buf, conn->request.host.size);
        }
        conn_read_server_reply_done(conn, CONN_CONNECT_ERROR);
        return;
    }

    conn->amountwritten =  0;
    s = vicompress->sockets;
    out = buffer_alloc(&vicompress->reqpool); 
    http_write_request(&conn->request, out, &(conn->clientip));

    /* If the request has a body, and we've read some of the
     * body already, then add the body to the "out" buffer.
     */
    contentlen = conn->request.buffer->size - conn->request.requestlength;
    if (contentlen > conn->request.contentlength) {
        contentlen = conn->request.contentlength;
    }
    if (conn->request.contentlength > 0 && contentlen > 0) {
        int offset = conn->request.requestlength;
        conn->amountwritten = contentlen;
        buffer_add(out, &conn->request.buffer->buf[offset], contentlen);
    }
    sockets_write(s, conn->serverfd, out, conn_write_request_cb, conn); 
}

/* Callback function invoked after we've finished writing the http
 * request to the origin server.  If the http request has content-length,
 * read the request body from the client.  Otherwise, start reading
 * the http reply from the server.
 *
 * fd - The socket connection to the server
 * data - The Connection structure
 * buffer - The http request we just wrote to the server
 * error - 0 if the write succeeded, else an errno value
 */
static void 
conn_write_request_cb(int fd, void *data, Buffer *buffer, int error)
{
    Connection *conn = data;
    Sockets *s = vicompress->sockets;

    if (error) {
        buffer_unref(buffer); buffer = NULL;
        conn->error = error;
        conn_read_server_reply_done(conn, CONN_SERVER_WRITE_ERROR);
        return;
    }

    if (conn->request.contentlength > 0 &&
        conn->amountwritten == 0 &&
        conn->request.expect100 == 1) {

        /* If the request has a "Expect: 100-continue" header, then
         * read the "HTTP/1.1 100 Continue" reply from the server.
         */
        buffer_reset(buffer);
        sockets_read_timeout(s, conn->serverfd, buffer, 5, 
                             conn_read_100continue_cb, conn);
    }
    else if (conn->amountwritten < conn->request.contentlength) {
        buffer_reset(buffer); 
        sockets_read(s, conn->clientfd, buffer, 
                     conn_read_requestbody_cb, conn);
    }
    else {
        /* We've written the entire request.  Now read the reply */
        Buffer *b;

        buffer_unref(buffer); buffer = NULL;
        conn->amountwritten = 0;
        b = buffer_alloc(&vicompress->replypool);
        http_init_reply(&conn->reply, b, &conn->parsedata);
        sockets_read(s, conn->serverfd, conn->reply.buffer, 
                     conn_read_reply_cb, conn);
    }
}


/* Callback function invoked after reading the "HTTP/1.1 100 Continue"
 * reply from the origin server.
 */
static void
conn_read_100continue_cb(int fd, void *data, Buffer *buffer, 
                         int error, int eof)
{
    Connection *conn;
    Sockets *s;
    int valid_reply = 0;

    conn = data;
    s = vicompress->sockets;

    valid_reply = \
        ((buffer->size >= 12) && 
         (strncmp(buffer->buf, "HTTP/1.", 7) == 0) &&
         (buffer->buf[buffer->size-1] == '\n') &&
         (buffer->buf[buffer->size-2] == '\n' || buffer->buf[buffer->size-3] == '\n') );

    /* If we got a valid reply, such as "HTTP/1.1 100 Continue", then
     * write it back to the client.
     */
    if (valid_reply) {
        sockets_write(s, conn->clientfd, buffer, conn_write_100continue_cb, conn);
    }
    else if (buffer->size > 0 || eof) {
        buffer_unref(buffer); buffer = NULL;
        conn_read_server_reply_done(conn, CONN_SERVER_READ_ERROR);
    }
    else {
        buffer_reset(buffer); 
        sockets_read(s, conn->clientfd, buffer, 
                     conn_read_requestbody_cb, conn);
    }
}

/* Callback function invoked after we've written a
 * HTTP/1.1 100 Continue reply back to the client.
 * 
 * Continue reading the request body from the client.
 */
static void
conn_write_100continue_cb(int fd, void *data, Buffer *buffer, int error)
{
    Connection *conn = data;
    Sockets *s = vicompress->sockets;

    if (error) {
        conn->error = error;
        buffer_unref(buffer); buffer = NULL;
        conn_read_server_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
        return;
    }
    buffer_reset(buffer);
    sockets_read(s, conn->clientfd, buffer, 
                 conn_read_requestbody_cb, conn);

}


/* Callback function invoked after we've read a chunk of the 
 * http request body from the client.  Write the chunk to
 * the server.
 *
 * fd - The socket connection with the client
 * data - The Connection structure
 * buffer - The http request body read from the client
 * error - 0 if the read succeeded, else an errno value
 * eof - 1 if the client closed the connection, else 0
 */
static void 
conn_read_requestbody_cb(int fd, void *data, Buffer *buffer, 
                         int error, int eof)
{
    Connection *conn;
    Sockets *s;

    conn = data;
    s = vicompress->sockets;

    if (error || eof) {
        buffer_unref(buffer); buffer = NULL;
        conn->error = error;
        conn_read_server_reply_done(conn, CONN_CLIENT_READ_ERROR);
        return;
    }

    if (buffer->size < buffer->capacity &&
        buffer->size + conn->amountwritten < conn->request.contentlength) {
        sockets_read(s, conn->clientfd, buffer, conn_read_requestbody_cb, conn);
        return;
    }
    conn->amountwritten += buffer->size;
    sockets_write(s, conn->serverfd, buffer, conn_write_request_cb, conn);
}

/* Callback function invoked after we've read some of the http reply
 * header from the origin server.  If we've read the entire http reply
 * header, call conn_read_server_reply_done(). Else, call sockets_read
 * to continue reading from the origin server.
 *
 * fd - The socket connection with the server
 * data - The Connection structure
 * buffer - The http reply header read from the origin server
 * error - 0 if the read succeeded, else an errno value
 * eof - 1 if the server closed the connection, else 0
 */
static void 
conn_read_reply_cb(int fd, void *data, Buffer *buffer, int error, int eof)
{
    Connection *conn = data;
    Sockets *s;
    int result;

    if (error) {
        conn->error = error;
        conn_read_server_reply_done(conn, CONN_SERVER_READ_ERROR);
        return;
    }
    s = vicompress->sockets;
    conn->eof = eof;

    assert(buffer == conn->reply.buffer); 
    result = http_parse_reply(&conn->reply, &conn->parsedata);
    if (result == PARSE_ERROR) {
        conn_read_server_reply_done(conn, CONN_BAD_REPLY);
        return;
    }
    else if (result == PARSE_PARTIAL && conn->eof) {
        if (conn->reply.buffer->size == 0) {
            conn_read_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else {
            conn_read_server_reply_done(conn, CONN_BAD_REPLY);
        }
        return;
    }
    else if (result == PARSE_PARTIAL && buffer->size == buffer->capacity) {
        /* If the reply size exceeds REPLY_SIZE (512), expand the buffer
         * to BODY_SIZE (4k).  If its more than that, return an error.
         */
        if (buffer->capacity < BODY_SIZE) {
            conn->reply.buffer = buffer_alloc(&vicompress->bodypool);
            buffer_add(conn->reply.buffer, buffer->buf, buffer->size);
            buffer_unref(buffer);
            buffer = conn->reply.buffer;
            sockets_read(s, fd, buffer, conn_read_reply_cb, conn);
            return;
        }
        else {
            conn_read_server_reply_done(conn, CONN_BAD_REPLY);
            return;
        }
    }
    else if (result == PARSE_PARTIAL) {
        sockets_read(s, fd, buffer, conn_read_reply_cb, conn);
        return;
    }
    else if (result == PARSE_DONE) {
        if (conn->reply.is_chunked) {
            chunked_init(&conn->chunked);
            conn->chunked_done = 0;
        }

        conn_read_server_reply_done(conn, CONN_SUCCESS);
    }
}

/* Write the http reply and body from the origin webserver to the client.
 * If GzipEncoder is not null, then compress the body of the response before
 * sending it to the client.  If CachedItem is not null, then store the reply
 * header and body into the memory cache.
 *
 * After the entire reply header and body has been written, the function
 * conn_write_server_reply_done() will be invoked.
 */
static void 
conn_write_server_reply(Connection *conn, GzipEncoder *gzip, CachedItem *item)
{
    Buffer *out;
    HttpReply *reply;
    Sockets *s;
    int offset, len;

    /* Create the serverread buffer now.
     * If we've read some of the reply body already,
     * copy it to the serverread buffer.
     */
    assert(conn->serverread == NULL);
    assert(conn->clientwrite == NULL);
    conn->serverread = buffer_alloc(&vicompress->bodypool);
    if (conn->chunk_encode_reply && gzip == NULL) {
        chunk_encode_start(conn->serverread);
    }
    offset = conn->reply.replylength;
    len = conn->reply.buffer->size - offset;
    if (len > 0) {
        buffer_add(conn->serverread, &(conn->reply.buffer->buf[offset]), 
                   len);
        conn->reply.buffer->size -= len;
    }

    conn->cacheditem = item;
    conn->gzip = gzip;
    if (gzip != NULL) {
        conn->gzipout = buffer_alloc(&vicompress->bodypool);
        if (conn->chunk_encode_reply) {
            chunk_encode_start(conn->gzipout);
        }
        if (conn->reply.is_chunked) {
            conn->gzip_dechunked = buffer_alloc(&vicompress->bodypool);
        }
        gzip_set_output(gzip, conn->gzipout);
    }

    reply = &conn->reply;
    conn->amountwritten = 0;
    conn->amountread = 0;
    out = buffer_alloc(&vicompress->replypool);

    /* Create the http reply header to send back to the client */
    if (Config.enable_sessions == 1 && 
        conn->webserverid != conn->request.webserverid) {

        http_write_reply(reply, &conn->request, out, 0, conn->chunk_encode_reply, 
                         conn->webserverid);
    }
    else {
        http_write_reply(reply, &conn->request, out, 0, conn->chunk_encode_reply, 0);
    }
    s = vicompress->sockets;
    /* Write the http reply header to the client */
    sockets_write(s, conn->clientfd, out, conn_write_reply_cb, conn);

}

/* Write the http reply and body from the origin webserver to the client.
 * This function is called when the http reply cannot be parsed, and we
 * don't know the content-length of the reply. In this case, keep writing
 * the server data until the server disconnects.
 *
 * After the entire reply header and body has been written, the function
 * conn_write_server_reply_done() will be invoked.
 */
static void conn_write_direct_reply(Connection *conn)
{
    assert(conn->chunk_encode_reply == 0);
    conn->cacheditem = NULL;
    if (conn->serverread == NULL) {
        conn->serverread = buffer_alloc(&vicompress->bodypool);
    }
    buffer_add(conn->serverread, conn->reply.buffer->buf, 
               conn->reply.buffer->size);
 
    conn_read_server_body_cb(conn->serverfd, conn, 
                             conn->serverread, 0, conn->eof);
}


/* Callback function invoked after we've written the http reply header
 * to the client.  If the content-length=0 or the server has closed
 * the connection (conn->eof == true), notify that the reply is done.  
 * Otherwise, read the http content from the server.
 *
 * fd - The socket connection with the client
 * data - The Connection structure
 * buffer - The http reply header written to the client
 * error - 0 if the write succeeded, else an errno value
 */
static void conn_write_reply_cb(int fd, void *data, Buffer *buffer, int error)
{
    Connection *conn;
    Sockets *s;

    conn = data;
    s = vicompress->sockets;

    buffer_unref(buffer); buffer = NULL;

    if (error) {
        conn->error = error;
        conn_write_server_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
        return;
    }

    if (conn->serverread->size > 0) {
        conn_read_server_body_cb(conn->serverfd, conn, 
                                 conn->serverread, 0, conn->eof);
    }
    else if (conn->reply.has_contentlength &&
             conn->reply.contentlength == 0) {
        conn_write_server_reply_done(conn, CONN_SUCCESS);
    }
    else if (!conn->reply.has_contentlength && conn->eof == 1) {
        conn_write_server_reply_done(conn, CONN_SUCCESS);
    }
    else if (conn->reply.contentlength > 0 && conn->eof == 1)  {
        conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
    } 
    else {
        conn->reading_from_server = 1;
        sockets_read(s, conn->serverfd, conn->serverread, 
                     conn_read_server_body_cb, conn);
    }
}

/* Callback function invoked after we've read a chunk of body data from
 * the origin server.  If the buffer is not yet full, and we haven't
 * reached the end of the page, then do another read from the origin server
 * (don't send the data to the client until the buffer is full). Otherwise,
 * send the data to the client.
 *
 * fd - The socket connection with the origin server
 * data - The Connection structure
 * buffer - The reply body read from the origin server
 * error - 0 if the read succeeded, else an errno value
 * eof - 1 if the server closed the connection, else 0
 */
static void 
conn_read_server_body_cb(int fd, void *data, Buffer *buffer, int error, int eof)
{
    Connection *conn;
    Sockets *s;
    int ret, can_read_more;
 
    conn = data;
    s = vicompress->sockets;
    assert(conn->serverread == buffer);

    conn->reading_from_server = 0;

    if (error) {
        conn->error = error;
    }
    conn->eof = eof;

    if (conn->error) {
        if (conn->writing_to_client) {
            return;
        }
        else if (error) {
            conn_write_server_reply_done(conn, CONN_SERVER_READ_ERROR); 
            return;
        }
        else {
            conn_write_server_reply_done(conn, CONN_CLIENT_WRITE_ERROR); 
            return;
        }
    }

    /* When Apache closes a connection, it waits for client to also close 
     * the connection before freeing the resources for that connection.
     * For good performance, close the server connection right when
     * an eof is detected. 
     */
    if (conn->eof && conn->serverfd != 0) {
        sockets_close(s, conn->serverfd);
        conn->serverfd = 0;
    }

    /* If there's space in the buffer, keep reading from the server, unless
     * - The server closed the connection (eof == true)
     * - We've read the entire contentlength
     * - We've read the entire chunked reply
     * If we can't read any more, then we write the data to the client.
     */
    can_read_more = 0;
    if (buffer->size < buffer->capacity) {
        can_read_more = 1;
        if (eof) {
            can_read_more = 0;
        }
        else if (conn->reply.has_contentlength &&
                 (buffer->size + conn->amountread >= conn->reply.contentlength)) {
            can_read_more = 0;
        }
        else if (conn->reply.is_chunked && chunked_is_done(&conn->chunked, buffer)) {
            can_read_more = 0;
        }
    }
    if (can_read_more) {
        conn->reading_from_server = 1;
        sockets_read(s, conn->serverfd, buffer, conn_read_server_body_cb, conn);
        return;
    }

    /* For chunked transfer-encodings
     * - If we encounter an error parsing chunks, disable chunked trasfer,
     *   and just forward the data to the client until the server connections closes.
     * - If the chunk size is right on the boundary, grow the buffer and
     *   read more data.
     * - If the chunked trasfer is done, set conn->chunked_done to true.
     */
    if (conn->reply.is_chunked) {
        ret = chunked_decode(&conn->chunked, buffer, conn->gzip_dechunked);
        if (ret == CHUNKED_ERROR) {
            conn->reply.is_chunked = 0;
        }
        else if (ret == CHUNKED_READSIZE) {
            if (buffer->size == buffer->capacity) {
                buffer_grow(buffer, buffer->capacity + 32);
            }
            conn->reading_from_server = 1;
            sockets_read(s, conn->serverfd, buffer,
                         conn_read_server_body_cb, conn);
            return;
        }
        else if (ret == CHUNKED_DONE) {
            conn->chunked_done = 1;
        }
    }

    conn->amountread += buffer->size;

    /* Finish chunk-encoding the buffer before writing it to client */
    if (conn->chunk_encode_reply && !conn->gzip) {
        int zerosize = (buffer->size == CHUNKED_HEADERSIZE);
        chunk_encode_end(buffer);

        /* If the reply is done, make sure it ends with  "0\r\n\r\n". */
        if (conn->eof) {
            conn->chunk_encode_done = 1;
            if (!zerosize) {
                buffer_add(buffer, "0\r\n\r\n", 5);
            }
        }
    }

    if (conn->gzip) {
        conn_write_gzippedbody(conn);
    }
    else if (conn->writing_to_client) {
        return;
    }
    else {
        conn_write_server_body(conn, buffer);
    }
}


/* Write the given buffer read from the origin server to the client. 
 * If this page is being cached, then also write it to the cache.
 * We read from the server and write to the client in parallel.
 * So if there's more data to read, then also read from the server.
 */
static void 
conn_write_server_body(Connection *conn, Buffer *buffer) 
{ 
    Sockets *s = vicompress->sockets;
    Buffer *temp;

    if (buffer->size == 0) {
        if (conn->reply.has_contentlength &&
            (conn->amountread < conn->reply.contentlength)) {
            conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else if (conn->reply.is_chunked && !conn->chunked_done) {
            conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else {
            conn_write_server_reply_done(conn, CONN_SUCCESS);
        }
        return;
    }

    conn->amountwritten += buffer->size;
    if (conn->cacheditem != NULL) {
        cacheditem_write(conn->cacheditem, buffer);
    }

    /* We read from the server and write to the client in parallel,
     * using two buffers, conn->serverread and conn->clientwrite.
     * 
     * - Switch the serverread and clientwrite buffers
     * - Write clientwrite to the client socket
     * - Read into serverread from the server socket (if there's more data to read).
     */
    assert(conn->writing_to_client == 0);
    assert(conn->reading_from_server == 0);
    temp = conn->clientwrite;
    if (temp == NULL) {
        temp = buffer_alloc(&vicompress->bodypool);
    }
    conn->clientwrite = conn->serverread;
    conn->serverread = temp;
    conn->writing_to_client = 1;
    sockets_write(s, conn->clientfd, conn->clientwrite, conn_write_server_body_cb, conn);

    /* See if we have more data to read */
    if ((conn->reply.has_contentlength && conn->amountwritten >= conn->reply.contentlength) ||
        (conn->reply.is_chunked && conn->chunked_done) ||
        (conn->eof)) {
        /* No more data to read */
    }
    else {
        conn->reading_from_server = 1;
        if (conn->chunk_encode_reply) {
            chunk_encode_start(conn->serverread);
        }
        sockets_read(s, conn->serverfd, conn->serverread, conn_read_server_body_cb, conn);
    }
}


/* Callback function invoked after writing a piece of body data to the client.
 *
 * If an error occurred, call conn_write_server_reply_done()
 *
 * If we're caching this page, allocate a new buffer (the previous one will
 * be stored in the cache).  Otherwise, re-use this buffer (conn->clientwrite).
 *
 * We read from the server and write to the client in parallel (see
 * conn_write_server_body).  If the parallel read operation is still in
 * progress, then return.
 *
 * If the parallel read operation is done, then
 * - If we've read some more data, call  conn_write_server_body()
 * - Else, we're done.  Call conn_write_server_reply_done()
 * 
 * fd - The socket connection with the client
 * data - The Connection structure
 * buffer - The reply body written to the client
 * error - 0 if the write succeeded, else an errno value
 */
static void 
conn_write_server_body_cb(int fd, void *data, Buffer *buffer, int error)
{
    Sockets *s;
    Connection *conn;
    unsigned long amount;

    conn = data;
    s = vicompress->sockets;
    assert(conn->clientwrite == buffer);
    conn->writing_to_client = 0;

    if (error) {
        conn->error = error;
    }

    if (conn->error) {
        if (conn->reading_from_server) {
            return;
        }
        else if (error) {
            conn_write_server_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
            return;
        }
        else {
            conn_write_server_reply_done(conn, CONN_SERVER_READ_ERROR);
            return;
        }
    }

    /* Update the clientwrite buffer */
    if (conn->cacheditem != NULL) {
        buffer_unref(conn->clientwrite);
        conn->clientwrite = NULL;
        amount = conn->reply.contentlength - conn->amountwritten;
        conn->clientwrite = buffer_alloc(&vicompress->bodypool);
    }
    else {
        buffer_reset(conn->clientwrite);
    }

    if (conn->reading_from_server) {
        return;
    }

    if (conn->serverread->size > 0) {
        conn_write_server_body(conn, conn->serverread);
        return;
    }
    else {
        if (conn->reply.has_contentlength &&
            (conn->amountread < conn->reply.contentlength)) {
            conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else if (conn->reply.is_chunked && !conn->chunked_done) {
            conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else {
            conn_write_server_reply_done(conn, CONN_SUCCESS);
        }
    }
}

/* Gzip the data read from the origin server.  If the data is 
 * chunk-encoded, then use the de-chunked data (in conn->gzip_dechunked).
 * The compression is done in a separate thread (gzip_thread). When 
 * the compression is done, conn_continue_write_gzippedbody() will 
 * be invoked.
 */
static void conn_write_gzippedbody(Connection *conn)
{
    GzipEncoder *gz;

    assert(conn->gzipout != NULL);
    gz = conn->gzip;
    if (conn->reply.is_chunked)
        gzip_add(gz, conn->gzip_dechunked, conn);
    else
        gzip_add(gz, conn->serverread, conn);

    /* When gzip_get_finished() is called, it will continue to 
     * conn_continue_write_gzippedbody() below
     */
}

/* gzip compression is done in a separate thread.  The main process
 * must periodically check to see if the gzip compression is done,
 * and then continue the connection.
 */
static void conn_check_finished_gzip(GzipPool *gzippool)
{
    GzipEncoder* gz[100];
    Connection *conn;
    int i, n;

    n = gzip_get_finished(gzippool, gz, 100);
    for (i = 0; i < n; i++) {
        conn = (Connection *)gz[i]->connection;
        assert(conn != NULL);
        conn_continue_write_gzippedbody(conn);
    }
}

/* The buffer data in conn->serverread has been gzipped.  The gzipped
 * data is stored in conn->gzipout.  It's possible that there is no
 * more gzipped data, and conn->gzipout is empty.  If there is gzip data,
 * write it to the client.  If we are caching this reply, write the
 * gzipped data to the cache also.
 */
static void conn_continue_write_gzippedbody(Connection *conn)
{
    Sockets *s;
    GzipEncoder *gz;
    Buffer *out;
    int replydone, has_output;

    s = vicompress->sockets;
    gz = conn->gzip; out = conn->gzipout;
    assert(out != NULL);

    /* Check whether we have finished reading the reply from the server.
     * - We've read the entire contentlength
     * - We've read the entire chunked reply
     * - The server closed the connection (eof == true)
     */
    replydone = 0;
    if ((conn->reply.has_contentlength && conn->amountread >= conn->reply.contentlength) ||
        (conn->reply.is_chunked && conn->chunked_done) ||
        (conn->eof)) {
        replydone = 1;
    }

    if (replydone)
        gzip_flush(gz);

    assert(conn->writing_to_client == 0);
    assert(conn->reading_from_server == 0);

    has_output = gzip_has_output(gz);
    if (conn->chunk_encode_reply && out->size == CHUNKED_HEADERSIZE)
        has_output = 0;

    if (has_output) {
        if (conn->chunk_encode_reply) {
            chunk_encode_end(out);
        }
        conn->amountwritten += out->size;
        if (conn->cacheditem != NULL) {
            cacheditem_write(conn->cacheditem, out);
        }
        conn->writing_to_client = 1;
        sockets_write(s, conn->clientfd, out, conn_write_gzippedbody_cb, conn);
    }
    /* If the reply is done, and we're chunk-encoding the output, write the
     * final "0\r\n\r\n" to the client, to indicate the end of chunked encoding.
     */
    else if (replydone && conn->chunk_encode_reply && !conn->chunk_encode_done) {
        conn->chunk_encode_done = 1;
        chunk_encode_end(out);
        conn->amountwritten += out->size;
        assert(conn->cacheditem == NULL);
        conn->writing_to_client = 1;
        sockets_write(s, conn->clientfd, out, conn_write_gzippedbody_cb, conn);
    } 
    /* We've written the entire gzipped reply to the client */
    else if (replydone) {
        if (conn->reply.has_contentlength &&
            (conn->amountread < conn->reply.contentlength)) {
            conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else if (conn->reply.is_chunked && !conn->chunked_done) {
            conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else {
            conn_write_server_reply_done(conn, CONN_SUCCESS);
        }
        return;
    }
    /* Read more data from the origin server */
    else {
        conn->reading_from_server = 1;
        buffer_reset(conn->serverread);
        if (conn->gzip_dechunked != NULL) {
            buffer_reset(conn->gzip_dechunked);
        }
        sockets_read(s, conn->serverfd, conn->serverread, 
                     conn_read_server_body_cb, conn);
    }
}

/* Callback function invoked after we've written the gzipped data to
 * the client. If we're caching this page, allocate a new buffer for
 * conn->gzipout, since the old conn->gzipout is stored in the cache.
 *
 * Tell the gzip encoder that the output buffer is now free.
 * It's possible more gzipped data may be available immediately.
 * If so, write the gzipped data to the client (and to the cache).
 *
 * fd - The socket connection with the client
 * data - The Connection structure
 * buffer - The gzipped data written to the client (conn->gzip->out)
 * error - 0 if the write succeeded, else an errno value
 */
static void 
conn_write_gzippedbody_cb(int fd, void *data, Buffer *buffer, int error)
{
    Sockets *s;
    Connection *conn;
    GzipEncoder *gz;
    int replydone, has_output;

    conn = data;
    s = vicompress->sockets;
    gz = conn->gzip;
    assert(conn->gzipout == buffer);
    conn->writing_to_client = 0;

    if (error) {
        conn->error = error;
        conn_write_server_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
        return;
    }

    if (conn->cacheditem != NULL) {
        buffer_unref(conn->gzipout); 
        conn->gzipout = buffer_alloc(&vicompress->bodypool);
    }
    else {
        buffer_reset(conn->gzipout);
    }

    if (conn->chunk_encode_reply) {
        chunk_encode_start(conn->gzipout);
    }
    gzip_set_output(gz, conn->gzipout);

    /* Check whether we have finished reading the reply from the server.
     * - We've read the entire contentlength
     * - We've read the entire chunked reply
     * - The server closed the connection (eof == true)
     */
    replydone = \
        (conn->reply.has_contentlength && conn->amountread >= conn->reply.contentlength) ||
        (conn->reply.is_chunked && conn->chunked_done) ||
        (conn->eof); 

    /* If there's more gzip data to write, then write it to the client */
    has_output = gzip_has_output(gz);
    if (conn->chunk_encode_reply && conn->gzipout->size == CHUNKED_HEADERSIZE)
        has_output = 0;

    if (has_output) {
        if (conn->chunk_encode_reply) {
            chunk_encode_end(conn->gzipout);
        }
        conn->amountwritten += conn->gzipout->size;
        if (conn->cacheditem != NULL) {
            cacheditem_write(conn->cacheditem, conn->gzipout);
        }
        conn->writing_to_client = 1;
        sockets_write(s, conn->clientfd, conn->gzipout, 
                      conn_write_gzippedbody_cb, conn);
    }
    /* If the reply is done, and we're chunk-encoding the output, write the
     * final "0\r\n\r\n" to the client, to indicate the end of chunked encoding.
     */
    else if (replydone && conn->chunk_encode_reply && !conn->chunk_encode_done) {
        conn->chunk_encode_done = 1;
        chunk_encode_end(conn->gzipout);
        assert(conn->cacheditem == NULL);
        sockets_write(s, conn->clientfd, conn->gzipout, 
                      conn_write_gzippedbody_cb, conn);
        return;
    }
    /* We've written the entire gzipped reply to the client */
    else if (replydone) {
        if (conn->reply.has_contentlength &&
            (conn->amountread < conn->reply.contentlength)) {
            conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else if (conn->reply.is_chunked && !conn->chunked_done) {
            conn_write_server_reply_done(conn, CONN_SERVER_CLOSED);
        }
        else {
            conn_write_server_reply_done(conn, CONN_SUCCESS);
        }
        return;
    }
    /* Read more data from the origin server */
    else {
        conn->reading_from_server = 1;
        buffer_reset(conn->serverread);
        if (conn->gzip_dechunked != NULL) {
            buffer_reset(conn->gzip_dechunked);
        }
        sockets_read(s, conn->serverfd, conn->serverread,
                     conn_read_server_body_cb, conn);
    }
}


/*** Cache functions ***/

/* Write the cached http reply and body to the client, using the
 * given cached item.  Invoke conn_write_cached_reply_done()
 * when the entire header and body have been written to the client.
 */
static void conn_write_cached_reply(Connection *conn, CachedItem *item)
{
    Buffer *out;
    Sockets *s;

    conn->cacheditem = item;
    cacheditem_httpreply(conn->cacheditem, &conn->reply);
    conn->cache_buf_index = 0;

    /* If the request is has the header
     *   Range: start-
     * then fill in the range end value now.
     */
    if (conn->request.is_range && conn->request.range_end == 0) {
        conn->request.range_end = conn->reply.contentlength;
    }

    out = buffer_alloc(&vicompress->replypool);

    conn->amountwritten = 0;
    http_write_reply(&conn->reply, &conn->request, out, 1, 0, 0); 

    s = vicompress->sockets;
    sockets_write(s, conn->clientfd, out, conn_write_cached_reply_cb, conn);
}

/* Read the next buffer from the cached body, and write it to the
 * client.  However, the client sent a "Range:" request.  Write only
 * the range specified by the client.  When we finish writing the
 * entire range (or an error occurs) call conn_write_cached_reply_done().
 */
static void conn_read_cached_body_range(Connection *conn)
{
    Buffer *buffer;
    int start, end, rangelength;
    Sockets *s;

    s = vicompress->sockets;

    start = conn->request.range_start;
    end = conn->request.range_end;
    rangelength = end - start + 1;
    assert(start < end && end > 0);

    /* If we've written the entire range, we're done */
    if (conn->amountwritten >= rangelength) {
        conn->cache_buf_index = 0;
        conn_write_cached_reply_done(conn, CONN_SUCCESS);
        return;
    }

    /* Get the next buffer to write */
    buffer = cacheditem_read(conn->cacheditem, &conn->cache_buf_index);
    if (buffer == NULL) {
        conn->cache_buf_index = 0;
        conn->error = EINVAL;
        conn_write_cached_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
        return;
    }

    /* If we've just started reading the cached item, then:
     *   - Find the starting buffer
     *   - Find the starting byte offset within that buffer.
     *   - Calculate the amount of buffer data to write.
     *     The amount written should not exceed the range length.
     * Else
     *   - We already have the correct buffer
     *   - The byte offset is 0
     *   - Calculate the amount of buffer data to write.
     *     The amount written should not exceed the range length.
     */
    if (conn->amountwritten == 0) {
        int byteoffset = 0, bufoffset = 0;
        while ((buffer != NULL) && (byteoffset + buffer->size < start)) {
            byteoffset += buffer->size;
            buffer = cacheditem_read(conn->cacheditem, &conn->cache_buf_index);
        }
        if (buffer == NULL) {
            conn->cache_buf_index = 0;
            conn->error = EINVAL;
            conn_write_cached_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
            return;
        }
        bufoffset = start - byteoffset;
        conn->range_buf.buf = &(buffer->buf[bufoffset]);
        conn->range_buf.size = MIN(buffer->size - bufoffset, rangelength);
    }
    else {
        conn->range_buf.buf = buffer->buf;
        conn->range_buf.size = MIN(rangelength - conn->amountwritten, buffer->size);
    }

    conn->amountwritten += conn->range_buf.size;
    sockets_write(s, conn->clientfd, &conn->range_buf,
                  conn_write_cached_body_cb, conn);
}


/* Callback function invoked after writing a cached http reply header
 * to the client.  Read the content body from the cache.
 */
static void 
conn_write_cached_reply_cb(int fd, void *data, Buffer *buffer, int error)
{
    Connection *conn;

    conn = data;
    buffer_unref(buffer); buffer = NULL;
    if (error) {
        conn->error = error;
        conn_write_cached_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
        return;
    }
    else
        conn_read_cached_body(conn);
}

/* Read the next buffer from the cached body, and write it to the
 * client.  The conn->cache_buf_index contains the index of the
 * next body buffer to read (the page has an array of body buffers).
 * If there's no more buffers, we're done.
 */
static void conn_read_cached_body(Connection *conn)
{
    Buffer *buffer;
    Sockets *s;

    /* If this is a range request, use conn_read_cached_body_range() */
    if (conn->request.is_range) {
        conn_read_cached_body_range(conn);
        return;
    }

    s = vicompress->sockets;
    buffer = cacheditem_read(conn->cacheditem, &conn->cache_buf_index);
    if (buffer != NULL) {
        conn->amountwritten += buffer->size;
        sockets_write(s, conn->clientfd, buffer, 
                      conn_write_cached_body_cb, conn);
        return;
    }
    else {
        conn->cache_buf_index = 0;
        conn_write_cached_reply_done(conn, CONN_SUCCESS);
    }
}

/* Callback function invoked after we've written cached data to
 * the client.
 */
static void 
conn_write_cached_body_cb(int fd, void *data, Buffer *buffer, int error)
{
    Connection *conn;

    conn = data;
    if (error) {
        conn->error = error;
        conn_write_cached_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
    }
    else
        conn_read_cached_body(conn);
}


/* Write an error reply back to the client.  Current, only status
 * 400 (bad request), 404 (dns lookup failed) and 503 (service
 * unavailable) are used.
 *
 *  Invoke conn_write_error_reply_done() when finished.
 */
static void conn_write_error_reply(Connection *conn, int status)
{
    Buffer *out;
    const char *statusmsg = NULL;
    Sockets *s = vicompress->sockets;
    int ret;

    /* Close the server connection */
    if (conn->serverfd != 0) {
        sockets_close(s, conn->serverfd);
        conn->serverfd = 0;
    }

    out = buffer_alloc(&vicompress->bodypool);
    if (status == ERROR_BAD_REQUEST) {
        statusmsg = "HTTP/1.1 400 Bad Request\r\n";
    }
    else if (status == ERROR_NOT_AVAILABLE) {
        statusmsg = "HTTP/1.1 503 Service Unavailable\r\n";
    }
    else if (status == ERROR_NOT_FOUND) {
        statusmsg = "HTTP/1.1 404 Not Found\r\n";
    }
    else {
        assert(0);
    }
    buffer_add(out, statusmsg, strlen(statusmsg));
    buffer_add(out, "Server: vicompress/1.0\r\n", sizeof("Server: vicompress/1.0\r\n")-1);
    buffer_add(out, "Content-type: text/html\r\n", sizeof("Content-type: text/html\r\n")-1);
    buffer_add(out, "Connection: close\r\n\r\n", sizeof("Connection: close\r\n\r\n")-1);

    if (Config.errorpage_size > 0) {
        buffer_add(out, Config.errorpage, Config.errorpage_size);
    }
    http_free_reply(&conn->reply);
    http_init_reply(&conn->reply, out, &conn->parsedata);
    ret = http_parse_reply(&conn->reply, &conn->parsedata);
    assert(ret == PARSE_DONE);
    s = vicompress->sockets;
    conn->amountwritten = Config.errorpage_size;
    conn->amountread = conn->amountwritten;
    sockets_write(s, conn->clientfd, out, conn_write_error_reply_cb, conn);
}

/* Callback function invoked after we've written an error reply 
 * to the client.
 */
static void 
conn_write_error_reply_cb(int fd, void *data, Buffer *buffer, int error)
{
    Connection *conn = (Connection *) data;
    buffer_unref(buffer); buffer = NULL;
    if (error) {
        conn->error = error;
        conn_write_error_reply_done(conn, CONN_CLIENT_WRITE_ERROR);
    }
    else
        conn_write_error_reply_done(conn, CONN_SUCCESS);
}

/* Write the list of cached URLs to the client. */
static void conn_write_cacheurls(Connection *conn, CacheUrls *list)
{
    Sockets *s;
    Buffer *out;
    const char *line;
    int ret;

    list->conn = conn;
    conn->cache_buf_index = 0;
    conn->amountread = 0;
    conn->amountwritten = 0;
    out = buffer_alloc(&vicompress->replypool);
    line = "HTTP/1.1 200 OK\r\n";
    buffer_add(out, line, strlen(line));
    line = "Content-type: text/plain\r\n";
    buffer_add(out, line, strlen(line));
    line = "Connection: close\r\n\r\n";
    buffer_add(out, line, strlen(line));

    http_init_reply(&conn->reply, out, &conn->parsedata);
    ret = http_parse_reply(&conn->reply, &conn->parsedata);
    assert(ret == PARSE_DONE);

    s = vicompress->sockets;
    sockets_write(s, conn->clientfd, out, conn_write_cacheurls_cb, list);
}

/* Callback function invoked after we've written the http header for the
 * _viewcache_ page to the client.
 */ 
static void conn_write_cacheurls_cb(int fd, void *data, Buffer *out, int error)
{
    CacheUrls *list;
    Connection *conn;
    Buffer *body;

    list = data;
    conn = list->conn;
    assert(out == conn->reply.buffer);

    if (error) {
        conn->error = error;
        conn_write_cacheurls_done(list, conn, CONN_CLIENT_WRITE_ERROR);
    }
    else {
        body = buffer_alloc(&vicompress->bodypool);
        conn_read_cacheurls(list, conn, body);
    }
}

/* Write the _viewcache_ page to the client.
 * Read more cached urls into the given buffer, and write the buffer
 * to the client.  conn->cache_buf_index contains the index of the last
 * url written to the client.
 */
static void conn_read_cacheurls(CacheUrls *list, Connection *conn, Buffer *out)
{
    Sockets *s;
    Buffer *url;
    int amount = 0;

    s = vicompress->sockets;
    buffer_reset(out);

    while (conn->cache_buf_index < list->size) {
        url = list->urls[conn->cache_buf_index];
        if (url == NULL) {
            conn->cache_buf_index += 1;
            continue;
        }
        /* Write http://<url>\n, one per line */
        amount = 7 + url->size + 1;
        if (amount + out->size > out->capacity && amount < BODY_SIZE)
            break;
        buffer_add(out, "http://", 7);
        buffer_add(out, url->buf, url->size);
        buffer_add(out, "\n", 1);
        conn->cache_buf_index += 1;
    }

    conn->amountwritten += out->size;
    conn->amountread = conn->amountwritten;
    sockets_write(s, conn->clientfd, out, conn_write_cacheurlsbody_cb, list);
}

/* Callback function invoked after we've written part of the _viewcache_
 * page to the client.
 */
static void 
conn_write_cacheurlsbody_cb(int fd, void *data, Buffer *out, int error)
{
    CacheUrls *list;
    Connection *conn;

    list = data;
    conn = list->conn;

    if (error) {
        buffer_unref(out);
        conn->error = error;
        conn_write_cacheurls_done(list, conn, CONN_CLIENT_WRITE_ERROR);
    }
    else if (conn->cache_buf_index >= list->size) {
        buffer_unref(out);
        conn->cache_buf_index = 0;
        conn_write_cacheurls_done(list, conn, CONN_SUCCESS);
    }
    else {
        conn_read_cacheurls(list, conn, out);
    }
}

/* Return the error string associated with the given codes. */
static const char* conn_status_string(int status) {
    const char *str[20];

    str[CONN_SUCCESS] = "Success";
    str[CONN_BAD_REQUEST] = "Bad HTTP Request from client";
    str[CONN_BAD_REPLY] = "Bad HTTP Reply from server";
    str[CONN_CLIENT_CLOSED] = "Client closed prematurely";
    str[CONN_SERVER_CLOSED] = "Server closed prematurely";
    str[CONN_CONNECT_ERROR] = "Connect failed";
    str[CONN_CLIENT_READ_ERROR] = "Error reading from client";
    str[CONN_CLIENT_WRITE_ERROR] = "Error writing to client";
    str[CONN_SERVER_READ_ERROR] = "Error reading from server";
    str[CONN_SERVER_WRITE_ERROR] = "Error writing to server";
    str[CONN_DNS_ERROR] = "DNS lookup failed";

    if (status >= 0 && status <= 10)
        return str[status];
    else 
        return "";
}


/* Return true if the url path is the special string _viewcache_,
 * and the client IP is from the local machine.
 */
static int isviewcache(sockaddr_all *client, Substring path)
{
    int i, family;
    struct hostent *ent;
    sockaddr_all *addr;
    unsigned char *ip;

    family = client->ipv4.sin_family;
    if (path.size == 12 && strncmp(path.buf, "/_viewcache_", 12) == 0) {

        ip = (unsigned char *)&(client->ipv4.sin_addr.s_addr);
        if (ip[0] == 127 && ip[1] == 0 && ip[2] == 0 && ip[3] == 1) {
            return 1;
        }

        for (i = 0; i < Config.num_listen; i++) {
            addr = &(Config.listenaddr[i]);

            if (client->ipv4.sin_family == addr->ipv4.sin_family &&
                family == AF_INET) {

                if (memcmp(&(client->ipv4.sin_addr), 
                           &(addr->ipv4.sin_addr), 4) == 0) {
                    return 1;
                }
            }
#ifdef HAVE_IPV6
            else if (client->ipv6.sin6_family == addr->ipv6.sin6_family &&
                     family == AF_INET6) {

                if (memcmp(&(client->ipv6.sin6_addr), 
                           &(addr->ipv6.sin6_addr), 16) == 0) {
                    return 1;
                }
            }
#endif
        }

        ent = gethostbyname(Config.myhostname);
        if (ent != NULL && family == AF_INET) {
            if (memcmp(&(client->ipv4.sin_addr), ent->h_addr_list[0], 4) == 0) {
                return 1;
            }
        }
    }
    return 0;
}

/* Write an entry to the accesslog.  The conn contains the client IP,
 * server IP, http request, and http reply, and url.
 * If CachedItem is not null, this was a cache hit.
 * isrequest is used to indicate whether an error occurred in the
 * request/reply (bad request, dns lookup failed, connect failed, etc).
 */
static void logaccess(AccessLog *log, Connection *conn, CachedItem *item,
                      int isrequesterr)
{
    unsigned long replysize, pregzip_size;
    int iscachehit, duration;

    if (log == NULL)
        return;

    duration = 0;
    if (Config.logformat == LOG_FORMAT_SQUID) {
        struct timeval end;
        gettimeofday(&end, NULL);
        duration = (end.tv_sec - conn->start.tv_sec) * 1000;
        duration += (end.tv_usec - conn->start.tv_usec) / 1000;
    }

    iscachehit = (item != NULL);
    if (iscachehit) {
        replysize = item->reply.contentlength;
        pregzip_size = item->orig_contentlength;
    }
    else {
        replysize = conn->amountwritten;
        pregzip_size = conn->amountread;
    }
    accesslog_write(log, &(conn->request), &(conn->reply), 
                                    &(conn->clientip), &(conn->serverip),
                                    replysize, pregzip_size, iscachehit, 
                                    isrequesterr, duration);
}


/* Invoked when a full http request has been read from the client.
 * If an error occurred, close the connection.  If the request is 
 * in the cache, write the cached reply back to the client.  Otherwise,
 * read the http reply from the origin webserver.
 */
static void conn_read_request_done(Connection *conn, int status)
{
    MemoryCache *cache;
    CachedItem *item;
    ErrorLog *errorlog;
    Buffer *url;

    cache = vicompress->cache;
    errorlog = vicompress->errorlog;
    url = conn->request.url;

    DEBUG_PRINTF(errorlog, &conn->clientip, 
        "Read http request from client: status=%s, url=%s",
        conn_status_string(status),
        url == NULL ? "" : url->buf);

    if (Config.enable_debug && conn->request.buffer != NULL) {
        DEBUG_PRINT(errorlog, &conn->clientip,
                    conn->request.buffer->buf, conn->request.buffer->size);
    }

    /* If an error occurred, close the connection. */
    if (status != CONN_SUCCESS) { 
        if (status == CONN_BAD_REQUEST) {
            DEBUG_PRINTF(errorlog, &conn->clientip,
                         "Writing error reply to client");
            conn->reply.status = ERROR_BAD_REQUEST;
            conn_write_error_reply(conn, ERROR_BAD_REQUEST);
        }
        else { 
            DEBUG_PRINTF(errorlog, &conn->clientip, 
                         "Closing connection");
            conn_close(conn); 
        }
        return;
    }

    /* If the URL is http://<hostname>/_viewcache_, send the
     * client a list of the cached urls in memory.
     */
    if (isviewcache(&conn->clientip, conn->request.path)) {
        CacheUrls *list = cacheurls_get(cache);
        DEBUG_PRINTF(errorlog, &conn->clientip, 
                     "Writing _viewcache_ page to client");
        conn_write_cacheurls(conn, list);
        return;
    }

    /* If the request is cachable, see if its in the cache */
    if (conn->request.is_cachable && Config.cache_memory > 0) {
        item = cache_get(cache, conn->request.url, conn->request.acceptgzip);
        if (item != NULL) {
            DEBUG_PRINTF(errorlog, &conn->clientip, 
                         "Writing cached reply to client");
            conn_write_cached_reply(conn, item);
            return;
        }
    }

    /* Else, read the reply from the origin server */
    conn_read_server_reply(conn);

}

/* Invoked when a full http reply header has been read from the origin webserver.
 * - If an error occurred, write an error reply back to the client.
 * - Determine if the reply can be gzipped
 * - Determine if the reply can be cached
 * - Determine if the reply can be chunk-encoded.
 * Then, write the server reply back to the client.
 */
 
static void conn_read_server_reply_done(Connection *conn, int status)
{
    MemoryCache *cache;
    CachedItem *item = NULL;
    GzipEncoder *gzip = NULL;
    LoadBalancer *balancer;
    ErrorLog *errorlog;
    Sockets *s;
    char debugstr[80];

    cache = vicompress->cache;
    errorlog = vicompress->errorlog;
    balancer = vicompress->loadbalancer;
    s = vicompress->sockets;

    DEBUG_PRINTF(errorlog, &conn->clientip, 
                 "Read http reply from server: status=%s", 
                 conn_status_string(status));

    if (Config.enable_debug && conn->reply.buffer != NULL) {
        DEBUG_PRINT(errorlog, &conn->clientip,
                    conn->reply.buffer->buf, conn->reply.replylength);
    }

    /* Check if an error occurred while trying to read the http reply
     * from the origin server.  If we have multiple backend webservers,
     * and we failed to connect (CONN_CONNECT_ERROR), then try connecting
     * to another backend webserver.
     */
    if (status != CONN_SUCCESS) {

        if (status == CONN_DNS_ERROR) {
            DEBUG_PRINTF(errorlog, &(conn->clientip),
                         "Writing error reply to client");
            conn->reply.status = ERROR_NOT_FOUND;
            conn_write_error_reply(conn, ERROR_NOT_FOUND);
        }
        else if (status == CONN_CONNECT_ERROR && Config.num_webservers > 1) {
            loadbalancer_connect_done(balancer, conn->webserverid, 1);
            conn->failed_connections++;

            if (conn->failed_connections >= 2) {
                /* After two connect failures, give up */
                DEBUG_PRINTF(errorlog, &(conn->clientip),
                             "Writing error reply to client");
                conn->reply.status = ERROR_NOT_AVAILABLE;
                conn_write_error_reply(conn, ERROR_NOT_AVAILABLE);
            }
            else {
                /* Try connecting to another backend webserver */
                sockaddr_all *addr;
                int id;

                id = loadbalancer_get_another_id(balancer, conn->webserverid, 
                                                 conn->request.host);
                addr = loadbalancer_getaddr(balancer, id);
                loadbalancer_connect_started(balancer, id);
                conn->webserverid = id;
                memcpy(&conn->serverip, addr, sizeof(sockaddr_all));
                conn_connect_to_server(conn); 
            } 
        }
        else if (status == CONN_CONNECT_ERROR && Config.num_webservers <= 1) {
            DEBUG_PRINTF(errorlog, &(conn->clientip),
                         "Writing error reply to client");
            conn->reply.status = ERROR_NOT_AVAILABLE;
            conn_write_error_reply(conn, ERROR_NOT_AVAILABLE);
        }

        else if (status == CONN_SERVER_CLOSED ||
                 status == CONN_SERVER_WRITE_ERROR ||
                 status == CONN_SERVER_READ_ERROR) {

            DEBUG_PRINTF(errorlog, &(conn->clientip),
                         "Writing error reply to client");
            conn->reply.status = ERROR_NOT_AVAILABLE;
            conn_write_error_reply(conn, ERROR_NOT_AVAILABLE);
        }
        else if (status == CONN_BAD_REPLY) {
            DEBUG_PRINTF(errorlog, &(conn->clientip),
                         "Writing direct reply to client");
            conn_write_direct_reply(conn);
        } 
        else {
            DEBUG_PRINTF(errorlog, &(conn->clientip),
                         "Closing connection");
            conn_close(conn); 
        }
        return;
    } /* status != CONN_SUCCESS */

    /* If the reply can be gzipped, mark the reply as "newly gzipped" */
    if (conn->request.acceptgzip && conn->reply.cangzip && 
        Config.enable_compression) {
        gzip = gzip_alloc(vicompress->gzippool);
    } else {
        gzip = NULL;
    }

    if (gzip != NULL)
        conn->reply.newlygzipped = 1;

    /* If the request stated "no caching", but the reply is cachable,
     * see if the reply is already in the cache.  If so, check whether
     * the cached reply matches the server's reply.  If the last-modified
     * dates are different, then mark the cached reply as expired.
     */
    if (conn->reply.is_cachable && !conn->request.is_cachable && 
        Config.cache_memory > 0) {
        CachedItem *check = cache_get(cache, conn->request.url, 1);
        if (check != NULL) {
            if (check->reply.lastmodified != conn->reply.lastmodified) {
                check->reply.expires = get_current_time() - 10;
            }
            cache_close(cache, check);
        }
    }

    item = NULL;

    /* Check if the reply can be cached.  If so, create a new CachedItem */
    if (conn->reply.is_cachable && conn->request.is_cachable && 
        Config.cache_memory > 0) {
        if (!conn->request.acceptgzip && conn->reply.cangzip && 
             Config.enable_compression) {

            /* If the reply can be compressed, but the client doesn't support
             * compression, then don't cache the reply.  We want to populate
             * the cache with compressed data, to save memory space.
             */
            item = NULL;
        }
        else {
            item = cache_create(cache, conn->request.url, &conn->reply);
        }
    }

    /* Check if the reply can be chunk-encoded, so that we can keep
     * the client connection persistent.
     */
    conn->chunk_encode_reply = 0;
    conn->chunk_encode_done = 0;
    if (conn->request.version == 1) { /* Need HTTP 1.1 to support chunk-encoding */

        if (item == NULL && conn->reply.newlygzipped) {
            conn->chunk_encode_reply = 1;
        }
        else if (item == NULL && !conn->reply.has_contentlength &&
                 !conn->reply.is_chunked) {
            conn->chunk_encode_reply = 1;
        }
    }

    if (Config.enable_debug) {
        debugstr[0] = 0;
        strcat(debugstr, "Writing");
        if (item != NULL) {
            strcat(debugstr, ", caching");
        }
        if (gzip != NULL) { 
            strcat(debugstr, ", compressing");
        }
        if (conn->chunk_encode_reply == 1) {
            strcat(debugstr, ", chunk-encoding");
        }
        strcat(debugstr, " server reply"); 

        DEBUG_PRINT(errorlog, &(conn->clientip), debugstr, strlen(debugstr));
    }

    conn_write_server_reply(conn, gzip, item);
} 

/* Invoked when a server reply has been completely written to the client.
 * If the transfer was successful, store the reply data in the memory cache.
 * Log the request to the accesslog.  Determine whether to keep the
 * client connection alive, or close it.
 */
static void conn_write_server_reply_done(Connection *conn, int status)
{
    MemoryCache *cache;
    GzipEncoder *gz;
    CachedItem *item;
    ErrorLog *errorlog;
    LoadBalancer *balancer;
    Sockets *sockets;
    int keepalive, maxclients;
    unsigned long newcontentlength;

    item = conn->cacheditem; conn->cacheditem = NULL;
    gz = conn->gzip; conn->gzip = NULL;
    cache = vicompress->cache;
    errorlog = vicompress->errorlog;
    balancer = vicompress->loadbalancer;
    sockets = vicompress->sockets;

    DEBUG_PRINTF(errorlog, &conn->clientip, 
                 "Wrote server reply: status=%s",
                 conn_status_string(status));

    if (Config.enable_compression)
        gzip_free(vicompress->gzippool, gz);

    if (conn->reply.newlygzipped)
        newcontentlength = conn->amountwritten;
    else
        newcontentlength = conn->reply.contentlength;

    if (Config.cache_memory > 0 && item != NULL) {
        if (status == CONN_SUCCESS && 
            conn->amountread == conn->reply.contentlength) 
            cache_create_done(cache, item, newcontentlength);
        else
            cache_create_aborted(cache, item);
    }

    logaccess(vicompress->accesslog, conn, NULL, 0);

    if (Config.num_webservers > 1) {
        loadbalancer_connect_done(balancer, conn->webserverid, 0);
    }

    /* Determine whether to keep the client connection alive, or close it */
    keepalive = 1;
    if (status != CONN_SUCCESS)
        keepalive = 0;
    else if (!conn->reply.has_contentlength && !conn->reply.is_chunked && !conn->chunk_encode_reply)
        keepalive = 0;
    else if (conn->reply.has_contentlength && 
             conn->amountwritten != conn->reply.contentlength)
        keepalive = 0;
    else if (conn->reply.is_chunked && !conn->chunked_done)
        keepalive = 0;
    else if (conn->reply.newlygzipped && !conn->chunk_encode_reply)
        keepalive = 0;
    else if (sockets->numclients > (2 * sockets->maxclients / 3))
        keepalive = 0;  /* Disable keepalive under heavy traffic */

    if (keepalive) {
        DEBUG_PRINTF(errorlog, &conn->clientip, 
                     "Keeping client connection alive");
        conn_keepalive_read(conn);
    }
    else {
        DEBUG_PRINTF(errorlog, &conn->clientip, 
                     "Closing connection");
        conn_close(conn);
    }
}

/* Invoked when a cached reply has been written to the client.
 * Log the request in the accesslog.  Then determine whether to
 * keepalive or close the client connection.
 */
static void conn_write_cached_reply_done(Connection *conn, int status)
{
    MemoryCache *cache;
    CachedItem *item;

    item = conn->cacheditem; conn->cacheditem = NULL;
    cache = vicompress->cache;
 
    DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, 
                 "Wrote cached reply: status=%s", 
                 conn_status_string(status));

    logaccess(vicompress->accesslog, conn, item, 0);

    cache_close(cache, item);

    if (vicompress->sockets->numclients > (2 * vicompress->sockets->maxclients / 3)) {
        /* Disable keepalive under heavy traffic */
        DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, 
                     "Closing connection");
        conn_close(conn);
    }
    else if (status == CONN_SUCCESS && conn->request.keepalive) {
        DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, 
                     "Keeping client connection alive");
        conn_keepalive_read(conn);
    }
    else {
        DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, 
                     "Closing connection");
        conn_close(conn);
    }
}

/* Invoked when an error reply has been completely written to the client.
 * Log the request in the accesslog and close the connection.
 */
static void conn_write_error_reply_done(Connection *conn, int status)
{

    DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, 
                 "Wrote error reply: status=%s",
                 conn_status_string(status));

    logaccess(vicompress->accesslog, conn, NULL, 1);
    DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, 
                 "Closing connection");
    conn_close(conn);
}

/* Invoked when the _viewcache_ page has been completely written
 * to the client.  Log the request in the accesslog, and close the
 * connection.
 */
static void 
conn_write_cacheurls_done(CacheUrls *list, Connection *conn, int status)
{

    cacheurls_free(list);

    DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, 
                 "Wrote _viewcache_ page: status=%s",
                 conn_status_string(status));

    logaccess(vicompress->accesslog, conn, 0, 0);
    DEBUG_PRINTF(vicompress->errorlog, &conn->clientip, 
                 "Closing connection");
    conn_close(conn);
}

static void update_log_stats(char *accesslog, char *dir, char *vicompresspath)
{
    char* args[4];
    char updatepath[256];
    char dirpath[256];

    memset(dirpath, 0, 256);
    strncpy(dirpath, vicompresspath, 255);
    sprintf(updatepath, "%s/update_log_stats", dirname(dirpath));
    if (fork() > 0) {
        return;
    }
    args[0] = "update_log_stats";
    args[1] = accesslog;
    args[2] = dir;
    args[3] = NULL;
    execv(updatepath, args);
}

/* Signal handler for SIGUSR1 signal.  This toggles the
 * debugging.
 */
static void change_enable_debug(int signum)
{
    Config.enable_debug = !Config.enable_debug;
}

/* Signal handling for the SIGINT signal.
 * Set stop_vicompress to 1 to stop vicompress in the main() loop.
 */
static int stop_vicompress = 0;
void vicompress_stop(int signum)
{
    stop_vicompress = 1;
}


int vicompress_start(int argc, char **argv)
{ 
    time_t start, curr, check_webservers, write_stats;
    int numcacheditems, maxclients;

    if (argc < 2 || *argv[1] == '-') {
        printf("vicompress: Version %s\n", VICOMPRESS_VERSION);
        printf("Usage: vicompress <path/to/vicompress.conf>\n");
#ifndef WIN32
        printf("       This program should not be directly invoked by the\n");
        printf("       end user.  Instead, use the script vicompress.sh to\n");
        printf("       start and stop the vicompress server.\n");
#endif
        exit(1);
    }

    set_current_time();
    if (config_init(argv[1]) == -1) {
        exit(1);
    }

    /* Run server in background. */
    if (fork() > 0) {
        exit(0);
    }

    /* After forking, create pid file. */
    create_pidfile();

    errno = 0;

    signal(SIGHUP, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGTERM, vicompress_stop);
    signal(SIGINT, vicompress_stop);
    signal(SIGUSR1, change_enable_debug);

    vicompress = (Vicompress *) malloc(sizeof(Vicompress));
    memset(vicompress, 0, sizeof(Vicompress));

    /* Initialize memory cache */
    if (Config.cache_memory > 0) {
        vicompress->cache = (MemoryCache *)malloc(sizeof(MemoryCache));
        cache_init(vicompress->cache, Config.cache_memory, 
                   Config.max_cacheditem_size);
    }

    /* Initialize compression */
    if (Config.enable_compression) {
        vicompress->gzippool = (GzipPool *) malloc(sizeof(GzipPool));
        gzip_init(vicompress->gzippool);
    }

    /* Intialize connection/socket structures */
    vicompress->sockets = (Sockets *) malloc(sizeof(Sockets));
    sockets_init(vicompress->sockets);
    vicompress->connpool = 
      (Connection *) malloc(sizeof(Connection) * Config.max_total_fd);
    memset(vicompress->connpool, 0, sizeof(Connection *) * Config.max_total_fd);

    /* Initialize buffer pools */
    maxclients = Config.max_total_fd / 2;

    /* Assume average webpage = 8 kb, and http reply headers + url = 1kb */
    numcacheditems = Config.cache_memory/(9*1024);

    /* Pool only 2/3 of the maximum number of items possible,
     * so that we don't waste memory under a light load.
     * If the pool runs out, it allocates memory using malloc(),
     * so that's not a problem.
     */
    bufferpool_init(&vicompress->reqpool, REQUEST_SIZE, maxclients*2/3);
    bufferpool_init(&vicompress->replypool, REPLY_SIZE, 
                    (maxclients + numcacheditems)*2/3);
    bufferpool_init(&vicompress->smallurlpool, SMALL_URL_SIZE, 
                    (maxclients + numcacheditems)/4 );
    bufferpool_init(&vicompress->largeurlpool, LARGE_URL_SIZE, 
                    (maxclients + numcacheditems)*2/3 );
    bufferpool_init(&vicompress->bodypool, BODY_SIZE, 
                    (maxclients + numcacheditems*8*1024/BODY_SIZE) * 2/3);
 
    if (start_listening(vicompress) == -1) {
        return -1;
    }

    /* Only after listening socket is created, change to user.
     * The listening socket may bind to a port (< 1024) requiring
     * root priviledge.
     */
    setgid(Config.group);
    setuid(Config.user);

    /* Only after user is changed, create the log files */
    if (Config.errorlog != NULL) {
        vicompress->errorlog = (ErrorLog*) malloc(sizeof(ErrorLog));
        errorlog_init(vicompress->errorlog, Config.errorlog);
    }
    else {
        vicompress->errorlog = NULL;
    }
    if (Config.accesslog != NULL) {
        vicompress->accesslog = (AccessLog *)malloc(sizeof(AccessLog));
        accesslog_init(vicompress->accesslog, Config.accesslog);
    }
    else {
        vicompress->accesslog = NULL;
    }

    /* Create the loadbalancer after the errorlog is created */
    if (Config.num_webservers > 0) {
        vicompress->loadbalancer = (LoadBalancer*) malloc(sizeof(LoadBalancer));
        loadbalancer_init(vicompress->loadbalancer, vicompress->sockets, 
                          vicompress->errorlog, Config.webservers, 
                          Config.num_webservers);
    }


    errorlog_printf(vicompress->errorlog, NULL, "vicompress started");
    errorlog_printf(vicompress->errorlog, NULL, 
                    "Maximum concurrent clients is %d", Config.max_total_fd/2);
    errorlog_printf(vicompress->errorlog, NULL, "Cache size is %d MB",
                    Config.cache_memory /MEGABYTE );
#ifdef WIN32
    service_running();
#endif

    start = get_current_time(); 
    check_webservers = write_stats = curr = start;

    /* Once the log files have been created, redirect stdin, stdout, stderr 
     * to /dev/null.  Do not close stdin, stdout, stderr.  We do not want 
     * any sockets to use file descriptor 0.  Just redirect them to /dev/null.
     */
    redirect_stdio_to_devnull();

    /* The main vicompress loop.
     * Check for socket events (new clients, reads available, etc).
     * Every 3 minutes, check if any "down" webservers have come back up.
     */
    while (!stop_vicompress) {
        sockets_checkevents(vicompress->sockets, 10);
        if (Config.enable_compression) {
            conn_check_finished_gzip(vicompress->gzippool);
        }
        if (Config.num_webservers > 1 && 
            (curr - check_webservers > CHECK_WEBSERVER_INTERVAL)) {
                     
            loadbalancer_check_webservers(vicompress->loadbalancer);
            check_webservers = curr;
        }
        if (Config.logstats != NULL &&
            (curr - write_stats > LOGSTATS_INTERVAL)) {

            update_log_stats(Config.accesslog, Config.logstats, argv[0]);
            write_stats = curr;
        }

        set_current_time();
        curr = get_current_time();
    }
    /* Vicompress was stopped by a signal */
    if (vicompress->accesslog) {
        accesslog_close(vicompress->accesslog);
    }
    if (vicompress->errorlog) {
        errorlog_printf(vicompress->errorlog, NULL, "vicompress shutting down\n");
        errorlog_close(vicompress->errorlog);
    }
#ifdef WIN32
    if (!is_service) {
        exit(0);
    }
#else
    delete_pidfile();
    exit(0);
#endif
    return 0;
}

/* Usage:
 *   vicompress /path/to/vicompress.conf
 */
#ifndef WIN32
int main(int argc, char **argv)
{
    int ret = vicompress_start(argc, argv);
    return ret;
}
#endif



