/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#ifndef _VICOMPRESS_H_
#define _VICOMPRESS_H_

#include <netinet/in.h>
#include "conf.h"
#include "queue.h"
#include "http.h"
#include "buffer.h"
#include "cache.h"
#include "sockets.h"
#include "gzipencoder.h"
#include "loadbalancer.h"
#include "currenttime.h"
#include "log.h"
#include "chunked.h"
#include "version.h"

/* 
  A connection includes
  - The client socket and address
  - The server socket and address (if the reply is not cached)
  - The Http request and reply
  - Buffers for reading and writing.

  The amountread/amountwritten values are unsigned longs.  These are
  32 bits for 32-bit executables, and 64 bits for 64-bit executables.
  So a 32-bit vicompress will only support uploads/downloads up to
  4 GB.
*/
struct _Connection
{
    CachedItem *cacheditem;      /* Cached item to use */
    GzipEncoder *gzip;           /* Gzip encoder used to compress response */
    int error;                   /* Error code of socket errors */
    int clientfd;                /* The client socket */
    sockaddr_all clientip;       /* The client IP address */
    HttpRequest request;         /* The client's http request */
    HttpReply reply;             /* The server or cached http reply */
    int cache_buf_index;         /* The buffer index of a cached item */
    Parsedata parsedata;         /* Used for parsing HTTP request/reply */
    int serverfd;                /* The server socket */
    int webserverid;             /* The id of the backend webserver */
    int failed_connections;      /* The number of failed connections */
    sockaddr_all serverip;       /* The server ip address */
    sockaddr_all serverip_keepalive;  /* Used to keep track of persistent server connection */
    Buffer *serverread;          /* The buffer to read server data from */
    Buffer *clientwrite;         /* The buffer for writing to the client */
    Buffer *gzipout;             /* Buffer to store the gzipped output */
    Buffer *gzip_dechunked;      /* Buffer to store de-chunked data for gzipping */
    Buffer range_buf;            /* Buffer for cached "Range:" requests */
    unsigned long amountread;    /* The amount of content read from the server */
    unsigned long amountwritten; /* The amount of content written to the client */
    struct timeval start;        /* Time when request starts. */
    ChunkedTransfer chunked;     /* Data needed to support chunked-encoding */
    boolean chunked_done;        /* True if the chunked transfer finished */
    boolean chunk_encode_reply;  /* True if vicompress should chunk-encode the reply data */
    boolean chunk_encode_done;   /* True if vicompress has finished the hunk-encoded reply */
    boolean eof;                 /* True if the server closed the connection */
    boolean writing_to_client;   /* True if we're writing to the client */
    boolean reading_from_server; /* True if we're reading from server */
};
typedef struct _Connection Connection;

/* The Vicompress server. */
struct _Vicompress
{
    Connection *connpool;        /* Array of connection structures */
    Sockets *sockets;            /* Manager for socket reads/writes */
    BufferPool reqpool;          /* The http request buffer pool */
    BufferPool replypool;        /* The http reply buffer pool */
    BufferPool smallurlpool;     /* The small url buffer pool */
    BufferPool largeurlpool;     /* The large url buffer pool */
    BufferPool bodypool;         /* The body buffer pool */
    MemoryCache *cache;          /* The memory cache */
    GzipPool *gzippool;          /* The pool of gzip encoders */
    AccessLog *accesslog;        /* The access log */
    ErrorLog *errorlog;          /* The error log */
    LoadBalancer *loadbalancer;  /* The load balancer */
};
typedef struct _Vicompress Vicompress;

int vicompress_start(int argc, char **argv);
void vicompress_stop(int signal);

#endif

