/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#ifdef WIN32

#include <pthread.h>
#include <pwd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include "epoll.h"

/* 
 * This file contains an implementation of cygwin functions
 * used by vicompress, so that we don't have a dependency
 * on cygwin.dll at runtime.
 */

#undef stderr
#define stderr (fdopen(2, "w"))
void win_fprintf(void *fp, const char *fmt, ...);

/* Socket function implementation.
 * 
 * A big difference between cygwin and Windows with regards to sockets is
 * that cygwin uses integer file descriptors, allocated in increasing
 * order (3, 4, 5, etc...) while Windows uses opaque SOCKET pointers.
 * The vicompress code uses the file descriptors as an index into the
 * Sockets->data array, so it requires that sockets be integer file
 * descriptors between 0 and max sockets (16384).
 *
 * In win32.h, all socket functions, (socket, connect, send, recv, etc)
 * are translated into win_socket, win_connect, etc.  When socket()
 * is called, it picks an int file descriptor (fd) to be associated
 * with the native SOCKET handle, and returns the fd.  The win_xxx
 * functions map the file descriptor (fd) to a SOCKET, and pass the
 * SOCKET as argument to the real function call (connect, send, recv, etc).
 */

/* Struct needed by WSAStartup(), to initialize sockets on Windows */
typedef struct WSAData {
    short wVersion;
    short wHighVersion;
    char  szDescription[256+1];
    char  szSystemStatus[128+1];
    short iMaxSockets;
    short iMaxUdpDg;
    char *lpVendorInfo;
    char  pad[64];
} WSADATA;

struct sockaddr {
    char pad[16];
};

struct hostent {
    const char  *h_name;
    char       **h_aliases;
    short        h_addrtype;
    short        h_length;
    char       **h_addr_list;
};

/* Windows sockets are pointers, not integer file descriptors */
typedef unsigned int* SOCKET;

#define INVALID_SOCKET   (SOCKET)(~0)  /* Invalid sockets are 0xFFFFFFFF */
#define SOCKET_ERROR     (-1)          /* Socket functions return -1 on error */
#define MAX_SOCKETS      (20000)       /* Maximum sockets we support on Windows */

/* socket/fd pairs are used to map SOCKET to integer file descriptors */
#define SOCKPAIR_LEN     (MAX_SOCKETS*2 + 3)
typedef struct _sockpair {
    SOCKET sock;
    int fd;
} sockpair;

/* Array used to map fd to SOCKET */
static SOCKET fd_to_socket[MAX_SOCKETS];

/* Array used to map SOCKET to fd */
static sockpair socket_to_fd[SOCKPAIR_LEN];

/* A bitmap used to find free entries quickly in the
 * fd_to_socket array.
 */
static unsigned int freefd[MAX_SOCKETS/32];

void __stdcall Sleep(int millisec);
int  __stdcall GetLastError();
int  __stdcall WSAGetLastError();
int  __stdcall WSAStartup(unsigned short version, WSADATA *result);
SOCKET __stdcall socket(int af, int type, int proto);
SOCKET __stdcall accept(SOCKET s, struct sockaddr *addr, int* addrlen);
int __stdcall listen(SOCKET s, int backlog);
int __stdcall bind(SOCKET s, const struct sockaddr *addr, int addrlen);
int __stdcall connect(SOCKET s, struct sockaddr *addr, int addrlen);
int __stdcall recv(SOCKET s, char *buf, int amount, int flags);
int __stdcall send(SOCKET s, char *buf, int amount, int flags);
int __stdcall shutdown(SOCKET s, int how);
int __stdcall getsockopt(SOCKET s, int level, int optname, void *optval, int *optlen);
int __stdcall setsockopt(SOCKET s, int level, int optname, void *optval, int optlen);
int __stdcall ioctlsocket(SOCKET s, int cmd, unsigned long *arg);
int __stdcall closesocket(SOCKET s);
int __stdcall gethostname(char *name, int namelen);
struct hostent* __stdcall gethostbyname(const char *name);
unsigned long __stdcall inet_addr(const char *addr);


/* Given a SOCKET, return the corresponding SOCKET/fd pair,
 * or NULL if the SOCKET wasn't found.
 */
static sockpair* get_sockpair(SOCKET s)
{
    uint32_t hash, orighash;
    sockpair *pair;

    orighash = hash = ((uint32_t) s) % SOCKPAIR_LEN;
    do {
        pair = &(socket_to_fd[hash]);
        if (pair->sock == s) {
            return pair;
        }
        hash = (hash + 1) % SOCKPAIR_LEN;
    }
    while (hash != orighash);
    return NULL;
}

/* Given a SOCKET, allocate an integer to act as a file descriptor
 * for this SOCKET.  The fd is an index into the fd_to_socket array.
 * Use the freefd[] bitmap to quickly locate an area in fd_to_socket
 * with free entries. Return -1 if there are no free file descriptors.
 */
static int allocfd(SOCKET s)
{
    int fd;
    uint32_t hash;
    sockpair *pair;

    for (fd = 0; fd < MAX_SOCKETS; fd += 32)  {
        if (freefd[fd/32] != 0xFFFFFFFF) {
            break;
        }
    }

    if (fd == MAX_SOCKETS) {
        return -1;
    }

    while (fd < MAX_SOCKETS && fd_to_socket[fd] != INVALID_SOCKET) {
        fd++;
    }
    if (fd == MAX_SOCKETS) {
        return -1;
    }
    fd_to_socket[fd] = s;
    freefd[fd/32] |= (1 << (fd % 32));

    hash = ( (uint32_t)s ) % SOCKPAIR_LEN;
    while (1) {    
        pair = &(socket_to_fd[hash]);
        if (pair->sock == INVALID_SOCKET) {
            pair->sock = s;
            pair->fd = fd;
            break;
        }
        else {
            hash = (hash + 1) % SOCKPAIR_LEN;
        }
    }
    return fd;
}

/* Mark the given file descriptor as free. Update the
 * fd_to_socket and socket_to_fd arrays, to show that
 * the fd is now free.
 */
static void free_fd(int fd)
{
    SOCKET s;
    unsigned int n;
    sockpair *pair;

    assert(fd >= 0 && fd < MAX_SOCKETS);
    s = fd_to_socket[fd];
    fd_to_socket[fd] = INVALID_SOCKET;
    n = freefd[fd/32];
    n = n & ~(1 << (fd % 32));
    freefd[fd/32] = n;
    pair = get_sockpair(s);
    if (pair != NULL) {
        pair->sock = INVALID_SOCKET;
        pair->fd = -1;
    }
}


static int my_errno = 0;
int *win_errno = &my_errno;
static int winsock_called = 0;

/* Initialize the socket library on Windows. Use WinSock 2.0.
 * Initialize the arrays that map SOCKET -> fd and fd -> SOCKET.
 */
void winsock_init()
{
    unsigned short version;
    WSADATA *result;
    int i, n;


    winsock_called = 1;
    win_errno = &my_errno;
    version = 2;
    result = malloc(sizeof(WSADATA));
    n = WSAStartup(version, result);
    assert(n == 0);

    (void)memset(fd_to_socket, 0, sizeof(SOCKET) * MAX_SOCKETS);
    (void)memset(socket_to_fd, 0, sizeof(sockpair) * SOCKPAIR_LEN);
    (void)memset(freefd, 0, sizeof(int) * MAX_SOCKETS/32);

    for (i = 0; i < MAX_SOCKETS; i++) {
        fd_to_socket[i] = INVALID_SOCKET;
    }
    for (i = 0; i < SOCKPAIR_LEN; i++) {
        socket_to_fd[i].sock = INVALID_SOCKET;
        socket_to_fd[i].fd = -1;
    }

    /* Don't use fd 0, 1, 2 for sockets */
    allocfd((SOCKET) 1);
    allocfd((SOCKET) 2);
    allocfd((SOCKET) 3);
}

int win_socket(int af, int type, int proto)
{
    SOCKET s;
    int fd;

    if (!winsock_called) {
        winsock_init();
    }

    s = socket(af, type, proto);

    if (s == INVALID_SOCKET) {
        *win_errno = WSAGetLastError();
        return -1;
    }

    fd = allocfd(s);
    if (fd == -1) {
        closesocket(s);
    }
    return fd;
}

int win_listen(int fd, int backlog)
{
    return listen(fd_to_socket[fd], backlog);
}

int win_bind(int fd, const struct sockaddr *addr, int addrlen)
{
    int ret;
    ret = bind(fd_to_socket[fd], addr, addrlen);
    if (ret == SOCKET_ERROR) {
        *win_errno = WSAGetLastError();
    }
    return ret;
}

int win_accept(int listenfd, struct sockaddr *addr, int* addrlen)
{
    int fd;
    SOCKET s;

    s = accept(fd_to_socket[listenfd], addr, addrlen);
    if (s == INVALID_SOCKET) {
        return -1;
    }
    fd = allocfd(s);
    return fd;
}


int win_connect(int fd, struct sockaddr *addr, int addrlen)
{
    int ret;

    ret = connect(fd_to_socket[fd], addr, addrlen);
    if (ret == SOCKET_ERROR) {
        *win_errno = WSAGetLastError();
    }
    else {
        *win_errno = 0;
    }
    return ret;
}

int win_recv(int fd, char *buf, int amount, int flags)
{
    int ret;

    ret = recv(fd_to_socket[fd], buf, amount, flags);
    if (ret == SOCKET_ERROR) {
        *win_errno = WSAGetLastError();
    }
    else {
        *win_errno = 0;
    }
    return ret;
}

int win_send(int fd, char *buf, int amount, int flags)
{
    int ret;

    ret = send(fd_to_socket[fd], buf, amount, flags);
    if (ret == SOCKET_ERROR) {
        *win_errno = WSAGetLastError();
    }
    else {
        *win_errno = 0;
    }
    return ret;
}

int win_shutdown(int fd, int how)
{
    SOCKET s = fd_to_socket[fd];
    return shutdown(s, 0);
}

int win_getsockopt(int fd, int level, int optname, void *optval, int *optlen)
{
    SOCKET s = fd_to_socket[fd];
    return getsockopt(s, level, optname, optval, optlen);
}

int win_setsockopt(int fd, int level, int optname, void *optval, int optlen)
{
    SOCKET s = fd_to_socket[fd];
    return setsockopt(s, level, optname, optval, optlen);
}


int win_close(int fd)
{
    SOCKET s = fd_to_socket[fd];
    closesocket(s);
    free_fd(fd);
    return 0;
}

#undef FIONBIO
#define FIONBIO 0x8004667E

/* Implement non-blocking sockets,  fnctl(fd, O_NONBLOCK, 1),
 * by using the Windows function ioctlsocket().
 */
int win_fcntl(int fd, int cmd, int arg)
{
    unsigned long n = 1;

    if (cmd != F_SETFL && arg != O_NONBLOCK) {
        return -1;
    }
    return ioctlsocket(fd_to_socket[fd], FIONBIO, &n);
}

int win_gethostname(char *name, int namelen) {
    return gethostname(name, namelen);
}

unsigned long win_inet_addr(const char *addr)
{
    return inet_addr(addr);
}

unsigned short win_htons(int x)
{
    return (x % 256) * 256 + (x/256);
}

unsigned short win_ntohs(unsigned short x)
{
    return (x % 256) * 256 + (x/256);
}

struct hostent* win_gethostbyname(const char *name)
{
    return gethostbyname(name);
}

/* epoll implementation
 *
 * Below is an emulation of the Linux epoll functions using the native
 * select() call on Windows.  In cygwin, the fd_set structures are simply
 * bitmaps, where each bit corresponds to an integer file descriptor.
 * On Windows, the fd_set struct is actually an array of SOCKETs.  So
 * we cannot use the cygwin structs/macros for fd_set structs.
 */

struct timeval {
    long tv_sec;
    long tv_usec;
};

/* In Windows, the fd_set struct is actually an array of SOCKET */
typedef struct _fdset {
    int count;
    SOCKET array[MAX_SOCKETS];
} fdset;

int __stdcall select(int unused, fdset *read, fdset *write, 
                     fdset *err, struct timeval *tv);

static fdset  readset;
static fdset  writeset;
static fdset  errorset;
static fdset  readsetcopy;
static fdset  writesetcopy;
static fdset  errorsetcopy;

static int   *fd_to_read_index;
static int   *fd_to_write_index;
static int   *fd_to_error_index;
static int   *read_index_to_fd;
static int   *write_index_to_fd;
static int   *error_index_to_fd;

static struct epoll_event *epollevents = NULL;
int arraysize = 0;

int epoll_create(int size)
{
    int i;

    arraysize = size+1;
    epollevents = (struct epoll_event*) calloc(size, sizeof(struct epoll_event));
    fd_to_read_index = (int*) calloc(size, sizeof(int));
    fd_to_write_index = (int*) calloc(size, sizeof(int));
    fd_to_error_index = (int*) calloc(size, sizeof(int));
    read_index_to_fd = (int*) calloc(size, sizeof(int));
    write_index_to_fd = (int*) calloc(size, sizeof(int));
    error_index_to_fd = (int*) calloc(size, sizeof(int));

    for (i = 0; i < size; i++) {
        epollevents[i].data.fd = -1;
        epollevents[i].events = 0;
        fd_to_read_index[i] = -1;
        fd_to_write_index[i] = -1;
        fd_to_error_index[i] = -1;
        read_index_to_fd[i] = -1;
        write_index_to_fd[i] = -1;
        error_index_to_fd[i] = -1;
    }
    for (i = 0; i < MAX_SOCKETS; i++) {
        readset.array[i] = INVALID_SOCKET;
        writeset.array[i] = INVALID_SOCKET;
        errorset.array[i] = INVALID_SOCKET;
    }
    readset.count = 0;
    writeset.count = 0;
    errorset.count = 0;
    return 1;
}

static void FDSET(int fd, fdset *set, int *fd_to_index, int *index_to_fd)
{
    fd_to_index[fd] = set->count;
    index_to_fd[set->count] = fd;
    set->array[set->count] = fd_to_socket[fd];
    set->count++;
}

static void FDCLR(int fd, fdset *set, int *fd_to_index, int *index_to_fd)
{
    int index, lastfd;

    index = fd_to_index[fd];
    fd_to_index[fd] = -1;

    if (index != (set->count - 1)) {
        /* Move the last entry in the set array
         * to this entry, and update the indexes.
         */
        lastfd = index_to_fd[set->count-1];
        set->array[index] = fd_to_socket[lastfd];
        fd_to_index[lastfd] = index;
        index_to_fd[index] = lastfd;
    }
    index_to_fd[set->count-1] = -1;
    set->array[set->count-1] = INVALID_SOCKET;
    set->count--;
}


int epoll_ctl(int epfd, int op, int fd, struct epoll_event *event)
{
    struct epoll_event *ep;
    int oldevents;

    if (fd < 0 || fd >= arraysize || fd >= MAX_SOCKETS) {
        return -1;
    }

    ep = &epollevents[fd];
    oldevents = ep->events;
    ep->events = event->events & (EPOLLIN | EPOLLOUT);
    ep->data = event->data;


    if (op == EPOLL_CTL_ADD) {
        assert(event->events & EPOLLIN);
        FDSET(fd, &readset, fd_to_read_index, read_index_to_fd);
        FDSET(fd, &errorset, fd_to_error_index, error_index_to_fd);
    }
    else if (op == EPOLL_CTL_MOD && ep->events == oldevents) {
    }
    else if (op == EPOLL_CTL_MOD && ep->events != oldevents) {
        if (oldevents & EPOLLIN) {
            FDCLR(fd, &readset, fd_to_read_index, read_index_to_fd);
            FDCLR(fd, &errorset, fd_to_error_index, error_index_to_fd);
        }
        if (oldevents & EPOLLOUT) {
            FDCLR(fd, &writeset, fd_to_write_index, write_index_to_fd);
            FDCLR(fd, &errorset, fd_to_error_index, error_index_to_fd);
        }
        if (ep->events & EPOLLIN) {
            FDSET(fd, &readset, fd_to_read_index, read_index_to_fd);
            FDSET(fd, &errorset, fd_to_error_index, error_index_to_fd);
        }
        if (ep->events & EPOLLOUT) {
            FDSET(fd, &writeset, fd_to_write_index, write_index_to_fd);
            FDSET(fd, &errorset, fd_to_error_index, error_index_to_fd);
        }
    }
    else if (op == EPOLL_CTL_DEL) {
        if (oldevents & EPOLLIN) {
            FDCLR(fd, &readset, fd_to_read_index, read_index_to_fd);
            FDCLR(fd, &errorset, fd_to_error_index, error_index_to_fd);
        }
        if (oldevents & EPOLLOUT) {
            FDCLR(fd, &writeset, fd_to_write_index, write_index_to_fd);
            FDCLR(fd, &errorset, fd_to_error_index, error_index_to_fd);
        }
    }
    return 0;
}

int epoll_wait(int epfd, struct epoll_event *events,
               int maxevents, int msec)
{
    int fd, index, i, n;
    struct epoll_event *ep;
    struct timeval tv;
    sockpair *pair;

    tv.tv_sec = msec / 1000;
    tv.tv_usec = (msec % 1000) * 1000;

    memcpy(&readsetcopy, &readset, sizeof(fdset));
    memcpy(&writesetcopy, &writeset, sizeof(fdset));
    memcpy(&errorsetcopy, &errorset, sizeof(fdset));

    n = select(0, 
               readsetcopy.count == 0 ? NULL : &readsetcopy, 
               writesetcopy.count == 0 ? NULL: &writesetcopy, 
               errorsetcopy.count == 0 ? NULL : &errorsetcopy, 
               &tv);

    if (n < 0) {
        return 0;
    }

    index = 0;
    for (i = 0; i < readsetcopy.count; i++) {
        if (index >= maxevents) {
            return index;
        }
        if (readsetcopy.array[i] == INVALID_SOCKET) {
            continue;
        }
        pair = get_sockpair(readsetcopy.array[i]);
        assert(pair != NULL);
        fd = pair->fd;
        assert(fd_to_socket[fd] == readsetcopy.array[i]);
        ep = &(epollevents[fd]);
        events[index].events = ep->events;
        events[index].data.fd = fd;
        index++;

    }
    for (i = 0; i < writesetcopy.count; i++) {
        if (index >= maxevents) {
            return index;
        }
        if (writesetcopy.array[i] == INVALID_SOCKET) {
            continue;
        }
        pair = get_sockpair(writesetcopy.array[i]);
        fd = pair->fd;
        assert(fd_to_socket[fd] == writesetcopy.array[i]);
        ep = &(epollevents[fd]);
        events[index].events = ep->events;
        events[index].data.fd = fd;
        index++;
    }
    for (i = 0; i < errorsetcopy.count; i++) {
        if (index >= maxevents) {
            return index;
        }
        if (errorsetcopy.array[i] == INVALID_SOCKET) {
            continue;
        }
        pair = get_sockpair(errorsetcopy.array[i]);
        fd = pair->fd;
        assert(fd_to_socket[fd] == errorsetcopy.array[i]);
        ep = &(epollevents[fd]);
        events[index].events = ep->events;
        events[index].data.fd = fd;
        index++;
    }
    return index;
}




/*
 * pthread function implementations
 */
struct _CRITICAL_SECTION {
    char lock[80];
};
typedef struct _CRITICAL_SECTION CRITICAL_SECTION;

static int lockindex = 0;
static CRITICAL_SECTION mylocks[16];

struct _pthread_wrapper_t {
    void* (*func)(void *);
    void *arg;
};
typedef struct _pthread_wrapper_t pthread_wrapper_t;

int pthread_attr_init(pthread_attr_t *attr)
{
    memset(attr, 0, sizeof(pthread_attr_t));
    return 0;
}

int pthread_attr_setdetachstate(pthread_attr_t *attr, int value)
{
    return 0;
}

static int thread_wrapper(void *args)
{
    void *result;
    pthread_wrapper_t *wrap = (pthread_wrapper_t *) args;

    result = wrap->func(wrap->arg);
    return 0; 
}

unsigned int __stdcall
CreateThread(void *attr, int stacksize, int(*)(void*), 
             void* arg, int flags, unsigned int *id);


int pthread_create(pthread_t *tid, const pthread_attr_t *attr, 
                   void *(*func)(void*), void *arg) 
{
    unsigned int threadid;
    unsigned int result;
    pthread_wrapper_t *wrap;

    wrap = (pthread_wrapper_t *)malloc(sizeof(pthread_wrapper_t));
    wrap->func = func;
    wrap->arg = arg;

    result = CreateThread(NULL, 0, thread_wrapper, wrap, 0, &threadid);
    if (result == 0) {
        return -1;
    }
    else {
        *tid = (pthread_t) result;
        return 0;
    } 
}

void __stdcall InitializeCriticalSection(CRITICAL_SECTION *x);
void __stdcall EnterCriticalSection(CRITICAL_SECTION *x);
void __stdcall LeaveCriticalSection(CRITICAL_SECTION *x);

int pthread_mutex_init(pthread_mutex_t *m, const pthread_mutexattr_t *attr)
{
    uint32_t *index = (uint32_t *)m;
    *index = lockindex;
    lockindex++;
    InitializeCriticalSection(&mylocks[*index]);
    return 0;
}

int pthread_mutex_lock(pthread_mutex_t *m) 
{
    uint32_t *index = (uint32_t *) m;
    EnterCriticalSection(&mylocks[*index]);
    return 0;
}

int pthread_mutex_unlock(pthread_mutex_t *m)
{
    uint32_t *index = (uint32_t *)m;
    LeaveCriticalSection(&mylocks[*index]);
    return 0;
}

uint32_t __stdcall CreateEventA(void *attr, int manual, int state, char *name);
int      __stdcall SetEvent(uint32_t event);
void     __stdcall WaitForSingleObject(unsigned int event, int timeout);

int pthread_cond_init(pthread_cond_t *cond, const pthread_condattr_t *attr)
{
    uint32_t* event = (uint32_t*) cond;
    *event = CreateEventA(NULL, 0, 0, NULL);
    return 0;
}

int pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *m)
{
    uint32_t *event = (uint32_t *) cond;
    uint32_t *index = (uint32_t *)m;

    LeaveCriticalSection(&mylocks[*index]);
    WaitForSingleObject(*event, 30);
    EnterCriticalSection(&mylocks[*index]);
    return 0;
}

int pthread_cond_signal(pthread_cond_t *cond)
{
    SetEvent((unsigned int) cond); 
    return 0;
}



/* Misc function implementations */

int chown(const char *file, uid_t u, gid_t g) { return 0; }
uid_t getuid() { return 0; }
gid_t getgid() { return 0; }
int setuid(uid_t x) {return 0;}
int setgid(gid_t x) {return 0;}

static struct passwd pw;
struct passwd* getpwnam(const char *username)
{
  memset(&pw, 0, sizeof(struct passwd));
  return &pw;
}

/* Change all forward slashes (/) to back slashes (\) in the filename */
static void forward_to_backslash(const char *filename)
{
    int i, len;
    char *name = (char *)filename;

    len = strlen(name);
    for (i = 0; i < len; i++) {
        if (name[i] == '/') {
            name[i] = '\\';
        }
    }
}

#undef O_BINARY
#define O_BINARY  0x8000
int win_open(const char *name, int flags, int mode)
{
    int ret;
    char *filename = (char *)name;
    
    if (strlen(filename) > 2 && filename[0] == '.' && filename[1] == '/') {
        filename = (char *) &name[2];
    }
    forward_to_backslash(filename);
    if (flags & O_RDONLY) {
        flags |= O_BINARY;
    }
    ret = open(filename, flags, mode);
    if (ret == -1) {
        *win_errno = GetLastError();
    }
    return ret;
}

void* win_fopen(const char *name, char *mode)
{
    FILE *result;
    char *filename = (char *)name;

    if (strlen(filename) > 2 && filename[0] == '.' && filename[1] == '/') {
        filename = (char *) &name[2];
    }
    forward_to_backslash(filename);
    if (strcmp(mode, "r") == 0) {
        mode = "rb";
    }
    result = fopen(filename, mode);
    if (result == NULL) {
        *win_errno = GetLastError();
    }
    return result;
}

struct win_stat {
    unsigned int st_dev;
    unsigned short st_ino;
    unsigned short st_mode;
    short st_nlink;
    short st_uid;
    short st_gid;
    unsigned int st_rdev;
    long st_size;
    long st_atimme;
    long st_mtimme;
    long st_ctimme;
};

int win_stat(const char *name, struct win_stat *info)
{
    int ret = stat(name, (struct stat *)info);
    if (ret != 0) {
        *win_errno = GetLastError();
    }
    return ret;
}

char *dirname(char *path)
{
    int i, len;
    len = strlen(path);

    for (i = len-1; i >= 0; i--) {
        if (path[i] == '\\') {
            path[i] = 0;
            return path;
        }
    }
    return "";
}


int strncasecmp(const char *s1, const char *s2, size_t n)
{
    int i;
    char c1, c2;

    /* In ascii, A = 65, a = 97, difference is 32 */
    for (i = 0; i < n; i++) {
        if (s1[i] == s2[i]) {
            continue;
        }
        c1 = s1[i];
        c2 = s2[i];

        if (c1 >= 'A' && c1 <= 'Z') {
            c1 = (char) (c1 + 32);
        }
        if (c2 >= 'A' && c2 <= 'Z') {
            c2 = (char) (c2 + 32);
        }
        if (c1 != c2) {
            return c1 < c2 ? -1 : 1;
        }
    }
    return 0;
}

struct tm* localtime_r(const time_t *sec, struct tm *date)
{
    struct tm *ret = localtime(sec);
    memcpy(date, ret, sizeof(struct tm));
    return NULL;
}

struct timezone {
    int tz_minuteswest;
    int tz_dsttime;
};

typedef struct _SYSTEMTIME {
    short wYear;
    short wMonth;
    short wDayOfWeek;
    short wDay;
    short wHour;
    short wMinute;
    short wSecond;
    short wMilliseconds;
} SYSTEMTIME;

void __stdcall GetSystemTime(SYSTEMTIME *t);

static int daysinmonth[] = {
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};


static time_t do_mktime(struct tm *tmval)
{
    int leapdays;
    time_t result;
    int i;

    result = 365*24*60*60*(tmval->tm_year - 70);
    result = result + tmval->tm_sec + tmval->tm_min*60 + tmval->tm_hour*3600;

    for (i = 0; i < tmval->tm_mon; i++) {
        result = result + daysinmonth[i]*24*60*60;
    }
    result = result + (tmval->tm_mday - 1)*24*60*60;

    leapdays = (tmval->tm_year - 69) / 4;
    if ((tmval->tm_year % 4) == 0) {
        if (tmval->tm_mon > 1) {
            leapdays++;
        }
    }
    result = result + leapdays*24*60*60;
    return result;
}

int gettimeofday(struct timeval *tv, struct timezone *tz)
{
    SYSTEMTIME sys;
    struct tm tmval;
    time_t thetime;

    GetSystemTime(&sys);
    tmval.tm_year = sys.wYear - 1900;
    tmval.tm_mon = sys.wMonth - 1;
    tmval.tm_mday = sys.wDay;
    tmval.tm_hour = sys.wHour;
    tmval.tm_min = sys.wMinute;
    tmval.tm_sec = sys.wSecond;
    thetime = do_mktime(&tmval);
    
    tv->tv_sec = thetime;
    tv->tv_usec = sys.wMilliseconds * 1000;
    return 0;
}


unsigned int sleep(unsigned int sec)
{
    Sleep(sec * 1000);
    return sec;
}

unsigned usleep(unsigned int usec)
{
    Sleep(usec);
    return usec;
}

void __assert(const char *file, int line, const char *msg) {
    win_fprintf(stderr, "assert failed: %s:%d %s\n", file, line, msg);
    exit(1);
}

static char errstring[64];
char* strerror(int err) {
    switch (err) {
        case 2: return "No such file or directory";
        case 3: return "No such file or directory";
        case 112: return "Disk is full";
        case 123: return "Invalid file name";
        case 10004: return "Interrupted system call";
        case 10024: return "Too many open files";
        default: sprintf(errstring, "errno %d", err); return &errstring[0];
    }
}

int win_isdigit(char c) {
    if (c >= '0' && c <= '9') {
        return 1;
    }
    else {
        return 0;
    }
}


typedef struct _STARTUPINFO {
    char data[72];
} STARTUPINFO;

typedef struct _PROCESS_INFORMATION {
    char data[20];
} PROCESS_INFORMATION;

int __stdcall 
CreateProcessA(void*, char *, void*, void*, int, int,
               char *, void*, STARTUPINFO*, PROCESS_INFORMATION*);

int fork() { return 0; }

int execv(const char *path, char * const argv[])
{
    STARTUPINFO info;
    PROCESS_INFORMATION pi;
    char env[32767];
    char command[1024];
    int i, n;

    memset(env, 0, sizeof(env));
    memset(&info, 0, sizeof(info));
    memset(&pi, 0, sizeof(pi));

    sprintf(command, "%s.exe", path);
    forward_to_backslash(command);

    i = 1;
    while (argv[i] != NULL) {
        strcat(command, " \"");
        strcat(command, argv[i]);
        strcat(command, "\"");
        i++;
    }

    n = CreateProcessA(NULL, command, NULL, NULL, 0, 0,
                       NULL, NULL, &info, &pi);
    if (n != 0) {
        return 1;
    }
    else {
        win_fprintf(stderr, "Error: Failed to execute program %s\n",
                    command);
        return -1;
    }
}


int __stdcall GetCurrentProcessId(void);

int getpid()
{
  return GetCurrentProcessId();
}

typedef struct _SYSTEM_INFO {
    int oemid;
    int pagesize;
    void* min_addr;
    void* max_addr;
    int* processor_mask;
    int num_processors;
    int processor_type;
    int allocation;
    short processor_level;
    short processor_revision;
} SYSTEM_INFO;

void __stdcall GetSystemInfo(SYSTEM_INFO *info);

int win_get_num_cpus()
{
    SYSTEM_INFO info;
    memset(&info, 0, sizeof(SYSTEM_INFO));
    GetSystemInfo(&info);
    if (info.num_processors <= 0) {
        info.num_processors = 1;
    }
    return info.num_processors;
}


/* Windows Event Log */

#define HKEY_LOCAL_MACHINE ((HKEY) 0x80000002)
#define KEY_WRITE                  0x00020006
#define REG_OPTION_NON_VOLATILE    0x01
#define EVENTLOG_ERROR_TYPE        0x01
#define EVENTID                    0x01
#define REG_SZ                     0x01


typedef void* HKEY;
typedef void* HANDLE;

int __stdcall RegCloseKey(HKEY hkey);
int __stdcall RegOpenKeyExA(HKEY hkey, char *subkey, int options, int access, HKEY *result);
int __stdcall RegCreateKeyExA(HKEY hkey, char *subkey, int reserved, char *class, int options,
                    int access, void *attr, HKEY *outkey, unsigned int *result);

int __stdcall RegSetValueExA(HKEY hkey, char *name, int reserved, int type, 
                   char *data, int datalen);
HANDLE __stdcall RegisterEventSourceA(char *server, char *source);
int __stdcall ReportEventA(HANDLE event, short type, short category, int eventid,
                 void* sid, short num_strings, int datalen, char **strings, char *data);
int __stdcall DeregisterEventSource(HANDLE h);
void __stdcall CloseHandle(HANDLE h);

/* Print the given message to the FILE* pointer p.  If the fp is stderr
 * (fd 2), then also print the message to the Windows Event Log.
 *
 * In order to write to the event log, we must:
 * - Have an event "source" in the Registry.  The source is located at:
 *    HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\EventLog\
 *    Application\ViCompress
 * - Get a handle to the ViCompress event source
 * - Write (Report) the message using the event source handle.
 */
void win_fprintf(void *fp, const char *fmt, ...)
{
    va_list args;
    char msg[256];
    char *msgs[2];
    HANDLE eventhandle;
    HKEY hkey;
    unsigned int result;
    const char *msgfile = "%SystemRoot%\\system32\\netmsg.dll";
    int ret;

    memset(msg, 0, sizeof(msg));
    va_start(args, fmt);
    vfprintf(fp, fmt, args);
    va_end(args);

    va_start(args, fmt);
    vsnprintf(msg, 256, fmt, args);
    va_end(args);

    if (fileno(fp) != 2) {
        return;
    }

    ret = RegOpenKeyExA(HKEY_LOCAL_MACHINE, 
            "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\ViCompress",
            0,
            KEY_WRITE,
            &hkey);

    if (ret != 0) {
        ret = RegCreateKeyExA(HKEY_LOCAL_MACHINE, 
            "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\ViCompress",
            0,
            NULL,
            REG_OPTION_NON_VOLATILE,
            KEY_WRITE,
            NULL,
            &hkey,
            &result);
        if (ret != 0) {
            return;
        } 
        ret = RegSetValueExA(hkey, "EventMessageFile", 0, REG_SZ, 
                             (char *)msgfile, strlen(msgfile) + 1);
        
        if (ret != 0) {
            return;
        }
    }

    RegCloseKey(hkey);
    eventhandle = RegisterEventSourceA(NULL, "ViCompress");
    if (eventhandle == NULL) {
        return;
    }
    msgs[0] = msg;
    msgs[1] = NULL;
    ReportEventA(eventhandle, EVENTLOG_ERROR_TYPE, 0, EVENTID, NULL, 1, 0,
                msgs, NULL);
    DeregisterEventSource(eventhandle);
    CloseHandle(eventhandle);
}



#endif /* WIN32 */


