/*
 * Copyright (c) 2003-2008 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/* 
 * Functions and structs for installing, uninstalling, starting, stopping,
 * a Windows Service.
 */

#ifdef WIN32
#include <string.h>
#include <stdio.h>
#include "vicompress.h"
#include "win32.h"

/* Service control codes */
#define SERVICE_CONTROL_STOP      0x01
#define SERVICE_CONTROL_PAUSE     0x02
#define SERVICE_CONTROL_CONTINUE  0x03
#define SERVICE_CONTROL_SHUTDOWN  0x05

/* Service states */
#define SERVICE_STOPPED        0x01
#define SERVICE_START_PENDING  0x02
#define SERVICE_STOP_PENDING   0x03
#define SERVICE_RUNNING        0x04

/* Service accepted controls */
#define SERVICE_ACCEPT_STOP            0x01
#define SERVICE_ACCEPT_PAUSE_CONTINUE  0x02
#define SERVICE_ACCEPT_SHUTDOWN        0x04

/* Service access levels */
#define SERVICE_ALL_ACCESS    0x000F01FF

/* Service type */
#define SERVICE_WIN32_OWN_PROCESS  0x10
#define SERVICE_WIN32              0x30

/* Start type */
#define SERVICE_AUTO_START         0x02
#define SERVICE_DEMAND_START       0x03

/* Error control types */
#define SERVICE_ERROR_NORMAL       0x01

/* Service Control Manager (SCM) access types */
#define SC_MANAGER_CREATE_SERVICE  0x00000001
#define SC_MANAGER_ALL_ACCESS      0x000F003F

/* Info levels for ChangeServiceConfig2 */
#define SERVICE_CONFIG_DESCRIPTION  0x01

/* ViCompress service strings */
#define VICOMPRESS_NAME  "ViCompress"
#define VICOMPRESS_EXE   "vicompress.exe"
#define VICOMPRESS_CONF  "vicompress.conf.txt"
#define VICOMPRESS_DESC  "An HTTP Compression and Caching Proxy Server"

typedef struct _SERVICE_STATUS {
    int dwServiceType;
    int dwCurrentState;
    int dwControlsAccepted;
    int dwWin32ExitCode;
    int dwServiceSpecificExitCode;
    int dwCheckPoint;
    int dwWaitHint;
} SERVICE_STATUS;

typedef struct _SERVICE_DESCRIPTION {
    char *desc;
} SERVICE_DESCRIPTION;

typedef void* SERVICE_STATUS_HANDLE;
typedef void* SC_HANDLE;
typedef void (*LPSERVICE_MAIN_FUNCTION)(int argc, char**args);
typedef void (*LPHANDLER_FUNCTION)(int code);

typedef struct _SERVICE_TABLE_ENTRY {
    char *lpServiceName;
    LPSERVICE_MAIN_FUNCTION lpServiceProc;
} SERVICE_TABLE_ENTRY;

void* malloc(int amount);
int GetLastError();
int __stdcall StartServiceCtrlDispatcherA(SERVICE_TABLE_ENTRY *table);
SERVICE_STATUS_HANDLE __stdcall RegisterServiceCtrlHandlerA(char *name, LPHANDLER_FUNCTION proc);
int __stdcall SetServiceStatus(SERVICE_STATUS_HANDLE handle, SERVICE_STATUS *status);
SC_HANDLE __stdcall OpenSCManagerA(char *machine, char *database, int access);
SC_HANDLE __stdcall OpenServiceA(SC_HANDLE SCManager, const char *name, int access);
int __stdcall CloseServiceHandle(SC_HANDLE handle);
int __stdcall DeleteService(SC_HANDLE handle);
int __stdcall ChangeServiceConfig2A(SC_HANDLE handle, int level, SERVICE_DESCRIPTION *desc);
SC_HANDLE __stdcall CreateServiceA(SC_HANDLE mgr, const char *name, const char *display,
                        int access, int service, int start,  int error,
                        char *exepath, char *group, int* tag,
                        const char *dependencies, char *nameptr, char *pass);


/* The status of vicompress (start_pending, running, stopped) */
static SERVICE_STATUS service_status;

/* The handle to the Service */
static SERVICE_STATUS_HANDLE service_handle = NULL;

/* The full path of the vicompress binary */
static char *vicompress_path = NULL;

/* The full path to the config file */
static char *configfile = NULL;

/* 1 if vicompress is running as a service, else 0 */
int is_service = 0;

/* The vicompress Service handler function.  Currently, only
 * the "stop" command is implemented.
 */
static void service_ctrl_handler(int code)
{
    switch (code) {
      case SERVICE_CONTROL_STOP:
      case SERVICE_CONTROL_SHUTDOWN:
          service_status.dwWin32ExitCode = 0;
          service_status.dwCurrentState  = SERVICE_STOPPED;
          service_status.dwCheckPoint    = 0;
          service_status.dwWaitHint      = 0;
          SetServiceStatus(service_handle, &service_status);
          vicompress_stop(SIGINT); 
          return;

      default:
          return;
   }
}


/* The callback function for starting a windows service.
 * Register the handler function (used to stop vicompress),
 * and then call vicompress_start().
 */
static void service_start(int argc, char **argv)
{ 
    char **params;

    service_status.dwServiceType        = SERVICE_WIN32;
    service_status.dwCurrentState       = SERVICE_START_PENDING;
    service_status.dwControlsAccepted   = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN;
    service_status.dwWin32ExitCode      = 0;
    service_status.dwServiceSpecificExitCode = 0;
    service_status.dwCheckPoint         = 1;
    service_status.dwWaitHint           = 1000;

    service_handle = RegisterServiceCtrlHandlerA(
        VICOMPRESS_NAME, service_ctrl_handler);

    if (service_handle == (SERVICE_STATUS_HANDLE)0) {
        fprintf(stderr, "Unable to Register Service Ctrl Handler\n");
        return;
    }

    SetServiceStatus(service_handle, &service_status);

    /* The Windows Service only passes the service name (ViCompress) as
     * arguments to service_start(), so ignore the argc/argv parameters.
     * We want to pass the vicompress.conf.txt path as argument to the
     * vicompress_start() routine.
     */
    params = malloc(sizeof(char *) * 3);
    params[0] = vicompress_path;
    params[1] = configfile;
    params[2] = NULL;

    if (vicompress_start(2, params) != 0) {
        service_status.dwCurrentState       = SERVICE_STOPPED;
        service_status.dwCheckPoint         = 0;
        service_status.dwWaitHint           = 0;
        service_status.dwWin32ExitCode      = -1;
        service_status.dwServiceSpecificExitCode = 0;
        SetServiceStatus(service_handle, &service_status);
        return;
    }
}

/* Update the status of the service, to indicate that it's
 * running successfully.
 */
void service_running() {
    service_status.dwCurrentState = SERVICE_RUNNING;
    service_status.dwWaitHint     = 0;
    service_status.dwCheckPoint++;
    SetServiceStatus(service_handle, &service_status);
}


/* Install vicompress as a Windows Service.  The install directory is
 * passed as argument.  Return 0 on success, -1 on error.
 * This function is called during installation (see vicompress.nsi).
 * At install time, the installer will invoke the following command:
 *   vicompress -i <install dir>
 * which will in turn call this function.
 */
int install_service(const char *dir)
{
    SC_HANDLE Service;
    SC_HANDLE SCManager;
    char cmd[512];
    SERVICE_DESCRIPTION desc = { VICOMPRESS_DESC };

    sprintf(cmd, "\"%s\\%s\" -s \"%s\\%s\"", 
            dir, VICOMPRESS_EXE, dir, VICOMPRESS_CONF);

    SCManager = OpenSCManagerA(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!SCManager) {
        return -1;
    }
    Service = CreateServiceA(SCManager,                 /* Manager */
                            VICOMPRESS_NAME,           /* Service name */
                            VICOMPRESS_NAME,           /* Display name */
                            SERVICE_ALL_ACCESS,        /* Access level */
                            SERVICE_WIN32_OWN_PROCESS, /* Service type */
                            SERVICE_DEMAND_START,      /* Start type */
                            SERVICE_ERROR_NORMAL,      /* Error control type */
                            cmd,                       /* Command exe */
                            NULL,                      /* Load group */
                            NULL,                      /* Tag */
                            "Tcpip\0Afd\0",            /* Dependencies */
                            NULL,                      /* System account */
                            NULL);                     /* Password */

    if (!Service) {
        CloseServiceHandle(SCManager);
        return -1;
    }
    else {
        ChangeServiceConfig2A(Service, SERVICE_CONFIG_DESCRIPTION, &desc);
        CloseServiceHandle(Service);
        CloseServiceHandle(SCManager);
        return 0;
    }
}

/* Uninstall vicompress as a Windows Service. Return 0 on success, -1 on error.
 * This function is called during un-installation (see vicompress.nsi).
 * At uninstall time, the uninstaller will invoke the following command:
 *   vicompress -u 
 * which will in turn call this function.
 */
int uninstall_service()
{
    SC_HANDLE Service;
    SC_HANDLE SCManager;
    int ret;

    SCManager = OpenSCManagerA(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!SCManager) {
        return -1;
    }
    Service = OpenServiceA(SCManager, VICOMPRESS_NAME, SERVICE_ALL_ACCESS);
    if (!Service) {
        return 0;
    }
    ret = DeleteService(Service);
    CloseServiceHandle(Service);
    CloseServiceHandle(SCManager);
    if (ret == 0) {
        return -1;
    } 
    else {
        return 0;
    }
}

/* Usage:
 *  vicompress -i <install direcotry>
 *    Install ViCompress as a Windows Service.
 *  vicompress -u
 *    Uninstall ViCompress as a Windows Service.
 *  vicompress -s </path/to/vicompress.conf.txt>
 *    Start vicompress from a Windows Service.
 *  vicompress </path/to/vicompress.conf.txt>
 *    Start vicompress from a command line
 */
int main(int argc, char **argv)
{
    SERVICE_TABLE_ENTRY table[] = {
        { VICOMPRESS_NAME, service_start },
        { NULL, NULL }
    };

    if (argc == 3 && strcmp(argv[1], "-i") == 0) {
        if (install_service(argv[2]) == -1) {
            printf("Error: Unable to install ViCompress as a Service\n");
            sleep(5);
            return -1;
        }
        return 0; 
    }
    else if (argc == 2 && strcmp(argv[1], "-u") == 0) {
        if (uninstall_service() == -1) {
            printf("Error: Unable to un-install ViCompress as a Service\n");
            sleep(5);
            return -1;
        }
        return 0;
    }
    else if (argc == 3 && strcmp(argv[1], "-s") == 0) {
        vicompress_path = argv[0];
        configfile = argv[2];
        is_service = 1;
        if (!StartServiceCtrlDispatcherA(table)) {
            return 1;
        }
        return 0;
    }
    else {
        return vicompress_start(argc, argv);
    }
}


#endif /* WIN32 */
