/*
 * Copyright (c) 2003-2007 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "http.h"

/* Find char c in buffer buf, between indexes offset and end.
 * Return the offset where it occurs, or -1 if not found.
 */
static int findchar(const char *buf, int offset, int end, char c)
{
    if (offset < 0)
        return -1;
    while (offset < end) {
        if (buf[offset] == c)
            return offset;
        offset++;
    }
    return -1;
}

/* Find the null-terminated string in buf, between indexes offset and end.
 * Return the offset where the string starts, or -1 if not found.
 */
static int findstring(const char *buf, int offset, int end, const char *string)
{
    int stringlen;
    stringlen = strlen(string);

    while (stringlen <= end - offset) {
        if (strncasecmp(&buf[offset], string, stringlen) == 0)
            return offset;
        offset++;
    }
    return -1;
}


/* Skip char c in the buf, between offset and end. Return the offset
 * where char c ends.
 */
static int skipchar(const char *buf, int offset, int end, char c)
{
    if (offset < 0)
        return -1;
    while (offset < end) {
        if (buf[offset] != c)
            return offset;
        offset++;
    }
    return offset;
}


static int request_is_complete(HttpRequest *req, char *buffer, int len)
{
    int i;
    int lastoffset = req->requestlength;
    if (lastoffset < 2)
        lastoffset = 2;

    for (i = lastoffset; i < len; i++) {
        if (buffer[i] == '\n' &&
                (buffer[i-1] == '\n' || buffer[i-2] == '\n')) {
            req->parsestatus = 1;
            req->requestlength = i+1;
            return 1;
        }
    }
    return 0;
}

static int request_parse_url(HttpRequest *req, char *buf, int len)
{
    int endline, offset1, offset2, amount;

    endline = findchar(buf, 0, len, '\n');
    if (endline == -1)
        return 0;

    if (findstring(buf, 0, endline, "cachable") > 0)
        req->iscachable = 1;
    if (findstring(buf, 0, endline, "cangzip") > 0)
        req->cangzip = 1;

    offset1 = findchar(buf, 0, len, ' ');
    offset2 = findchar(buf, offset1+1, len, ' ');
    amount = offset2 - (offset1 + 1);
    if (amount > 31)
        amount = 31;
    memcpy(req->url, &buf[offset1+1], amount);
    req->url[amount] = 0;
    return 1;
}

int request_parse(HttpRequest *request, char *buf, int len)
{
    int result;

    if (request->parsestatus == 0) {
        result = request_is_complete(request, buf, len);
        if (result == -1)
            return PARSE_ERROR;
        else if (result == 0)
            return PARSE_PARTIAL;
        else
            request->parsestatus = 1;
    }
    if (request->parsestatus == 1) {
        result = request_parse_url(request, buf, len);
        if (result == 0)
            return PARSE_ERROR;
        else
            return PARSE_DONE;
    }
    return PARSE_ERROR;
} 


static int reply_is_complete(HttpReply *reply, char *buffer, int len)
{
    int i;
    int lastoffset = reply->replylength;
    if (lastoffset < 2)
        lastoffset = 2;

    for (i = lastoffset; i < len; i++) {
        if (buffer[i] == '\n' &&
            (buffer[i-1] == '\n' || buffer[i-2] == '\n')) {
            reply->parsestatus = 1;
            reply->replylength = i+1;
            return 1;
        }
    }
    return 0;
}

static int reply_parse_status(HttpReply *reply, char *buf, int len)
{
    int endline, offset;
    endline = findchar(buf, 0, len, '\n');
    if (endline == -1)
        return 0;
    if (strncmp(buf, "HTTP/1.", strlen("HTTP/1.")) != 0)
        return 0;
    offset = findchar(buf, strlen("HTTP/1.1"), endline, ' ');
    if (offset == -1)
        return 0;
    offset = skipchar(buf, offset, endline, ' ');
    if (offset == endline)
        return 0;
    reply->status = atoi(&buf[offset]);
    if (reply->status <= 0 || reply->status > 700)
        return 0;
    return 1;
}

static char reply_is_cached(char *buf, int len)
{
    int offset;
    offset = findstring(buf, 0, len, "Set-Cookie");
    if (offset == -1)
        return 1;
    else
        return 0;
}

static char reply_is_gzipped(char *buf, int len)
{
    int offset;
    offset = findstring(buf, 0, len, "Content-Encoding: gzip");
    if (offset == -1)
        return 0;
    else
        return 1;
}
  

static int reply_get_contentlength(char *buf, int len)
{
    int offset;
    offset = findstring(buf, 0, len, "Content-length:");
    if (offset == -1)
        return -1;
    offset += strlen("Content-length:");
    return atoi(&buf[offset]);
}

static unsigned long reply_get_crc(char *buf, int len)
{
    int offset;
    offset = findstring(buf, 0, len, "CRC:");
    if (offset == -1)
        return 0;
    offset += strlen("CRC:");
    return atol(&buf[offset]);
}

int reply_parse(HttpReply *reply, char *buffer, int len)
{
    int result;
    if (reply->parsestatus == 0) {
        result = reply_is_complete(reply, buffer, len);
        if (result == -1)
            return PARSE_ERROR;
        if (result == 0)
            return PARSE_PARTIAL;
        reply->parsestatus = 1;
    }
    if (reply->parsestatus == 1) {
        result = reply_parse_status(reply, buffer, len);
        if (result == 0)
            return PARSE_ERROR;
        reply->contentlength = reply_get_contentlength(buffer, len);
        reply->iscached = reply_is_cached(buffer, len);
        reply->isgzipped = reply_is_gzipped(buffer, len);
        reply->crc = reply_get_crc(buffer, len);
    }
    return PARSE_DONE;
}


