/*
 * Copyright (c) 2003-2007 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/***

  Usage: stress_client <proxyip> <proxyport> <numclients> 
                        <duration> <gzipped cached pages> 
                        <cached pages> <gzipped pages> <other pages>

  proxyip:              IP address of Vicompress
  proxyport:            Listening port of Vicompress
  numclients:           Number of concurent HTTP clients
  duration:             Duration of test, in minutes
  gzipped cached pages: Number of pages that can be gzipped and cached
  cached pages:         Number of pages that can be only cached.
  gzipped pages:        Number of pages that can be only gzipped.
  other pages:          Number of non-gzipped, non-cached pages.

The stress_server sends the same page (sample.html) for every request
it receives.  The stress_client will create random URLs such as:

http://hostname/123456.html
http://hostname/925234.cachable.html
http://hostname/655231.cangzip.html
http://hostname/235814.cachable.cangzip.html

If the URL contains the word "cachable", then the server will make
the reply cachable (it will include a Last-Modified header).

If the URL contains the word "cangzip", then the server will send
the header "Content-type: text/html", so that the reply can be gzipped.

The server also sends the CRC (checksum) of the reply data in the
http reply header.  This stress_client independently computes the 
CRC on the reply data and verifies that it matches the one in the
reply header.  For gzipped replies, the reply data is uncompressed,
and the CRC is computed on the uncompressed reply.

At the end of the test, the following statistics will be printed:

Starting vicompress with temp.conf
Start stress server....
Starting stress client, will run for 10 minutes.....

Duration                             10 minutes
Clients                              10
Total requests/sec                   71.67
    cached and gzipped                8.35
    cached non-gzipped                3.61
    gzipped non-cached               27.62
    other                            32.09
total average response time         138.82 msec
cached and gzipped response time     31.45 msec
cached non-gzipped response time     81.67 msec
gzipped non-cached response time    142.81 msec
other response time                 169.77 msec
Stress client done


***/


#define FD_SETSIZE 8192
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <assert.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <zlib.h>
#include <stdarg.h>
#include "http.h"

#define MAX_CLIENTS       4096
#define BUF_SIZE          1024
#define GZIP_HEADER_SIZE  10
#define COOL_DOWN_TIME    12

#define WAIT_CONNECT    1
#define WAIT_WRITE      2
#define WAIT_REPLY      3
#define WAIT_BODY       4

#define PARSE_ERROR     1
#define PARSE_PARTIAL   2
#define PARSE_DONE      3

#define CONNECT_TIMEOUT 30
#define WRITE_TIMEOUT   30
#define READ_TIMEOUT    30

struct _Config {
    int clients;
    int gzipped_and_cached_pages;
    int cached_pages;
    int gzipped_pages;
    int other_pages;
    int total_pages;
    int duration;
    struct sockaddr_in proxy;
};
typedef struct _Config Config;

struct _Stats {
    unsigned long total_other;
    unsigned long total_gzipped_and_cached;
    unsigned long total_cached;
    unsigned long total_gzipped;
    double other_time;
    double gzipped_and_cached_time;
    double cached_time;
    double gzipped_time;
};
typedef struct _Stats Stats;

typedef struct _StressClient StressClient;

struct _Client {
    int fd;
    char waitstatus;
    time_t timeout;
    int readoffset;
    int writeoffset;
    int requestlength;
    int totalread;
    int totaluncompressed;
    char url[32];
    HttpReply reply;
    long id;
    char read_gzip_header;
    unsigned long crc;
    char buffer[BUF_SIZE];
    struct timeval start;
    struct timeval end;
    StressClient *stress;
    z_stream zstream;
};
typedef struct _Client Client;

struct _StressClient {
    Config config;
    Stats stats;
    Client clients[MAX_CLIENTS];
    time_t currtime, endtime, prevtime;
    fd_set readset, writeset, errorset;
    fd_set readset2, writeset2, errorset2;
    int maxfd, maxfd2;
};

static void wait_connect(Client *c);
static void wait_write(Client *c);
static void wait_read_reply(Client *c);
static void wait_read_body(Client *c);

static void start_client(StressClient *s);
static void end_client(Client *c, int error);
static void do_connect(Client *c, struct sockaddr_in *proxy);
static void create_request(Config *c, Client *client);
static void write_request(Client *c);
static void read_reply(Client *c);
static void read_body(Client *c);
static void process_body(Client *c, int len, int err);
static void process_gzipped_body(Client *c, int len, int err);
static void check_crc(Client *c);
static void store_stats(Client *c);
static void print_stats(StressClient *s);


static void printerr(Client *client, const char *fmt, ...)
{
    struct sockaddr_in addr;
    int addrlen;
    char msg[256];
    va_list args;

    addrlen = sizeof(addr);
    getsockname(client->fd, (void *)&addr, &addrlen);
    fprintf(stderr, "ERROR client (%d) url %s: ", ntohs(addr.sin_port), client->url);

    memset(msg, 0, sizeof(msg));
    va_start(args, fmt);
    vsnprintf(msg, 256, fmt, args);
    va_end(args); 

    fprintf(stderr, msg); 
}

static int ignoreErrno(int error)
{
    switch (error) {
    case EINPROGRESS:
    /*  case EWOULDBLOCK: */
    case EAGAIN:
    case EALREADY:
    case EINTR:
        return 1;
    default:
        return 0;
    }
}


static void initzlib(Client *client)
{
    int n;

    if (!client->reply.isgzipped)
        return;

    memset(&(client->zstream), 0, sizeof(z_stream));
    client->zstream.next_in = (unsigned char *)client->buffer;
    client->zstream.avail_in = GZIP_HEADER_SIZE;
    n = inflateInit2(&(client->zstream), -MAX_WBITS);
    if (n != Z_OK) {
        printerr(client, "Error in inflateInit2 %d\n", n);
        memset(&(client->zstream), 0, sizeof(z_stream));
    }
    client->read_gzip_header = 1;
}

static void freezlib(Client *client)
{
    if (!client->reply.isgzipped)
        return;

    inflateEnd(&(client->zstream));
    client->zstream.avail_in = client->zstream.avail_out = 0;
    client->zstream.next_in = client->zstream.next_out = NULL;
}


static void end_client(Client *client, int error)
{
    StressClient *s = client->stress;
    int fd = client->fd;

    if (error == 0) {
        check_crc(client);
    } 
    freezlib(client); 
    gettimeofday(&client->end, NULL);
    store_stats(client);

    if (error) {
        close(fd);
        client->fd = -1;
        return;
    }
    if (s->currtime >= s->endtime) {
        close(client->fd);
        return;
    }

    /* If possible, keep the connection alive.  If no content-length
     * was given in the reply header, then the vicompress proxy will
     * close the connection when the reply is done, so we'll have to 
     * create a new connection.
     */
    if (client->reply.contentlength > 0) {
        memset(client, 0, sizeof(Client));
        client->fd = fd;
        client->stress = s;
        create_request(&client->stress->config, client);
        write_request(client);
    }
    else {
        close(fd);
        memset(client, 0, sizeof(Client));
        start_client(s);
    }
}

static void start_client(StressClient *s)
{
    Client *client;
    int fd;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    assert(fd > 0);
    fcntl(fd, F_SETFL, O_NONBLOCK);
    client = &s->clients[fd];
    memset(client, 0, sizeof(Client));
    client->fd = fd; 
    client->stress = s;

    client->timeout = s->currtime + CONNECT_TIMEOUT;
    do_connect(client, &s->config.proxy);
}

static void do_connect(Client *client, struct sockaddr_in *proxy)
{
    int result, error;

    result = connect(client->fd, (struct sockaddr*)proxy, 
                     sizeof(struct sockaddr_in));
    error = errno;
    if (result == 0 || error == EISCONN) {
        create_request(&client->stress->config, client);
        write_request(client);
        return;
    }
    else if (result == -1 && (error == EINPROGRESS || error == EALREADY)) {
        wait_connect(client);
    }
    else {
        printerr(client, "Failed to connect: %s\n", strerror(error));
        end_client(client, 1);
    } 
}


int iscachable(Config *config, long id)
{
    int start, end;

    start = 0; 
    end = config->gzipped_and_cached_pages;
    if (id >= start && id < end)
        return 1;

    start = config->gzipped_and_cached_pages;
    end = start + config->cached_pages;
    if (id >= start && id < end)
        return 1;
    else
        return 0;
}

int cangzip(Config *config, long id)
{
    int start, end;

    start = 0; 
    end = config->gzipped_and_cached_pages;
    if (id >= start && id < end)
        return 1;

    start = config->gzipped_and_cached_pages +
              config->cached_pages;
    end = start + config->gzipped_pages;
    if (id >= start && id < end)
        return 1;
    else
        return 0;
}


static void create_request(Config *config, Client *client)
{
    char *suffix;

    client->id = random() % config->total_pages;
    assert(client->id >= 0 && client->id < config->total_pages);

    if (iscachable(config, client->id) && cangzip(config, client->id))
        suffix = ".cachable.cangzip";
    else if (iscachable(config, client->id))
        suffix = ".cachable";
    else if (cangzip(config, client->id))
        suffix = ".cangzip";
    else
        suffix = "";


    client->buffer[0] = 0;
    sprintf(client->buffer, "GET /%ld%s.html HTTP/1.0\r\n", 
            client->id, suffix);
    strcat(client->buffer, 
"Host: localhost\r\n\
Connection: keep-alive\r\n\
Accept-Encoding: gzip\r\n\r\n");

    client->requestlength = strlen(client->buffer);
    sprintf(client->url, "/%ld%s.html", client->id, suffix);
    client->timeout = client->stress->currtime + WRITE_TIMEOUT;
    gettimeofday(&client->start, NULL);
} 

static void write_request(Client *client)
{
    int offset, len, result;

    offset = client->writeoffset;
    len = client->requestlength - offset;
    result = send(client->fd, &client->buffer[offset], len, 0);
    if (result > 0)
        client->writeoffset += result;

    if (result == -1 && !ignoreErrno(errno)) {
        printerr(client, "Failed to write request: %s\n", strerror(errno));
        end_client(client, 1);
        return;
    }
    else if (client->writeoffset == client->requestlength) {
        client->timeout = client->stress->currtime + READ_TIMEOUT;
        wait_read_reply(client);
        return;
    }
    else {
        wait_write(client);
    }
}

static void read_reply(Client *client)
{
    int len, offset, result;

    offset = client->readoffset;
    len = recv(client->fd, &client->buffer[offset], BUF_SIZE - offset, 0);
    if (len > 0)
        client->readoffset += len;

    if (len == 0 || (len == -1 && !ignoreErrno(errno))) {
        printerr(client, "Failed in reading reply: len %d %s\n", len, strerror(errno));
        end_client(client, 1);
        return;
    }
    else if (len == -1 && ignoreErrno(errno)) {
        wait_read_reply(client);
        return;
    }
    result = reply_parse(&client->reply, client->buffer, client->readoffset);
    if (result == PARSE_ERROR) {
        printerr(client, "Failed to parse reply %s\n", client->buffer);
        end_client(client, 1);
        return;
    }
    else if (result == PARSE_PARTIAL) {
        wait_read_reply(client);
        return;
    }
    else if (result == PARSE_DONE) {
        int bodylen = client->readoffset - client->reply.replylength;
        client->readoffset = 0;

        if (bodylen > 0) {
            memmove(client->buffer, &client->buffer[client->reply.replylength], bodylen);
            if (client->reply.isgzipped)
                process_gzipped_body(client, bodylen, 0);
            else
                process_body(client, bodylen, 0);
        }
        else {
            wait_read_body(client);
        }
    }
}

static void check_crc(Client *client) 
{
    if (client->reply.contentlength >= 0 &&
        client->totalread != client->reply.contentlength) {
        printerr(client, "Server closed prematurely, read %d of %d bytes\n",
              client->totalread, client->reply.contentlength);
        return;
    }
    if (client->crc != client->reply.crc) {
        printerr(client, "The computed CRC %ld != %ld\n", 
                 client->crc, client->reply.crc);
        return;
    }
}


static void read_body(Client *client)
{
    int len, err;

    err = 0;
    len = recv(client->fd, client->buffer, BUF_SIZE, 0);
    err = errno;
    if (len == -1 && err == EAGAIN)
        wait_read_body(client); 
    else if (client->reply.isgzipped)
        process_gzipped_body(client, len, err);
    else 
        process_body(client, len, err);
}

static void process_body(Client *client, int len, int err)
{
    if (len == -1) {
        printerr(client, "Error while reading body %s\n", strerror(err));
        end_client(client, 1);
        return;
    }
    else if (len == 0) {
        end_client(client, 0);
        return;
    }
    else if (len > 0) {
        client->crc = crc32(client->crc, (unsigned char *)client->buffer, len);
        client->totalread += len;
 
        if (client->totalread == client->reply.contentlength) { 
            end_client(client, 0);
        }
        else {
            wait_read_body(client);
        }
    }
}

static int uncompress_data(Client *client, int inputlen, 
                           char *dest, int destlen, int isend)
{
    int n;

    client->zstream.next_in = (unsigned char *) &(client->buffer[0]);
    client->zstream.avail_in = inputlen;
    client->zstream.next_out = (unsigned char *) dest;
    client->zstream.avail_out = destlen;

    if (isend)
        n = inflate(&(client->zstream), Z_FINISH);
    else
        n = inflate(&(client->zstream), 0);

    if (n != Z_OK && n != Z_STREAM_END)
        return -1;

    n = destlen - client->zstream.avail_out;
    client->zstream.avail_in = 0;
    return n; 
}


static void process_gzipped_body(Client *client, int len, int err)
{
    char plaindata[65535];
    int n;

    if (len == -1) {
        printerr(client, "Error reading body: %s\n", strerror(err));
        end_client(client, 1);
        return;
    }
    else if (len == 0) {
        n = uncompress_data(client, 0, plaindata, 65535, 1);
        if (n == -1) {
            printerr(client, "Error uncompressing data: compressed %d bytes, uncompressed %d bytes\n",
                client->totalread, client->totaluncompressed);
            end_client(client, 1);
            return;
        }
        else {
            client->crc = crc32(client->crc, (unsigned char *)plaindata, n);
            client->totaluncompressed += n;
            end_client(client, 0);
            return;
        }
    }
    else if (len > 0 && !client->read_gzip_header) {
        if (len >= GZIP_HEADER_SIZE) {
            client->totalread += GZIP_HEADER_SIZE;
            initzlib(client);
            if (len > GZIP_HEADER_SIZE) {
                memmove(client->buffer, &(client->buffer[GZIP_HEADER_SIZE]), 
                        len - GZIP_HEADER_SIZE);
                process_gzipped_body(client, len - GZIP_HEADER_SIZE, 0);
            }
            else {
                wait_read_body(client);
            }
        }
        else {
            assert(0);
            wait_read_body(client);
        }
    }
    else if (len > 0 && client->read_gzip_header) {
        client->totalread += len;
        n = uncompress_data(client, len, plaindata, 65535, 0);
        if (n == -1) {
            printerr(client, "Error uncompressing data: compressed %d bytes, uncompressed %d bytes\n",
                client->totalread, client->totaluncompressed);
            end_client(client, 1);
            return;
        }
        else {
            client->crc = crc32(client->crc, (unsigned char *)plaindata, n);
            client->totaluncompressed += n;
 
            if (client->totalread == client->reply.contentlength) 
                end_client(client, 0); 
            else
                wait_read_body(client);
        }
    }
}



static void wait_connect(Client *c)
{
    StressClient *s = c->stress;
    c->waitstatus = WAIT_CONNECT;
    FD_SET(c->fd, &s->writeset2);
    FD_SET(c->fd, &s->errorset2);
    if (s->maxfd2 < c->fd)
        s->maxfd2 = c->fd;
}

static void wait_write(Client *c)
{
    StressClient *s = c->stress;
    c->waitstatus = WAIT_WRITE;
    FD_SET(c->fd, &s->writeset2);
    FD_SET(c->fd, &s->errorset2);
    if (s->maxfd2 < c->fd)
        s->maxfd2 = c->fd;
}

static void wait_read_reply(Client *c)
{
    StressClient *s = c->stress;
    c->waitstatus = WAIT_REPLY;
    FD_SET(c->fd, &s->readset2);
    FD_SET(c->fd, &s->errorset2);
    if (s->maxfd2 < c->fd)
        s->maxfd2 = c->fd;
}

static void wait_read_body(Client *c)
{
    StressClient *s = c->stress;
    c->waitstatus = WAIT_BODY;
    FD_SET(c->fd, &s->readset2);
    FD_SET(c->fd, &s->errorset2);
    if (s->maxfd2 < c->fd)
        s->maxfd2 = c->fd;
}

static void checkTimeouts(StressClient *s)
{
    int i;
    Client *client;
    const char* msg[12];

    memset(msg, 0, sizeof(msg));
    msg[WAIT_CONNECT] = "waiting to connect";
    msg[WAIT_WRITE] = "waiting to write";
    msg[WAIT_REPLY] = "waiting to read reply";
    msg[WAIT_BODY] = "waiting to read body";

    for (i = 3; i <= s->maxfd; i++) {
        client = &s->clients[i];
        if (client->fd <= 0)
            continue;
        if (client->timeout > 0 && client->timeout < s->currtime) {
            FD_CLR(i, &s->readset2); 
            FD_CLR(i, &s->writeset2); 
            FD_CLR(i, &s->errorset2);
            printerr(client, "Timeout %s\n", msg[(int)client->waitstatus]);
            end_client(client, 1);
        }
    }
}

static void run(StressClient *s)
{
    int i, sleepcount;

    s->currtime = time(NULL);
    s->prevtime = s->currtime;
    s->endtime = s->currtime + s->config.duration; 
    sleepcount = 0;

    for (i = 0; i < s->config.clients; i++)
        start_client(s);

    while (s->currtime < s->endtime + COOL_DOWN_TIME) {
        Client *c;
        int n;
        char readevent, writeevent;
        struct timeval tv;

        s->currtime = time(NULL);

        if (s->currtime - s->prevtime > 3) {
            s->prevtime = s->currtime;
            checkTimeouts(s);
        }

        tv.tv_sec = 1; tv.tv_usec = 0; 
        memcpy(&s->readset, &s->readset2, sizeof(fd_set));
        memcpy(&s->writeset, &s->writeset2, sizeof(fd_set));
        memcpy(&s->errorset, &s->errorset2, sizeof(fd_set));
        s->maxfd = s->maxfd2;

        if (sleepcount == 4) {
            usleep(5);
            sleepcount = 0;
        }
        else {
            sleepcount++;
        }
        n = select(s->maxfd+1, &s->readset, &s->writeset, &s->errorset, &tv);
        for (i = 3; i <= s->maxfd; i++) {
            c = &s->clients[i];
            if (c->fd <= 0)
                continue;

            readevent = FD_ISSET(i, &s->readset) || FD_ISSET(i, &s->errorset);
            writeevent = FD_ISSET(i, &s->writeset) || FD_ISSET(i, &s->errorset); 

            if (readevent && c->waitstatus == WAIT_REPLY) {
                FD_CLR(i, &s->readset2); FD_CLR(i, &s->errorset2);
                read_reply(c);
            }
            else if (readevent && c->waitstatus == WAIT_BODY) {
                FD_CLR(i, &s->readset2); FD_CLR(i, &s->errorset2);
                read_body(c);
            }
            else if (writeevent && c->waitstatus == WAIT_CONNECT) {
                FD_CLR(i, &s->writeset2); FD_CLR(i, &s->errorset2);
                do_connect(c, &s->config.proxy);
            }
            else if (writeevent && c->waitstatus == WAIT_WRITE) {
                FD_CLR(i, &s->writeset2); FD_CLR(i, &s->errorset2);
                write_request(c);
            }
        } 
    }
}

static void store_stats(Client *c)
{
    Stats *stats;
    long msec;
    
    if (c->stress->currtime >= c->stress->endtime)
        return;

    stats = &c->stress->stats;
    msec = (c->end.tv_usec - c->start.tv_usec)/1000;
    msec = msec + (c->end.tv_sec - c->start.tv_sec) * 1000;

    if (c->reply.iscached && c->reply.isgzipped) {
        stats->total_gzipped_and_cached += 1;
        stats->gzipped_and_cached_time += msec;
    }
    else if (c->reply.iscached) {
        stats->total_cached += 1;
        stats->cached_time += msec;
    }
    else if (c->reply.isgzipped) {
        stats->total_gzipped += 1;
        stats->gzipped_time += msec;
    }
    else {
        stats->total_other += 1;
        stats->other_time += msec;
    } 
}

static void print_stats(StressClient *stress)
{
    long total;
    double totalmsec, duration;
    Stats *s; 

    s = &stress->stats;
    total = s->total_gzipped_and_cached + s->total_cached + 
                    s->total_gzipped + s->total_other;

    duration = stress->config.duration;

    printf("Duration                          %5d minutes\n", 
           stress->config.duration / 60);
    printf("Clients                           %5d\n", stress->config.clients);
    printf("Total requests/sec                %8.2f\n", total/duration);
    printf("    cached and gzipped            %8.2f\n", 
           s->total_gzipped_and_cached / duration);
    printf("    cached non-gzipped            %8.2f\n",
           s->total_cached / duration);
    printf("    gzipped non-cached            %8.2f\n", 
           s->total_gzipped / duration);
    printf("    non-gzipped non-cached        %8.2f\n", 
           s->total_other / duration);

    totalmsec = s->gzipped_and_cached_time + s->cached_time +
                s->gzipped_time + s->other_time;

    /* For average response times, don't divide by zero */
    if (s->total_gzipped_and_cached == 0)
        s->total_gzipped_and_cached += 1;
    if (s->total_cached == 0)
        s->total_cached += 1;
    if (s->total_gzipped == 0)
        s->total_gzipped += 1;
    if (s->total_other == 0)
        s->total_other += 1;

    printf("total average response time       %8.2f msec\n", totalmsec / total);
    printf("    cached and gzipped            %8.2f msec\n",
                  s->gzipped_and_cached_time / s->total_gzipped_and_cached);
    printf("    cached non-gzipped            %8.2f msec\n",
                  s->cached_time / s->total_cached);
    printf("    gzipped non-cached            %8.2f msec\n",
                  s->gzipped_time / s->total_gzipped);
    printf("    non-gzipped non-cached        %8.2f msec\n",
                  s->other_time / s->total_other); 
}

int main(int argc, char **argv)
{
    StressClient *s;
 
    if (argc < 9) {
        printf("Usage: stress_client <proxyip> <proxyport> <numclients> <duration> <gzipped cached pages> <cached pages> <gzipped pages> <other pages>\n");
        printf("  proxyip:              IP address of Vicompress\n");
        printf("  proxyport:            Listening port of Vicompress\n");
        printf("  numclients:           Number of concurent HTTP clients\n");
        printf("  duration:             Duration of test, in minutes\n");
        printf("  gzipped cached pages: Number of pages that can be gzipped and cached\n");
        printf("  cached pages:         Number of pages that can be only cached.\n");
        printf("  gzipped pages:        Number of pages that can be only gzipped.\n");
        printf("  other pages:          Number of non-gzipped, non-cached pages.\n");
        exit(1);
    } 
    s = malloc(sizeof(StressClient));
    memset(s, 0, sizeof(StressClient));
    s->config.proxy.sin_family = AF_INET;
    s->config.proxy.sin_addr.s_addr = inet_addr(argv[1]);
    s->config.proxy.sin_port = htons((unsigned short) atoi(argv[2]));
    assert(s->config.proxy.sin_port != 0);

    s->config.clients = atoi(argv[3]);
    assert(s->config.clients > 0 && s->config.clients < MAX_CLIENTS);

    s->config.duration = atoi(argv[4]) * 60;
    assert(s->config.duration > 0);

    s->config.gzipped_and_cached_pages = atoi(argv[5]);
    assert(s->config.gzipped_and_cached_pages >= 0);

    s->config.cached_pages = atoi(argv[6]);
    assert(s->config.cached_pages >= 0);

    s->config.gzipped_pages = atoi(argv[7]);
    assert(s->config.gzipped_pages >= 0);

    s->config.other_pages = atoi(argv[8]);
    assert(s->config.other_pages >= 0);

    s->config.total_pages = s->config.gzipped_and_cached_pages +
                            s->config.cached_pages +
                            s->config.gzipped_pages +
                            s->config.other_pages;

    signal(SIGPIPE, SIG_IGN);

    run(s);
 
    print_stats(s);
    return 0;
}

