/*
 * Copyright (c) 2003-2007 Visolve
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */

/**

Usage: stress_server <port> <file> <duration>
  port:     The port to listen on
  file:     The html file used to respond to all requests
  duration: The time to run the server, in minutes

The stress server will send the same reply body (sample.html) for
every request.  

If the URL contains the word "cachable", the server will include a
"Last-Modified" header in the reply, so that it can be cached.

If the URL contains the word "cangzip", the server will include a
"Content-type: text/html" header in the reply, so that it can be gzipped.

The server also sends the CRC (checksum) of the reply body in the
reply http header.  The stress client uses the CRC to make sure the
reply data it downloaded is correct.

**/


#define FD_SETSIZE  8192
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <zlib.h>
#include <fcntl.h>
#include <assert.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include "http.h"

#define MAX_CLIENTS       4096
#define BUF_SIZE          1024

#define WAIT_ACCEPT     1
#define WAIT_READ       2
#define WAIT_REPLY      3
#define WAIT_BODY       4

#define PARSE_ERROR     1
#define PARSE_PARTIAL   2
#define PARSE_DONE      3

#define CONNECT_TIMEOUT 30
#define WRITE_TIMEOUT   30
#define READ_TIMEOUT    40

typedef struct _StressServer StressServer;
struct _Client {
    int fd;
    char waitstatus;
    time_t timeout;
    int readoffset;
    int writeoffset;
    int replylength;
    int amountwritten;
    HttpRequest request;
    char buffer[BUF_SIZE];
    StressServer *server;
};
typedef struct _Client Client;

struct _StressServer {
    int port;
    int listenfd;
    int duration;
    Client clients[MAX_CLIENTS];
    time_t currtime, endtime;
    fd_set readset, writeset, errorset;
    fd_set readset2, writeset2, errorset2;
    int maxfd, maxfd2;
    const char *filename;
    char *page;
    int pagesize;
    unsigned long page_crc;
};

static void wait_read(Client *c);
static void wait_write_reply(Client *c);
static void wait_write_body(Client *c);

static void accept_client(StressServer *s);
static Client* init_client(StressServer *s, int fd);
static void read_request(Client *c);
static void create_reply(Client *c);
static void write_reply(Client *c);
static void write_body(Client *c);
static void close_client(Client *c);
static void run(StressServer *s);


static void loadpage(StressServer *s)
{
    struct stat info;
    int fd, offset, n;

    n = stat(s->filename, &info); assert(n == 0);
    s->pagesize = info.st_size;
    s->page = malloc(s->pagesize);
    fd = open(s->filename, O_RDONLY);
    assert(fd > 0);

    offset = 0;
    n = read(fd, &s->page[offset], s->pagesize - offset);
    while (n > 0) {
        offset += n;
        n = read(fd, &s->page[offset], s->pagesize - offset); 
    }
    close(fd);
    s->page_crc = crc32(0, NULL, 0);
    s->page_crc = crc32(s->page_crc, (unsigned char *)s->page, s->pagesize);
}
    

static int ignoreErrno(int error)
{
    switch (error) {
    case EINPROGRESS:
    /*  case EWOULDBLOCK: */
    case EAGAIN:
    case EALREADY:
    case EINTR:
        return 1;
    default:
        return 0;
    }
}


static void accept_client(StressServer *s)
{
    int fd, i;
    struct sockaddr_in addr;
    int len;

    for (i = 0; i < 100; i++) {
        len = sizeof(addr);
        fd = accept(s->listenfd, (struct sockaddr*) &addr, &len); 
        if (fd != -1) {
            Client *c = init_client(s, fd);
            read_request(c);
        }
        else if (fd == -1 && errno != EINTR) {
            return;
        }
    }
}

static void close_client(Client *c)
{
    close(c->fd);
    memset(c, 0, sizeof(Client));
}

static Client* init_client(StressServer *s, int fd)
{
    Client *c = &s->clients[fd];
    memset(c, 0, sizeof(Client));
    c->fd = fd;
    fcntl(fd, F_SETFL, O_NONBLOCK);
    c->server = s;
    return c;
}

static void read_request(Client *c)
{
    int offset, n, eof, result, error;

    offset = c->readoffset;
    n = recv(c->fd, &c->buffer[offset], BUF_SIZE - offset, 0);
    error = (n == -1 && !ignoreErrno(errno));
    if (n > 0)
        c->readoffset += n;
    eof = (n == 0);

    result = request_parse(&c->request, c->buffer, c->readoffset);
    if (result == PARSE_ERROR) {
        fprintf(stderr, "ERROR stress_server client %s: Error parsing request: %s\n", 
                c->request.url, c->buffer);
        close_client(c);
    }
    else if (result == PARSE_PARTIAL && !eof) {
        wait_read(c);
    }
    else if (result == PARSE_PARTIAL && eof) {
        if (c->readoffset > 0) {
            fprintf(stderr, "ERROR stress_server client %s: Client closed during request\n",
                    c->request.url);
        }
        close_client(c);
    }
    else if (result == PARSE_DONE) {
        c->readoffset = 0;
        create_reply(c);
        write_reply(c);
    } 
}

static void create_reply(Client *c)
{
    memset(c->buffer, 0, BUF_SIZE);
    sprintf(c->buffer, 
"HTTP/1.1 200 OK\r\n\
Date: Thu, 24 Jul 2003 21:33:02 GMT\r\n\
Content-Length: %d \r\n\
Connection: keep-alive\r\n\
CRC: %ld \r\n\
Set-Cookie: id=0\r\n",
          c->server->pagesize, c->server->page_crc);

    if (c->request.iscachable)
        strcat(c->buffer, "Last-Modified: Tue, 22 Jul 2003 12:33:02 GMT\r\n");
    else
        strcat(c->buffer, "Pragma: no-cache\r\n");

    if (c->request.cangzip)
        strcat(c->buffer, "Content-type: text/html\r\n");

    strcat(c->buffer, "\r\n");
    c->replylength = strlen(c->buffer); 
}

static void write_reply(Client *client)
{
    int offset, len, result;

    offset = client->writeoffset;
    len = client->replylength - offset;
    result = send(client->fd, &client->buffer[offset], len, 0);
    if (result > 0)
        client->writeoffset += result;

    if (result == -1 && !ignoreErrno(errno)) {
        fprintf(stderr, "ERROR stress_server client %s: Error in writing reply: %s\n",
                client->request.url, strerror(errno));
        close_client(client);
        return;
    }
    else if (client->writeoffset == client->replylength) {
        client->writeoffset = 0;
        client->amountwritten = 0;
        wait_write_body(client);
        return;
    }
    else {
        wait_write_reply(client);
    }
}

static void write_body(Client *client)
{
    int len, result;
    char *data;

    data = &(client->server->page[ client->amountwritten ]);
    len = client->server->pagesize - client->amountwritten;
    if (len > 4096)
        len = 4096;

    result = send(client->fd, data, len, 0);
    if (result > 0) {
        client->amountwritten += result;
    }

    if (result == -1 && !ignoreErrno(errno)) {
        fprintf(stderr, "ERROR stress_server client %s: Error writing reply body: %s\n",
                client->request.url, strerror(errno));
        close_client(client);
        return;
    }
    else if (client->amountwritten == client->server->pagesize) {
        close_client(client);
    }
    else {
        wait_write_body(client);
    }
}

static void wait_write_reply(Client *c)
{
    StressServer *s = c->server;
    c->waitstatus = WAIT_REPLY;
    FD_SET(c->fd, &s->writeset2);
    FD_SET(c->fd, &s->errorset2);
    if (s->maxfd2 < c->fd)
        s->maxfd2 = c->fd;
}

static void wait_write_body(Client *c)
{
    StressServer *s = c->server;
    c->waitstatus = WAIT_BODY;
    FD_SET(c->fd, &s->writeset2);
    FD_SET(c->fd, &s->errorset2);
    if (s->maxfd2 < c->fd)
        s->maxfd2 = c->fd;
}

static void wait_read(Client *c)
{
    StressServer *s = c->server;
    c->waitstatus = WAIT_READ;
    FD_SET(c->fd, &s->readset2);
    FD_SET(c->fd, &s->errorset2);
    if (s->maxfd2 < c->fd)
        s->maxfd2 = c->fd;
}

static void createListen(StressServer *s)
{
    struct sockaddr_in addr;
    int fd, n, port;

    port = s->port;
    fd = socket(AF_INET, SOCK_STREAM, 0);
    assert(fd > 0);
    s->listenfd = fd;
    n = 1;
    setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &n, 4);
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons((unsigned short)port);
    addr.sin_addr.s_addr = INADDR_ANY;
    bind(fd, (struct sockaddr *)&addr, sizeof(addr));
    listen(fd, 4096);
    fcntl(fd, F_SETFL, O_NONBLOCK); 
    FD_SET(fd, &s->readset2);
    FD_SET(fd, &s->errorset2);
    s->maxfd2 = fd;
}

static void run(StressServer *s)
{
    int i, sleepcount;

    loadpage(s);
    createListen(s);
    s->currtime = time(NULL);
    s->endtime = s->currtime + s->duration;
    sleepcount = 0;

    while (s->currtime < s->endtime) {
        Client *c;
        int n;
        char readevent, writeevent;
        struct timeval tv;

        s->currtime = time(NULL);
        tv.tv_sec = 1; tv.tv_usec = 0;
        memcpy(&s->readset, &s->readset2, sizeof(fd_set));
        memcpy(&s->writeset, &s->writeset2, sizeof(fd_set));
        memcpy(&s->errorset, &s->errorset2, sizeof(fd_set));
        s->maxfd = s->maxfd2;

        if (sleepcount == 4) {
            usleep(5);
            sleepcount = 0;
        }
        else {
            sleepcount++;
        }

        n = select(s->maxfd+1, &s->readset, &s->writeset, &s->errorset, &tv);
        for (i = 3; i <= s->maxfd; i++) {
            c = &s->clients[i];
            if (c->fd <= 0 && i != s->listenfd)
                continue;
            readevent = FD_ISSET(i, &s->readset) || FD_ISSET(i, &s->errorset);
            writeevent = FD_ISSET(i, &s->writeset) || FD_ISSET(i, &s->errorset);

            if (readevent && i == s->listenfd) {
                accept_client(s);
            }
            else if (readevent && c->waitstatus == WAIT_READ) {
                FD_CLR(i, &s->readset2); FD_CLR(i, &s->errorset2);
                read_request(c);
            }
            else if (writeevent && c->waitstatus == WAIT_REPLY) {
                FD_CLR(i, &s->writeset2); FD_CLR(i, &s->errorset2);
                write_reply(c);
            }
            else if (writeevent && c->waitstatus == WAIT_BODY) {
                FD_CLR(i, &s->writeset2); FD_CLR(i, &s->errorset2);
                write_body(c);
            }
        }
    }
}

int main(int argc, char **argv)
{
    StressServer *s;

    if (argc < 4) {
        printf("Usage: stress_server <port> <file> <duration>\n");
        printf("  port:     The port to listen on\n");
        printf("  file:     The html file used to respond to all requests\n");
        printf("  duration: The time to run the server, in minutes\n");
        exit(1);
    }
    s = (StressServer *)malloc(sizeof(StressServer));
    memset(s, 0, sizeof(StressServer));
    s->port = atoi(argv[1]);
    assert(s->port > 0 && s->port < 65536);
    s->filename = argv[2];
    s->duration = atoi(argv[3]) * 60;
    assert(s->duration > 0); 
    signal(SIGPIPE, SIG_IGN);
    run(s);
    return 0;
}

